
/*
 *	binary.c: Algorithms for binary image operators
 * excluding some large ones:
		Morphological operators: morphop.c
		Skeletonization: morphop.c
		Watershed segmentation: watershed.c
 */



/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "defs.h"

#include "application.h"
#include "binary.h"
#include "dlgbox.h"
#include "functions.h"
#include "filter.h"
#include "findclust.h"
#include "graphdsp.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "roi.h"
#include "secondwnd.h"
#include "segment.h"
#include "special1.h"






/*
	Alternate 2D region-growing algorithm. 
	The segmentation function "region grow" (2D and 3D) is in segmentation.c.

	This function is a helper function that allows to fill some contigugous regions
	as needed in other functions. 

	This function requires a mask of type B_CHAR where background is 0
	and foreground is 255. No other values are allowed.

	NOTE: Region growing is incredibly sped up when the scan directions
   alternate per sweep */


void rg_regiongrow (image_type* img, vertex seed, int z)
{
unsigned char* p;
long idx;
int itercnt, changecnt;
int x,y;

	g_assert (img->imgtype==B_CHAR);
	p = img->data;

	itercnt=0; idx = img->xmax * img->ymax * z;
	p[seed.x + seed.y*img->xmax + idx]=30;			/* Set the seed pixel */

	do
	{
		itercnt++; changecnt=0;
		dp (2, "Iteration %d: ", itercnt);

		if (itercnt & 0x01)
		{
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
				{
					idx = x + img->xmax* (y + img->ymax*z);
					if (p[idx]==0)
					{
						if (((x>0) && (p[idx-1]==30))
							|| ((x<img->xmax-1) && (p[idx+1]==30))
							|| ((y>0) && (p[idx-img->xmax]==30))
							|| ((y<img->ymax-1) && (p[idx+img->xmax]==30)))
						{
							p[idx] = 30;
							changecnt++;
						}
					}
				}
		}
		else
		{
			for (y=img->ymax-1; y>=0; y--)
				for (x=img->xmax-1; x>=0; x--)
				{
					idx = x + img->xmax* (y + img->ymax*z);
					if (p[idx]==0)
					{
						if (((x>0) && (p[idx-1]==30))
							|| ((x<img->xmax-1) && (p[idx+1]==30))
							|| ((y>0) && (p[idx-img->xmax]==30))
							|| ((y<img->ymax-1) && (p[idx+img->xmax]==30)))
						{
							p[idx] = 30;
							changecnt++;
						}
					}
				}
		}

	
		dp (2, "%d pixels\n",changecnt);

	}
	while (changecnt>0);


}


/***********************************************************************************

 #####    ###    #       #               #   #    ###    #       #####    ###   
 #         #     #       #               #   #   #   #   #       #       #   #  
 #         #     #       #               #   #   #   #   #       #       #      
 ####      #     #       #               #####   #   #   #       ####     ###   
 #         #     #       #               #   #   #   #   #       #           #  
 #         #     #       #               #   #   #   #   #       #       #   #  
 #        ###    #####   #####           #   #    ###    #####   #####    ###   

***********************************************************************************/


/* 
   Hole-filling algorithm. We assume background==0
   and everything else !=0. We region-grow a mask from a background 
   boundary point which covers all background pixels that are
   connected with the seed point. Any zero regions
   that are left over must be patched and will be filled
   with the histogram mean.
   TODO: This needs to be improved for RGB images.
*/



/* Move along the boundary strips to find a pixel ==0 that we assume to
	be a background pixel. Used as seed for fill_holes
*/

vertex find_background_pixel (image_type *img, int z)
{
long idx;
int x,y;
vertex v;
char found;
float buf;


	idx = img->xmax * img->ymax * z;		/* Point to top left of current slice */
	x=0; y=0; found=0;
	while ( (!found) && (x<img->xmax))
	{
		if (!nonzero (img, x,0,z))
		{
			v.x=x; v.y=0;
			found=1;
		}
		else if (!nonzero (img, x,img->ymax-1,z))
		{
			v.x=x; v.y=img->ymax-1;
			found=1;
		}
		x++;
	}

	if (found) return v;			/* Got a seed pt along top or bottom strip */

	y=0;
	while ( (!found) && (y<img->ymax))
	{
		if (!nonzero (img, 0,y,z))
		{
			v.x=0; v.y=y;
			found=1;
		}
		else if (!nonzero (img, img->xmax-1,y,z))
		{
			v.x=img->xmax-1; v.y=y;
			found=1;
		}
		y++;
	}

	if (found) return v;			/* Got a seed pt along left or right strip */

	v.x=-1; v.y=-1;			/* No seed point found */
	return v;

}






int fill_holes (image_type* img, int z)
{
image_type mask;
float avg, buf; int cnt;
int x,y; long idx;
unsigned char* p;
vertex seed;

	if (allocate_image (&mask, B_CHAR, img->xmax, img->ymax, 1)==-1)
		fatal_memerr();
	p = mask.data;

	seed = find_background_pixel (img, z);
	if ( (seed.x==-1) || (seed.y==-1) )
		return -1;								/* Do nothing if no background seed pixel exists */

	/* Create the mask and determine mean value */
	
	avg=0; cnt=0; idx = z * img->xmax * img->ymax;
	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
			if (nonzero (img,x,y,z))
			{
				buf = readbuf_flt (img,x,y,z);
				avg+=buf; cnt++;
				p[x + mask.xmax*y] = 255;
			}
		}
	if (cnt>0) avg/=cnt; else avg=1.0;
		
	/* Region-grow the mask, setting all background pixels to 30 */
	
	rg_regiongrow (&mask, seed, 0);
	
	/* Stuff all holes of the original image where the mask is still zero */

	switch (img->imgtype)	
	{
		case B_CHAR:
		{
				unsigned char* p1 = img->data;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						if (p[x + mask.xmax*y]==0)
						{
							p1[x + img->xmax*y + idx] = (int)avg;
						}
					}
				break;
		}
		case B_SHORT:
		{
				short* p1 = img->data;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						if (p[x + mask.xmax*y]==0)
						{
							p1[x + img->xmax*y + idx] = (int)avg;
						}
					}
				break;
		}
		case B_FLOAT:
		{
				float* p1 = img->data;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						if (p[x + mask.xmax*y]==0)
						{
							p1[x + img->xmax*y + idx] = avg;
						}
					}
				break;
		}
		case B_RGB:
		{
				rgbtriplet* p1 = img->data;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						if (p[x + mask.xmax*y]==0)
						{
							p1[x + img->xmax*y + idx] = (rgbtriplet){(int)avg,(int)avg,(int)avg};
						}
					}
				break;
		}
	}

	freebuf (&mask);
	return 0;
}



/* Non-GUI and GUI access */

void fill_holes_main ()
{
int z, err;

	err = 0;
	undo_nexus (&mainimg);
	for (z=0; z<mainimg.zmax; z++)
	{
		err |= fill_holes (&mainimg, z);
	}

	if (err) statusbar_display ("Error: No background found. Fill-holes FAILED");
}








/*********************************************************************************/


/*
	Clean boundary:
	Remove all features that are connected to the boundary strip.
	Zero/nonzero assumptions as in rg_regiongrow.
	Operates on a 2D basis for all slices
*/




/* Note that the floodfill algorithm is recursive and thus stack-heavy.
   At this time we need to abort it at level 130,000 to prevent a segfault.
   Note that floodfill is a 2D operation only.
   Note that the mask MUST be a B_CHAR.
   
   Floodfill fills a contiguous area of nonzero pixels that requires
   the seed point (x,y) as its member.
*/

static int floodfill_recursion_level;
static int floodfill_breakflg;


void feature_floodfill (image_type* img, image_type *mask, int x, int y, int z)
{

	if (floodfill_recursion_level > 130000) floodfill_breakflg=1;
	if (floodfill_breakflg) return;

	/* Abort recursion if out of range */
	if ((x<0) || (y<0) || (x>=img->xmax) || (y>=img->ymax)) return;
	floodfill_recursion_level++;
	
	/* Only consider data points of nonzero value that have not been visited in the mask */
	
	if ( (nonzero (img,x,y,z) ) &&
		((unsigned char*)mask->data)[x + img->xmax*y] ==0)
	{
		((unsigned char*)mask->data)[x + img->xmax*y] =255;		/* Mark this pixel as visited */

		/* and recurse to its eight neighbors */
		
		feature_floodfill (img, mask, x-1,y, z);		
		feature_floodfill (img, mask, x+1,y, z);
		feature_floodfill (img, mask, x,y-1, z);
		feature_floodfill (img, mask, x,y+1, z);
		feature_floodfill (img, mask, x-1,y-1, z);		
		feature_floodfill (img, mask, x+1,y-1, z);
		feature_floodfill (img, mask, x-1,y+1, z);
		feature_floodfill (img, mask, x+1,y+1, z);

	}
	floodfill_recursion_level--;
}



void feature_floodfill_init (image_type* img, image_type *mask, int x, int y, int z)
{
	floodfill_breakflg=0; floodfill_recursion_level=0;

	dp (2, "Flood-filling cluster at %d,%d\n",x,y);
	feature_floodfill (img, mask, x, y, z);
}

	
	
void clean_boundary_features (image_type *img)
{
long idx;
int x,y,z;
char found;
float buf;
image_type mask;
unsigned char *p;
unsigned char zeroelem[4] = {0,0,0,0};


	idx = img->xmax * img->ymax * z;		/* Point to top left of current slice */
	x=0; y=0; found=0;

	if (allocate_image (&mask, B_CHAR, img->xmax, img->ymax, 1)==-1)
		fatal_memerr();
	p = mask.data;
	

	for (z=0; z<img->zmax; z++)				/* Go over all slices */
	{

		/* At each slice, we need to ensure the mask is blank */

		for (y=0; y<mask.ymax; y++)
			for (x=0; x<mask.xmax; x++)
				p[x + mask.xmax*y]=0;

		for (x=0; x<img->xmax; x++)			/* Top and bottom margins */
		{
			if (nonzero (img, x,0,z))
			{
				feature_floodfill_init (img, &mask, x,0,z);
			}
			if (nonzero (img, x,img->ymax-1,z))
			{
				feature_floodfill_init (img, &mask, x,img->ymax-1,z);
			}
		}

		for (y=0; y<img->ymax; y++)			/* Left and right margins */
		{
			if (nonzero (img, 0,y,z))
			{
				feature_floodfill_init (img, &mask, 0,y,z);
			}
			if (nonzero (img, img->xmax-1,y,z))
			{
				feature_floodfill_init (img, &mask, img->xmax-1,y,z);
			}
		}

		/* We filled all features. Now blank out all elements in img that have a pixel in mask */

		for (y=0; y<mask.ymax; y++)
			for (x=0; x<mask.xmax; x++)
				if (p[x + mask.xmax*y]) writebuf (img, zeroelem, x,y,z);
	}



}



void clean_boundary_features_main ()
{
	undo_nexus (&mainimg);
	clean_boundary_features (&mainimg);
}



/*************************************************************

	Center: Center features of an image (i.e. shift the
	centroid in the middle of the image area)
	Depending on mass, we use nonzero pixels, equally weighted
	or we weight them by gray value (mass=1)

**************************************************************/


vertex get_centroid (image_type* img, int slice, int mass)
{
int x,y;
float cogx, cogy, val, buf;
vertex v;

	val=0.0; cogx=0.0; cogy=0.0;
	
	if (mass)
	{
	
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				buf = readbuf_flt (img,x,y,slice);
				val+=buf;
				cogx+=buf*x;
				cogy+=buf*y;
			}
		if (val>0)
		{
			cogx/=val;
			cogy/=val;
		}
		else
		{
			cogx=img->xmax/2;
			cogy=img->ymax/2;
		}
	}
	else
	{
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				buf = readbuf_flt (img,x,y,slice);
				if (buf!=0.0)
				{
					val+=1.0;
					cogx+=x;
					cogy+=y;
				}
			}
		if (val>0)
		{
			cogx/=val;
			cogy/=val;
		}
		else
		{
			cogx=img->xmax/2;
			cogy=img->ymax/2;
		}
	}
	v.x = (int)(0.5+cogx);
	v.y = (int)(0.5+cogy);
	return v;
}			




void center_feat (image_type* img, int z, int mass)
{
vertex v;
int x,y,x1,y1,dx,dy;
image_type temp;

	if (allocate_image (&temp, img->imgtype, img->xmax, img->ymax, 1)==-1)
		fatal_memerr();


	v = get_centroid (img, z, mass);
	dp (1, "Center image: Centroid at %d,%d\n",v.x,v.y);

	dx = v.x - img->xmax/2; /* Negative: Shift right */
	dy = v.y - img->ymax/2;
	if ((dx==0) &&(dy==0)) return;


	switch (img->imgtype)
	{
		case B_CHAR:
		{
			unsigned char c;
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
				{
					x1 = x+dx; y1 = y+dy; c = 0;
					if ((x1>=0) && (y1>=0) && (x1<img->xmax) && (y1<img->ymax)) readbuf (img, &c, x1,y1,z);
					writebuf (&temp, &c, x,y,0);
				}
			copybuf (&temp, img, 0,0,0, 0,0,z, img->xmax * img->ymax);
			break;
		}
		case B_SHORT:
		{
			short s;
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
				{
					x1 = x+dx; y1 = y+dy; s = 0;
					if ((x1>=0) && (y1>=0) && (x1<img->xmax) && (y1<img->ymax)) readbuf (img, &s, x1,y1,z);
					writebuf (&temp, &s, x,y,0);
				}
			copybuf (&temp, img, 0,0,0, 0,0,z, img->xmax * img->ymax);
			break;
		}
		case B_FLOAT:
		{
			float f;
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
				{
					x1 = x+dx; y1 = y+dy; f = 0.0;
					if ((x1>=0) && (y1>=0) && (x1<img->xmax) && (y1<img->ymax)) readbuf (img, &f, x1,y1,z);
					writebuf (&temp, &f, x,y,0);
				}
			copybuf (&temp, img, 0,0,0, 0,0,z, img->xmax * img->ymax);
			break;
		}
		case B_RGB:
		{
			rgbtriplet rgb;
			for (y=0; y<img->ymax; y++)
				for (x=0; x<img->xmax; x++)
				{
					x1 = x+dx; y1 = y+dy; rgb.red=0; rgb.green=0; rgb.blue=0;
					if ((x1>=0) && (y1>=0) && (x1<img->xmax) && (y1<img->ymax)) readbuf (img, &rgb, x1,y1,z);
					writebuf (&temp, &rgb, x,y,0);
				}
			copybuf (&temp, img, 0,0,0, 0,0,z, img->xmax * img->ymax);
			break;
		}
	}

	freebuf (&temp);

}







/****************************************************************************************************/







