


/*************************
	DICOM import module.

	NOTE: This module is based on 
	libdicom by Tony Voet.
	All restrictions and license requirements
	for libdicom apply.

 *************************/


/*		CRYSTAL IMAGE 0.9.0 BETA

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include "dicom.h"

char dicom_version[]="libdicom 0.31",**dicom_transfer_syntax=0L;

static void	dicom_transfer(void);
static void	dicom_vr(void);
static void	dicom_encapsulated(int);
static void	dicom_sequence(int);
static void	dicom_endian(void);
static int	dicom_vm(void);
static void	dicom_swap(void *,U8);
static int	dicom_check(int);

static void	mdc_dicom_endian(void);
	
static ELEMENT	element;
static FILE	*stream=0L;
static long	position;
static int	meta;
static enum
{
  LITTLE=1,
  BIG=2,
  IMPLICIT=4,
  EXPLICIT=8
}
syntax,endian,filesyntax,pixelsyntax;

#if MEDCON_INTEGRATED
/* eNlf: routine for setting the stream from outside the library    */
/* eNlf: in MedCon this library doesn't have to open or close steam */         
/********
 * init *
 ********/
void dicom_init(FILE *fp, char *report)
{
 	stream = fp;					/* Give file handle here to log to a file */
	dicom_log_string = report;		/* or give buffer pointer here to log to a file */
}									/* or give NULL pointers to log to stderr */
#endif

/********
 * open *
 ********/

int dicom_open(const char *file)
{
  U16	magic=0x1234;
  char  vr[2];
#if MEDCON_INTEGRATED
  char  buffer[512];
#else
  char	*dot,*tmp,buffer[512];
#endif 

  dicom_log(DEBUG,"dicom_open()");

#if !MEDCON_INTEGRATED
  if (!file)
  {
    dicom_log(ERROR,"No file given");
    return -1;
  }

  dot=strrchr(file,'.');
  if (dot)
    if (!strcmp(dot,".gz") || !strcmp(dot,".Z"))
    {
      tmp=tmpnam(0L);

      sprintf(buffer,"gzip -cd %.435s > %.64s",file,tmp);
      if (system(buffer))
      {
        dicom_log(ERROR,"Unable to uncompress file");
        unlink(tmp);
        return -2;
      }

      stream=fopen(tmp,"rb");
      unlink(tmp);

      if (!stream)
      {
        dicom_log(ERROR,"Unable to open temporary file");
        return -3;
      }
    }

  if (!stream)
  {
    stream=fopen(file,"rb");
    if (!stream)
    {
      dicom_log(ERROR,"Unable to open file");
      return -4;
    }
  }
#else
  if (!stream)
  {
    dicom_log(ERROR,"Bad null stream");
    return -4;
  }
#endif

	/* DICOM starts with 128 bytes of zeroes followed by the string DICM 
		so read 128+4 = 132 bytes and check for DICM */


	fread(buffer,1,132,stream);
	if (dicom_check(0))
		return -5;
	
	if (!strncmp(buffer+128,"DICM",4))
	{
		buffer[128]=0;
		dicom_log(INFO,"Dicom preamble");
	
		meta=-1;
		syntax=LITTLE|EXPLICIT;
	
		/* watch out for LITTLE|IMPLICIT */
		fread(&element.group,2,2,stream);
		dicom_swap(&element.group,2);
		dicom_swap(&element.element,2);
		fread(vr,1,2,stream);
		element.vr=(*vr<<8)|vr[1];
	
		if (element.vr != UL) 
		{
			syntax=LITTLE|IMPLICIT;  /* weird */
			dicom_log(INFO,"Syntax: Little:implicit");
		}
		else
			dicom_log(INFO,"Syntax: Little:explicit");

	
		fseek(stream,132,SEEK_SET);
	
	}
	else	/* This is the older ACR/NEMA without preamble */
	{
		rewind(stream);
		
		meta=0;
		dicom_log(INFO,"No Dicom preamble");
		
		if (*buffer) 
		{
			if (buffer[5]) 
			{
				syntax=LITTLE|EXPLICIT;
				dicom_log(INFO,"Syntax: Little:explicit");
			}
			else
			{
				syntax=LITTLE|IMPLICIT;
				dicom_log(INFO,"Syntax: Little:implicit");
			}
		}
		else
		{
			if (buffer[4]) 
			{
				syntax=BIG|EXPLICIT;
				dicom_log(INFO,"Syntax: Big:explicit");
			}
			else
			{
				syntax=BIG|IMPLICIT;
				dicom_log(INFO,"Syntax: Big:implicit");
			}
		}
	}

	filesyntax=syntax; pixelsyntax=syntax;
	
	if ( *((U8*)&magic)==0x12 )
		endian=BIG;
	else
		endian=LITTLE;
	
	dicom_encapsulated(-1);
	dicom_sequence(-1);
	
	return 0;
}

/***********
 * element *
 ***********/

ELEMENT *dicom_element(void)
{
  long	rewind;
  U16	tmp;
  char	vr[2];
  char buf[256];

	dicom_log(DEBUG,"dicom_element()");
	
	if (!stream)
		return 0L;
	
	position=ftell(stream);
	
	fread(&element.group,2,2,stream);
	if (dicom_check(-1))
		return 0L;
	dicom_swap(&element.group,2);
	dicom_swap(&element.element,2);

	if (meta)
		if (element.group>=0x0008)
		{
			meta=0;
			dicom_transfer();
		
			fseek(stream,position,SEEK_SET);  
			return dicom_element();
		}

	if (syntax & IMPLICIT || element.group==0xFFFE)
	{
		dicom_vr();		/* Implicit requires that we use our own dictionary */
	
		fread(&element.length,4,1,stream);
		dicom_swap(&element.length,4);
	}
	else
	{
		fread(vr,1,2,stream);			/* Explicit - get VR from stream */
		element.vr=(*vr<<8)|vr[1];
	
		switch(element.vr)
		{
			case OB :
			case OW :
			case SQ :
			case UN :
			case UT :
				fseek(stream,2,SEEK_CUR);
			
				fread(&element.length,4,1,stream);
				dicom_swap(&element.length,4);
				break;
		
			default :
				fread(&tmp,2,1,stream);
				dicom_swap(&tmp,2);
				element.length=tmp;
		}
	}

	sprintf (buf, "Dicom element at filepos %08x is %04x:%04x and VR=%c%c", position, (int)element.group, (int)element.element, element.vr>>8, element.vr & 0xff);
	dicom_log (INFO, buf);

	if (dicom_check(0))
		return 0L;

	dicom_encapsulated(0);
	dicom_sequence(0);
	
	if (element.group==0x0002)
		if (element.element==0x0010)
		{
			rewind=ftell(stream);
			if (dicom_load(UI))
				return 0L;
			fseek(stream,rewind,SEEK_SET);
			dicom_transfer_syntax=element.value.UI;
		}
	
	return &element;
}



/********
 * skip *
 ********/

int dicom_skip(void)
{
	dicom_log(DEBUG,"dicom_skip()");
	
	if (!stream)
	{
		dicom_log(WARNING,"Stream closed - attempt to skip");
		return -1;
	}
	
	if (element.vr==SQ || element.length==0xFFFFFFFF)
		return 0;
	
	if (element.group==0xFFFE)
		if (!element.encapsulated)
			return 0;
	
	fseek(stream,element.length,SEEK_CUR);
	
	return dicom_check(0);
}


/********
 * load *
 ********/

int dicom_load(VR vr)
{
  dicom_log(DEBUG,"dicom_load()");

  if (!stream)
  {
    dicom_log(WARNING,"Stream closed - attempt to load");
    return -1;
  }

  if (element.vr==UN)
    element.vr=vr;

  if (element.vr==SQ || element.length==0xFFFFFFFF)
    return 0;

  if (element.group==0xFFFE)
    if (!element.encapsulated)
      return 0;

  if (!element.length)
    element.value.UN=0L;
  else
  {
    /* eNlf: - allocate an extra 4 bytes, otherwise the bit.c   */
    /* eNlf: routines like source.u++ go beyond the boundaries  */
    /* eNlf: - memset the allocated buffer for sure             */
    element.value.UN=malloc(element.length + 4);
    if (!element.value.UN)
    {
      dicom_log(ERROR,"Out of memory");
      dicom_close();
      return -2;
    }
    memset(element.value.UN,0,element.length + 4);
    fread(element.value.UN,1,element.length,stream);
    if (dicom_check(0))
    {
      eNlfSafeFree(element.value.UN);
      return -3;
    }

    mdc_dicom_endian();

  }

  return dicom_vm();
}
/**************
 * MDC endian *
 **************/
/* 
 * fix endian, take care of special pixel syntax
 */
void mdc_dicom_endian(void)
{
	if ((element.group==0x7FE0) && (element.element == 0x0010)) 
	{
		syntax=pixelsyntax;
		dicom_endian();
		syntax=filesyntax;
	}
	else
	{
		dicom_endian();
	}
}

/************
 * MDC load *
 ************/
/* eNlf: BEGIN -- changes for integration in MedCon */
/*
   Routine for MedCon, at the end the tags are not handled by 
   dicom_vm() so we can pass the tag through our MdcDoTag()
   routine and get the header info we need */
int mdc_dicom_load(VR vr)
{
  dicom_log(DEBUG,"dicom_load()");

  if (!stream)
  {
    dicom_log(WARNING,"Stream closed - attempt to load");
    return -1;
  }

  if (element.vr==UN)
    element.vr=vr;

  if (element.vr==SQ || element.length==0xFFFFFFFF)
    return 0;

  if (element.group==0xFFFE)
    if (!element.encapsulated)
      return 0;

  if (!element.length)
    element.value.UN=0L;
  else
  {
    /* eNlf: allocate an extra 4 bytes - see also dicom_load() */
    element.value.UN=malloc(element.length + 4);
    if (!element.value.UN)
    {
      dicom_log(ERROR,"Out of memory");
      dicom_close();
      return -2;
    }
    memset(element.value.UN,0,element.length + 4);
    fread(element.value.UN,1,element.length,stream);
    if (dicom_check(0))
    {
      eNlfSafeFree(element.value.UN);
      return -3;
    }

    mdc_dicom_endian();

  }

  return 0;
}
/* eNlf: END   -- changes for integration in MedCon */

/*********
 * clean *
 *********/

void dicom_clean(void)
{
  U32	i;
  char	*c;

  dicom_log(DEBUG,"dicom_clean()");

  switch(element.vr)
  {
  case PN :
    for (i=0; i<element.vm; i++)
      for (c=element.value.PN[i]; *c; c++)
        if (*c=='^')
          *c=' ';

  case AE :
  case AS :
  case CS :
  case DA :
  case DS :
  case DT :
  case IS :
  case LO :
  case SH :
  case TM :
  case UI :
    for (i=0; i<element.vm; i++)
    {
      for (c=element.value.AE[i]; *c; c++)
        if (*c==' ' || *c=='\t')
          element.value.AE[i]++;
        else
          break;

      for (; *c; c++);
      c--;

      for (; c>=element.value.AE[i]; c--)
        if (*c==' ' || *c=='\t')
          *c=0;
        else
          break;
    }
    break;

  default:
    break;
  }
}

/*********
 * close *
 *********/

int dicom_close(void)
{
  dicom_log(DEBUG,"dicom_close()");

  if (!stream)
    return 0;

  eNlfSafeFree(dicom_transfer_syntax);
  dicom_transfer_syntax=0L;

#if ! MEDCON_INTEGRATED 

  if (fclose(stream))
  {
    dicom_log(WARNING,"Unable to close file");
    stream=0L;
    return -1;
  }

  stream=0L;

#else
  fseek(stream,0,SEEK_SET);
#endif


  return 0;
}

/************
 * transfer *
 ************/

static void dicom_transfer(void)
{
char buf[256];

	dicom_log(DEBUG,"dicom_transfer()");
	
	if (!dicom_transfer_syntax)
	{
		dicom_log(WARNING,"No transfer syntax found");
		return;
	}

	sprintf (buf, "Dicom transfer syntax %s", *dicom_transfer_syntax);
	dicom_log (INFO, buf);
	
	if (strncmp(*dicom_transfer_syntax,"1.2.840.113619.5.2",18) == 0)
	{
		syntax=LITTLE|IMPLICIT;
		filesyntax=syntax; pixelsyntax=BIG|IMPLICIT;
		dicom_log (INFO, "...Transfer syntax set to Little:implicit, file syntax Big:implicit");
		return;
	}
	if (strncmp(*dicom_transfer_syntax,"1.2.840.10008.1.2",17))
	{
		dicom_log(WARNING,"Transfer syntax is not DICOM");
		return;
	}
	
	if ((*dicom_transfer_syntax)[17]!='.') 
	{
		syntax=LITTLE|IMPLICIT;
		filesyntax=syntax; pixelsyntax=syntax;
	}
	else
	{
		switch((*dicom_transfer_syntax)[18])
		{
			case '1' :
			case '4' :
				break;
			case '2' :
				syntax=BIG|EXPLICIT;
				filesyntax=syntax; pixelsyntax=syntax;
				dicom_log (INFO, "...Transfer syntax set to Big:implicit");
				break;
			default :
				dicom_log(WARNING,"Unknown transfer syntax");
				dicom_log(WARNING,*dicom_transfer_syntax);
		}
	}
	


}

/******
 * vr *
 ******/

static void dicom_vr(void)
{
  static DICTIONARY data[]=
  {
    #include "dictionary.SQ"
  };

  dicom_log(DEBUG,"dicom_vr()");

  element.vr=dicom_private(data,&element)->vr;
}

/****************
 * encapsulated *
 ****************/

static void dicom_encapsulated(int reset)
{
  static int encapsulated;

  dicom_log(DEBUG,"dicom_encapsulated()");

  if (reset)
  {
    encapsulated=0;
    return;
  }

  element.encapsulated=encapsulated;

  if (encapsulated)
    if (element.group==0xFFFE)
      if (element.element==0xE0DD)
        encapsulated=0;

  if (element.length==0xFFFFFFFF)
    if (element.vr!=SQ && element.group!=0xFFFE)
      encapsulated=-1;
}

/************
 * sequence *
 ************/

static void dicom_sequence(int reset)
{
  static U32	length[0x100];
  static U8	sequence;

  dicom_log(DEBUG,"dicom_sequence()");

  if (reset)
  {
    sequence=0;
    return;
  }

  element.sequence=sequence;

  if (sequence)
    if (length[sequence]!=0xFFFFFFFF)
    {
      *length=ftell(stream)-position;
      if (element.length!=0xFFFFFFFF)
        if (element.group!=0xFFFE || element.element!=0xE000)
          *length+=element.length;

      if (*length>length[sequence])
      {
        dicom_log(WARNING,"Incorrect sequence length");
        sequence--;
      }
      else
        length[sequence]-=*length;

      if (!length[sequence])
        sequence--;
    }

  if (element.vr==SQ) {
    if (sequence!=0xFF)
    {
      sequence++;
      length[sequence]=element.length;
    }
    else
      dicom_log(WARNING,"Deep sequence hierarchy");
  }

  if (element.group==0xFFFE)
    if (element.element==0xE0DD) {
      if (!element.encapsulated) {
        if (sequence)
          sequence--;
        else
          dicom_log(WARNING,"Incorrect sequence delimiter");
      }
    }
}

/**********
 * endian *
 **********/

static void dicom_endian(void)
{
  U32	i;
  U8	*s;

  dicom_log(DEBUG,"dicom_endian()");

  if (syntax & endian)
    return;

  switch(element.vr)
  {
  case AT :
  case OW :
  case SS :
  case US :
    s=element.value.UN;
    for (i=element.length>>1; i; i--,s+=2)
      dicom_swap(s,2);
    return;

  case SL :
  case UL :
  case FL :
    s=element.value.UN;
    for (i=element.length>>2; i; i--,s+=4)
      dicom_swap(s,4);
    return;

  case FD :
    s=element.value.UN;
    for (i=element.length>>3; i; i--,s+=8)
      dicom_swap(s,8);
    return;

  default:
    return;
  }
}

/******
 * vm *
 ******/

static int dicom_vm(void)
{
  U32	i;
  char	*c,**table,*s,*d;

  dicom_log(DEBUG,"dicom_vm()");

  switch(element.length)
  {
  case 0 :
    element.vm=0;
    return 0;

  case 0xFFFFFFFF :
    element.vm=1;
    return 0;
  }

  switch(element.vr)
  {
    case LT :
    case OB :
    case OW :
    case SQ :
    case ST :
    case UT :
    default :
      element.vm=1;
      return 0;

    case SS :
    case US :
      element.vm=element.length>>1;
      return 0;

    case AT :
    case FL :
    case SL :
    case UL :
      element.vm=element.length>>2;
      return 0;

    case FD :
      element.vm=element.length>>3;
      return 0;

    case AE :
    case AS :
    case CS :
    case DA :
    case DS :
    case DT :
    case IS :
    case LO :
    case PN :
    case SH :
    case TM :
    case UI :
      element.vm=1;
      c=element.value.UN;
      for (i=element.length; i; i--,c++)
        if (*c=='\\')
          element.vm++;

      element.value.UN=realloc(element.value.UN,element.vm*sizeof(char*)
      +element.length+1);
      if (!element.value.UN)
      {
	dicom_log(ERROR,"Out of memory");
	dicom_close();
	return -1;
      }

      c=element.value.LT+element.vm*sizeof(char*);

      s=element.value.LT+element.length;
      d=c+element.length;
      for (i=element.length; i; i--)
        *--d=*--s;

      table=element.value.AE;
      *table++=c;

      for (i=element.length; i; i--,c++)
        if (*c=='\\')
        {
          *c=0;
          *table++=c+1;
        }

      *c=0;

      if (!(element.length&1))
        if (*--c==' ')
          *c=0;

      return 0;
  }
}

/********
 * swap *
 ********/

static void dicom_swap(void *v,U8 n)
{
  U8 i,*b,*e,tmp;

  if (syntax & endian)
    return;

  b=v;
  e=b+n-1;

  for (i=n>>1; i; i--)
  {
    tmp=*b;
    *b++=*e;
    *e--=tmp;
  }
}

/*********
 * check *
 *********/

static int dicom_check(int expected)
{
  if (ferror(stream))
  {
    dicom_log(ERROR,"Error while reading file");
    dicom_close();
    return -1;
  }

  if (feof(stream))
  {
    if (!expected)
      dicom_log(ERROR,"Unexpected end of file");
    dicom_close();
    return -2;
  }

  return 0;
}



/*************************
 * libdicom by Tony Voet - bit.c *
 *************************/

union
{
  U32	*u32;
  U16	*u16;
  U8	*u8;
}
source;

U32	cache32;
U16	cache16;
U8	cache8;

int left;

/*******
 * bit *
 *******/

void dicom_bit(void *data)
{
  dicom_log(DEBUG,"dicom_bit()");

  source.u32=data;
  left=0;
}

/**********
 * 8 skip *
 **********/

void dicom_8_skip(int bit)
{
  if (!bit)
    return;

  if (bit<left)
  {
    cache8<<=bit;
    left-=bit;
  }
  else
  {
    bit-=left;

    cache8=*source.u8++;
    left=8;

    if (bit)
      dicom_8_skip(bit);
  }
}

/***********
 * 16 skip *
 ***********/

void dicom_16_skip(int bit)
{
  if (!bit)
    return;

  if (bit<left)
  {
    cache16<<=bit;
    left-=bit;
  }
  else
  {
    bit-=left;

    cache16=*source.u16++;
    left=16;

    if (bit)
      dicom_16_skip(bit);
  }
}

/***********
 * 32 skip *
 ***********/

void dicom_32_skip(int bit)
{
  if (!bit)
    return;

  if (bit<left)
  {
    cache32<<=bit;
    left-=bit;
  }
  else
  {
    bit-=left;

    cache32=*source.u32++;
    left=32;

    if (bit)
      dicom_32_skip(bit);
  }
}

/**********
 * 8 read *
 **********/

U32 dicom_8_read(int bit)
{
  U32 result;

  if (!bit)
    return 0;

  if (bit<left)
  {
    result=cache8>>(8-bit);
    cache8<<=bit;
    left-=bit;
  }
  else
  {
    result=cache8>>(8-left);
    bit-=left;

    cache8=*source.u8++;
    left=8;

    if (!bit)
      return result;

    result<<=bit;
    result|=dicom_8_read(bit);
  }

  return result;
}

/***********
 * 16 read *
 ***********/

U32 dicom_16_read(int bit)
{
  U32 result;

  if (!bit)
    return 0;

  if (bit<left)
  {
    result=cache16>>(16-bit);
    cache16<<=bit;
    left-=bit;
  }
  else
  {
    result=cache16>>(16-left);
    bit-=left;

    cache16=*source.u16++;
    left=16;

    if (!bit)
      return result;

    result<<=bit;
    result|=dicom_16_read(bit);
  }

  return result;
}

/***********
 * 32 read *
 ***********/

U32 dicom_32_read(int bit)
{
  U32 result;

  if (!bit)
    return 0;

  if (bit<left)
  {
    result=cache32>>(32-bit);
    cache32<<=bit;
    left-=bit;
  }
  else
  {
    result=cache32>>(32-left);
    bit-=left;

    cache32=*source.u32++;
    left=32;

    if (!bit)
      return result;

    result<<=bit;
    result|=dicom_32_read(bit);
  }

  return result;
}



/*************************
 * libdicom by Tony Voet - dictionary.c *
 *************************/


/*********
 * query *
 *********/

DICTIONARY *dicom_query(ELEMENT *element)
{
  static DICTIONARY data[]=
  {
    #include "dictionary.data"
  };

  dicom_log(DEBUG,"dicom_query()");

  if (!element)
  {
    dicom_log(ERROR,"No element given");
    return 0L;
  }

  return dicom_private(data,element);
}

/***********
 * private *
 ***********/

DICTIONARY *dicom_private(DICTIONARY *data,ELEMENT *e)
{
  static DICTIONARY *d;

  dicom_log(DEBUG,"dicom_private()");

  if (!data)
  {
    dicom_log(ERROR,"No dictionary given");
    return 0L;
  }

  if (!e)
  {
    dicom_log(ERROR,"No element given");
    return 0L;
  }

  for (d=data; d->group!=0xFFFF; d++)
  {
    if (e->group<d->group)
      continue;

    if (e->group>d->group_last)
      continue;

    switch(d->group_match)
    {
    case ANY :
      break;

    case EVEN :
      if (e->group&1)
        continue;
      break;

    case ODD :
      if (!(e->group&1))
        continue;
    }

    if (e->element<d->element)
      continue;

    if (e->element>d->element_last)
      continue;

    switch(d->element_match)
    {
    case ANY :
      break;

    case EVEN :
      if (e->element&1)
        continue;
      break;

    case ODD :
      if (!(e->element&1))
        continue;
    }

    break;
  }

  return d;
}


/*************************
 * libdicom by Tony Voet - image.c *
 *************************/

/* eNlf: BEGIN - comment out unwanted stuff */
/*
#include "jpeglib.h"
*/
/* eNlf: END   - comment out unwanted stuff */

/*******
 * new *
 *******/

IMAGE *dicom_new(int rgb,U16 frames,U16 w,U16 h)
{
  IMAGE *image;

  dicom_log(DEBUG,"dicom_new()");

  image=malloc(sizeof(IMAGE));
  if (!image)
  {
    dicom_log(ERROR,"Out of memory");
    return 0L;
  }

  image->rgb=rgb;
  image->frames=frames;
  image->w=w;
  image->h=h;

  if (rgb)
    image->data.rgb=malloc(frames*w*h*3);
  else
    image->data.gray=malloc(frames*w*h*2);

  if (!image->data.rgb)
  {
    dicom_log(ERROR,"Out of memory");

    eNlfSafeFree(image);
    return 0L;
  }

  return image;
}

/********
 * read *
 ********/

/* The main reader function. Reads a file of given name (see transfer syntax!)
	into image; returns slice count in images */


int dicom_read (const char *file,IMAGE **image,int *images,int parametric)
{
  SINGLE	*single;
  IMAGE		*new, *tmp;

	dicom_log(INFO ,"Dicom_read");

	if (!file)
	{
    	dicom_log(ERROR,"No file given");
    	return -1;
	}

	if (!image || !images)
	{
    	dicom_log(ERROR,"Argument missing");
    	return -2;
	}

	if (dicom_open(file)) return -3;

	for (*image=0L,*images=0;;)
	{
		single=dicom_single();		/* read one single slice */
		if (!single) break;

    	new=dicom_transform(single,parametric);		/* Transform to one of our allowed types */
    	if (new)
    	{
			if (*image) 
				tmp=realloc(*image,(*images+1)*sizeof(IMAGE));
			else
				tmp=malloc(sizeof(IMAGE));

			if (!tmp)
			{
				dicom_log(ERROR,"Error reallocating memory");
				eNlfSafeFree(new->data.rgb);
			}
			else
			{
        		*image=tmp;
        		memcpy(*image+*images,new,sizeof(IMAGE));
        		(*images)++;
			}
		}

		dicom_single_free();
	}

	if (*images==0)
	{
    	dicom_log(ERROR,"No images found");
    	/* eNlf: BEGIN -- changes for integration in MedCon */
    	dicom_close();
    	/* eNlf: END   -- changes for integration in MedCon */
    	return -4;
	}

	return 0;
}



/********
 * free *
 ********/

void dicom_free(IMAGE *image,int images)
{
  int i;

  dicom_log(DEBUG,"dicom_free()");

  if (!image)
    return;

  for (i=0; i<images; i++)
    eNlfSafeFree(image[i].data.rgb);

  eNlfSafeFree(image);
}

/*********
 * write *
 *********/

int dicom_write(const char *file,const IMAGE *image)
{
  dicom_log(DEBUG,"dicom_write()");

  if (!file)
  {
    dicom_log(ERROR,"No file given");
    return -1;
  }

  if (!image)
  {
    dicom_log(ERROR,"No image given");
    return -2;
  }

  dicom_log(EMERGENCY,"DICOM write is not implemented yet");

  return -3;
}
/* eNlf: BEGIN - comment out unwanted stuff */
/***************
 * write ascii *
 ***************/
/*
int dicom_write_ascii(const char *file,const IMAGE *image,int width)
{
  static char gray[]=
    "$@B%8&WM#*oahkbdpqwmZO0QLCJUYXzcvunxrjft/\\|()1{}[]?-_+~<>i!lI;:,\"^`'. ";

  FILE	*stream;
  IMAGE	*zoom;
  U16	frame,x,y,*pixel;

  dicom_log(DEBUG,"dicom_write_ascii()");

  if (!image)
  {
    dicom_log(ERROR,"No image given");
    return -1;
  }

  if (!file)
    stream=stdout;
  else
  {
    stream=fopen(file,"wb");
    if (!stream)
    {
      dicom_log(ERROR,"Unable to create ascii file");
      return -2;
    }
  }

  zoom=dicom_zoom(image,width,width*image->h/(image->w<<1),-1);
  if (!zoom)
    return -3;

  dicom_gray(zoom);
  dicom_max(zoom);

  pixel=zoom->data.gray;

  for (frame=zoom->frames; frame; frame--)
  {
    for (y=zoom->h; y; y--)
    {
      for (x=zoom->w; x; x--,pixel++)
	putc(gray[(int) 69**pixel/0xFFFF],stream);
      puts("");
    }
    puts("");
  }

  dicom_free(zoom,1);

  return 0;
}
*/

/**************
 * write jpeg *
 **************/
/*
int dicom_write_jpeg(const char *file,const IMAGE *image,int quality)
{
  struct jpeg_compress_struct	cinfo;
  struct jpeg_error_mgr		jerr;
  JSAMPROW			line,target;
  FILE				*stream;
  U16				*source,l;

  dicom_log(DEBUG,"dicom_write_jpeg()");

  if (!file)
  {
    dicom_log(ERROR,"No file given");
    return -1;
  }

  if (!image)
  {
    dicom_log(ERROR,"No image given");
    return -2;
  }

  if (!image->rgb)
  {
    line=malloc(image->w*2);
    if (!line)
    {
      dicom_log(ERROR,"Out of memory");
      return -3;
    }
  }

  stream=fopen(file,"wb");
  if (!stream)
  {
    dicom_log(ERROR,"Unable to create jpeg file");
    return -4;
  }

  cinfo.err=jpeg_std_error(&jerr);
  jpeg_create_compress(&cinfo);
  jpeg_stdio_dest(&cinfo,stream);

  cinfo.image_width=image->w;
  cinfo.image_height=image->h*image->frames;

  if (image->rgb)
  {
    cinfo.input_components=3;
    cinfo.in_color_space=JCS_RGB;
  }
  else
  {
    cinfo.input_components=1;
    cinfo.in_color_space=JCS_GRAYSCALE;
  }

  jpeg_set_defaults(&cinfo);
  jpeg_set_quality(&cinfo,quality,-1);
  jpeg_start_compress(&cinfo,-1);

  while (cinfo.next_scanline<cinfo.image_height)
  {
    if (image->rgb)
      line=image->data.rgb+cinfo.next_scanline*image->w*3;
    else
    {
      source=image->data.gray+cinfo.next_scanline*image->w;
      target=line;

      for (l=image->w; l; l--)
        *target++=*source++>>8;
    }

    jpeg_write_scanlines(&cinfo,&line,1);
  }

  if (!image->rgb)
    eNlfSafeFree(line);

  jpeg_finish_compress(&cinfo);
  fclose(stream);
  jpeg_destroy_compress(&cinfo);

  return 0;
}
*/
/*************
 * write eps *
 *************/
/*
int dicom_write_eps(const char *file,const IMAGE *image)
{
  dicom_log(DEBUG,"dicom_write_eps()");

  if (!file)
  {
    dicom_log(ERROR,"No file given");
    return -1;
  }

  if (!image)
  {
    dicom_log(ERROR,"No image given");
    return -2;
  }

  dicom_log(EMERGENCY,"DICOM write EPS is not implemented yet");

  return -3;
}
*/
/*eNlf: END  - comment out unwanted stuff */


/*************************
 * libdicom by Tony Voet - log.c *
 *************************/



/* eNlf: BEGIN -- change for compilation error on Red Hat 6.0 */
/* static FILE     *stream=stderr;                            */
/* The above statement fails: initializer not constant        */
/* eNlf: END   -- change for compilation error on Red Hat 6.0 */
static char	*program=NULL;

/************
 * log name *
 ************/

void dicom_log_name(char *name)
{
  program=strrchr(name,'/');

  if (program)
    program++;
  else
    program=name;
}

/************
 * log open *
 ************/

int dicom_log_open(const char *file)
{
  if (!file)
  {
    dicom_log(ERROR,"No file given");
    return -1;
  }

  stream=fopen(file,"a");

  if (!stream)
  {
    stream=stderr;
    dicom_log(ERROR,"Unable to open log file");
    return -1;
  }

  return 0;
}

/*******
 * log *
 *******/

void dicom_log(CONDITION condition,const char *message)
{
  time_t	t;
  char		tmp[2048];

  static char *explination[]=
  {
    "emergency",
    "alert",
    "critical",
    "error",
    "warning",
    "notice",
    "info",
    "debug"
  };

	if (condition>dicom_log_level)
      return;

	time(&t);
	strftime(tmp,32,"%b %d %T",localtime(&t));


	if ((stream == NULL) && (dicom_log_string==NULL))
	{
		fprintf(stderr,"%s %s[%u]: %s: %s\n",
    				  tmp,
    				  program ? program : "log",
    				  (unsigned int) getpid(),
    				  explination[condition],
    				  message);
	}
	else if (dicom_log_string!=NULL)
	{
		sprintf(tmp,"%s: %s\n",
    				  explination[condition],
    				  message);
		if (strlen(dicom_log_string)<16000) strcat (dicom_log_string, tmp);
	}
	else
	{
		fprintf(stream,"%s %s[%u]: %s: %s\n",
    				  tmp,
    				  program ? program : "log",
    				  (unsigned int) getpid(),
    				  explination[condition],
    				  message);
	}

}

/*************
 * log close *
 *************/

int dicom_log_close(void)
{
  if (stream==stderr)
  {
    dicom_log(NOTICE,"Attempt to close stderr");
    return -1;
  }

  if (fclose(stream))
  {
    stream=stderr;
    dicom_log(WARNING,"Unable to close log");
    return -2;
  }

  stream=stderr;

  return 0;
}

/*************************
 * libdicom by Tony Voet - process.c *
 *************************/


/*******
 * max *
 *******/

void dicom_max(IMAGE *image)
{
  U32	length,l;
  U16	min,max,*pixel;

  dicom_log(DEBUG,"dicom_max()");

  if (!image)
  {
    dicom_log(WARNING,"No image given");
    return;
  }

  if (image->rgb)
  {
    dicom_log(WARNING,"Color image");
    return;
  }

  length=image->frames*image->w*image->h;
  pixel=image->data.gray;

  min=*pixel;
  max=min;

  for (l=length; l; l--,pixel++)
  {
    if (*pixel<min)
    {
      min=*pixel;
      continue;
    }
    if (*pixel>max)
      max=*pixel;
  }

  if (min==max)
    return;

  if (min==0)
    if (max==0xFFFFU)
      return;

  pixel=image->data.gray;

  for (l=length; l; l--, pixel++)
    *pixel=0xFFFFUL*(*pixel-min)/(max-min);
}

/**********
 * invert *
 **********/

void dicom_invert(IMAGE *image)
{
  U32	l;
  U16	*pixel;

  dicom_log(DEBUG,"dicom_invert()");

  if (!image)
  {
    dicom_log(WARNING,"No image given");
    return;
  }

  if (image->rgb)
  {
    dicom_log(WARNING,"Color image");
    return;
  }

  pixel=image->data.gray;

  for (l=image->frames*image->w*image->h; l; l--, pixel++)
    *pixel=0xFFFFU-*pixel;
}

/*******
 * voi *
 *******/

void dicom_voi(IMAGE *image,U16 min,U16 max)
{
  U32	l;
  U16	*pixel;

  dicom_log(DEBUG,"dicom_voi()");

  if (min==0)
    if (max==0xFFFFU)
      return;

  if (!image)
  {
    dicom_log(WARNING,"No image given");
    return;
  }

  if (image->rgb)
  {
    dicom_log(WARNING,"Color image");
    return;
  }

  pixel=image->data.gray;

  for (l=image->frames*image->w*image->h; l; l--,pixel++)
  {
    if (*pixel<=min)
    {
      *pixel=0;
      continue;
    }

    if (*pixel>=max)
    {
      *pixel=0xFFFFU;
      continue;
    }

    *pixel=0xFFFFUL*(*pixel-min)/(max-min);
  }
}

/********
 * gray *
 ********/

void dicom_gray(IMAGE *image)
{
  U32	length,l;
  U16	*target;
  U8	*source;

  dicom_log(DEBUG,"dicom_gray()");

  if (!image)
  {
    dicom_log(WARNING,"No image given");
    return;
  }

  if (!image->rgb)
    return;

  length=image->frames*image->w*image->h;

  source=image->data.rgb;
  target=image->data.gray;

  for (l=length; l; l--,source+=3)
    *target++=77UL*source[0]+151UL*source[1]+29UL*source[2];

  image->rgb=0;

  target=realloc(image->data.gray,2*length);
  if (!target)
    dicom_log(WARNING,"Error reallocating memory");
  else
    image->data.gray=target;

  dicom_max(image);
}

/*******
 * hsv *
 *******/

void dicom_hsv(U16 h,U16 s,U16 v,U8 *rgb)
{
  float	hue,saturation,f;
  int	i;
  U8	value,m,n;

  hue=h*6.0/65536.0;
  saturation=s/65535.0;
  value=v>>8;

  i=hue;

  f=hue-i;
  if (!(i&1))
    f=1.0-f;

  m=value*(1.0-saturation);
  n=value*(1.0-saturation*f);

  switch(i)
  {
  case 0 :
    rgb[0]=value;
    rgb[1]=n;
    rgb[2]=m;
    break;

  case 1 :
    rgb[0]=n;
    rgb[1]=value;
    rgb[2]=m;
    break;

  case 2 :
    rgb[0]=m;
    rgb[1]=value;
    rgb[2]=n;
    break;

  case 3 :
    rgb[0]=m;
    rgb[1]=n;
    rgb[2]=value;
    break;

  case 4 :
    rgb[0]=n;
    rgb[1]=m;
    rgb[2]=value;
    break;

  case 5 :
    rgb[0]=value;
    rgb[1]=m;
    rgb[2]=n;
  }
}

/*********
 * merge *
 *********/

IMAGE *dicom_merge(const IMAGE *anatomic,const IMAGE *parametric,U16 saturation)
{
  IMAGE	*zoom,*merge;
  U16	bar,*value,*hue,frame,x,y;
  U8	*target;

  dicom_log(DEBUG,"dicom_merge()");

  if (!anatomic || !parametric)
  {
    dicom_log(ERROR,"Image missing");
    return 0L;
  }

  if (anatomic->rgb || parametric->rgb)
  {
    dicom_log(ERROR,"Wrong image type");
    return 0L;
  }

  if (anatomic->frames!=parametric->frames)
  {
    dicom_log(ERROR,"Wrong number of frames");
    return 0L;
  }

  zoom=dicom_zoom(parametric,anatomic->w,anatomic->h,-1);
  if (!zoom)
    return 0L;

  bar=anatomic->w>>5;

  merge=dicom_new(-1,anatomic->frames,anatomic->w+(bar<<1),anatomic->h);
  if (!merge)
  {
    dicom_free(zoom,1);
    return 0L;
  }

  value=anatomic->data.gray;
  hue=zoom->data.gray;
  target=merge->data.rgb;

  for (frame=anatomic->frames; frame; frame--)
    for (y=0; y<anatomic->h; y++)
    {
      for (x=anatomic->w; x; x--)
      {
	dicom_hsv(2UL*(0xFFFFU-*hue)/3U,*hue?saturation:0,*value,target);

	value++;
	hue++;
	target+=3;
      }

      for (x=3*bar; x; x--)
	*target++=0;

      for (x=bar; x; x--)
      {
	dicom_hsv(0xAAAAUL*y/(anatomic->h-1),saturation,0xFFFFU,target);
	target+=3;
      }
    }

  dicom_free(zoom,1);

  return merge;
}


/*************************
 * libdicom by Tony Voet - single.c *
 *************************/


static S32 dicom_pixel(const ELEMENT *);

static SINGLE single;

/**********
 * single *
 **********/

/* SINGLE reads one single image (tile?) form the DICOM file
	and reads all group-28 data (image format, color interpretation etc) */


SINGLE *dicom_single(void)
{
  char buf[256];
  ELEMENT	*e;
  S32		length;
  U32		i;
  char		*interpretation[]=
  {
		"MONOCHROME2",
		"MONOCHROME1",
		"PALETTE COLOR",
		"RGB",
		"HSV",
		"ARGB",
		"CMYK",
		"UNKNOWN"
  };
  char		*colorname [3] = { "Red", "Green", "Blue" };

	dicom_log (INFO,"Reading single DICOM slice");
	
	memset(&single,0,sizeof(SINGLE));
	single.frames=1;
	single.samples=1;

	for (;;)
	{
		e=dicom_element();				/* Get all elements from DICOM file (again) */
		if (!e)
			break;
	
		if (e->group==0x0028)			/* Group 0x28: Format description */
		{
			if (e->element==0x0002)		/* Element 0x02: Samples per pixel */
			{
				if (dicom_load(US))
					break;
				single.samples=*e->value.US;
				eNlfSafeFree(e->value.US);
				sprintf (buf, "   0028:0002: %d samples per pixel", single.samples);
				dicom_log (INFO, buf);
				continue;
			}
		
			if (e->element==0x0004)		/* Element 0x04: Photometric interpretation */
			{
				if (dicom_load(CS))
					break;
				dicom_clean();
		
				for (single.photometric=0; single.photometric<UNKNOWN;
				single.photometric++)
				if ( !strncmp(*e->value.CS,interpretation[single.photometric],
					strlen(interpretation[single.photometric])) )
					break;
		
				sprintf (buf, "   0028:0004: Photometric interpretation %s", interpretation[single.photometric]);
				dicom_log (INFO, buf);
				if (single.photometric==UNKNOWN)
				dicom_log(WARNING,"Unknown Photometric Interpretation");
		
				eNlfSafeFree(e->value.CS);
				continue;
			}

			if (e->element==0x0006)		/* Element 0x06: Planar configuration */
			{
				if (dicom_load(US))
					break;
				single.planar=*e->value.US;
				eNlfSafeFree(e->value.US);
				continue;
			}

			if (e->element==0x0008)		/* Element 0x08: Number of frames */
			{
				if (dicom_load(IS))
					break;
				dicom_clean();
				single.frames=atoi(*e->value.IS);
				eNlfSafeFree(e->value.IS);
				sprintf (buf, "   0028:0008: %d frames", single.frames);
				dicom_log (INFO, buf);
				continue;
			}

			if (e->element==0x0010)		/* Element 0x10: Rows (y size or height) */
			{
				if (dicom_load(US))
					break;
				single.h=*e->value.US;
				eNlfSafeFree(e->value.US);
				sprintf (buf, "   0028:0010: %d rows", single.h);
				dicom_log (INFO, buf);
				continue;
			}

			if (e->element==0x0011)		/* Element 0x11: Columns (x size or width) */
			{
				if (dicom_load(US))
					break;
				single.w=*e->value.US;
				eNlfSafeFree(e->value.US);
				sprintf (buf, "   0028:0010: %d columns", single.w);
				dicom_log (INFO, buf);
				continue;
			}

			if (e->element==0x0100)		/* Element 0x0100: Bits allocated */
			{				sprintf (buf, "   0028:0010: %d rows", single.h);
				dicom_log (INFO, buf);

				if (dicom_load(US))
					break;
				single.alloc=*e->value.US;
				eNlfSafeFree(e->value.US);
				sprintf (buf, "   0028:0100: %d bits allocated", single.alloc);
				dicom_log (INFO, buf);
				continue;
			}

			if (e->element==0x0101)		/* Element 0x0101: Bits stored */
			{
				if (dicom_load(US))
					break;
				single.bit=*e->value.US;
				eNlfSafeFree(e->value.US);
				sprintf (buf, "   0028:0101: %d bits stored", single.bit);
				dicom_log (INFO, buf);
				continue;
			}

			if (e->element==0x0102)		/* Element 0x0102: High bit starting with 0; determines bit order? */
			{
				if (dicom_load(US))
					break;
				single.high=*e->value.US;
				eNlfSafeFree(e->value.US);
				sprintf (buf, "   0028:0102: High bit %d", single.high);
				dicom_log (INFO, buf);
				continue;
			}

			if (e->element==0x0103)		/* Element 0x0103: Pixel representation */
			{
				if (dicom_load(US))
					break;
				single.sign=*e->value.US;
				eNlfSafeFree(e->value.US);
				sprintf (buf, "   0028:0103: Pixel representation %d", single.sign);
				dicom_log (INFO, buf);
				continue;
			}

			if (0x1101<=e->element && e->element<=0x1103)	/* Elements 0x1101 through 0x1103  */
			{												/* RedPaletteColorLookupTableDescriptor */
				if (dicom_load(US))							/* GreenPaletteColorLookupTableDescriptor */
					break;									/* BluePaletteColorLookupTableDescriptor */

				if (e->vm!=3)
					dicom_log(WARNING,"Wrong VM for PaletteColorLookupTableDescriptor");
				else
				{
					i=e->element-0x1101;
					single.clut[i].size=e->value.US[0];
					single.clut[i].threshold.u16=e->value.US[1];
					single.clut[i].bit=e->value.US[2];
					sprintf (buf, "   0028:110x: %s Palette color lookup descriptor size %d, thresh %d, bit %d",
										colorname[i], single.clut[i].size, single.clut[i].threshold.u16, single.clut[i].bit);
					dicom_log (INFO, buf);
				}
			
				eNlfSafeFree(e->value.US);
        		continue;
  			}

			if (0x1201<=e->element && e->element<=0x1203)	/* Palette Data */
			{
				if (dicom_load(US))
					break;
				single.clut[e->element-0x1201].data.u16=e->value.US;
				sprintf (buf, "   0028:120x: Palette data %d", (int)e->value.US);
				dicom_log (INFO, buf);
				continue;
			}
		}				/* end e->group==0x0028) */

		if (!(e->group&1))
			if (0x7F00<=e->group && e->group<0x7FFF)
				if (e->element==0x0010)
				{
					/* first fix bad VR values (OB instead of OW) */
					/* necessary for pixel image data endian swap */
					if (e->vr==OB && single.alloc==16) 
					{
						dicom_log(WARNING,"Incorrect OB value representation (fixed)");
						e->vr=OW;
					}

					length=dicom_pixel(e);
					if (length<0)
						break;
				
					if (length!=single.frames*single.w*single.h*single.samples*single.alloc>>3)
						dicom_log(WARNING,"Incorrect PixelData length");
				
					return &single;
				}
				
				if (dicom_skip())
				break;
	}						/* End outer for;; loop */

	dicom_single_free();

	return 0L;
}

/***************
 * single free *
 ***************/

void dicom_single_free(void)
{
int i;
	
	dicom_log(DEBUG,"dicom_single_free()");
	
	
	/* KLUDGE! We get a crash here, so we disable freeeing the CLUTs.
		Obviously, we get a memory leak. FIXME! */
	
	/* for (i=0; i<3; i++)
		eNlfSafeFree(single.clut[i].data.u16); */
	
	eNlfSafeFree(single.data);
	
	memset(&single,0,sizeof(SINGLE));
}

/*********
 * pixel *
 *********/

static S32 dicom_pixel(const ELEMENT *e)
{
  int error;

	dicom_log(DEBUG,"dicom_pixel()");
	
	if (e->length!=0xFFFFFFFF)
	{
		if (single.alloc==16)
			error=dicom_load(OW);
		else
			error=dicom_load(OB);
	
		if (error)
			return -1;
	
		single.data=e->value.OW;
	
		return e->length;
	}
	
	if (dicom_skip())
		return -2;
	
	dicom_log(EMERGENCY,"Encapsulated PixelData is not implemented yet");
	
	return -3;
}



/*************************
 * libdicom by Tony Voet - transform.c *
 *************************/


static U16 dicom_clut(const CLUT *,U16);

/*********
 * alloc *
 *********/

int dicom_alloc(SINGLE *single)
{
  U32	length,l;
  U16	magic=0x1234,*data,*d;
  int	high,bit,low;

  dicom_log(DEBUG,"dicom_alloc()");

  if (!single)
  {
    dicom_log(ERROR,"No image given");
    return -1;
  }

  if (single->alloc>16)
    dicom_log(WARNING,"Large BitsAllocated");

  length=single->frames*single->w*single->h*single->samples;

  data=malloc(length*2);
  if (!data)
  {
    dicom_log(ERROR,"Out of memory");
    return -2;
  }

  high=single->alloc-single->high-1;
  bit=single->bit;
  low=single->high+1-bit;

  d=data;

  dicom_bit(single->data);

  if ( *((U8*)&magic)==0x12 )
    for (l=length; l; l--)
    {
      dicom_32_skip(high);
      *d++=dicom_32_read(bit);
      dicom_32_skip(low);
    }
  else
    if (single->alloc==16)
      for (l=length; l; l--)
      {
	dicom_16_skip(high);
	*d++=dicom_16_read(bit);
	dicom_16_skip(low);
      }
    else
      for (l=length; l; l--)
      {
	dicom_8_skip(high);
	*d++=dicom_8_read(bit);
	dicom_8_skip(low);
      }

  eNlfSafeFree(single->data);
  single->data=data;

  single->alloc=16;
  single->high=single->bit-1;

  return 0;
}

/********
 * sign *
 ********/

int dicom_sign(SINGLE *single)
{
  int	edge,i;
  U32	length,l;
  U16	*d;

  dicom_log(DEBUG,"dicom_sign()");

  if (!single)
  {
    dicom_log(ERROR,"No image given");
    return -1;
  }

  if (!single->sign)
    return 0;

  if (single->alloc!=16)
  {
    dicom_log(ERROR,"BitsAllocated != 16");
    return -2;
  }

  if (single->high!=single->bit-1)
    dicom_log(WARNING,"Wrong HighBit");

  edge=1<<(single->bit-1);

  length=single->frames*single->w*single->h*single->samples;
  d=single->data;

  for (l=length; l; l--,d++)
    if (*d<edge)
      *d+=edge;
    else
      *d-=edge;

  switch(single->photometric)
  {
  case PALETTE_COLOR :
  case ARGB :
    for (i=0; i<3; i++)
      if (single->clut[i].threshold.u16<edge)
        single->clut[i].threshold.u16+=edge;
      else
        single->clut[i].threshold.u16-=edge;

    for (i=0; i<3; i++)
      if (!single->clut[i].data.u16)
	dicom_log(ERROR,"Missing CLUT");
      else
      {
        edge=1<<(single->clut[i].bit-1);
        d=single->clut[i].data.u16;

        for (l=single->clut[i].size; l; l--,d++)
	  if (*d<edge)
	    *d+=edge;
	  else
	    *d-=edge;
      }
    break;

  default :
    break;
  }

  single->sign=0;

  return 0;
}

/**********
 * planar *
 **********/

int dicom_planar(SINGLE *single)
{
  int	i,j;
  U32	length,l;
  U16	*frame_s,*frame_d,*s,*d;

  dicom_log(DEBUG,"dicom_planar()");

  if (!single)
  {
    dicom_log(ERROR,"No image given");
    return -1;
  }

  if (single->samples<=1)
    return 0;

  if (!single->planar)
    return 0;

  if (single->alloc!=16)
  {
    dicom_log(ERROR,"BitsAllocated != 16");
    return -2;
  }

  length=single->w*single->h;

  frame_d=malloc(length*single->samples*2);
  if (!frame_d)
  {
    dicom_log(ERROR,"Out of memory");
    return -3;
  }

  for (i=0; i<single->frames; i++)
  {
    frame_s=(U16*)single->data+i*length*single->samples;
    s=frame_s;

    for (j=0; j<single->samples; j++)
    {
      d=frame_d+j;

      for (l=length; l; l--)
      {
        *d=*s++;
        d+=single->samples;
      }
    }

    memcpy(frame_s,frame_d,length*single->samples*2);
  }

  eNlfSafeFree(frame_d);

  single->planar=0;

  return 0;
}

/*********
 * shift *
 *********/

int dicom_shift(SINGLE *single)
{
  int	shift,i;
  U32	length,l;
  U16	*d;

  dicom_log(DEBUG,"dicom_shift()");

  if (!single)
  {
    dicom_log(ERROR,"No image given");
    return -1;
  }

  if (single->photometric==MONOCHROME1 || single->photometric==MONOCHROME2)
    return 0;

  if (single->alloc!=16)
  {
    dicom_log(ERROR,"BitsAllocated != 16");
    return -2;
  }

  switch(single->photometric)
  {
  default :
    shift=15-single->high;

    if (!shift)
      return 0;

    length=single->frames*single->w*single->h*single->samples;
    d=single->data;

    for (l=length; l; l--)
      *d++<<=shift;

    single->high=15;
    break;

  case ARGB :
    shift=15-single->high;

    if (shift)
    {
      length=single->frames*single->w*single->h;
      d=single->data;

      for (l=length; l; l--)
      {
        d++;
	*d++<<=shift;
	*d++<<=shift;
	*d++<<=shift;
      }

      single->high=15;
    }

  case PALETTE_COLOR :
    for (i=0; i<3; i++)
    {
      shift=16-single->clut[i].bit;

      if (!shift)
	continue;

      d=single->clut[i].data.u16;

      for (l=single->clut[i].size; l; l--)
	*d++<<=shift;

      single->clut[i].bit=16;
    }
  }

  return 0;
}

/*************
 * transform *
 *************/

/* This function converts the raw pixel data to one of the two allowed output formats:
		16-bit grayscale or 8-bit RGB. No other formats are supported at this time. */


IMAGE *dicom_transform(SINGLE *single, int parametric)
{
  static IMAGE image;

  U32	length,l;
  U16	*s;			/* Source image raw data */
  U8	*d;			/* Destination RGB data */

	dicom_log(INFO,"dicom_transform of image to parametric");
	
	if (!single)
	{
		dicom_log(ERROR,"No image given");
		return 0L;
	}
	
	if (dicom_alloc(single))
		return 0L;

	/* eNlf: BEGIN -- don't make positive, need original pixel values */
	/*  if (dicom_sign(single)) */
	/*    return 0L;            */
	/* eNlf: END   -- don't make positive, need original pixel values */
	
	if (dicom_planar(single))
		return 0L;
	
	if (dicom_shift(single))
		return 0L;

	memset(&image,0,sizeof(IMAGE));
	
	image.frames=single->frames;
	image.w=single->w;
	image.h=single->h;

	switch(single->photometric)
	{
		case MONOCHROME1 :
		case MONOCHROME2 :
			image.rgb=0;
			image.data.gray=single->data;
			single->data=0L;
		
			if (parametric)
				return &image;
		
			dicom_max(&image);
			
			if (single->photometric==MONOCHROME1)
			dicom_invert(&image);
		
			return &image;
		
		case PALETTE_COLOR :
		case ARGB :
			if (!single->clut[0].data.u16 ||
				!single->clut[1].data.u16 ||
				!single->clut[2].data.u16)
			{
				dicom_log(ERROR,"Missing CLUT");
				return 0L;
			}
			break;
		
		default :
			break;
	}

	image.rgb=-1;
	image.data.rgb=malloc(image.frames*image.w*image.h*3);
	if (!image.data.rgb)
	{
		dicom_log(ERROR,"Out of memory");
		return 0L;
	}

	length=image.frames*image.w*image.h;
	
	s=single->data;
	d=image.data.rgb;

	switch(single->photometric)
	{
		case PALETTE_COLOR :
			for (l=length; l; l--)
			{
				*d++=dicom_clut(single->clut,  *s)>>8;
				*d++=dicom_clut(single->clut+1,*s)>>8;
				*d++=dicom_clut(single->clut+2,*s)>>8;
				s++;
			}
			break;
		
		case RGB :
			for (l=length*3; l; l--)
				*d++=*s++>>8;
			break;
		
		case HSV :
			for (l=length; l; l--)
			{
				dicom_hsv(s[0],s[1],s[2],d);
				s+=3;
				d+=3;
			}
			break;
		
		case ARGB :
			for (l=length; l; l--)
			if (*s)
			{
				*d++=dicom_clut(single->clut,  *s)>>8;
				*d++=dicom_clut(single->clut+1,*s)>>8;
				*d++=dicom_clut(single->clut+2,*s)>>8;
				s+=4;
			}
			else
			{
				s++;
				*d++=*s++>>8;
				*d++=*s++>>8;
				*d++=*s++>>8;
			}
			break;
		
		case CMYK :
			for (l=length; l; l--)
			{
				*d++=(0xFFFF-*s++)>>8;
				*d++=(0xFFFF-*s++)>>8;
				*d++=(0xFFFF-*s++)>>8;
				s++;
			}
			break;
		
		default :
			break;
	}
	
	return &image;
}

/********
 * clut *
 ********/

static U16 dicom_clut(const CLUT *clut,U16 i)
{
	if (i<=clut->threshold.u16)
		return clut->data.u16[0];
	
	i-=clut->threshold.u16;
	
	if (i>=clut->size-1)
		return clut->data.u16[clut->size-1];
	
	return clut->data.u16[i];
}



/*************************
 * libdicom by Tony Voet - zoom.c *
 *************************/


static void dicom_zoom_gray(const IMAGE *,U16 *,U16 *);
static void dicom_zoom_rgb(const IMAGE *,U8 *,U8 *);

static void dicom_hyper_gray(const IMAGE *,U16 *,U16 *);
static void dicom_hyper_rgb(const IMAGE *,U8 *,U8 *);

static IMAGE *zoom;

/********
 * zoom *
 ********/

IMAGE *dicom_zoom(const IMAGE *image,int w,int h,int hyper)
{
  U16	frame;
  int	size;
  void	*source,*target;

  dicom_log(DEBUG,"dicom_zoom()");

  if (!image)
  {
    dicom_log(ERROR,"No image given");
    return 0L;
  }

  zoom=dicom_new(image->rgb,image->frames,w,h);
  if (!zoom)
    return 0L;

  if (zoom->rgb)
    size=3;
  else
    size=2;

  if (w==image->w && h==image->h)
  {
    memcpy(zoom->data.rgb,image->data.rgb,zoom->frames*w*h*size);
    return zoom;
  }

  source=image->data.rgb;
  target=zoom->data.rgb;

  for (frame=zoom->frames; frame; frame--)
  {
    if (!hyper || (w<image->w && h<image->h))
      if (zoom->rgb)
        dicom_zoom_rgb(image,source,target);
      else
        dicom_zoom_gray(image,source,target);
    else
      if (zoom->rgb)
        dicom_hyper_rgb(image,source,target);
      else
        dicom_hyper_gray(image,source,target);

    source=(U8*)source+image->w*image->h*size;
    target=(U8*)target+w*h*size;
  }

  return zoom;
}

/*************
 * zoom gray *
 *************/

static void dicom_zoom_gray(const IMAGE *image,U16 *source,U16 *target)
{
  float	x,y,sx,sy;
  U16	*line;

  dicom_log(DEBUG,"dicom_zoom_gray()");

  sx=(float) image->w/zoom->w;
  sy=(float) image->h/zoom->h;

  for (y=sy/2.0; y<image->h; y+=sy)
  {
    line=source+image->w*(int)y;
    for (x=sx/2.0; x<image->w; x+=sx)
      *target++=line[(int)x];
  }
}

/************
 * zoom rgb *
 ************/

static void dicom_zoom_rgb(const IMAGE *image,U8 *source,U8 *target)
{
  float	x,y,sx,sy;
  int	i;
  U8	*line;

  dicom_log(DEBUG,"dicom_zoom_rgb()");

  sx=(float) image->w/zoom->w;
  sy=(float) image->h/zoom->h;

  for (y=sy/2.0; y<image->h; y+=sy)
  {
    line=source+3*image->w*(int)y;
    for (x=sx/2.0; x<image->w; x+=sx)
    {
      i=3*(int)x;
      *target++=line[i];
      *target++=line[i+1];
      *target++=line[i+2];
    }
  }
}

/**************
 * hyper gray *
 **************/

static void dicom_hyper_gray(const IMAGE *image,U16 *source,U16 *target)
{
  float	x,y,sx,sy,dx,dy;
  int	ix,iy;
  U16	*line,*next;

  dicom_log(DEBUG,"dicom_hyper_gray()");

  sx=(float) image->w/zoom->w;
  sy=(float) image->h/zoom->h;

  for (y=sy/2.0; y<image->h; y+=sy)
  {
    iy=(int)(y+0.5)-1;

    line=source+image->w*iy;
    next=line+image->w;

    for (x=sx/2.0; x<image->w; x+=sx)
    {
      ix=(int)(x+0.5)-1;

      dx=x-ix-0.5;
      dy=y-iy-0.5;

      if (x<0.5)
        dx=1.0;
      if (image->w-0.5<x)
        dx=0.0;
      if (y<0.5)
        dy=1.0;
      if (image->h-0.5<y)
        dy=0.0;

      *target++= (1.0-dx)*(1.0-dy)*line[ix] + dx*(1.0-dy)*line[ix+1] +
        (1.0-dx)*dy*next[ix] + dx*dy*next[ix+1] + 0.5;
    }
  }
}

/*************
 * hyper rgb *
 *************/

static void dicom_hyper_rgb(const IMAGE *image,U8 *source,U8 *target)
{
  float	x,y,sx,sy,dx,dy;
  int	ix,iy,i;
  U8	*line,*next;

  dicom_log(DEBUG,"dicom_hyper_rgb()");

  sx=(float) image->w/zoom->w;
  sy=(float) image->h/zoom->h;

  for (y=sy/2.0; y<image->h; y+=sy)
  {
    iy=(int)(y+0.5)-1;

    line=source+3*image->w*iy;
    next=line+3*image->w;

    for (x=sx/2.0; x<image->w; x+=sx)
    {
      ix=(int)(x+0.5)-1;

      dx=x-ix-0.5;
      dy=y-iy-0.5;

      if (x<0.5)
        dx=1.0;
      if (image->w-0.5<x)
        dx=0.0;
      if (y<0.5)
        dy=1.0;
      if (image->h-0.5<y)
        dy=0.0;

      ix*=3;

      for (i=3; i; i--)
      {
        *target++= (1.0-dx)*(1.0-dy)*line[ix] + dx*(1.0-dy)*line[ix+4] +
          (1.0-dx)*dy*next[ix] + dx*dy*next[ix+4] + 0.5;

        ix++;
      }
    }
  }
}


