

/*
 *  functions.c - some functions used all over the place
 * This is a two-part package. Part 1 (this one) contains more
 * image-related functions and helper functions, and
 * part 2 contains more numerical algorithms and methods.
 */


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <fftw3.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_rng.h>


#include "defs.h"
#include "functions.h"
#include "functions2.h"
#include "dlgbox.h"
#include "imgbuf.h"




/* We need a GSL random generator */

const gsl_rng_type *gsl_randgen_type;
gsl_rng *gsl_randgen;



/*****************************************************

	An important help for debugging. dp (debug-print)
	is equivalent to fprintf (stderr, ...), but
	only prints if the priority level pri of the message
	exceeds the currently selected debug level.
	If debuglevel == 0, no messages are printed.
*/



void dp (int pri, char *format,...)
{
va_list args;
char buf[1024];

	if (pri > debuglevel) return;
    va_start(args, format);
    vsnprintf(buf,1020,format,args);
    fprintf (stderr,"%s",buf);
}



/********************************************************

	Critical memory allocations lead to program exit

*********************************************************/


void fatal_memerr ()	/* Abort program when critical memory allocation failed */
{
	ShowMessageC ("Fatal error - no memory. Program exits","That sucks!");
	exit (1);
}


/********************************************************

	Small function to determine whether a string is empty 
   (null string or only blanks)

********************************************************/


char isempty (const char* s)
{
int i;

	i=0;
	while ((i<strlen(s)) && ((s[i]==' ') || (s[i]=='\t') || (s[i]=='\n'))) i++;
	if (i<strlen(s)) return 0; else return 1;
}



/********************************************************

	Rather than torturing ourselves with the dirname
	and basename functions in <string.h> or perhaps
	<libgen.h>
	let's define our own, copied from string.c

********************************************************/


char *basename (const char *filename)
{
	/* strrchr returns a pointer to the last occurrence of the character c in the string s */
	char *p = strrchr (filename, '/');

	/* If null pointer, return the whole filename, else return string starting at p+1 */
	/* Note that the return value is a pointer inside the source string, so it must not be freed */

	return (p ? p + 1 : (char *) filename);
}

/* For dirname, we have to roll our own */

char *dirname (const char *filename)
{
char *result, *p;

	if (isempty(filename))
	{
		result = calloc (2, sizeof (char));
		return result;						/* Empty string */
	}

	/* strrchr returns a pointer to the last occurrence of the character c in the string s */
	p = strrchr (filename, '/');

	/* If this is the null pointer, the filename does not contain a directory component */

	if (!p)
	{
		result = calloc (2, sizeof (char));
		return result;						/* Empty string */
	}

	/* Length is p-filename, so copy this number of bytes */

	result = calloc (strlen(filename), sizeof(char));
	strncpy (result, filename, p - filename);

	return result;
}


/* Compose_filename returns a full filename with path as follows:

	1. If filename already contains a path (dirname returns non-empty string)
		then the filename is returned unchanged
	2. If the filename does not contain a path, concatenate with imager_cwd

	TODO: This has been sloppily and hastily written, and it causes memory leaks!
*/


char *compose_filename ( const char *filename)
{
char *dirp;

	dirp = dirname (filename);			/* Directory or empty */
	if (!isempty (dirp))				/* This means the filename already contains a path */
	{
		dirp = strdup (filename);
		return dirp;					/* In that case return the pointer to the file name */
	}

	/* We get here if the file name does not contain a path */

	if (imager_cwd && (strlen(imager_cwd)>0) && (!isempty(imager_cwd)))
	{
		dirp = calloc (strlen(filename)+strlen(imager_cwd)+10, sizeof (char));
		strcpy (dirp, imager_cwd);
		if (imager_cwd[strlen(imager_cwd)-1] == '/')
			strcat (dirp, filename);				/* Slash provided, add fname to fname2 */
		else
		{
			strcat (dirp, "/");					/* We provide our own slash */
			strcat (dirp, filename);
		}
		return dirp;
	}

	/* This last point is reached when imager_cwd is empty, agaqin return the file name */

	dirp = strdup (filename);
	return dirp;					/* In that case return the pointer to the file name */

}



/*************************************************************/

/* Initialize whatever we need for using GSL components
*/


void	init_gsl_fn ()
{

	/* Prepare a random generator */
	
	gsl_rng_env_setup();
	gsl_randgen_type = gsl_rng_default;
	gsl_randgen = gsl_rng_alloc (gsl_randgen_type);


	
	
}



/*************************************************************/


/* Function returns whether a slice is all-zero.
	Returns 1 if all components are zero or 0 if not */


char slice_is_zero (image_type* img, int slice)
{
char* p;
char b;
long i,l,e;

	l = img->xmax * img->ymax * DATALEN(img->imgtype);
	e = l*slice;
	b=0;
	p = img->data;
	
	for (i=e; i<e+l; i++) b |= p[i];
	
	if (b!=0) return 0; 
	
	dp (2, "Zero slice %d detected\n",slice);
	return 1;
}
	


/*************************************************************/

/* Function returns image min & max values as stored in min/max metadata */


void imgminmax (image_type* img, float* mmin, float* mmax)
{
float min,max;

	if (img->imgtype==B_FLOAT)
	{
		min = img->minval.fmin;
		max = img->maxval.fmax;
	}
	else
	{
		min = img->minval.imin;
		max = img->maxval.imax;
	}
	*mmin = min; *mmax = max;
}



/*************************************************************/

/* This function returns the nearest smaller or equal power of two
	image size. What exactly is returned depends on the flags:

	Bit 0 set: Returns the dyadic x-size. Example: An image has 512x1532
			pixels. 512 is a power of 2, and 512 is returned.
			Example 2: The image has 1532x512 pixels instead. The next smaller
			power of two is 1024, and 1024 is returned. 
	Bit 1 is set: Returns the dyadic y-size. Example: An image has 512x1532
			pixels. The next smaller power of two of the y-size of 1532 
			is 1024, and 1024 is returned. 
	Bits 0 and 1 are both set (flags == 0x03): Both the dyadic x- and y-
			sizes are computed and the *smaller* of the two is returned.
			Example: Image size is 512x1532. The dyadic sizes are 512 and
			1024, and 512 is returned.
	Bit 2 is set: This inverts the logic of bits 0 and 1 in that dyadic size
			must be *larger* or equal. In the above example (512x1532), a flag
			value of 0x05 returns 512, and a flag value of 0x06 returns 2048 (!).
			Consequently, if both x- and y-flags are set (flag value 0x07),
			the larger size is returned, in this example 2048.
*/


int dyadic_image_size (image_type *img, unsigned char flags)
{
int xs, ys;


	xs = 1;
	ys = 1;

	if (flags & 0x04)			/* Dyadic size greater or equal */
	{
		while (xs < img->xmax) xs <<= 1;
		while (ys < img->ymax) ys <<= 1;
	}
	else						/* Dyadic size less or equal */
	{
		while (xs < img->xmax) xs <<= 1;
		if (xs > img->xmax) xs >>= 1;
		while (ys < img->ymax) ys <<= 1;
		if (ys > img->ymax) ys >>= 1;
	}

	if ( (flags & 0x03) == 0)			/* uhh... */
		return 0;
	else if ((flags & 0x03) == 1)		/* Want only x size */
		return xs;
	else if ((flags & 0x03) == 2)		/* Want only y size */
		return ys;
	else 								/* Min or max of both sizes */
	{
		if (flags & 0x04)
			return MAX(xs,ys);
		else
			return MIN(xs,ys);
	}

}




/*************************************************************/


/* Function swaps quadrants I - III and II - IV, often used
	in Fourier-like functions when the origin of a signed coordinate
	system goes to the center if the image.
	Function acts on a slice-by-slice basis.
*/


void swap_quadrants (image_type* img)
{
int z;
int x,y, xm,ym, xmh,ymh, xodd, yodd;
unsigned char element1[4], element2[4];


	/* Get image size and determine its center (and whether it is an odd-sized image) */

	xm = img->xmax;	xodd = xm & 0x01;
	ym = img->ymax; yodd = ym & 0x01;
	xmh = xm >> 1;
	ymh = ym >> 1;

	for (z=0; z < img->zmax; z++)
	{

		/* Swap quadrants II and III -- for odd-sized dimensions this is the odd-sized quadrant */

		for (y=0; y < ymh+yodd; y++)
			for (x=0; x < xmh+xodd; x++)
			{
				readbuf (img, element1, x, y, z);
				readbuf (img, element2, x+xmh, y+ymh, z);
				writebuf (img, element2, x, y, z);
				writebuf (img, element1, x+xmh, y+ymh, z);
			}

		/* Swap quadrants I and IV -- for odd-sized dimensions this is the even-sized quadrant */

		for (y=0; y < ymh; y++)
			for (x=0; x < xmh; x++)
			{
				readbuf (img, element1, x+xmh+xodd, y, z);
				readbuf (img, element2, x, y+ymh+yodd, z);
				writebuf (img, element2, x+xmh+xodd, y, z);
				writebuf (img, element1, x, y+ymh+yodd, z);
			}

	}

	img->uid = next_uid++;		/* Invalidate the cache */
}



/*********************************************+

	Color model conversions
	
***********************************************/


/* Convert a rgb pixel to hsv */

hsvtriplet rgb_to_hsv (rgbtriplet rgb)
{
hsvtriplet hsv;
int h,s,v,m;
	
	if ((rgb.red==0) && (rgb.green==0) && (rgb.blue==0))
	{
		hsv = (hsvtriplet){0,0,0};
	}
	else
	{
	
		if ((rgb.red>=rgb.green) && (rgb.red>=rgb.blue))
		{
			v = rgb.red; m = (rgb.green<rgb.blue? rgb.green : rgb.blue);
			h = (int) (60.0*(rgb.green-rgb.blue) / (float)(v-m));
		}
		else if ((rgb.green>rgb.red) && (rgb.green>=rgb.blue))
		{
			v = rgb.green; m = (rgb.red<rgb.blue? rgb.red : rgb.blue);
			h = (int) (120.0+60.0*(rgb.blue-rgb.red) / (float)(v-m));
		}
		else
		{
			v = rgb.blue; m = (rgb.red<rgb.green? rgb.red : rgb.green);
			h = (int) (240.0+60.0*(rgb.red-rgb.green) / (float)(v-m));
		}
		s = (int) (255.0 *(1.0 - (double) m / (double) v));
		hsv.hue= (h<0?  h+360 : h);
		hsv.sat=s & 0xff;
		hsv.val=v & 0xff;
	}
	return hsv;
}


rgbtriplet hsv_to_rgb (hsvtriplet hsv)
{
int sector;
float f, s;
int p,q,t;
rgbtriplet rgb;

	if( hsv.sat == 0 ) 	/* Sat==0 means achromatic (=grey) with r=b=g */
	{
		rgb.red = rgb.green = rgb.blue = hsv.val;
	}
	else
	{

		if (hsv.hue<0) hsv.hue+=360;
		sector = floor (hsv.hue / 60.0);	/* color sectors 0 through 5 */
		f = hsv.hue/60.0 - sector;			/* fractional part of h */
		s = (float)hsv.sat / 255.0;			/* Let sat go from 0 to 1 */
		p = (int)(0.5+hsv.val * ( 1 - s ));
		q = (int)(0.5+hsv.val * ( 1 - s * f ));
		t = (int)(0.5+hsv.val * ( 1 - s * ( 1 - f )));

		switch (sector)
		{
			case 0:
				rgb.red = hsv.val;
				rgb.green = t;
				rgb.blue = p;
				break;
			case 1:
				rgb.red = q;
				rgb.green = hsv.val;
				rgb.blue = p;
				break;
			case 2:
				rgb.red = p;
				rgb.green = hsv.val;
				rgb.blue = t;
				break;
			case 3:
				rgb.red = p;
				rgb.green = q;
				rgb.blue = hsv.val;
				break;
			case 4:
				rgb.red = t;
				rgb.green = p;
				rgb.blue = hsv.val;
				break;
			case 5:
			default:
				rgb.red = hsv.val;
				rgb.green = p;
				rgb.blue = q;
				break;
		}
	}
	return rgb;
}


rgbtriplet rgb_to_ycrcb (rgbtriplet rgb)
{
float y, cr, cb;
int i;
rgbtriplet result;

		y  = 0.299*rgb.red + 0.587*rgb.green + 0.114*rgb.blue;
		cr = -0.169*rgb.red - 0.331*rgb.green + 0.5*rgb.blue;
		cb = 0.5*rgb.red - 0.419*rgb.green - 0.081*rgb.blue;

		/* Pack back into uchar. Note that Cr and Cb are packed as signed short with offset */

		i = IROUND(y); if (i>255) i=255; else if (i<0) i=0;
		result.red=i;		/* Integer Y component */
		i = IROUND(127.0+cr); if (i>255) i=255; else if (i<0) i=0;
		result.green=i;		/* Integer Cr component */
		i = IROUND(127.0+cb); if (i>255) i=255; else if (i<0) i=0;
		result.blue=i;		/* Integer Cb component */
		return result;
}





/******************************************************************/



/* Create a MIP projection of a stack. This is an image operation
	and not a visualization operation (as it used to be).
	img is a stack, and the MIP is returned in img.
	The data type is unchanged.
*/



void makemip (image_type* img)
{
int x,y,z;
image_type tempimg;
long idx, zofs;
unsigned char d;



	/* Allocate a bitmap to take MIP data. Same data type as img, but only one slice. */

	if (allocate_image (&tempimg, img->imgtype, img->xmax, img->ymax, 1)==-1)
		fatal_memerr ();


	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{

			idx = x + y*img->xmax;
			zofs = img->xmax * img->ymax;
			switch (img->imgtype)
			{
				case B_CHAR:
				{
					unsigned char* p = img->data;
					unsigned char* p1 = tempimg.data;
					unsigned char max;
					p += idx;					/* points to current pixel "column" */
					max = *p;					/* first slice's data */
					if (img->zmax>0)
						for (z=1; z<img->zmax; z++)
						{
							p+= zofs;
							if (max < *p) max = *p;
						}
					p1[idx] = max;
					break;
				}
				case B_SHORT:
				{
					short* p = img->data;
					short* p1 = tempimg.data;
					short max;
					p += idx;
					max = *p;
					if (img->zmax>0)
						for (z=1; z<img->zmax; z++)
						{
							p+= zofs;
							if (max < *p) max = *p;
						}
					p1[idx] = max;
					break;
				}
				case B_FLOAT:
				{
					float* p = img->data;
					float* p1 = s_img.data;
					float max;
					p += idx;
					max = *p;
					if (img->zmax>0)
						for (z=1; z<img->zmax; z++)
						{
							p+= zofs;
							if (max < *p) max = *p;
						}
					p1[idx] = max;
					break;
				}
				case B_RGB:			/* Note: Channels treated individually for RGB! */
				{
					rgbtriplet* p = img->data;
					rgbtriplet* p1 = s_img.data;
					rgbtriplet mip;
					p += idx;
					mip.red   = p->red;
					mip.green = p->green;
					mip.blue  = p->blue;
					if (img->zmax>0)
						for (z=1; z<img->zmax; z++)
						{
							p+= zofs;
							if ( p->red   > mip.red)   mip.red   = p->red;
							if ( p->green > mip.green) mip.green = p->green;
							if ( p->blue  > mip.blue)  mip.blue  = p->blue;
						}
					p1[idx]=mip;
					break;
				}
			}
		}

	handback (img, &tempimg);
}






/******************************************************************/




