

/*
	imgbuf.c - core functions to handle data in an image buffer
*/



/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */



#include "imgbuf.h"
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "functions.h"
#include "functions2.h"
#include "application.h"
#include "interpolate.h"


/* Allocate memory for image of dimensions x,y,z - returns a pointer (img)
   and an error code. NOTE - we need two functions. One creates an image,
   but keeps the display structure empty. The other just allocates the 
   buffer, but leaves the display stuff (pixbuf, drawable, parent...)
   untouched. */


int allocate_image (image_type* img, TYPE_SELECTOR imgtype, int x, int y, int z)
{
int e;

	img->xmax = x; img->ymax=y; img->zmax=z;
	img->imgtype = imgtype;
	img->pixbuf = NULL; 				/* Likely, this new image has no image window associated to it */
	img->parent = NULL;					/* therefore, null out the display variables */
	img->drawable = NULL;
	img->width=0; img->height = 0;
	img->uid = next_uid++;
	e = DATALEN (imgtype);

	img->data = calloc (x*y*z, e);
	if (img->data==NULL) return -1;	/* Mem alloc problem */
	return 0;
}


int allocate_imgbuf (image_type* img, TYPE_SELECTOR imgtype, int x, int y, int z)
{
int e;

	img->xmax = x; img->ymax=y; img->zmax=z;
	img->imgtype = imgtype;
	img->uid = next_uid++;
	e = DATALEN (imgtype);

	img->data = calloc (x*y*z, e);
	if (img->data==NULL) return -1;	/* Mem alloc problem */
	return 0;
}



/* Dpulicate an image */

int clone_image (image_type* src, image_type* dest)
{
int e;
long idx;

	if (allocate_image (dest, src->imgtype, src->xmax, src->ymax, src->zmax))
		return -1;	/* failed */
		
	if (src->imgtype == B_FLOAT)
	{
		dest->minval.fmin = src->minval.fmin;
		dest->maxval.fmax = src->maxval.fmax;
	}
	else
	{
		dest->minval.imin = src->minval.imin;
		dest->maxval.fmax = src->maxval.fmax;
	}
	
	e = DATALEN(src->imgtype);
	idx = e * src->xmax * src->ymax * src->zmax;
	memcpy (dest->data, src->data, idx);
	dest->uid = next_uid++;
	return 0;
}


/* MOVE an image from src to dest, free dest beforehand. This is often
   used to hand back a result into the original image.
	If dest is the main image, retain its GUI infrastructure, i.e. don't overwrite */

void handback (image_type* dest, image_type* src)
{
	freebuf (dest);
	dest->data = src->data; 
	dest->xmax = src->xmax;
	dest->ymax = src->ymax;
	dest->zmax = src->zmax;
	dest->uid = next_uid++;
	dest->imgtype = src->imgtype;
	make_minmax (dest);
}



/* Retrieve one data element from coordinates (x,y,z) */

void readbuf (image_type* img, void* element, int x, int y, int z)
{
char* p1;
short* p2;
rgbtriplet* p3;
float* p4;

	p1=(char*)img->data; 
	p2=(short*)img->data; 
	p3=(rgbtriplet*)img->data; 
	p4=(float*)img->data;
	switch (img->imgtype)
	{
		case B_CHAR:	
				*(unsigned char*)element = (unsigned char)p1[x+img->xmax*(y+img->ymax*z)];
				break;
		case B_SHORT:
				*(short*)element = (short)p2[x+img->xmax*(y+img->ymax*z)];
				break;
		case B_RGB:
				*(rgbtriplet*)element = (rgbtriplet)p3[x+img->xmax*(y+img->ymax*z)];
				break;
		case B_FLOAT:
				*(float*)element = (float)p4[x+img->xmax*(y+img->ymax*z)];
	}

}


/* Read one arbitrary image element, return as float.
	NOTE: RGB gets converted to a single gray value using YUV weight factors
	NOTE: Simple edge handling is implemented. If any out-of-bounds element is accessed,
		the function will continue the last valid value
*/

double readbuf_flt (image_type* img, int x, int y, int z)
{
unsigned char* p1;
short* p2;
rgbtriplet* p3;
float* p4;
rgbtriplet rgb;

	if (x<0) x=0; else if (x>=img->xmax) x=img->xmax-1;
	if (y<0) y=0; else if (y>=img->ymax) y=img->ymax-1;
	if (z<0) z=0; else if (z>=img->zmax) z=img->zmax-1;

	p1=(char*)img->data; 
	p2=(short*)img->data; 
	p3=(rgbtriplet*)img->data; 
	p4=(float*)img->data;
	switch (img->imgtype)
	{
		case B_CHAR:	
				return (double)p1[x+img->xmax*(y+img->ymax*z)];
				break;
		case B_SHORT:
				return (double)p2[x+img->xmax*(y+img->ymax*z)];
				break;
		case B_RGB:
				rgb = (rgbtriplet)p3[x+img->xmax*(y+img->ymax*z)];
				return REDWEIGHT*rgb.red + GREENWEIGHT*rgb.green + BLUEWEIGHT*rgb.blue;
				break;
		case B_FLOAT:
				return p4[x+img->xmax*(y+img->ymax*z)];
	}
	return 0.0;	/* Silence a compiler warning */
}


/* Read one arbitrary image element, return as uchar, scaling as necessary
	NOTE: RGB gets converted to a single gray value using YUV weight factors
*/

unsigned char readbuf_char (image_type* img, int x, int y, int z)
{
unsigned char* p1;
short* p2;
rgbtriplet* p3;
float* p4;
rgbtriplet rgb;
float f;

	if (x<0) x=0; else if (x>=img->xmax) x=img->xmax-1;
	if (y<0) y=0; else if (y>=img->ymax) y=img->ymax-1;
	if (z<0) z=0; else if (z>=img->zmax) z=img->zmax-1;

	p1=img->data; p2=img->data; p3=img->data; p4=img->data;
	switch (img->imgtype)
	{
		case B_CHAR:	
				return p1[x+img->xmax*(y+img->ymax*z)];
				break;
		case B_SHORT:
				return IROUND(255.0*(p2[x+img->xmax*(y+img->ymax*z)] - img->minval.imin)
					/ (double)(img->maxval.imax - img->minval.imin));
				break;
		case B_RGB:
				rgb = (rgbtriplet)p3[x+img->xmax*(y+img->ymax*z)];
				return IROUND(REDWEIGHT*rgb.red + GREENWEIGHT*rgb.green + BLUEWEIGHT*rgb.blue);
				break;
		case B_FLOAT:
				return IROUND(255.0* (p4[x+img->xmax*(y+img->ymax*z)]-img->minval.fmin)
					/ (img->maxval.fmax - img->minval.fmin));
	}
	return 0.0;	/* Silence a compiler warning */

}



/* Read one arbitrary image element from an RGB image. Fails for non-RGB.
	TODO: Scan through all of the code to see where we can simplify using this function.
*/


rgbtriplet readbuf_rgb (image_type* img, int x, int y, int z)
{
rgbtriplet* p;
rgbtriplet rgb;

	if (x<0) x=0; else if (x>=img->xmax) x=img->xmax-1;
	if (y<0) y=0; else if (y>=img->ymax) y=img->ymax-1;
	if (z<0) z=0; else if (z>=img->zmax) z=img->zmax-1;

	g_assert (img->imgtype==B_RGB);			/* Only RGB allowed or we die a bad death */

	p=img->data;

	return p[x+img->xmax*(y+img->ymax*z)];

}



/* Another variation on the above theme, but this one can
	interpolate between points, i.e. it allows double x,y,z coords.
	Interpolation always returns float types.
	Furthermore, interpolation is only valid within a slice */



double ireadbuf (image_type* img, double x, double y, int z, interpol_method intp)
{
double a1,a2,a3,a4,a5,a6;	/* 4 nearest neighbors */
double xval[4];				/* 4 row interpolations in bi-cubic */
double A[4];				/* Cubic interpolation coefficents */
int x1,x2,y1,y2,i;

	if (intp==INTP_NN)		/* Nearest neighbor interpolation */
	{
		x1 = floor (0.5+x);
		y1 = floor (0.5+y);
		return readbuf_flt (img, x1,y1,z);
	}
	else if (intp==INTP_BILIN)	/* Bi-linear interpolation */
	{
		x1=floor (x); x2=floor (x+1);	/* Determine the 4 neighbors */
		y1=floor (y); y2=floor (y+1);
		a1 = readbuf_flt (img,x1,y1,z);
		a2 = readbuf_flt (img,x2,y1,z);
		a3 = readbuf_flt (img,x1,y2,z);
		a4 = readbuf_flt (img,x2,y2,z);
		
		x = x-x1; y = y-y1;	/* Should always be in range 0...1 */
		a5 = (1-x)*a1 + x*a2;
		a6 = (1-x)*a3 + x*a4;
		return (1-y)*a5 + y*a6;
	}
	else if (intp==INTP_BICUB)	/* Bi-Cubic interpolation */
	{
		x1=floor (x); y1=floor (y);	/* Integer base point */
		x = x-x1; y = y-y1;	/* Should always be in range 0...1 */

		cub_coeff (A, x, 1);
		for (i=-1; i<=2; i++)
			xval[i+1] =   readbuf_flt (img, x1-1, y1+i, z) * A[0]
						+ readbuf_flt (img, x1  , y1+i, z) * A[1]
						+ readbuf_flt (img, x1+1, y1+i, z) * A[2]
						+ readbuf_flt (img, x1+2, y1+i, z) * A[3];

		/* same weight factors, but now in Y direction */
		
		cub_coeff (A, y, 1);
		return (xval[0]*A[0] + xval[1]*A[1] + xval[2]*A[2] + xval[3]*A[3]);

	}
	else if (intp==INTP_SPLINE2)
		return spline_interpol (img, x,y,z, 2);
	else if (intp==INTP_SPLINE3)
		return spline_interpol (img, x,y,z, 3);
	else if (intp==INTP_SPLINE4)
		return spline_interpol (img, x,y,z, 4);
	else if (intp==INTP_SPLINE5)
		return spline_interpol (img, x,y,z, 5);
	
	return 0.0;	/* Silence a compiler warning */

}


/* The only time when we cannot use the above is when we need RGB data, so there... */
/* TODO: RGB is presently unable to do the new spline interpolation, defaults to bicubic */

rgbtriplet rgbireadbuf (image_type* img, double x, double y, int z, interpol_method intp)
{
double a1,a2,a3,a4;	/* 4 nearest neighbors */
double xval[4][3];
double A[4];			/* Cubic interpolation coeffs */
double buf;
int x1,x2,y1,y2;
int i1,i2,i3,i4,i,k1,k2;
int ired, igreen, iblue;		/* Intermediate values */
rgbtriplet* p;
rgbtriplet rgb;


	g_assert (img->imgtype == B_RGB);
	p = img->data;
	if (z<0) z=0; else if (z>=img->zmax) z=img->zmax-1;
	ired = 0; igreen = 0, iblue = 0;
	
	if (intp==INTP_NN)		/* Nearest neighbor interpolation */
	{						/* This one is straightforward and returns the result directly */
		x1 = IROUND(x);
		y1 = IROUND(y);
		x1 = CLAMP (x1,0,img->xmax-1);
		y1 = CLAMP (y1,0,img->ymax-1);
		return p[ x1 + img->xmax*(y1 + img->ymax*z)];
	}
	else if (intp==INTP_BILIN)	/* Bi-linear interpolation */
	{
		x1=floor (x); x2=x1+1;	/* Determine the 4 neighbors */
		y1=floor (y); y2=y1+1;
		x = x-x1; y = y-y1;	/* Should always be in range 0...1 */
		x1 = CLAMP (x1,0,img->xmax-1);		/* Clamp _AFTER_ computing x,y */
		y1 = CLAMP (y1,0,img->ymax-1);
		x2 = CLAMP (x2,0,img->xmax-1);
		y2 = CLAMP (y2,0,img->ymax-1);

		/* Don't frown - I'm trying to reduce the number of multiplications */
		i1 = i2 = i3 = i4 = img->xmax * img->ymax * z;
		k1 = img->xmax*y1; k2 = img->xmax*y2; 			/* Requires clamped values */
		i1+=k1+x1; i2+=k1+x2; i3+=k2+x1; i4+=k2+x2;

		a1 = p[i1].red;
		a2 = p[i2].red;
		a3 = p[i3].red;
		a4 = p[i4].red;
		ired = IROUND ((1.0-y)*((1.0-x)*a1 + x*a2) + y*((1.0-x)*a3 + x*a4));

		a1 = p[i1].green;
		a2 = p[i2].green;
		a3 = p[i3].green;
		a4 = p[i4].green;
		igreen = IROUND ((1.0-y)*((1.0-x)*a1 + x*a2) + y*((1.0-x)*a3 + x*a4));

		a1 = p[i1].blue;
		a2 = p[i2].blue;
		a3 = p[i3].blue;
		a4 = p[i4].blue;
		iblue = IROUND ((1.0-y)*((1.0-x)*a1 + x*a2) + y*((1.0-x)*a3 + x*a4));
	}
	else		/* Let this default to bi-cubic if our preferences indicate more exotic interpolation schemes */
				/* else if (intp==INTP_BICUB)			/* Bi-Cubic interpolation */
	{
		x1=floor (x); y1=floor (y);		/* Integer base point */
		x = x-x1; y = y-y1;				/* Should always be in range 0...1 */

		/* four weight factors in X direction; try to minimize mults */
		
		cub_coeff (A, x, 1);

		/* The closest coordinates - note a slightly different edge behavior than non-rgb */

		i1 = i2 = i3 = i4 = x1 + img->xmax * img->ymax * z;		/* Indices into slice combined with x offsets */
		if (x1>0) i1-=1;
		if (x1<img->xmax-1) i3+=1;
		if (x1<img->xmax-2) i4+=2;

		for (i=-1; i<=2; i++)
		{
			y2 = y1+i; if (y2<0) y2=0; else if (y2>=img->ymax) y2=img->ymax-1;
			y2 *= img->xmax;
			xval[i+1][0] = p[i1+y2].red*A[0] +   p[i2+y2].red*A[1] +   p[i3+y2].red*A[2] +   p[i4+y2].red*A[3];
			xval[i+1][1] = p[i1+y2].green*A[0] + p[i2+y2].green*A[1] + p[i3+y2].green*A[2] + p[i4+y2].green*A[3];
			xval[i+1][2] = p[i1+y2].blue*A[0] +  p[i2+y2].blue*A[1] +  p[i3+y2].blue*A[2] +  p[i4+y2].blue*A[3];
		}

		/* same weight factors, but now in Y direction */
		
		cub_coeff (A, y, 1);

		buf =  (xval[0][0]*A[0] + xval[1][0]*A[1] + xval[2][0]*A[2] + xval[3][0]*A[3]);
		ired = IROUND(buf);
		buf = (xval[0][1]*A[0] + xval[1][1]*A[1] + xval[2][1]*A[2] + xval[3][1]*A[3]);
		igreen = IROUND(buf);
		buf = (xval[0][2]*A[0] + xval[1][2]*A[1] + xval[2][2]*A[2] + xval[3][2]*A[3]);
		iblue = IROUND(buf);
	}

	rgb.red = CLAMP (ired, 0, 255);
	rgb.green = CLAMP (igreen, 0, 255);
	rgb.blue = CLAMP (iblue, 0, 255);

	return rgb;

}


/* Determine if a data element is nonzero - exact method */
/* Any point outside of the image area counts as zero */
/* TODO: Scan through the entire code and eliminate 
	all fabs(blah) < 1e-8 or similar checks (lot of work) */

unsigned char nonzero (image_type* img, int x, int y, int z)
{
int e,i; long idx;
unsigned char* p;
unsigned char b;
float *pflt;
float buf;

	if ((x<0) || (y<0) || (z<0) || (x>=img->xmax) || (y>=img->ymax) || (z>=img->zmax)) return 0;

	/* Float is handled separately, because we allow a tolerance */

	if (img->imgtype == B_FLOAT)
	{
		pflt = img->data;
		buf = pflt[x + img->xmax*(y + img->ymax*z)];
		return (!F_ISZERO(buf));
	}

	/* All other types follow here */

	e = DATALEN(img->imgtype);
	idx = (x + img->xmax*(y + img->ymax*z) )*e; 
	p = img->data;

	b=0;
	for (i=0; i<e; i++) b |= p[idx+i];
	return b!=0;
}



/* Store one data element at coords x,y,z */

void writebuf (image_type* img, void* element, int x, int y, int z)
{
int e; long idx;
unsigned char* p;

	if ((x<0) || (y<0) || (z<0) || (x>=img->xmax) || (y>=img->ymax) || (z>=img->zmax)) return;
	e = DATALEN(img->imgtype);
	idx = (x + img->xmax*(y + img->ymax*z) )*e; 
	p = img->data;
	memcpy (p+idx,element, e);
	
}


/* Copy data between buffers of the same type. Useful for copying
   entire buffers or scanlines. (x1,y1,z1) define the source
   coordinates, (x2,y2,z2) the dest coordinates.  */

void copybuf (image_type* src, image_type* dest, int x1, int y1, int z1, 
				int x2, int y2, int z2, long len)
{
long idx1, idx2;
int e;
unsigned char* p1; unsigned char* p2;

	g_assert (src->imgtype == dest->imgtype);
	e=DATALEN(src->imgtype);
	p1=src->data; p2=dest->data;
	idx1=e*(x1 + src->xmax*(y1 + src->ymax*z1));
	idx2=e*(x2 + dest->xmax*(y2 + dest->ymax*z2));
	len *=e;
	while (len--) { p2[idx2++] = p1[idx1++]; }
	
}



/* Copy selected slices (a substack from z1 to z2) between images of the same type.
   It is assumed that z1 of src becomes z=0 of dest */

void copybuf_z (image_type* src, image_type* dest, int z1, int z2)
{
long idx1, idx2, len;
int e;
char* p1; char* p2;

	g_assert (src->imgtype == dest->imgtype);
	g_assert (dest->zmax > (z2-z1) );	/* enough space? */
	e=DATALEN(src->imgtype);
	p1=src->data; p2=dest->data;
	idx1=e*src->xmax*src->ymax*z1; 	/* offset to slice z1 */
	idx2=0;
	len =e*(1+z2-z1)*src->xmax*src->ymax;
	while (len--) { p2[idx2++] = p1[idx1++]; }
	make_minmax (dest);
	
}


/* Free data associated with this image */

void freebuf (image_type* img)
{

	if (img->data) free (img->data);
	img->data = NULL;
	img->xmax = img->ymax = img->zmax = 0;
	img->imgtype = B_CHAR;
	img->maxval.imax=0;
	img->minval.imin=0;
	img->uid = 0;

}



/***************************************************************************************************/

