

/*
 *	macro.c: Macro interpreter and ancilliary functions
 */


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */



#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "application.h"
#include "defs.h"
#include "dlgbox.h"
#include "filter.h"
#include "functions.h"
#include "functions2.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "interface.h"
#include "tiffstuff.h"

#include "macro.h"
#include "macro1.h"
#include "macro2.h"


/* We need MAXPREC to be one higher than the number of precedence levels */

#define MAXPREC	4



/*****************************************************************

	This is part 1 of two parts. In this source file, we find
	the actual macro interpreter, including
	- Tokenizer
	- Dynamic menu handling
	- Macro run-time code with variable handling
	- and the arithmetic expression parser
	
	Part 2 contains the functions that are associated with
	the keywords and get called during macro execution.
	
******************************************************************/


/* Lookaheads and prototypes for the token handler functions */

void exec_macro (GtkMenuItem *menuitem, gpointer  user_data);


/* NOTE: for proper operation, the token should be 128+place,
   meaning: keywordlist[n].token == 128+n,
   meaning: Sort ascendingly, don't have holes in token numbers.
   If a token has a negative number of parameters it's a dyadic operator
   and the absvalue is the precedence
   We don't need "macro" and "procedure" here, because they are parsed away
   in the tokenization process.
   Warning, the = operator is hardwired. Don't change its number */

#define MINUS_TOK	154
#define EQUALS_TOK	130
#define ENDMAC_TOK	128
#define ENDPROC_TOK	129


static keywordtype keywordlist [128] = 
{
	{	"ENDMACRO",			128, 0, MT_NULL, (*mac_endmacro)},
	{	"ENDPROC",	 		129, 0, MT_NULL, (*mac_endmacro)},
	{	"=",				130,-1, MT_NULL, (*mac_equals)},	/* Dual use. Hardwired as assignment operator, otherwise equals */
	{	"ENDFOR",			131, 0, MT_NULL, NULL},
	{	"FOREACH",			132, 2, MT_NULL, (*mac_foreach)},
	{	"FOR",				133, 2, MT_NULL, (*mac_forloop)},
	{	"ENDIF",			134, 0, MT_NULL, NULL},
	{	"ELSE",				135, 0, MT_NULL, NULL},
	{	"IF",				136, 1, MT_NULL, (*mac_if)},
	{	"ENDWHILE",			137, 0, MT_NULL, NULL},
	{	"WHILE",			138, 1, MT_NULL, (*mac_while)},
	{	"UNTIL",			139, 1, MT_NULL, NULL},
	{	"REPEAT",			140, 0, MT_NULL, (*mac_repeat)},
	{	"Placeholder",		141, 0, MT_NULL, NULL},
	{	"Placeholder",		142, 0, MT_NULL, NULL},
	{	"Placeholder",		143, 0, MT_NULL, NULL},
	{	"Placeholder",		144, 0, MT_NULL, NULL},
	
	{	"DISPOSE",			145, 1, MT_NULL, (*mac_dispose)},
	{	"LOAD",				146, 1, MT_IMG, (*mac_loadimg)},
	{	"IMPORT",			147, 6, MT_IMG, (*mac_import)},
	{	"SAVE",				148, 1, MT_NULL, (*mac_saveimg)},
	{	"MESSAGE",			149, 1, MT_NULL, (*mac_message)},
	{	"PRINT",			150, 1, MT_NULL, (*mac_print)},
	{	"DISPLAY",			151, 1, MT_NULL, (*mac_display)},
	{	"MORPH",			152, 2, MT_IMG,  (*mac_morph)},		/* contains OR, must be first */
	{	"+",				153,-2, MT_NULL, (*mac_add)},
	{	"-",				154,-2, MT_NULL, (*mac_subtract)},
	{	"*",				155,-3, MT_NULL, (*mac_mult)},
	{	"/",				156,-3, MT_NULL, (*mac_div)},
	{	"<>",				157,-1, MT_NULL, (*mac_notequals)},
	{	"!=",				158,-1, MT_NULL, (*mac_notequals)},
	{	"<=",				159,-1, MT_NULL, (*mac_le)},
	{	">=",				160,-1, MT_NULL, (*mac_ge)},
	{	"<",				161,-1, MT_NULL, (*mac_lt)},
	{	">",				162,-1, MT_NULL, (*mac_gt)},
	{	"OR",				163,-1, MT_NULL, (*mac_or)},
	{	"AND",				164,-1, MT_NULL, (*mac_and)},
	{	"XOR",				165,-1, MT_NULL, (*mac_xor)},
	{	"CONVOLVE",			166, 2, MT_IMG, (*mac_convolve)},
	{	"FREICHEN",			167, 2, MT_IMG, (*mac_freichen)},
	{	"FOURIERFILT",		168, 6, MT_IMG, (*mac_fourierfilt)},
	{	"SKELETONIZE",		169, 2, MT_IMG, (*mac_skeletonize)},
	{	"INVERT",			170, 1, MT_IMG, (*mac_invert)},
	{	"THRESHOLD2",		171, 1, MT_IMG, (*mac_threshold2)},
	{	"THRESHOLD",		172, 1, MT_IMG, (*mac_threshold)},
	{	"AUTO_THRESHOLD", 	173, 2, MT_REAL, (*mac_auto_threshold)},
	{	"OTSU_THRESHOLD",	174, 2, MT_REAL, (*mac_otsu_threshold)},
	{	"CONVERT_IMAGE",	175, 3, MT_NULL, (*mac_convert_image)},
	{	"CROP",				176, 5, MT_IMG, (*mac_crop)},
	{	"ROTATE",			177, 2, MT_IMG, (*mac_rotate)},
	{	"IMGINFO",			178, 2, MT_REAL, (*mac_imginfo)},
	{	"RUNLENGTH",		179, 3, MT_IMG, (*mac_runlength)},
	{	"QUANTILE",			180, 2, MT_REAL, (*mac_quantile)},
	{	"SUBSTACK",			181, 3, MT_IMG, (*mac_substack)},
	{	"PROFILE",			182, 7, MT_NULL, (*mac_profile)},
	{	"REMOVE_BK",		183, 3, MT_IMG, (*mac_remove_bk)},
	{	"FILL_HOLES",		184, 2, MT_IMG, (*mac_fill_holes)},
	{	"IMG_HISTO",		185, 8, MT_REAL, (*mac_img_histo)},
	{	"HISTOGRAM",		186, 4, MT_REAL, (*mac_histogram)},
	{	"RGBTOSTACK",		187, 1, MT_IMG, (*mac_rgb2stack)},
	{	"IMGTORGB",			188, 3, MT_IMG, (*mac_imgtorgb)},
	{	"SETPARM",			189, 2, MT_NULL, (*mac_setparm)},
	{	"DIMENSION",		190, 5, MT_REAL, (*mac_dimension)},
	{	"CALL",				191, 1, MT_NULL, (*mac_callproc)},		/* Call another macro */
	{	"FFT",				192, 2, MT_IMG, (*mac_fft)},
	{	"INVFFT",			193, 1, MT_IMG, (*mac_invfft)},
	{	"SETROI",			194, 1, MT_NULL, (*mac_setroi)},
	{	"RESIZE",			195, 3, MT_IMG, (*mac_resize)},
	{	"SKELETON_STATS", 	196, 1, MT_STRING, (*mac_skeletonstats)},
	{	"DIM",				197, 1, MT_NULL, (*mac_dimarray)},
	{	"TRIM_IMAGE",		198, 6, MT_NULL, (*mac_trim_image)},
	{	"INT",				199, 1, MT_INT, (*mac_int)},
	{	"ROUND",			200, 1, MT_INT, (*mac_round)},
	{	"LOG10",			201, 1, MT_REAL, (*mac_log10)},
	{	"LOG",				202, 1, MT_REAL, (*mac_log)},
	{	"EXP",				203, 1, MT_REAL, (*mac_exp)},
	{	"POWER",			204, 2, MT_REAL, (*mac_power)},
	{	"MAINIMAGE",		205, 0, MT_IMG, (*mac_mainimg)},
	{	"GRAPH",			206, 1, MT_NULL, (*mac_graph)},
	{	"KLTRANSFORM",		207, 2, MT_IMG, (*mac_kltransform)},
	{	"CLUSTER2D",		208, 7, MT_IMG, (*mac_cluster2d) },
	{	"CLUSTER3D",		209, 6, MT_IMG, (*mac_cluster3d) },
	{	"CLUSTERCOUNT",		210, 6, MT_IMG, (*mac_clustercount) },
	{	"CENTER",			211, 3, MT_IMG, (*mac_center) },
	{	"STRLEN",			212, 1, MT_INT, (*mac_strlen) },
	{	"SUBSTRING",		213, 3, MT_STRING, (*mac_substring) },
	{	"STRPOS",			214, 3, MT_INT, (*mac_strpos) },
	{	"STRNUM",			215, 1, MT_STRING, (*mac_strnum) },
	{	"TESTPOINT",		216, 1, MT_NULL, (*mac_testpoint) },
	{	"WATERSHED",		217, 3, MT_IMG, (*mac_watershed) },
	{	"BOUNDARY_LENGTH",	218, 2, MT_INT, (*mac_boundary_len) },
	{	"REGIONGROW",		219, 0, MT_IMG, (*mac_regiongrow) },
	{	"FILLROI",			220, 2, MT_NULL, (*mac_fillroi) },
	{	"MEASURE",			221, 3, MT_REAL, (*mac_measure) },
	{	"HISTEQ",			222, 3, MT_IMG, (*mac_histeq) },
	{	"ROSEPLOT",			223, 3, MT_REAL, *mac_roseplot },
	{	"HYSTERESIS_THRESHOLD",	224, 3, MT_IMG, *mac_hysteresis_threshold },
	{	"ADAPTIVE_THRESHOLD", 	225, 5, MT_REAL, (*mac_adaptive_threshold)},
	{	"LACUNARITY",		226, 3, MT_REAL, *mac_lacunarity },
	{	"COMBINE",			227, 3, MT_IMG, *mac_combine },
	{	"HILLCLIMB",		228, 5, MT_IMG, *mac_hillclimb },
	{	"DWT2D",			229, 4, MT_IMG, *mac_dwt },
	{	"WVDENOISE",		230, 4, MT_IMG, *mac_wv_denoise },
	{	"MIP",				231, 1, MT_IMG, *mac_mk_mip },
	{	NULL,			232, 0, MT_NULL, NULL },
	{	NULL,			233, 0, MT_NULL, NULL },
	{	NULL,			234, 0, MT_NULL, NULL },
	{	NULL,			235, 0, MT_NULL, NULL },
	{	NULL,			236, 0, MT_NULL, NULL },
	{	NULL,			237, 0, MT_NULL, NULL },
	{	NULL,			238, 0, MT_NULL, NULL },
	{	NULL,			239, 0, MT_NULL, NULL },
	{	NULL,			240, 0, MT_NULL, NULL },
	{	NULL,			241, 0, MT_NULL, NULL },
	{	NULL,			242, 0, MT_NULL, NULL },
	{	NULL,			243, 0, MT_NULL, NULL },
	{	NULL,			244, 0, MT_NULL, NULL },
	{	NULL,			245, 0, MT_NULL, NULL },
	{	NULL,			246, 0, MT_NULL, NULL },
	{	NULL,			247, 0, MT_NULL, NULL },
	{	NULL,			248, 0, MT_NULL, NULL },
	{	NULL,			249, 0, MT_NULL, NULL },
	{	NULL,			250, 0, MT_NULL, NULL },
	{	NULL,			251, 0, MT_NULL, NULL },
	{	NULL,			252, 0, MT_NULL, NULL },
	{	NULL,			253, 0, MT_NULL, NULL },
	{	NULL,			254, 0, MT_NULL, NULL },
	{	NULL,			255, 0, MT_NULL, NULL },


};

static int	keywordlist_size = sizeof(keywordlist)/sizeof(keywordtype);



void macro_error (int errnumber, int pos)
{
char s[1024];
char n[256];

	switch (errnumber)
	{
		case 1:	strcpy (s, "Syntax error");
				break;
		case 2:	strcpy (s, "( expected");
				break;
		case 3:	strcpy (s, "Function called as procedure");
				break;
		case 4:	strcpy (s, "\" expected");
				break;
		case 5:	strcpy (s, "; expected");
				break;
		case 6:	strcpy (s, "Type mismatch error");
				break;
		case 7: strcpy (s, "Variable memory space full");
				break;
		case 8: strcpy (s, "Dyadic operator expected");
				break;
		case 9: strcpy (s, "Function or monadic operator expected");
				break;
		case 10: strcpy (s, "File open failed in macro");
				break;
		case 11: strcpy (s, "Out of memory error");
				break;
		case 12: strcpy (s, "Image size mismatch error");
				break;
		case 13: strcpy (s, "Division by zero error");
				break;
		case 14: strcpy (s, "Illegal image type for save");
				break;
		case 15: strcpy (s, "Illegal function parameter");
				break;
		case 16: strcpy (s, "Undefined variable");
				break;
		case 17: strcpy (s, "Image type mismatch error");
				break;
		case 18: strcpy (s, "Submacro or procedure not found");
				break;
		case 19: strcpy (s, "Image is not a FFT image");
				break;
		case 20: strcpy (s, "Illegal use of array subscript");
				break;
		case 21: strcpy (s, "Array subscript out of bounds");
				break;
		case 22: strcpy (s, "Array has same name as previously declared variable");
				break;
		case 23: strcpy (s, "No main image open");
				break;
		case 24: strcpy (s, "FOR without ENDFOR,\nWHILE without ENDWHILE,\nor IF without ENDIF");
				break;
		case 25: strcpy (s, "Illegal value in function call");
				break;
		case 26: strcpy (s, "Untested function should not be used");
				break;
		case 27: strcpy (s, "Import file problems");
				break;
		case 28: strcpy (s, "Macro aborted by user");
				break;
		case 29: strcpy (s, "Directory scan failed in macro");
				break;
		default:strcpy (s, "Unknown or unimplemented error");
				
	}
	statusbar_display ("Macro aborted");
	
	strncat (s, "\nin macro\n", 1000);
	strncat (s, current_macro, 1000);
	sprintf (n, "%d",pos-current_macro_pos);
	strncat (s, "\nnear position ", 1000);
	strncat (s, n, 1000);
	
	/* clean up, then display the error message */
	
	
	if (!immd_mode) mac_endmacro (NULL,0);
	ShowMessageC (s, "Macro execution error");

}



/*****************************************************************

	Parsing and string handling functions for macro interpreter
	
******************************************************************/



/* Test a char for space / tab / newline */

char is_space (char c)
{
	if ((c==' ') || (c=='\t') || (c=='\n')) return 1;
	return 0;
}


/* Test for numerical, must be dot or [0-9] for first digit, then
   can be any of +, -, dot, 0..9, e */

char isnum1 (char c)
{
	if ((c=='.') || ((c>='0') && (c<='9')) ) return 1;
	return 0;
}

char isnum2 (char c)
{
	if ((c=='-') || (c=='+') || (c=='.') || ((c>='0') && (c<='9')) ) return 1;
	if ((c=='e') || (c=='E')) return 1;
	return 0;
}


/* Test for alphanumerical and alpha / num, respectively */

char isalpha1 (char c)
{
	if ((c>='a') && (c<='z')) return 1;
	if ((c>='A') && (c<='Z')) return 1;
	if ((c=='_')) return 1;
	return 0;
}

char isalpha2 (char c)
{
	if ((c>='a') && (c<='z')) return 1;
	if ((c>='A') && (c<='Z')) return 1;
	if ((c=='_') || ((c>='0') && (c<='9'))) return 1;
	return 0;
}


/* Extract a numerical value from a string. We assume that
   the first char has already been validated, so use isnum2. */

num_type extract_num (char* buf, int* idx)
{
int i,k;
char s[256];
num_type n;

	i = *idx; k=0; memset (s, 0, sizeof(s));
	n.isint=1;
	do
	{
		if (buf[i]=='.') n.isint=0;
		if ((buf[i]=='e') || (buf[i]=='E'))
		{
			if (k==0) macro_error (1,i); else n.isint=0;
			
			/* Right after an "E", a minus sign is allowed.
			   If we have a minus token, convert it back. */
			
			if ((unsigned char)(buf[i+1])==MINUS_TOK) buf[i+1]='-';
			
		}
		s[k++] = buf[i++];
	} 
	while (isnum2(buf[i]));
	*idx=i;	
	
	if (n.isint)
		n.intvalue = atoi(s);
	else
		n.realvalue = atof(s);
	
	return n;
}


/* Extract a string, " delimited. We assume that
   idx points to the opening " . Result must point
   to a valid memory location. Strings are limited to
   maxlen chars. */

void extract_string (char* buf, int* idx, char* result, int maxlen)
{
int j,k;

	j = *idx;
	if (buf[j++]!='"') macro_error (4,j);
	
	/* Extract the string */
	
	k=0;
	while ((buf[j]!='"') && (k<maxlen)) result[k++]=buf[j++];
	result[k]=0;
	while (buf[j]!='"') j++;
	*idx = j+1;	/* skip over the closing " */
}



/* If we run across a comment symbol, #, we need to advance to next eoln */

int skipcomment (char* src, int *idx, int len)
{
int i;

	i = *idx;
	if (src[i]=='#')
	{
		while ((src[i]!='\n') && (i<len)) i++;
		*idx=i+1;
		return 1;	/* Skipped comment */
	}
	return 0; 		/* No comment skipped */	

}

/* Extract an identifier from a string - returns -1 if at end of line
   or an index to the next blank character */

int nextident (char* dest, char* src, int idx, int len)
{
int i,j, flg;
char s[255];

	i = idx;
	memset (s, 0, sizeof (s));
	
	/* Skip possible blank spaces or tabs */
	
	flg = 1;
	while ((flg) && (i < len))
	{
		if (is_space(src[i])) i++; else flg=0;
	}
	if (i>=len) return -1;


	/* The first letter has to be alpha or SYNERR */
	
	if (!isalpha1 (src[i])) return -2;

	/* Advance until eoln, spaces or tabs, collect chars as keyword */

	flg=1; j=0;
	while ((flg) && (i < len))
	{
		if (isalpha2(src[i]))
			s[j++]=src[i++];
		else
			flg=0;
	}

	/* Find a colon (again ignoring spaces in between) */

	while ((i < len) && (is_space(src[i]))) i++;
	if (i>=len) return -1;
	if (src[i] != ':') return -2;
	i++;	/* Point to the next character after the colon. */
	
	memcpy (dest, s, strlen(s)+1);	/* Include closing zero */
	
	if (i>=len) return -1;

	return i;

}


/* Parse a keyword (find in keyword list). If found, return the token#
   and advance idx by the length of the token. Otherwise, return -1 */

int parsenext (char* src, int* idx, int len)
{
int i, token, flg;

	/* Skip possible blank spaces or tabs */
	
	flg = 1; i=*idx;
	while ((flg) && (i < len))
	{
		if (is_space(src[i])) i++; else flg=0;
	}
	if (i>=len) return -1;

	token=0;
	do
	{
		if (strncasecmp (&src[i], keywordlist[token].keywd, strlen(keywordlist[token].keywd))==0)
		{
			*idx += strlen(keywordlist[token].keywd);
			return keywordlist[token].token;
		}
		token++;
	}
	while (token < keywordlist_size);

	return -1;					/* Not found error */
}


/* Eliminate comments, tabs, linefeeds etc */

void bufprep (char* buf, int len)
{
int i;

	i=0;
	while (i<len)
	{
		if (buf[i]=='#')
		{
			while ((i<len) && (buf[i]!='\n'))
			{
				buf[i]=' '; i++;
			}
		}
		if ((buf[i]==0x09) || (buf[i]==0x0d) || (buf[i]==0x0a) || (buf[i]=='\n') || (buf[i]=='\t')) buf[i]=' ';
		i++;
	}
}


/* Parse the keywords, replace keywords by tokens. Also eliminate multiple spaces */


int parse (char* buf, int len)
{
char* dest;
int i,j,k, token, err;


	/* The dest buffer holds as much data as the source buffer - at worst */
	
	dest = (char*) calloc (len, sizeof (char));
	
	i=0; j=0; err=0;
	while (i<len)
	{
	
		/* See if we have a string here */
		
		if (buf[i]=='"')	/* then copy verbatim, include spaces etc */
		{
			dest[j++] = buf[i++];	/* opening " */
			while ((i<len) && (buf[i]!='"')) dest[j++] = buf[i++];
			dest[j++] = buf[i++];	/* closing " */
			if (i>len) return -1;
		}
		else
		{		
			/* Try to find a keyword at the current position */
	
			k=0; token=0;
			do 
			{
				if ((keywordlist[k].keywd!=NULL) &&
					(strncasecmp (&buf[i], keywordlist[k].keywd, strlen(keywordlist[k].keywd))==0))
				{
					token = keywordlist[k].token;
				}
				else
					k++;
			}
			while ((token==0) && (k<keywordlist_size));

			if (token)	/* yes we found one */
			{
				dest[j++] = token;
				i += strlen(keywordlist[k].keywd);
			}
			else		/* Not found, copy verbatim */
			{
				if (/*(j>0) && (dest[j-1] == ' ') && */(buf[i]==' '))
				{
					i++;	/* do not copy (multiple) blank spaces */
				}
				else
				{
					if ((buf[i] & 0x80) !=0)			/* Unusual ASCII symbol... */
					{
						dp (1, "Parser warning: Illegal symbol %d ignored\n", buf[i]);
						i++;				/* Ignore it */
						err |=1;			/* and prepare an error flag */
					}
					else
						dest[j++] = buf[i++];
				}
			}
		}
	}
	
	/* hand back the parsed result */

	for (i=0; i<len; i++) buf[i]=dest[i];	
	free (dest);
	return err;

}


/* the opposite: Replace tokens by text strings.
	The parsed macro text is in buf; text will be copied from
	idx1 to idx2 and the result placed in list */


void listbuf (unsigned char* buf, int idx1, int idx2, char* list, int lim)
{
int i,j,k;
char s[512];

	k=0; list[0]=0;
	for (i=idx1; i<=idx2; i++)
	{
		if (buf[i] >= 128)
		{
			j=0; 
			while ((j<128) && (buf[i]!=keywordlist[j].token)) j++;
			if (j<128)
			{
				snprintf (s, 255, " %s ", keywordlist[j].keywd);
				strncat (list, s, lim);
			}
			else
				dp (1, "WARNING: Listbuf: Error finding token %d\n",buf[i]);
		}
		else
		{
			snprintf (s, 255, "%c",buf[i]);
			strncat (list,s, lim);
		}
	}
}


int forward_find_semicolon (unsigned char *buf, int idx1, int idx2)
{
int i,found;

	i=idx1; found=0;

	while ((!found) && (i<idx2))
	{
		i++;
		if (buf[i]==ENDPROC_TOK) found=1;
		if (buf[i]==ENDMAC_TOK) found=1;
		if (buf[i]==';') found=1;
		if (buf[i]=='"')
		{
			do i++; while (buf[i]!='"');	/* Skip to closing quotes */
		}
	}
	return i;
}
		




/*****************************************************************

	Variable handling
	
******************************************************************/

/* findvar returns the index into the variable table or -1 if not there */


int findvar (char* varname)
{
int i;

	i=0;
	while (i<var_max)
	{
		if (!strncasecmp (varname, vars[i].name,16))
			return i;
		else 
			i++;
	}
	return -1;

}


/* addvar enters a variable into the list */

int addvar (variable_type v)
{
	if (var_max >= 127) macro_error (7,0);

	vars[var_max] = nullvar;
	strncpy (vars[var_max].name, v.name, 16);
	vars[var_max].type = v.type;
	
	if (v.type==MT_REAL)
	{
		vars[var_max].numvalue = v.numvalue;	/* copy numerical value */
		if (v.dimension==0)
		{
			vars[var_max].dimension=0;
			vars[var_max].arrayptr=NULL;
			dp (2, "added numerical variable %s at position %d, value %f\n",
				vars[var_max].name,var_max, vars[var_max].numvalue);
		}
		else		/* enter an array - memory is set aside by DIM */
		{
			vars[var_max].dimension=v.dimension;
			vars[var_max].numelems = v.numelems;
			vars[var_max].arrayptr = v.arrayptr;
			vars[var_max].xmax=v.xmax; vars[var_max].ymax=v.ymax; vars[var_max].zmax=v.zmax; 
			dp (2, "added numerical array %s at position %d, pointer %08lX\n",
				vars[var_max].name,var_max, (long)vars[var_max].arrayptr);
		}
	}
	if (v.type==MT_INT)
	{
		vars[var_max].intvalue = v.intvalue;	/* copy numerical value */
		vars[var_max].dimension=0;
		vars[var_max].arrayptr=NULL;
			dp (2, "added integer variable %s at position %d, value %li\n",
				vars[var_max].name,var_max, vars[var_max].intvalue);
	}
	else if (v.type==MT_IMG)
	{
		vars[var_max].numvalue = 0.0;
		vars[var_max].dimension=0;
		memcpy (&vars[var_max].imgptr, &v.imgptr, sizeof (v.imgptr));
		dp (2, "added image variable %s at position %d, size %lix%li, pointer %08lX\n",
			vars[var_max].name,var_max, vars[var_max].imgptr.xmax,vars[var_max].imgptr.ymax, (long)vars[var_max].imgptr.data);
	}	
	else if (v.type==MT_STRING)
	{
		vars[var_max].numvalue = 0.0;
		vars[var_max].dimension=0;
		vars[var_max].strg = v.strg;
		dp (2, "added string variable %s at position %d, pointer %08lX, string %s\n",
			vars[var_max].name,var_max, (long)vars[var_max].strg, vars[var_max].strg);
	}
	
	var_max++;
	return var_max-1;
}


int deletevar (variable_type v)
{
int i,j;

	i = findvar (v.name);
	if (i<0) return -1;
	
	if (vars[i].type==MT_IMG) 
	{
		dp (2, "Freeing image at pointer %08lX\n",(long)vars[i].imgptr.data);
		freebuf (&vars[i].imgptr);
	}
	else if (vars[i].type==MT_STRING)
	{
		if (vars[i].strg)
		{
			dp (2, "Freeing string buffer at pointer %08lX\n",(long)vars[i].strg);
			free (vars[i].strg); vars[i].strg=NULL;
		}
		else
			dp (1, "ERROR: Null pointer in string variable!\n");
	}
	else if (vars[i].type==MT_REAL)
	{
		if ((vars[i].dimension) && (vars[i].arrayptr))
		{
			dp (2, "freeing numerical array at %08lX\n",(long)vars[i].arrayptr);
			free (vars[i].arrayptr);
			vars[i].arrayptr=NULL;
		}
	}
	/* else if (vars[i].type==MT_INT) has nothing to free */
	
	dp (2, "Disposing of variable %s at position %d\n",vars[i].name, i);
	
	if ((var_max==1) || (i==var_max-1))
	{
		vars[i]=nullvar;
	}
	else
	{
		for (j=i+1; j<var_max; j++)
			vars[j-1] = vars[j];
		vars[var_max-1] = nullvar;
	}
	var_max--;
	return 0;
}


void free_all_vars()
{
int i;

	if (var_max>0)
	{
		for (i=0; i<var_max; i++)
		{
			dp (2, "Freeing variable %s ",vars[i].name);
			if (vars[i].type==MT_IMG)
			{
				dp (2, "image (freeing buffer at pointer %08lX, too)", (long)vars[i].imgptr.data);
				freebuf (&vars[i].imgptr);
			}
			else if (vars[i].type==MT_STRING)
			{
				dp (2, "string (freeing buffer at pointer %08lX, too)", (long)vars[i].strg);
				free (vars[i].strg);
			}
			else
			{
				if (vars[i].type==MT_REAL)
					dp (2, "Numerical value %f",vars[i].numvalue);
				else
					dp (2, "Integer value %li",vars[i].intvalue);
				if ((vars[i].dimension) && (vars[i].arrayptr))
				{
					dp (2, "Freeing array at %08lX, too ",(long)vars[i].arrayptr);
					free (vars[i].arrayptr);
				}
			}
			vars[i] = nullvar;

			dp (2, "  --freed\n");
		}
	}
	var_max=0;
}


/* Storevar updates a variable value or adds the variable if new */

void storevar (variable_type v)
{
int i;

	i = findvar (v.name);
	if (i<0)				/* create new */
	{
		addvar (v);
	}
	else
	{
		if (v.type!=vars[i].type)	/* Type mismatch error */ 
			macro_error (6,i);
		
		if (v.type==MT_REAL)
		{
			vars[i].numvalue = v.numvalue;
			dp (2, "Numerical variable %s updated to %f\n",v.name, vars[i].numvalue);
		}
		else if (v.type==MT_INT)
		{
			vars[i].intvalue = v.intvalue;
			dp (2, "Integer variable %s updated to %li\n",v.name, vars[i].intvalue);
		}
		else if (v.type==MT_IMG)
		{
			dp (2, "Freeing image at pointer %08lX\n",(long)vars[i].imgptr.data);
			freebuf (&vars[i].imgptr);
			memcpy (&vars[i].imgptr, &v.imgptr, sizeof (v.imgptr));
			dp (2, "Image variable %s reassigned. Pointer %08lX\n",v.name, (long)v.imgptr.data);
		}
		else if (v.type==MT_STRING)
		{
			dp (2, "Freeing string at pointer %08lX\n",(long)vars[i].strg);
			free (vars[i].strg);
			vars[i].strg = v.strg;
			dp (2, "String variable %s reassigned. Pointer %08lX\n",v.name, (long)v.strg);
		}
		
	}

}



/* This function interprets an array subscript, e.g.
   image[120,6,0] and returns the subscript and the 
   dimension. For example [100,0] returns x=100, y=0, 
   z=undefined and dim=2.
   We expect the index to point just past the opening [,
   and this function will skip beyond the closing ]
*/

subscript_type extract_subscript (char* buf, int* idx)
{
int i;
variable_type v;
subscript_type s;

	i = *idx;
	v = expr (buf, &i, MT_INT);	/* We must have at least ONE dimension */
	s.x = v.intvalue;
	s.y=0; s.z=0;	/* Have them at defined value */
	s.dim=1;
	
	if (buf[i]==',')
	{
		i++;
		v = expr (buf, &i, MT_INT);	
		s.y = v.intvalue;
		s.dim++;
		if (buf[i]==',')
		{
			i++;
			v = expr (buf, &i, MT_INT);
			s.z = v.intvalue;
			s.dim++;
		}
	}
	
	/* There must be a ']' here... */
	
	if (buf[i++]!=']') macro_error (1,i);
	*idx=i;
	return s;
		
}



/* Extract a variable name from a string. We assume that
   the first char has already been validated. */

variable_type extract_var (char* buf, int* idx)
{
int i,k;
char s[256];
variable_type v;
subscript_type d;

	i = *idx; k=0; memset (s, 0, sizeof(s));
	d.dim=0; v = nullvar;
	
	do
	{
		s[k++] = buf[i++];
	} 
	while (isalpha2(buf[i]));
	
	/* Check if it is a subscripted variable */
	
	if (buf[i]=='[')	/* Yes, get subscript */
	{
		i++;
		d = extract_subscript(buf, &i);
	}
	
	*idx=i;	
	
	/* A special case is the "user" array, which we trap here. */
	
	if (!strcmp (s,"user"))
	{
		/* Must be subscripted */
		
		if (d.dim==0) macro_error (1,i);
		if (d.dim>1) macro_error (20,i);
		if ((d.x<0) || (d.x>99)) macro_error (21,i);
		v.type=MT_REAL;
		v.numvalue = mac_user[d.x];
		return v;
	}


	i = findvar (s);
	if (i<0)		/* not found, create generic variable with zero value */
	{
		if (d.dim>0) macro_error (20,i);	/* Can't generate array variable here */
		strncpy (v.name, s, 16);
		v.type = MT_REAL;
		v.numvalue = 0.0;
		v.dimension=0;
		i=addvar (v);
	}
	
	/* get the value */
	
	v.type = vars[i].type;
	strncpy (v.name, vars[i].name, 16);

	/* Case 1. The variable is an image. Return the image (cloned!) if
	   direct use; if subscripted, return the pixel value */

	if (v.type==MT_IMG)
	{
		if (d.dim==0)	/* Full image */
			clone_image (&vars[i].imgptr, &v.imgptr);
		else
		{
			if (d.dim==1) macro_error (21,i);
			if ((d.x<0) || (d.y<0) || (d.z<0) 
				|| (d.x>vars[i].imgptr.xmax-1) 
				|| (d.y>vars[i].imgptr.ymax-1) 
				|| (d.z>vars[i].imgptr.zmax-1))
				macro_error (21,i);

			/* Now the type is really MT_REAL */
			
			v.numvalue = readbuf_flt (&vars[i].imgptr,d.x,d.y,d.z);
			v.type=MT_REAL;
				
		}
	}
	else if (v.type==MT_STRING)
	{
		/* Case 2. The variable is a string. Clone the string for direct
		   use or return the ASCII value as INT (1-dimensional array) */

		if (d.dim==0)
		{
			v.strg = calloc (strlen (vars[i].strg)+1, sizeof (char));
			strcpy (v.strg, vars[i].strg);
		}
		else
		{
			if (d.dim!=1) macro_error (20,i);
			if ((d.x<0) || (d.x > strlen (vars[i].strg))) macro_error (21,i);
			v.intvalue = vars[i].strg[d.x];
			v.type = MT_INT;
		}
	}
	else if (v.type==MT_REAL)
	{
		/* Case 3. The variable is a number array. Subscripting is
		allowed only in the case of arrays. */
		
		if ((d.dim==0) && (vars[i].dimension==0))
		{
			v.numvalue = vars[i].numvalue;
			v.type = MT_REAL;
		}
		else if (d.dim==vars[i].dimension)		/* Dimensions must match */
		{
			int idx = d.x + vars[i].xmax*(d.y + vars[i].ymax*d.z);
			if (idx > vars[i].numelems) macro_error (21,i);
			v.numvalue = vars[i].arrayptr[idx];
			v.type = MT_REAL;
		}
		else
			macro_error (20,i);
	}
	else 
	{
		/* Case 4. The only remaining case is a plain integer variable */
		
		v.intvalue = vars[i].intvalue;
		v.type = MT_INT;
	}
	
	return v;
}



void macro_swapvar (variable_type *x, variable_type *y)
{
variable_type v;

	memcpy (&v, x, sizeof (variable_type));
	memcpy (x, y, sizeof (variable_type));
	memcpy (y, &v, sizeof (variable_type));

}


int img_match_size (image_type* img1, image_type* img2)
{

	if ((img1->xmax==img2->xmax) && (img1->ymax==img2->ymax) && (img1->zmax==img1->zmax)) return 1;
	return 0;
}

/*****************************************************************

	The following two functions handle the dynamic menu and the
	DSO registry of new functions

	load_macro() allows to choose a text file and interprets
	it, copying it into the macro buffer. In addition, the
	macro will be added to the menu.
	
	delete_macros() erases all macros from the menu and
	frees the buffer.
	
******************************************************************/



/* If a new macro is loaded and something of the same name exists,
   delete the old macro first */

void delete_single_macro (char* macroname)
{
int i,j;

	if (nummacros)
	{
		i=0;
		while ((i<nummacros) && (strcasecmp (macroname, macronames[i]))) i++;
		if ((i<nummacros) && (!strcasecmp (macroname, macronames[i])))
		{
			dp (1, "Duplicate macro: Deleting macro %s\n",macronames[i]);
			if (macromenutbl[i]) gtk_widget_destroy ((GtkWidget*)macromenutbl[i]);
			macromenutbl[i]=NULL;
			if (macrodef[i]) free (macrodef[i]);
			macrodef[i]=NULL;
			if (macronames[i]) free (macronames[i]);
			macronames[i]=NULL;
			
			for (j=i+1; j<nummacros; j++)
			{
				macromenutbl[j-1] = macromenutbl[j];
				macrodef[j-1] = macrodef[j];
				macronames[j-1] = macronames[j];
			}
			nummacros--;
			macromenutbl[nummacros]=NULL;
			macronames[nummacros]=NULL;
			macrodef[nummacros]=NULL;
		}
	}

}


/* Delete all macros (Menu: Macro -> Delete macros) */

void delete_macros ()
{
int i;

	if (nummacros)
	{
		for (i=0; i<nummacros; i++)
		{
			if (macromenutbl[i]) gtk_widget_destroy ((GtkWidget*)macromenutbl[i]);
			macromenutbl[i]=NULL;
			if (macrodef[i]) free (macrodef[i]);
			macrodef[i]=NULL;
			if (macronames[i]) free (macronames[i]);
			macronames[i]=NULL;
		}
	}
	nummacros = 0;


}




void load_macro (const char *macfile)
{
FILE* MACF;
char buf[32768];		/* to buffer the input file */
char isproc;			/* Whether a macro (add menu entry) or procedure */
char macname[256];
int buflen,i,j,macrosize,a,b,cnt,parserr,err;
char* transferpointer;
char* toparse;			/* The section of the buffer to parse */


	if (nummacros>=MACMAX)
	{
		ShowMessage ("Macro memory full.");
		return;
	}
	
	if (!macfile)
	{
		ChooseFile ("Select text file with macro definition", window1, 0, (gchar **)&macfile);
		if (!macfile) return;
	}

	buflen=0; memset (buf, 0, sizeof (buf));
	MACF = fopen (macfile, "rb");
	if (MACF)
	{
		buflen=fread (buf, sizeof (char), sizeof (buf), MACF);
		dp (1, "Macro: %d bytes read from file %s\n",buflen, macfile);
		fclose (MACF);
	}
	// g_free (macfile);  // causes crash with -m command-line option
							// Now we have a memory leak with the interactive load.
	
	if (buflen==0)
	{
		ShowMessage ("Macro: No data read. File error?");
		return;
	}
	else if (buflen>=sizeof(buf))
	{
		ShowMessage ("Macro: Can't handle more than 32 kB");
		return;
	}
	
	bufprep (buf, buflen);
		
	/* find an occurrence of the keyword "MACRO" and its matching ENDMACRO
	   or the occurrence of "PROCEDURE" and "ENDPROC" */

	a=0; cnt=0; isproc=0; err=0;
	while (a<strlen(buf))
	{
	
		isproc=-1;
		while ((a<strlen(buf)) && (isproc==-1))
		{
			if (!strncasecmp(&buf[a],"macro",5)) isproc=0;
			else if (!strncasecmp(&buf[a],"procedure",9)) isproc=1;
			else
				a++;
		}

		if (isproc)
		{
			b=a+9;
			while ((b<strlen(buf)) && (strncasecmp(&buf[b],"endproc",7))) b++;
		}
		else
		{
			b=a+5;
			while ((b<strlen(buf)) && (strncasecmp(&buf[b],"endmacro",8))) b++;
		}
		
		if (a+5<=strlen(buf))
		{
			if (b>=strlen(buf))
			{
				ShowMessage ("Macro: No matching 'endmacro' or 'endproc' found");
				return;
			}

			if (isproc) i=a+9; else i=a+5;
			
			/* Now extract the macro name. */
			j = nextident (macname, buf, i, strlen(buf));
			if (j<0)	/* error occured extracting the name */
			{
				ShowMessage ("Macro: Syntax error in macro definition");
				return;
			}

			/* at this point, we have an approximate idea
			   of the buffer size */

			macrosize = b-a+20;
			dp (1, "Macro name %s has approximately %d bytes\n",macname,macrosize);

			/* Parse the buffer to identify all tokens, saves run time */
			
			toparse = (char*)calloc (macrosize, sizeof(char));
			if (isproc)
				for (i=j; i<=b+8; i++) toparse[i-j] = buf[i];
			else
				for (i=j; i<=b+9; i++) toparse[i-j] = buf[i];
			
			parserr = parse (toparse, strlen (toparse));
			if (parserr>0) err |= parserr;
			if (parserr<0)
			{
				ShowMessage ("Parse error. Macro parsing aborted");
			}

			/* Check if a macro of this name already exists and delete it */
			
			delete_single_macro (macname);

			/* Allocate memory and copy the actual macro. Please use calloc
			   here, because we want the zeroes ending the string */

			macrosize = strlen (toparse);
			dp (1, "Macro %s has exactly %d bytes\n",macname,macrosize);
			macrodef[nummacros] = calloc (macrosize+10,1);

			transferpointer = macrodef[nummacros];
			for (i=0; i<=macrosize; i++)
				transferpointer[i]= toparse[i];

			free (toparse);
			
			/* Also copy the macro name */
			
			macronames[nummacros] = calloc (strlen(macname)+10,sizeof(char));
			transferpointer = macronames[nummacros];
			strcpy (transferpointer, macname);

			/* Enter the new macro into the dynamic menu (not if only a proc) */

			if (!isproc)
			{
				macromenutbl[nummacros] = Menuitem (window1, mmacros_menu, NULL, macname, "m_delmacro");
				gtk_signal_connect (GTK_OBJECT (macromenutbl[nummacros]), "activate",
        				GTK_SIGNAL_FUNC (exec_macro), &macroopts[nummacros]);
			}
			else
				macromenutbl[nummacros] = NULL;

			nummacros++; cnt++;
		}
		a=b+8;
	}
	if (cnt==0)
		ShowMessage ("Macro: No macro definition found ");
	if (err & 1)
		ShowMessage ("Parse warning: Illegal characters ignored. Check your input file.");
}


int register_dso_macro (char* kwd, int numparams, MT_TYPE type, void* apointer)
{
int i;

	i=0;
	while ((i<keywordlist_size) && (keywordlist[i].keywd!=NULL)) i++;
	if (i<keywordlist_size)
	{
		keywordlist[i].keywd=kwd;
		keywordlist[i].numparms = numparams;
		keywordlist[i].type = type;
		keywordlist[i].apointer = apointer;
		dp (2, "DSO/Macro: Registered keyword %s as token %d\n",kwd,keywordlist[i].token);
		return 0;
	}
	else
		return -1;	/* Ran out of space */

}

/*****************************************************************

	This is the top level of the macro interpreter

	exec_macro () gets executed whenever one of the dynamically
	loaded menu items gets activated. exec_macro () invokes
	the actual macro interpreter.
	
******************************************************************/


/* This is the arithmetic expression handler.
   Unfortunately, this is the "dumb" version which does
   right-to-left evaluation without operator precedence.
   It is deprecated and should not be used.
*/

int expr_recursion_level;

variable_type old_expr (unsigned char* buf, int* idx, MT_TYPE expctype)
{
int i, isexpr, status, tkx;
variable_type x,y;
num_type f;
char strbuf[1024];
variable_type (*mac_funcpointer)(unsigned char* s, int* i) = NULL;
variable_type (*mac_dyadic)(variable_type x, variable_type y) = NULL;


	i = *idx; isexpr=1; status=0; expr_recursion_level++;
	x=nullvar; y=nullvar;
	
	do
	{
		if ((buf[i]==';') || (buf[i]==',') || (buf[i]==']'))
			isexpr=0; 		/* these always end an expression */
		else if (buf[i]==')')
		{
			isexpr=0;
		}
		else if (buf[i]=='(')		/* Open parentheses, recurse expr */
		{
			i++;					/* skip over parentheses */
			x = expr (buf, &i, MT_NULL);	/* Compute expression in () recursively */
			status=1;				/* Treat like a constant or variable */
			if (buf[i++]!=')') macro_error (1,i);	/* Skip over the left-over ) */
		}
		else if (buf[i]=='"')		/* String constant */
		{
			if ((expctype!=MT_NULL) && (expctype != MT_STRING)) macro_error (6,i);
			if (status==1) macro_error (8,i);
			extract_string (buf, &i, strbuf, 1023);
			x.type = MT_STRING;
			x.strg = calloc (strlen (strbuf)+1, sizeof (char));
			strcpy (x.strg, strbuf);
			status=1;				/* now we expect a dyadic operator */
		}
		else if (isnum1(buf[i]))		/* Numerical constant */
		{
			if ((expctype!=MT_NULL) && (expctype != MT_REAL)) macro_error (6,i);
			if (status==1) macro_error (8,i);
			f = extract_num (buf, &i);
			if (f.isint)
			{
				x.type = MT_INT; 
				x.intvalue = f.intvalue;
			}
			else
			{
				x.type = MT_REAL; 
				x.numvalue = f.realvalue;
			}
			status=1;				/* now we expect a dyadic operator */
		}
		else if (isalpha1(buf[i]))	/* variable reference */
		{
			if (status==1) macro_error (8,i);
			x = extract_var (buf, &i);
			if ((expctype!=MT_NULL) && (x.type != expctype)) macro_error (6,i);
			status=1;				/* now we expect a dyadic operator */
		}
		else if (buf[i]>=128)		/* function reference */
		{
			tkx = buf[i]-128;
			if (status==1)			/* must be dyadic function */
			{
				if (keywordlist[tkx].numparms>=0) macro_error (8,i);
				i++;	/* skip over the operator token */
				y = expr (buf,&i,MT_NULL);	/* don't check types here */
				
				/* Check the type. If types match OR if the left-hand
				   side (x) is an image and the right-hand side is a
				   numvalue, it is acceptable */
				 
				if ((x.type != y.type) && ( (x.type!=MT_IMG) || (y.type!=MT_REAL)))
					macro_error (6,i);
				
				/* now execute the dyadic function */
				mac_dyadic = keywordlist[tkx].apointer;
				if (mac_dyadic) x = mac_dyadic (x,y);
				status=0;
			}
			else					/* must be a monadic function */
			{
				i++;	/* skip over the operator token */
				if (keywordlist[tkx].numparms<0) macro_error (9,i);
				else if (keywordlist[tkx].numparms>0)
				{
					if (buf[i++] != '(') macro_error (2,i);
				}
				mac_funcpointer = keywordlist[tkx].apointer;
				if (mac_funcpointer) x=mac_funcpointer (buf, &i);
				status=1;
			}
				
		}
		else
			macro_error (1,i);		/* Syntax error */
	
	}
	while (isexpr);

	*idx = i; expr_recursion_level--;
	return x;
}


/********************************************************************

This is the full-blown expression parser with operator precedence.
The idea is to keep lower-priority operations on a stack while
immediately executing higher-priority operations. As a general
rule, operands and operators must occur alternatingly. Whenever we
read an operator, we check the previous operator's precedence and
either enter the operation into the stack (rising precedence) or
execute off the stack (lower or equal precedence). We start with
a precedence of 0. In an example, this might look like:

Operation: 15 + 3 * 2 - 2;

Object	Action	Stack.X  Stack.Y   Stack.Z	comments

   15			15							Read an operand into X
   +	Enter			15 (+)				Precendence 2 is higher than 0
   3			3		15 (+)				Read next operand into X
   *	Enter			3  (*)		15 (+)	Precedence of 3 is higher than 2
   2			2		3  (*)		15 (+)	Read next operand into X
   -	Exec	6		15 (+)				Precedence drops, so execute all higher-precedence operations
   		Enter			6  (-)		15 (+)	Make room for the next operation
   2			2		6  (-)		15 (+)
   End	Exec	4		15 (+)				Note: Y-X. Execute whatever is
   		Exec	19							left on the stack. Result in X.
		
Note that we need one more stack register than precedence levels. */



/* for debugging: Print the opstack's contents */

void print_opstack (exprstack_type opstk[], int opstack_last)
{
int i;

	for (i=0; i<=opstack_last; i++) 
	{
		if (opstk[i].v.type==MT_REAL)
			dp (2, "   Op %d Value %f, operation %s\n", i,opstk[i].v.numvalue,keywordlist[opstk[i].op].keywd);
		else if (opstk[i].v.type==MT_INT)
			dp (2, "   Op %d Value %li, operation %s\n", i,opstk[i].v.intvalue,keywordlist[opstk[i].op].keywd);
		else if (opstk[i].v.type==MT_STRING)
			dp (2, "   Op %d Value %s, operation %s\n", i,opstk[i].v.strg,keywordlist[opstk[i].op].keywd);
		else if (opstk[i].v.type==MT_IMG)
			dp (2, "   Op %d (image), operation %s\n", i,keywordlist[opstk[i].op].keywd);
	}
}


/* ENTER function - shift the op stack up by one element; associate an operation
   with the second-lowest element. The lowest element is blank */

int expr_enter (exprstack_type opstk[], int opstack_last, int opidx, int prec)
{
int i;

	for (i=MAXPREC-1; i>=0; i--)
		opstk[i+1] = opstk[i];
	opstk[1].op = opidx;
	opstk[1].precedence = prec;
	opstk[0].op=0;
	opstk[0].precedence=0;
	opstk[0].v = nullvar;
	return opstack_last+1;

}


/* EXEC function - execute the lowest dyadic function on the stack.
   The lowest elements become invalid, the stack gets shifted down
   and the result is stored in the lowest element */

int expr_exec (exprstack_type opstk[], int opstack_last, int bufidx)
{
variable_type x;
int i,tkx;			/* Index into token list */
variable_type (*mac_dyadic)(variable_type x, variable_type y) = NULL;

	x = nullvar;
	if (opstk[1].op > 0)	/* If not, no operation in stack */
	{
		tkx = opstk[1].op;
		if (keywordlist[tkx].numparms>=0) macro_error (8,bufidx);

		/* now execute the dyadic function */
		mac_dyadic = keywordlist[tkx].apointer;
		if (mac_dyadic) x = mac_dyadic (opstk[1].v, opstk[0].v);
		
		/* Pop the stack and store x */
		
		for (i=1; i<=MAXPREC; i++)
			opstk[i-1] = opstk[i];
		
		opstk[0].v = x;
	}
	/* Should there be an error call here ? */
	
	return opstack_last-1;

}


/* Get one object from the buffer - this may be either a variable,
   a numerical or string constant, or a function.
   Note. We allocate memory here, so the caller is responsible for freeing it */


variable_type expr_getobject (unsigned char* buf, int* idx)
{
variable_type x;
int i,tkx;
num_type f;
char strbuf[1024];
variable_type (*mac_funcpointer)(unsigned char* s, int* i) = NULL;

	i = *idx; 
	x=nullvar;
	
	if (buf[i]=='"')		/* String constant */
	{
		extract_string (buf, &i, strbuf, 1023);
		x.type = MT_STRING;
		x.strg = calloc (strlen (strbuf)+1, sizeof (char));
		strcpy (x.strg, strbuf);
	}
	else if (isnum1(buf[i]))		/* Numerical constant */
	{
		f = extract_num (buf, &i);
		if (f.isint)
		{
			x.type = MT_INT; 
			x.intvalue = f.intvalue;
		}
		else
		{
			x.type = MT_REAL; 
			x.numvalue = f.realvalue;
		}
	}
	else if (isalpha1(buf[i]))	/* variable reference */
	{
		x = extract_var (buf, &i);
	}
	else if (buf[i]>=128)		/* monadic function reference */
	{
		tkx = buf[i]-128;
		if (keywordlist[tkx].numparms<0) macro_error (9,i);	/* No dyadic op allowed */
		i++;	/* skip over the operator token */
		if (keywordlist[tkx].numparms>0)
		{
			if (buf[i++] != '(') macro_error (2,i);
		}
		mac_funcpointer = keywordlist[tkx].apointer;
		if (mac_funcpointer) x=mac_funcpointer (buf, &i);
	}
	else
		macro_error (1,i);		/* Syntax error */
	
	/* and return the buffer pointer as well as the object */
	
	*idx=i;
	
	return x;
}



variable_type new_expr (unsigned char* buf, int* idx, MT_TYPE expctype)
{
int i, j, isexpr,status,tkx,prev_prec,precedence;
variable_type x;
exprstack_type opstk[MAXPREC+1];		/* The operator stack. One more element than precedence levels */
int opstack_last;


	i = *idx; isexpr=1; status=0; prev_prec=0; expr_recursion_level++;
	opstack_last=0; x=nullvar;
	for (j=0; j<=MAXPREC; j++) 
	{
		opstk[j].v=nullvar;
		opstk[j].op=0;
		opstk[j].precedence=0;
	}
	
	do
	{
		if ((buf[i]==';') || (buf[i]==',') || (buf[i]==']'))
			isexpr=0; 		/* these always end an expression */
		else if (buf[i]==')')
		{
			isexpr=0;
		}
		else if (buf[i]=='(')		/* Open parentheses, recurse expr */
		{
			i++;					/* skip over parentheses */
			x = new_expr (buf, &i, MT_NULL);	/* Compute expression in () recursively */
			if (buf[i++]!=')') macro_error (1,i);	/* Skip over the left-over ) */
			opstk[0].v = x;
			status=1;		/* We need to follow this by a dyadic operator */
		}
		else if ((buf[i]==MINUS_TOK) && (status==0)) /* Trap monadic minus */
		{
			dp (2, "Monadic minus at position %d\n",i);
			i++;		/* Skip over the token */
			if ((isnum1 (buf[i])) || (isalpha1(buf[i])))		/* Constant or variable ref */
			{
				x = expr_getobject (buf, &i);
				if ((x.type!=MT_INT) && (x.type!=MT_REAL)) macro_error (1,i);
				x.intvalue = -x.intvalue; x.numvalue = -x.numvalue;
				opstk[0].v=x;
				status=1;
			}
			else if (buf[i]=='(')
			{
				i++;								/* skip over parentheses */
				x = new_expr (buf, &i, MT_NULL);	/* Compute expression in () recursively */
				if (buf[i++]!=')') macro_error (1,i);	/* Skip over the left-over ) */
				if ((x.type!=MT_INT) && (x.type!=MT_REAL)) macro_error (1,i);
				x.intvalue = -x.intvalue; x.numvalue = -x.numvalue;
				opstk[0].v = x;
				status=1;		/* We need to follow this by a dyadic operator */
			
			}
			else
				macro_error (1,i);
		}
		else if ((isnum1(buf[i])) || (isalpha1(buf[i])) || (buf[i]=='"'))		/* Operand */
		{
			if (status!=0) macro_error (8,i);
			x = expr_getobject (buf, &i);
			opstk[0].v = x;
			status=1;		/* We need to follow this by a dyadic operator */
		}
		else if (buf[i]>=128)		/* function reference; monadic OR dyadic */
		{
			tkx = buf[i]-128;
			if (status==0)			/* we expect a monadic function */
			{
				if (keywordlist[tkx].numparms<0) macro_error (9,i);
				x = expr_getobject (buf, &i);
				opstk[0].v = x;
				status=1;
			}
			else					/* We expect a dyadic operator */
			{
				if (keywordlist[tkx].numparms>=0) macro_error (8,i);
				precedence = abs(keywordlist[tkx].numparms);
				dp (2, "Operator %s, precedence %d\n", keywordlist[tkx].keywd, precedence);
				i++; /* Skip over the operator token */

				/* Now distinguish between two cases. If the new precedence is
				   higher than the previous operator's precedence, ENTER the operation
				   Otherwise, EXEC off the stack until the stack's operator
				   precedence drops below the current one; then ENTER the last op */
				
				if (precedence > prev_prec)
				{
					opstack_last=expr_enter (opstk,opstack_last,tkx,precedence);
					prev_prec = precedence;
					dp (2, "ENTER: %d elements on opstack\n",opstack_last);
				}
				else
				{
					while (opstk[1].precedence >= precedence)
					{
						dp (2, "Executing %s\n", keywordlist[opstk[1].op].keywd);
						opstack_last=expr_exec (opstk,opstack_last,i);
					}
					opstack_last=expr_enter (opstk,opstack_last,tkx,precedence);
					dp (2, "ENTER: %d elements on opstack\n",opstack_last);
					prev_prec = precedence;
				}
				
				status=0;	/* done - expect the next object */

			}
		}

	}
	while (isexpr);
	
	/* If there is anything left in the op stack, we need to execute it */
	
	dp (2, "POST: %d opstack elements\n",opstack_last);

	if (opstack_last>0)
	{
		print_opstack (opstk, opstack_last);
	
		while (opstack_last>0)
		{
			dp (2, "POST: Executing %s\n", keywordlist[opstk[1].op].keywd);
			opstack_last=expr_exec (opstk,opstack_last,i);
			print_opstack (opstk, opstack_last);
		}
		
		x = opstk[0].v;
	}
	
	
	/* If a REAL was requested and all we have is an INT, fix it.
	   Otherwise, check if we got the expected type. Error if not. */
	
	if ((expctype==MT_REAL) && (x.type==MT_INT))
	{
		x.numvalue = (float)x.intvalue;
		x.type=MT_REAL;
	}
	else if ((expctype!=MT_NULL) && (expctype!=x.type))
		macro_error (6,i);

	/* This is a bad hack. Expr always returns scalars, so set dimension=0 
	   The problem should be fixed by now. */
	
	if (x.dimension>0) dp (1, "Warning: Expr returned non-scalar. Resetting.\n");
	x.dimension=0;

	*idx = i; expr_recursion_level--;
	return x;

}

/* Select the expr algorithm */

variable_type expr (unsigned char* buf, int* idx, MT_TYPE expctype)
{
	return new_expr (buf, idx, expctype);
}


/*******************************************************************/


/* Execute one macro until we find the closing ; */

int exec_macro_single (unsigned char* buf, int* idx)
{
int i,j, tkx, varidx, aidx;
char varname[16];
variable_type v;
subscript_type d;
MT_TYPE vartype;
variable_type (*mac_funcpointer)(unsigned char* s, int* i) = NULL;
char clistbuf[2048];
gchar yesno;

	

	if (debuglevel>=2) 
	{
		i = *idx; j = forward_find_semicolon (buf,i,strlen(buf));
		listbuf (buf, i, j, clistbuf, 2040);
		dp (2, "Single macro %s, index %d\n",clistbuf,*idx);
	}

	/* test if we have a keyword or variable */
	
	i = *idx; expr_recursion_level=0; 
	varidx=-1; 	/* stop a compiler warning */
	tkx = buf[i];
	if (tkx >= 128)	/* keyword */
	{
		i++; tkx-=128;
		
		/* if it is an immediate-execution keyword (aka procedure),
		   its return type has to be MT_NULL */
		
		if (keywordlist[tkx].type != MT_NULL) macro_error (3,i);
		
		/* After a procedure token, there MUST be an opening parentheses
		   if the proc needs one or more parameters */
		
		if ((keywordlist[tkx].numparms>0) && (buf[i++] != '(')) macro_error (2,i);
		
		/* so it _is_ a procedure - go call it */
		
		mac_funcpointer = keywordlist[tkx].apointer;
		if (mac_funcpointer) mac_funcpointer (buf, &i);
		

	}
	else if (isalpha1 (buf[i]))	/* start of a variable assignment of the type varname = expr */
	{
		j=0; v = nullvar; vartype = MT_NULL;
		while (isalpha2(buf[i]))
		{
			if (j<15) varname[j++]=buf[i];
			i++;
		}
		varname[j]=0;
		
		/* Are we targeting an array element? */
		
		d.dim=0;
		if (buf[i]=='[')
		{
			i++; d = extract_subscript(buf,&i); /* if so, d.dim>0 */
			vartype=MT_REAL;					/* Array element always a # */
			
			/* In this case, we need to make sure that the variable exists */

			varidx = findvar (varname);
			if (varidx<0) macro_error (20,i);
		}
		else		/* Check if the variable exists to match type */
		{
			varidx = findvar (varname);
			if (varidx>=0) vartype = vars[varidx].type;
		}
		
		
		/* Skip over the required equals token and evaluate right-hand side */
		
		if (buf[i++]!=EQUALS_TOK) macro_error(1,i);
		v = expr (buf, &i, vartype);
		
		/* and now store the variable. If subscripted, we need to differentiate
		   between the basic types. Otherwise simply store (update) */

		if (d.dim>0)		/* Element of an array */
		{
			if (vars[varidx].type==MT_IMG)	/* Set a pixel of the image */
			{
				unsigned char e1;
				short e2;
				float e3;
				rgbtriplet e4;
				
				if ((d.x<0) || (d.x > vars[varidx].imgptr.xmax)
						|| (d.y<0) || (d.y > vars[varidx].imgptr.ymax)
						|| (d.y<0) || (d.y > vars[varidx].imgptr.ymax))
					macro_error (21,i);
				
				if (vars[varidx].imgptr.imgtype==B_CHAR)
				{
					e1 = (int)v.numvalue; writebuf (&vars[varidx].imgptr, &e1, d.x,d.y,d.z);
				}
				else if (vars[varidx].imgptr.imgtype==B_SHORT)
				{
					e2 = (int)v.numvalue; writebuf (&vars[varidx].imgptr, &e2, d.x,d.y,d.z);
				}
				else if (vars[varidx].imgptr.imgtype==B_FLOAT)
				{
					e3 = v.numvalue; writebuf (&vars[varidx].imgptr, &e3, d.x,d.y,d.z);
				}
				else if (vars[varidx].imgptr.imgtype==B_RGB)
				{
					e4.red=e4.green=e4.blue = (int)v.numvalue; 
					writebuf (&vars[varidx].imgptr, &e4, d.x,d.y,d.z);
				}
			}
			else if (vars[varidx].type==MT_STRING)	/* Set a string element */
			{
				if (d.dim>1) macro_error (20,i);
				if ((d.x<0) || (d.x>=strlen(vars[varidx].strg))) macro_error (21,i);
				vars[varidx].strg[d.x] = ((int)v.numvalue) & 0x0FF;
			}
			else if (vars[varidx].type==MT_REAL)	/* Set a real array element */
			{
				if (d.dim!=vars[varidx].dimension) macro_error (20,i);
				aidx = d.x + vars[varidx].xmax*(d.y + vars[varidx].ymax*d.z);
				if ((aidx<0) || (aidx>=vars[varidx].numelems)) macro_error (21,i);
				vars[varidx].arrayptr[aidx]=v.numvalue;
				dp (2, "Updated array element %d,%d,%d of %s to %f\n",
						d.x,d.y,d.z,vars[varidx].name, v.numvalue);
			}

		}
		else				/* Complete variable */
		{
			strncpy (v.name, varname, 16);
			storevar (v);
		}
	
	}
	else
		macro_error (1,i);

	/* Update the string index. We should find a ; here */
	
	if (buf[i]!=';') macro_error (5,i);
	*idx=i+1;

	/* Interactive element. Pulse the progress bar & query the master abort button */
	
	indicate_progress (-1);		/* Pulse the progress bar */
	if (master_abort)			/* Did somebody press "ABORT"? */
	{
		master_abort=0;
		yesno_dialog (NULL, "Do you want to abort the macro?\nPress OK to abort,\nor press CANCEL to continue.", &yesno);
		if (yesno==GTK_RESPONSE_APPLY)
		{
			reset_progress ();
			macro_error (28, i);
		}
	}

	return 0;

}



void exec_macro (GtkMenuItem *menuitem, gpointer  user_data)
{
int macronum,i,j,k, rtncode;
unsigned char* buf;

	memset (&nullvar, 0, sizeof (nullvar));
	
	/* Allocate a log buffer for the messages. 128k should be enough for everyone */
	mac_logbuffer = calloc (128000,sizeof (char));
	if (!mac_logbuffer)
		fatal_memerr ();
	else
		mac_logsize = 128000;

	macrolevel=0; forlevel=0; iflevel=0; immd_mode=0; mac_has_output=0;
	macrorun=1;				/* Flag that we are inside a macro */
	master_abort = 0;		/* in case there is an "abort" pending */
	macronum = *(int*)user_data;
	buf = macrodef[macronum];
	sprintf (mac_logbuffer, "Output of macro %s\n", (char*)macronames[macronum]);
	dp (1, "Executing macro %d: %s\n",macronum,(char*)macronames[macronum]);
	i=0;
	do
	{
		j=i+1; k=0; current_macro_pos = i;
		while ((j<strlen(buf)) && (buf[j]!=';')) j++;
		listbuf (buf,i,j,current_macro, 2040);
		
		rtncode=exec_macro_single (buf, &i);

	}
	while ( (i<strlen(buf)) && (rtncode==0));

	reset_progress ();
	macrorun=0;
	dp (1, "\n");

}



/*******************************************************************


	GUI for immediate execution mode


*******************************************************************/


void ix_macro_exec (char* buf)
{
int i,k,j,rtncode;


	i=0;
	do
	{
		j=i+1; k=0; current_macro_pos = i;
		while ((j<strlen(buf)) && (buf[j]!=';')) j++;
		if (j>=strlen(buf)) strcat (buf,";");
		listbuf (buf,i,j,current_macro, 2040);
		
		rtncode=exec_macro_single (buf, &i);
	}
	while ( (i<strlen(buf)) && (rtncode==0));

	macrorun=0;
	dp (1, "\n");
}


GtkEntry* ix_line;
GtkWidget *ix_view2;

void ixCancelClicked (GtkButton * button, gpointer data )
{

	free_all_vars();	/* clear all variables and the logbuffer */
	
	if (mac_logbuffer)
	{
		free (mac_logbuffer);
		mac_logbuffer=NULL;
	}

	/* destroy the window pointed at by data */
	gtk_widget_destroy ( GTK_WIDGET ( data ) );
}


void ixSaveClicked (GtkButton *button, gpointer user_data )
{
gchar* filename;
FILE* FP;
char* p1;
int i;
GtkWidget *view;
GtkTextBuffer *buffer;
GtkTextIter start;
GtkTextIter end;
gchar *text;




	ChooseFile ("Save text as:", window1, 1, &filename);
	if (filename)
	{
		FP = fopen (filename, "rt");
		view = (GtkWidget*)(user_data);
		buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (view));

		/* Obtain iters for the start and end of points of the buffer, then get the text info */
		gtk_text_buffer_get_start_iter (buffer, &start);
		gtk_text_buffer_get_end_iter (buffer, &end);
		text = gtk_text_buffer_get_text (buffer, &start, &end, FALSE);

		FP = fopen (filename, "wt");
		fprintf (FP, "%s", text);
		fclose (FP);
		/* Do I need to free text? */;
		statusbar_display ("Text file written");
		g_free (filename);
	}

}



/* We used to capture the keypress event, but this is no longer possible,
	since we'd disconnect the default keypress handler. The entry would appear 
	"dead" and not accept keypresses. Therefore, this code is obsolete:
	
void ixLine_keypress (GtkWidget *widget, GdkEventKey *event, gpointer user_data)
GtkEntry *entry;
GtkTextBuffer *buffer;
char buf[1024], list[1024], s[1024];
GdkWChar* p;
int i;
GtkTextIter start;
GtkTextIter end;
gchar *text;


	if ((event->keyval==65293)		/ Return key /
		|| (event->keyval==65421))	/ Num-Enter key /
	{
		etc...
*/


/* Instead, pressing "enter" initiates an "active" signal: */

void ixLineChanged (GtkWidget * widget, gpointer data )
{
GtkEntry *entry;
GtkTextBuffer *buffer;
unsigned char list[2048], ptext[2048], s[2080];
//GdkWChar* p;
int i;
GtkTextIter start;
GtkTextIter end;
gchar *text;

	dp (3, "Macro: Entry received an 'activate' event\n");

	/* First, extract the text from the text entry */
	//entry = (GtkWidget*)(user_data);
	entry = GTK_ENTRY(widget);
	text = (gchar *)gtk_entry_get_text (entry);
	strcpy (ptext, text);

	/* Next, execute the line in immediate-mode */
	
	immd_mode=1;
	mac_logbuffer[0]=0;
	parse (ptext, strlen(ptext));
	listbuf (ptext, 0, strlen(ptext), list, 2040);
	dp (1, "Executing single command %s\n", list);
	sprintf (s, "> %s\n", list);
	ix_macro_exec (ptext);
	immd_mode=0;

	/* Housekeeping. Erase the entry line, add text to log window */
	
// 		gtk_text_buffer_delete (buffer, &start, &end);
// 		gtk_entry_set_buffer (entry, buffer);

	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (ix_view2));
	gtk_text_buffer_get_start_iter (buffer, &start);
	gtk_text_buffer_get_end_iter (buffer, &end);
	gtk_text_buffer_insert (buffer, &end, s, -1);
	gtk_text_buffer_get_end_iter (buffer, &end);
	gtk_text_buffer_insert (buffer, &end, mac_logbuffer, -1);
	text[0]=0;
	gtk_entry_set_position (entry, 0);
	gtk_entry_set_text (entry, text);	

}


void macro_gui ( )
{
  GtkWidget *ix_window;
  GtkWidget *vbox4;
  GtkWidget* hbox4;
  GtkWidget *ix_scrolledwindow1;
  GtkWidget *ix_okbutton;
  GtkWidget* ix_savebutton;
  GtkWidget* hs1, *hs2;
  GtkWidget *label1, *label2;
  GtkTextBuffer *buffer;
  GtkTextIter start;
  GtkTextIter end;

  
  const gchar* hellomsg = "ImageBasic ready.\n";
  
  	/* Allocate a log buffer for the messages. */
	mac_logbuffer = calloc (65000,sizeof (char));
	if (!mac_logbuffer) 
		fatal_memerr();
	else
		mac_logsize = 65000;
	mac_has_output=0;

	ix_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (ix_window), "ix_window", ix_window);
	gtk_widget_set_usize (ix_window, 320,300);
	gtk_window_set_title (GTK_WINDOW (ix_window), "ImageBasic immediate execution");

	vbox4 = vbox_in_window (ix_window, "vbox4");

	/* The scrolled window for history display */

	label1 = MakeLabel (ix_window,vbox4,"History / Output","label1",-1,-1);
	ix_scrolledwindow1 = gtk_scrolled_window_new (NULL, NULL);
	gtk_widget_ref (ix_scrolledwindow1);
	gtk_object_set_data_full (GTK_OBJECT (ix_window), "ix_scrolledwindow1", ix_scrolledwindow1,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (ix_scrolledwindow1);
	gtk_box_pack_start (GTK_BOX (vbox4), ix_scrolledwindow1, TRUE, TRUE, 0);

	ix_view2 = gtk_text_view_new ();
	gtk_widget_ref (ix_view2);
	gtk_object_set_data_full (GTK_OBJECT (ix_window), "ix_entry2", ix_view2,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (ix_view2);
	gtk_container_add (GTK_CONTAINER (ix_scrolledwindow1),ix_view2);
	gtk_text_view_set_editable (GTK_TEXT_VIEW (ix_view2), FALSE);
	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (ix_view2));
	gtk_text_buffer_get_start_iter (buffer, &start);
	gtk_text_buffer_get_end_iter (buffer, &end);
	gtk_text_buffer_insert (buffer, &end, hellomsg, -1);

	hs1 = MakeHsep (ix_window,vbox4,"hs1",-1,-1);

	/* Entry field for the command(s) */

	label2 = MakeLabel (ix_window,vbox4,"Command","label2",-1,-1);
	ix_line = GTK_ENTRY(Entry_in_table (ix_window,vbox4,NULL,"ix_line",-1,-1));

	/* Buttons */

	hs2 = MakeHsep (ix_window,vbox4,"hs1",-1,-1);
  	hbox4 = gtk_hbox_new (FALSE, 0);
	gtk_widget_ref (hbox4);
	gtk_object_set_data_full (GTK_OBJECT (ix_window), "hbox4", hbox4,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (hbox4);
	gtk_box_pack_start (GTK_BOX (vbox4), hbox4, FALSE, FALSE, 2);

	ix_savebutton = Button_in_table (ix_window, hbox4, "Save text", "ix_savebutton", -1,-1,NULL);
	ix_okbutton = Button_in_table (ix_window, hbox4, "Cancel", "ix_okbutton", -1,-1,ixCancelClicked);
	gtk_widget_set_usize (ix_okbutton, 48, 22);

	/* We need to connect the save button separately, because we have to pass
	   the char* to the handler function. */

	gtk_widget_set_events (ix_view2, GDK_EXPOSURE_MASK
										| GDK_ENTER_NOTIFY_MASK
										| GDK_LEAVE_NOTIFY_MASK
										| GDK_BUTTON_PRESS_MASK
										| GDK_BUTTON_RELEASE_MASK
										| GDK_POINTER_MOTION_MASK
										| GDK_POINTER_MOTION_HINT_MASK
										| GDK_KEY_PRESS_MASK);
										
	gtk_signal_connect (GTK_OBJECT (ix_savebutton), "clicked",
                    	GTK_SIGNAL_FUNC (ixSaveClicked), ix_view2);
 	gtk_signal_connect (GTK_OBJECT (ix_window), "destroy",
                    	GTK_SIGNAL_FUNC (ixCancelClicked), ix_window);
 	gtk_signal_connect (GTK_OBJECT (ix_line), "activate",
                    	GTK_SIGNAL_FUNC (ixLineChanged), ix_line);
//  	gtk_signal_connect (GTK_OBJECT (ix_line), "key_press_event",
//                     	GTK_SIGNAL_FUNC (ixLine_keypress), ix_line);

	gtk_widget_show (ix_window);

}


void immediate_exec ()
{

	macro_gui ();
}




