


/*
 *	macro.c: Macro interpreter and ancilliary functions
 */


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */




#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "application.h"
#include "backgnd.h"
#include "binary.h"
#include "defs.h"
#include "dlgbox.h"
#include "fft.h"
#include "fileio.h"
#include "filter.h"
#include "filter_adaptive.h"
#include "findclust.h"
#include "fractals.h"
#include "functions.h"
#include "functions2.h"
#include "graphdsp.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "imgmath.h"
#include "import.h"
#include "interface.h"
#include "measure.h"
#include "miscop.h"
#include "morphop.h"
#include "processm.h"
#include "roi.h"
#include "segment.h"
#include "special1.h"
#include "statop.h"
#include "threshold.h"
#include "tiffstuff.h"
#include "viz.h"
#include "watershed.h"
#include "wvlt_filt.h"

#include "macro1.h"
#include "macro.h"






/*****************************************************************

	Here are the functions that handle the tokens.
	
	All functions obtain the buffer
	and an index to the first character of the parameter list
	(i.e. just after the opening parentheses). They are responsible
	for getting the required parameters and advancing the index
	to past the closing parentheses (so it now points to the ;)
	
	Note that any procedure or function that takes an image expr
	as parameter is responsible for disposing of the image buffer-
	expr allocates a "dangling" image that needs to be passed thru
	(as return variable) or needs to be freed. In many cases, the
	actual function performs this job. Sometimes (particularly in
	cases where the return type is MT_REAL) it needs to be
	freed inside the function.
	
******************************************************************/

/*
	testpoint procedure
	syntax:	testpoint (intexpr);
	Parameter: A testpoint number between 0 and 10
	Testpoint is a debug function. It allows to set a breakpoint
	right here in this function to break the program at a specific
	point in a macro. 
*/

variable_type mac_testpoint (unsigned char* s, int* i)
{
variable_type r;
int j,k;

	j=*i; k=0;
	r = expr (s, &j, MT_INT);
	

	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;

	switch (r.intvalue)
	{
		case 0:
			dp (1, "Macro: Tespoint 0 reached\n");	/* Set breakpoint here */
			break;
		case 1:
			dp (1, "Macro: Tespoint 1 reached\n");	/* Set breakpoint here */
			break;
		case 2:
			dp (1, "Macro: Tespoint 2 reached\n");	/* Set breakpoint here */
			break;
		case 3:
			dp (1, "Macro: Tespoint 3 reached\n");	/* Set breakpoint here */
			break;
		case 4:
			dp (1, "Macro: Tespoint 4 reached\n");	/* Set breakpoint here */
			break;
		case 5:
			dp (1, "Macro: Tespoint 5 reached\n");	/* Set breakpoint here */
			break;
		case 6:
			dp (1, "Macro: Tespoint 6 reached\n");	/* Set breakpoint here */
			break;
		case 7:
			dp (1, "Macro: Tespoint 7 reached\n");	/* Set breakpoint here */
			break;
		case 8:
			dp (1, "Macro: Tespoint 8 reached\n");	/* Set breakpoint here */
			break;
		case 9:
			dp (1, "Macro: Tespoint 9 reached\n");	/* Set breakpoint here */
			break;
		default:
			dp (1, "Illegal testpoint set\n");
	}

	
	return nullvar;
}


/*
	print procedure
	syntax:	print (expr, expr,...);
	This command prints data into the output window.
	Example: print ("Thresholding by ",auto_threshold(img,0));
*/

variable_type mac_print (unsigned char* s, int* i)
{
char msg[1024];
int j;
variable_type v;
char no_lf;

	j = *i; no_lf=0;

	/* Check if string constant or expr */
	
	while (s[j]!=')')	/* end of log var list */
	{
		v = expr (s,&j, MT_NULL);
		if (v.type==MT_REAL)
			sprintf (msg,"%f",v.numvalue);
		else if (v.type==MT_INT)
			sprintf (msg,"%li",v.intvalue);
		else if (v.type==MT_STRING)
		{
			snprintf (msg,1020, "%s", v.strg);
			free (v.strg);
		}
		else if (v.type==MT_IMG)
		{
			snprintf (msg,1020," [image %s]",v.name);
			freebuf (&v.imgptr);
		}
			
		MAC_ADDLOG (msg);
		
		if ((s[j]==',') && (s[j+1]==')')) no_lf=1;
		if (s[j]==',') j++;
	}
	if (!no_lf) 
	{
		strcpy (msg, "\n");
		MAC_ADDLOG (msg); 	/* always end with a newline */
	}
	
	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;

	return nullvar;

}


/*
	message procedure
	syntax:	message (expr, expr, ...);
	message prints a message to stderr,
	otherwise works like print.
	Example: message ("I'm done with this.");
	Note: The Message function is useful for debugging,
	because the message appears immediately on the console.
*/

/* Note: Don't use db here because we want to display unconditionally */

variable_type mac_message (unsigned char* s, int* i)
{
variable_type v;
int j;
char no_lf;

	j = *i; no_lf=0;
	fprintf (stderr, "Macro: ");

	/* Check if string constant or expr */
	
	while (s[j]!=')')	/* end of log var list */
	{
		v = expr (s,&j, MT_NULL);
		if (v.type==MT_REAL)
			fprintf (stderr,"%f",v.numvalue);
		else if (v.type==MT_INT)
			fprintf (stderr,"%li",v.intvalue);
		else if (v.type==MT_STRING)
		{
			fprintf (stderr, "%s", v.strg);
			free (v.strg);
		}
		else if (v.type==MT_IMG)
		{
			fprintf (stderr," [image %s]",v.name);
			freebuf (&v.imgptr);
		}
			
		if ((s[j]==',') && (s[j+1]==')')) no_lf=1;
		if (s[j]==',') j++;
	}
	if (!no_lf) fprintf (stderr, "\n");

	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;
	
	return nullvar;

}



/*
	dispose procedure
	syntax:	dispose (varname);
	Parameter: Variable to dispose of.
	Dispose deletes a variable from the variable list and
	frees its buffer if it's an image.
*/

variable_type mac_dispose (unsigned char* s, int* i)
{
variable_type v;
int j,k;

	j=*i; k=0; v = nullvar;
	if (!isalpha1(s[j])) macro_error (1,j);
	do
	{
		v.name[k++] = s[j++];
	} 
	while ((k<16) && (isalpha2(s[j])));
	

	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;

	deletevar (v);
	
	return nullvar;
}

/*
	load function
	syntax 1:	load ("interactive")
	Parameter: String keyword "interactive"
	Opens the file selection window and lets user choose file
	Syntax 2:   load ("filename"); 
	Parameter: String constant file name.
	Opens the file given in the parameter
*/


variable_type mac_loadimg (unsigned char* s, int* i)
{
int j;
variable_type v;
int tiffreport_bak;
char *fname, *fname2;

	/* Extract the file name */
	
	j = *i;
	v = expr (s, &j, MT_STRING);

	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;

	if (strncasecmp (v.strg,"interactive",11)==0)
	{
		ChooseFile ("Select file to load:", window1, 0, &fname);
	}
	else
	{
		fname = g_malloc (strlen(v.strg)+strlen(imager_cwd)+10);		/* g_malloc -- must match that of ChooseFile */
		strcpy (fname, v.strg);

		/* New in 1.0.0: Concat with imager_cwd to ensure default directory is accessed */

		fname2 = compose_filename (fname);
		free (fname);
		fname = strdup (fname2);
		free (fname2);

	}

	dp (1, "Macro: Loading image %s\n",fname);
	tiffreport_bak = tiffreport; tiffreport=0;

	/* Load the file. Memory allocation is done by load_file */

	if (internal_file_read (fname, &v.imgptr) <=0)
		macro_error (10,j);
		
	free (v.strg); v.strg=NULL;
	g_free (fname);
	v.type = MT_IMG;
	tiffreport = tiffreport_bak;
	make_minmax (&v.imgptr);
	
	dp (2, "Macro load: New image variable has pointer %08X\n", (long)v.imgptr.data);

	return v;
}



/*
	save procedure
	syntax 1:	save (imgexpr, "interactive")
	Parameters: Image expression, string keyword "interactive"
	Saves the image computed under imgexpr using the file save dialog.
	Syntax 2:   save (imgexpr, "fname", "ftype");
	Parameters: Image expression, string expression filename, string expr for file type,
		may be any of "tiff", "vsrt", "raw", "opendx".
	Saves the image computed under imgexpr using the filename constant and the
	given file type. Always saves all slices; no compression for TIFF.
	Syntax 3:   save (imgexpr, "fname", "ftype", expr1, expr2, expr3);
	Parameters: Image expression, string expression filename, string expression for file type,
		may be any of "tiff", "vsrt", "pnm", "raw", "opendx".
		expr1 is the start slice; expr2 is the last slice; expr 3 is the tiff compression.
	Examples:
	save (thisimage, "thisimage.tif", "tiff", 0,39,0);
	save (myimage, "interactive");
	save (astack, "dxstack.dx", "opendx");
	
	Caveat: If imgexpr is a expression, the format will always be B_FLOAT
	and vsrt will cause an error. Only if a variable is provided,
	the original image type will be retained.
*/


variable_type mac_saveimg (unsigned char* s, int* i)
{
char *fname, *fname2;
char savetype[11];
int err,j,stslice,enslice, compression_mode;
variable_type v1,v;

	/* Get the image to save - may be an expr. Check for comma after that */
	
	v = expr (s, i, MT_IMG);
	j = *i;
	if (s[j++]!=',') macro_error (1,j);

	/* Extract the file name or the "interactive" instruction */

	v1 = expr (s, &j, MT_STRING);
	fname = g_malloc (strlen(v1.strg)+strlen(imager_cwd)+10);		/* g_malloc -- must match that of ChooseFile */
	strcpy (fname, v1.strg);

	free (v1.strg);
	/* extract_string (s, &j, fname, 1023); */

	if (strncasecmp (fname,"interactive",11)==0)
	{
		save_main (&v.imgptr);

		/* Skip to closing parentheses, update pointer. We're done */

		while (s[j]!=')') j++;
		*i = j+1;
		freebuf (&v.imgptr);
	}
	else	/* Non-interactive. Get additional information */
	{

		/* First, like mac_load, concat the current PWD */

		fname2 = compose_filename (fname);
		dp (2, "Macro: Saving a file to %s\n", fname2);

		/* Get additional paramteres if provided */

		if (s[j++]!=',') macro_error (1,j);
		v1 = expr (s, &j, MT_STRING);
		strncpy (savetype, v1.strg, 10);
		free (v1.strg);
		/* extract_string (s, &j, savetype, 10); */
		if (s[j]==')')		/* Done - use defaults */
		{
			stslice = 0; enslice = v.imgptr.zmax-1;
			compression_mode=0;
		}
		else
		{
			if (s[j++]!=',') macro_error (1,j);
			v1 = expr (s,&j,MT_INT); stslice = v1.intvalue;
			if (s[j++]!=',') macro_error (1,j);
			v1 = expr (s,&j,MT_INT); enslice = v1.intvalue;
			if (s[j++]!=',') macro_error (1,j);
			v1 = expr (s,&j,MT_INT); compression_mode = v1.intvalue;
			while (s[j]!=')') j++;
		}
		
		if (!strncasecmp (savetype, "tiff",4))
		{
			err=tiffwrite (fname2, &v.imgptr, compression_mode, "Created with Crystal Image", stslice, enslice);
		}
		else if (   (   (!strncasecmp (savetype, "vsrt",4) || !strncasecmp (savetype, "vflt",4))   )
					&& ( (v.imgptr.imgtype==B_SHORT) || (v.imgptr.imgtype==B_FLOAT) ) )
		{
			err=vsrtwrite (&v.imgptr, fname2, stslice, enslice);
		}
		else if ((!strncasecmp (savetype, "pnm",3)) && ((v.imgptr.imgtype==B_RGB) || (v.imgptr.imgtype==B_CHAR)))
		{
			err=pnm_write (&v.imgptr, fname2, stslice, enslice);
		}
		else if (!strncasecmp (savetype, "raw",3))
		{
			err=raw_write (&v.imgptr, fname2, stslice, enslice);
		}
		else if (!strncasecmp (savetype, "opendx",6))
		{
			err=DX_write (&v.imgptr, fname2, stslice, enslice,1.0);
		}
		else
			macro_error (14,j);
				
		*i = j+1;
		freebuf (&v.imgptr);
		g_free (fname2);

	}

	g_free (fname);

	return nullvar;
}


/*
	import function
	syntax:		import ("filename", xsz, ysz, zsz, ofs, "type", swap, ascii);
	Parameter: 	String filename
				size of image xsz, ysz, zsz (int)
				offset (bytes to skip) ofs (int)
				"type": string: one of "char", "short", "float", "rgb"
				swap: 1 if shorts need to be lsb/msb swapped (little-endian), 0 otherwise (big-endian)
				ascii: 1 if import data is ASCII, 0 otherwise
						if ascii, the ofs and swap parameters are ignored.

*/


variable_type mac_import (unsigned char* s, int* i)
{
int j, err;
variable_type v,r;
char imgtype_s[256];
char *fname;
import_params params;

	/* Extract the file name, make fully qualified path with imager_cwd */
	
	j = *i;
	v = expr (s, &j, MT_STRING);
	fname = compose_filename(v.strg);
	free (v.strg);
	strncpy (params.filename, fname, 1020);		/* iffy due to the fixed length of params.filename */
	free (fname);

	/* Extract the dimesions and the offset */

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); params.xdim = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); params.ydim = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); params.zdim = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); params.ofs = r.intvalue;

	/* Extract the image type */

	if (s[j++]!=',') macro_error (1,j);
	v = expr (s, &j, MT_STRING);
	strncpy (imgtype_s, v.strg, 200);
	free (v.strg);
	if (!strncasecmp (imgtype_s,"char",4)) 			
	{ 
		params.src_type=iB_CHAR; params.imgtype=B_CHAR; 
	}
	else if (!strncasecmp (imgtype_s,"short",5)) 	
	{ 
		params.src_type=iB_SHORT; params.imgtype=B_SHORT; 
	}
	else if (!strncasecmp (imgtype_s,"float",5)) 	
	{ 
		params.src_type=iB_FLOAT; params.imgtype=B_FLOAT; 
	}
	else if (!strncasecmp (imgtype_s,"rgb",3)) 	 	
	{ 
		params.src_type=iB_RGB; params.imgtype=B_RGB; 
	}
	else if (!strncasecmp (imgtype_s,"double",6)) 	
	{ 
		params.src_type=iB_DOUBLE; params.imgtype=B_FLOAT; 
	}
	else macro_error (15,j);
	if (params.zdim==0) params.zdim=1;

	/* Finally, extract the swap and ascii flags */

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); params.imp_swab = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); params.is_ascii = r.intvalue;
	
	/* Skip to closing parentheses, update pointer. */

	while (s[j]!=')') j++;
	*i = j+1;

	/* Got it all. Now import. */
	
	v = nullvar; v.type = MT_IMG;
	if (params.is_ascii)
	{
		dp (2, "Macro: Importing ASCII file %s\n",params.filename);
		err = do_import_ascii (&v.imgptr, params);
	}
	else
	{
		dp (2, "Macro: Importing %s\n",params.filename);
		err = do_import (&v.imgptr, params);
	}

	if (err) macro_error (27,j);
	return v;

}




/*
	display procedure
	syntax:		display (imgexpr);
				display (imgexpr, strexpr);
	Parameter: Image expression to display
	display places an image into the program's main image buffer
	and opens a display window.
	strexpr allows to set the window title.
	The main image remains there even after the macro finished.
*/


variable_type mac_display (unsigned char* s, int* i)
{
variable_type v,r;
int j;
char sbuf[256];

	v = expr (s, i, MT_IMG);
	j = *i;

	strcpy (sbuf, "Macro Display");
	if (s[j]==',') 			/* Optional window title */
	{
		j++;
		r = expr (s,&j,MT_STRING);
		strncpy (sbuf, r.strg, 250);
		free (r.strg);
	}


	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;

	if (file_ok) close_file();
	if (clone_image (&v.imgptr, &mainimg)==-1)
		macro_error (11,j);
	make_minmax (&mainimg);
	freebuf (&v.imgptr);
	prepare_display_window (sbuf);
	file_ok = 1; 
	make_imlib_buffer_from_main_image (0);
	redraw (1);
	return nullvar;

}


/*
	convert_image procedure
	Syntax: convert_image (imgvar, "newtype", expand);
	This procedure converts an image to a different type.
	"newtype" may be any of "float", "char", "short", or "rgb".
	expand is a flag that tells whether to expand the range 
	(expand==1) or keep the values (expand==0)
	Example: convert_image (myimg, "float",0);

	A special case exists where "newtype" is "hsv". In this case,
	the original image *must* be a RGB image, and it gets converted
	to its HSV interpretation.
*/

variable_type mac_convert_image (unsigned char* s, int* i)
{
variable_type v,r,q;
int j,k,vidx;
char newtype_s[11];
char expand, target;
TYPE_SELECTOR newtype;

	j = *i; k=0; memset (v.name, 0, sizeof(v.name));
	if (!isalpha1(s[j])) macro_error (1,j);
	
	do
	{
		v.name[k++] = s[j++];
	} 
	while ((k<16) && (isalpha2(s[j])));
	
	vidx = findvar (v.name);
	if (vidx<0)		/* not found error */
		macro_error (16,j);
	
	strncpy (v.name, vars[vidx].name, 16);
	v.type = vars[vidx].type;
	if (v.type!=MT_IMG) macro_error (6,j);
	v.imgptr = vars[vidx].imgptr;
	if (s[j++]!=',') macro_error (1,j);

	q = expr (s, &j, MT_STRING);
	strncpy (newtype_s, q.strg, 10);
	free (q.strg);
	/* extract_string (s, &j, newtype_s, 10); */
	newtype=v.imgptr.imgtype;

	target=0;
	if (!strncasecmp (newtype_s,"char",4)) 
	{
		newtype=B_CHAR; target = 1;
	}
	else if (!strncasecmp (newtype_s,"short",5)) 
	{
		newtype=B_SHORT;
		target = 2;
	}
	else if (!strncasecmp (newtype_s,"float",5)) 
	{
		newtype=B_FLOAT;
		target = 3;
	}
	else if (!strncasecmp (newtype_s,"rgb",3)) 
	{
		newtype=B_RGB;
		target = 4;
	}
	else if (!strncasecmp (newtype_s,"hsv",3)) 
	{
		newtype=B_RGB;
		target = 5;
	}
	else 
		macro_error (15,j);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); expand = r.intvalue;
	
	while (s[j]!=')') j++;
	*i = j+1;

	if (target==5)				/* RGB to HSV special case */
	{
		if (v.imgptr.imgtype != B_RGB) macro_error (17,j);
		dataconv_hsvconv (&v.imgptr);
	}
	else
		dataconv_execute (&v.imgptr, newtype, expand);

	vars[vidx]=v;	/* We need to update the variable in the list
						because we lose v upon exit */

	return nullvar;
}





/*
	profile procedure
	syntax:	profile (imgexpr,x1,y1,z1,x2,y2,z2);
	Parameter: Image expression to display
	start coordinates (x,y,z) and end coordinates (x,y,z) of line
	Profile opens a graph window and displays the pixel values
	along a straight line from (x1,y1,z1) to (x2,y2,z2),
	interpolating in-plane, but not between planes.
*/


variable_type mac_profile (unsigned char* s, int* i)
{
variable_type v,r;
int j,x1,y1,z1,x2,y2,z2;

	v = expr (s, i, MT_IMG);
	j = *i;

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); x1 = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); y1 = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); z1 = r.intvalue;

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); x2 = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); y2 = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); z2 = r.intvalue;

	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;

	view_profile_xyz (&v.imgptr, x1,y1,z1,x2,y2,z2);

	freebuf (&v.imgptr);
	return nullvar;

}



/*
	setparm procedure
	syntax:	setparm ("parameter", value);
	Parameter: Name of the parameter to set
	value: Value to set the parameter to.

	The following parameters may be set:
	"interpolation": 0 = nearest neighbor, 1 = bi-liniear, 2= bi-cubic
					 3 = Spline 2nd order, 4 = Spline 3rd, 5 = Spline 4th,
					 6 = Spline 5th.

*/

variable_type mac_setparm (unsigned char* s, int* i)
{
char parm[25];
int j,k;
variable_type r;


	j = *i;
	r = expr (s, &j, MT_STRING);
	strncpy (parm, r.strg, 24);
	free (r.strg);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_REAL); /* Needs to be real because we don't know the final type yet */

	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;

	if (!strncasecmp (parm,"interpolation",13))
	{
		k = (int) r.numvalue;
		if (k==0) intp_pref = INTP_NN;
		else if (k==1) intp_pref = INTP_BILIN;
		else if (k==2) intp_pref = INTP_BICUB;
		else if (k==3) intp_pref = INTP_SPLINE2;
		else if (k==4) intp_pref = INTP_SPLINE3;
		else if (k==5) intp_pref = INTP_SPLINE4;
		else if (k==6) intp_pref = INTP_SPLINE5;
		else macro_error (15,j);
	}
	/* else if chain continues */
	else
		macro_error (15,j);

	return nullvar;

}



/*
	setroi procedure
	syntax:	setroi ();
			setroi (x1,y1,r);
			setroi (x1,y1,x2,y2);
			setroi (x1,y1,x2,y2,x3,y3...xn,yn);

	Defines a region of interest. There are three different
	parameter lists:
	1) If the parameter list is empty, the ROI will be deleted
	2) If the parameter list contains 4 parameters, they constitute 
		a rectangular ROI
	3) If the parameter list contains 3 parameters, they constitute 
		center and radius of a circle
	4) If the parameter list contains 2n parameters (n>2), they constitute
		a polygonal ROI
	
	Note: the macro Display command may reset the ROI, so don't use SETROI
	before a DISPLAY.
*/

variable_type mac_setroi (unsigned char* s, int* i)
{
int x1,y1,x2,y2,n,j,k;
variable_type r;


	j = *i;
	if (s[j]==')')		/* Empty parameter list */
	{
		current_roi.roi_mode=ROI_NONE;		
		*i = j+1;
	}
	else
	{
	
		r = expr (s, &j, MT_INT);
		x1 = r.intvalue;
		if (s[j++]!=',') macro_error (1,j);	/* Must be x,y pairs */
		r = expr (s, &j, MT_INT);
		y1 = r.intvalue;
		if (s[j++]!=',') macro_error (1,j);	/* Must radius or 2 vertices or more */
		r = expr (s, &j, MT_INT);
		x2 = r.intvalue;


		/* Check for closing parentheses - then X2 is a radius,
		   do a circular ROI and exit immediately */
		
		if (s[j]==')') 
		{
			for (k=0; k<=16; k++)
			{
				current_roi.xi[k] = (int)(0.5 + x1 + x2*cos(2*PI*k/16));
				current_roi.yi[k] = (int)(0.5 + y1 + x2*sin(2*PI*k/16));
			}
			current_roi.numvertices=17;
			current_roi.xi[current_roi.numvertices]=current_roi.xi[0];
			current_roi.yi[current_roi.numvertices]=current_roi.yi[0];
			
			/* Set the bounding box */
	
			current_roi.bbox_x1 = x1-x2;
			current_roi.bbox_x2 = x1+x2;
			current_roi.bbox_y1 = y1-x2;
			current_roi.bbox_y2 = y1+x2;

			current_roi.roi_mode=ROI_CIRCLE; current_roi.xc = x1; current_roi.yc = y1;
			current_roi.radius=x2;

			*i=j+1;
			return nullvar;
		}
		
		if (s[j++]!=',') macro_error (1,j);	/* Must be x,y pairs */
		r = expr (s, &j, MT_INT);
		y2 = r.intvalue;
		
		/* Tentatively assume a rectangular ROI */
		
		current_roi.roi_mode = ROI_RECT;
		current_roi.xi[0]=x1;
		current_roi.xi[1]=x2;
		current_roi.xi[2]=-1;
		current_roi.yi[0]=y1;
		current_roi.yi[1]=y2;
		current_roi.yi[2]=-1;

		current_roi.bbox_x1 = x1;	/* Bounding box is the same */
		current_roi.bbox_x2 = x2;
		current_roi.bbox_y1 = y1;
		current_roi.bbox_y2 = y2;
		
		n=2;
		while (s[j]==',')		/* even more vertices!!! */
		{
			j++;
			current_roi.roi_mode=ROI_POLY;
			r = expr (s, &j, MT_INT);
			x1 = r.intvalue;
			if (s[j++]!=',') macro_error (1,j);	/* Must be x,y pairs */
			r = expr (s, &j, MT_INT);
			y1 = r.intvalue;
			current_roi.xi[n]=x1;
			current_roi.yi[n]=y1;
			n++;
			current_roi.numvertices=n;
			
		}
		if (current_roi.roi_mode==ROI_POLY)	/* Finalize ROI */
		{
			current_roi.xi[current_roi.numvertices]=current_roi.xi[0];
			current_roi.yi[current_roi.numvertices]=current_roi.yi[0];
			current_roi.numvertices++;
			current_roi.xi[current_roi.numvertices]=-1;
			current_roi.yi[current_roi.numvertices]=-1;
			current_roi.bbox_x1 = current_roi.bbox_x2 = current_roi.xi[0];
			current_roi.bbox_y1 = current_roi.bbox_y2 = current_roi.yi[0];
			for (k=1; k<current_roi.numvertices; k++)
			{
				if (current_roi.bbox_x1 > current_roi.xi[k]) current_roi.bbox_x1 = current_roi.xi[k];
				if (current_roi.bbox_x2 < current_roi.xi[k]) current_roi.bbox_x2 = current_roi.xi[k];
				if (current_roi.bbox_y1 > current_roi.yi[k]) current_roi.bbox_y1 = current_roi.yi[k];
				if (current_roi.bbox_y2 < current_roi.yi[k]) current_roi.bbox_y2 = current_roi.yi[k];
			}
		
		}
	
		while (s[j]!=')') j++;
		*i = j+1;
	}

	return nullvar;

}



/*
	fillroi procedure
	syntax:	fillroi (imgvar,expr);
			fillroi (imgvar,expr,slice);

	Fills a region of interest with a constant value
	parameter list:
		imgvar: Image to work on
		expr: Value to use
		slice: Optional parameter - if given, fills only in this slice
*/


variable_type mac_fillroi (unsigned char* s, int* i)
{
variable_type v,r;
int j,k,vidx,x,y,z,z0,z1,oneslice;
float fillval;
int bx1,bx2,by1,by2;
unsigned char c;
short sh;
rgbtriplet rgb;
long idx;


	j = *i; k=0; memset (v.name, 0, sizeof(v.name));
	if (!isalpha1(s[j])) macro_error (1,j);
	
	do
	{
		v.name[k++] = s[j++];
	} 
	while ((k<16) && (isalpha2(s[j])));
	
	vidx = findvar (v.name);
	if (vidx<0)		/* not found error */
		macro_error (16,j);
	
	strncpy (v.name, vars[vidx].name, 16);
	v.type = vars[vidx].type;
	if (v.type!=MT_IMG) macro_error (6,j);
	v.imgptr = vars[vidx].imgptr;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s, &j, MT_REAL); fillval = r.numvalue;

	oneslice=-1; z0=0; z1=v.imgptr.zmax-1;
	if (s[j]==',')
	{
		j++;
		r = expr (s, &j, MT_INT); oneslice = r.intvalue;
		z0=oneslice; z1=oneslice;
	}


	while (s[j]!=')') j++;
	*i = j+1;

	if (current_roi.roi_mode==ROI_NONE) return nullvar; 	/* No ROI, no fill */
	
	roi_bbox (&v.imgptr,&bx1,&by1,&bx2,&by2);
	bx1 = MAX(0,bx1); by1 = MAX(0,by1);
	bx2 = MIN (v.imgptr.xmax-1,bx2); by2 = MIN (v.imgptr.ymax-1,by2);

	sh = (int) (fillval+0.499);
	if (sh>255) c=255; else if (sh<0) c=0; else c=sh;
	rgb = (rgbtriplet){c,c,c};

	for (y=by1; y<=by2; y++)
		for (x=bx1; x<=bx2; x++)
		{
			if (in_roi (x,y))
				for (z=z0; z<=z1; z++)
				{
					idx = x + v.imgptr.xmax*(y + v.imgptr.ymax*z);
					switch (v.imgptr.imgtype)
					{
						case B_CHAR:
							((unsigned char*)v.imgptr.data)[idx]=c;
							break;
						case B_SHORT:
							((short*)v.imgptr.data)[idx]=sh;
							break;
						case B_FLOAT:
							((float*)v.imgptr.data)[idx]=fillval;
							break;
						case B_RGB:
							((rgbtriplet*)v.imgptr.data)[idx]=rgb;
							break;
					}
				}
		}
	make_minmax (&v.imgptr);
	return nullvar;
}





/*
	dim procedure
	Syntax: dim (varname[subscript] {,varname[subscript]...} ):
	Dimension creates floating-point arrays. Arrays may be multi-
	dimensional, and multiple arrays may be declared using
	Dimension, separated by commas.
	Example: dim (ary1[100], twodimary[64,64]);
	
*/

variable_type mac_dimarray (unsigned char* s, int* i)
{
variable_type v;
int j,k;
subscript_type d;

	j = *i; 
	
	while (s[j]!=')')	/* closing parentheses - end of list */
	{
	
		/* extract the variable name */
		k=0; v=nullvar;
		if (!isalpha1(s[j])) macro_error (1,j);
		do
		{
			v.name[k++] = s[j++];
		} 
		while ((k<16) && (isalpha2(s[j])));
		
		/* Must not yet exist at this time */
		
		if (findvar (v.name)>=0) macro_error (22,j);
		
		/* The next char *must* be a [ */
		
		if (s[j++]!='[') macro_error (1,j);
		d = extract_subscript (s, &j);
	
		/* must have a ',' or ')' - check for it. */
		
		if ((s[j]!=',') && (s[j]!=')')) macro_error (1,j);
		if (s[j]==',') j++; /* preserve j if pointing at ) */
		
		dp (2, "Dimensioning array %s with ",v.name);
		if (d.dim==1) dp (2, " %d elements\n",d.x);
		if (d.dim==2) dp (2, " %dx%d elements\n",d.x,d.y);
		if (d.dim==3) dp (2, " %dx%dx%d elements\n",d.x,d.y,d.z);

		v.dimension=d.dim;
		if (d.dim==1) v.numelems=d.x;
		else if (d.dim==2) v.numelems=d.x*d.y;
		else if (d.dim==3) v.numelems=d.x*d.y*d.z;
		v.xmax=d.x; v.ymax=d.y; v.zmax=d.z;
		
		/* get the memory for this array - only to be freed by dispose or endmacro */
		
		v.type=MT_REAL; v.arrayptr=NULL;
		v.arrayptr = calloc (v.numelems+5,sizeof (float));
		if (!v.arrayptr) fatal_memerr();

		/* and enter the variable into the list */
		
		addvar (v);
		
	}
	*i=j+1;
	return nullvar;
}



/*
	graph procedure
	Syntax: 	graph (intexpr, numarray1, numarray2);
				graph (intexpr, numarray);
	Purpose: To plot x-y data stored in the array(s)
	This procedure plots numerical data. The first syntax requires
	two one-dimensional arrays numarray1 and numarray2 for the x-
	and y-data, respectively. The second syntax requires an n-dimensional
	array. The first subscript indexes the datapoints, and the second
	subscript determines the data set; 0=X-data; 1=first y-data, etc.
	Thus, with a m by n array it is possible to plot n-1 curves
	with up to m data points.
	Intexpr determines the number of data points to be plotted. Intexpr 
	MUST be less than m.
	All numarrays must be variable references. Expressions are not allowed.
	
*/

variable_type mac_graph (unsigned char* s, int* i)
{
variable_type v1,v2;
int j,k,numpts,numsets,mode,x,y;
float *xdata, *ydata;

	j = *i; 
	
	/* Get the number of data points */
	v1 = expr (s, &j, MT_INT);
	numpts = v1.intvalue;
	if (s[j++]!=',') macro_error (1,j);	/* Must be x,y pairs */

	/* extract the variable name for the first array */
	k=0; v1=nullvar; v2=nullvar;
	if (!isalpha1(s[j])) macro_error (1,j);
	do
	{
		v1.name[k++] = s[j++];
	} 
	while ((k<16) && (isalpha2(s[j])));

	/* Must exist at this time and have dimension>0 */

	k = findvar (v1.name);
	if (k<0) macro_error (16,j);
	v1 = vars[k];

	/* First or second syntax? */

	if (s[j]==',')
	{
		j++; mode=1;

		k=0; v2=nullvar;
		if (!isalpha1(s[j])) macro_error (1,j);
		do
		{
			v2.name[k++] = s[j++];
		} 
		while ((k<16) && (isalpha2(s[j])));

		/* Must exist at this time and have dimension>0 */

		k = findvar (v2.name);
		if (k<0) macro_error (16,j);
		v2 = vars[k];

		/* Two arrays means that both have to be one-dimensional */

		if ((v1.dimension!=1) || (v2.dimension!=1))
			macro_error (20,j);
		numsets=1;
	}
	else
	{
		mode=0;
		if (v1.dimension != 2) macro_error (20,j);
		numsets = v1.ymax-1; if (numsets<1) macro_error (20,j);
	}

	while (s[j]!=')') j++;
	*i = j+1;

	/* Now copy the data and display the graph */

	xdata = (float*) calloc (numpts+10, sizeof(float));
	ydata = (float*) calloc (numpts*numsets+10, sizeof(float));

	if (mode==0)	/* multi-dimensional array */
	{
		for (y=1; y<=numsets; y++)
			for (x=0; x<numpts; x++)
				ydata[x + numpts*(y-1)]=v1.arrayptr[x + v1.xmax*y];
				
		for (x=0; x<numpts; x++)
			xdata[x]=v1.arrayptr[x];
			
		multi_graphme (xdata, ydata, numpts, numsets, " ", " ", "Output of Macro Run", 0);
	}
	else
	{
		for (x=0; x<numpts; x++)
		{
			xdata[x]=v1.arrayptr[x];
			ydata[x]=v2.arrayptr[x];
		}
		graphme (xdata, ydata, numpts, " ", " ", "Output of Macro Run", 0);
	
	}
	
	free (xdata);
	free (ydata);

	return nullvar;
}





/*
	trim_image procedure
	Syntax: trim_image (imgvar, newmin, newmax, offset, shift, mask, cyclex, cycley);
	This procedure manipulates an image in various ways:
	- it changes its min and max values if they are provided
	   (if both are zero, min/max is left alone)
	- it adds or subtracts an offset (much like dyadic add, but preserves type)
	- it can do bit-shift operations and mask away bits
	- it can cyclically shift an image
	
	Parameters are:
		imgvar:		The image to operate on. This is a procedure, not a function.
		newmin, newmax: (real) values to serve as new intensity scaling values (if both are zero, min/max is left alone)
		offset: (integer) Offset to subtract from pixel values.
		shift: (int) bit-wise shift of each pixel to the right (negative: to the left)
		mask: (int) bitmask to apply to each pixel, e.g. to mask out sign bits
		cyclex, cycley: (int) Number of bits to cycle the image right or down
	
	Example: trim_image (myimg, 0, 0, 0, 4, 4095, 0, 0);
*/

variable_type mac_trim_image (unsigned char* s, int* i)
{
variable_type v,r;
int j,k,vidx;
float newmin, newmax;
int rshift, mask, offs, cyclex, cycley, z;


	j = *i; k=0; memset (v.name, 0, sizeof(v.name));
	if (!isalpha1(s[j])) macro_error (1,j);
	
	do
	{
		v.name[k++] = s[j++];
	} 
	while ((k<16) && (isalpha2(s[j])));
	
	vidx = findvar (v.name);
	if (vidx<0)		/* not found error */
		macro_error (16,j);
	
	strncpy (v.name, vars[vidx].name, 16);
	v.type = vars[vidx].type;
	if (v.type!=MT_IMG) macro_error (6,j);
	v.imgptr = vars[vidx].imgptr;
	if (s[j++]!=',') macro_error (1,j);

	r = expr (s, &j, MT_REAL); newmin = r.numvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s, &j, MT_REAL); newmax = r.numvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s, &j, MT_INT); offs = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s, &j, MT_INT); rshift = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s, &j, MT_INT); mask = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s, &j, MT_INT); cyclex = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s, &j, MT_INT); cycley = r.intvalue;
	
	while (s[j]!=')') j++;
	*i = j+1;

	if (!mask) mask = 0xffff;	/* covers both char and short */

	execute_trim_image (&v.imgptr, offs, newmin, newmax, rshift, mask);
	
	if (cyclex || cycley)
	{
		for (z=0; z<v.imgptr.zmax; z++)
			cycle_image (&v.imgptr, cyclex, cycley, z);
	}
	
	vars[vidx]=v;	/* We need to update the variable in the list
						because we lose v upon exit */
	
	return nullvar;
}






/*************************************************************************


#######
#        #    #  #    #   ####    #####     #     ####   #    #   ####
#        #    #  ##   #  #    #     #       #    #    #  ##   #  #
#####    #    #  # #  #  #          #       #    #    #  # #  #   ####
#        #    #  #  # #  #          #       #    #    #  #  # #       #
#        #    #  #   ##  #    #     #       #    #    #  #   ##  #    #
#         ####   #    #   ####      #       #     ####   #    #   ####


**************************************************************************/


/* Let'start with a few multimode and type conversion operators */


/*
	int   function
	round function
	syntax:	intvar = int (expr);
	Parameter: Expression with real result
	Int and round convert the result of the expression to integer.
	Int truncates the result and round rounds it.
	Example: t = round (4.0/3.0);

*/

variable_type mac_int (unsigned char* s, int* i)
{
variable_type v;
int j;

	j=*i;
	v = expr (s,&j, MT_REAL);
	while (s[j]!=')') j++;
	*i = j+1;
	
	v.type=MT_INT; v.intvalue = (int)v.numvalue;
	return v;
}

variable_type mac_round (unsigned char* s, int* i)
{
variable_type v;
int j;

	j=*i;
	v = expr (s,&j, MT_REAL);
	while (s[j]!=')') j++;
	*i = j+1;
	
	v.type=MT_INT; v.intvalue = (int)(0.5+v.numvalue);
	return v;
}


/*
	log   function
	log10 function
	syntax:	var = log (expr);
	log and log10 compute the logarithms of expr. Expr may be of type
	- image: The result is the pixel-by-pixel log of the expr image
	- real: The result is the log of the scalar expr
	- int: Same as real, but the result is real.

*/

variable_type mac_log (unsigned char* s, int* i)
{
variable_type v;
int j;

	j=*i;
	v = expr (s,&j, MT_NULL);

	if (v.type==MT_REAL)
		v.numvalue = SAFELOG(v.numvalue);
	else if (v.type==MT_INT)
	{
		v.numvalue = SAFELOG ((float)v.intvalue);
		v.type=MT_REAL;
	}
	else if (v.type==MT_IMG)
		logexp (&v.imgptr, 0, 0.0);
	else
		macro_error (6,j);

	while (s[j]!=')') j++;
	*i = j+1;
	
	return v;
}

variable_type mac_log10 (unsigned char* s, int* i)
{
variable_type v;
int j;

	j=*i;
	v = expr (s,&j, MT_NULL);

	if (v.type==MT_REAL)
		v.numvalue = SAFELOG10(v.numvalue);
	else if (v.type==MT_INT)
	{
		v.numvalue = SAFELOG10 ((float)v.intvalue);
		v.type=MT_REAL;
	}
	else if (v.type==MT_IMG)
		logexp (&v.imgptr, 1, 0.0);
	else
		macro_error (6,j);

	while (s[j]!=')') j++;
	*i = j+1;
	
	return v;
}


/*
	exp   function
	syntax:	var = exp (expr);
	Raise e to the power of expr.
	If expr is an
	- image: The result is the pixel-by-pixel exp of the expr image
	- real: The result is the exp of the scalar expr
	- int: Same as real, but the result is real.

*/


variable_type mac_exp (unsigned char* s, int* i)
{
variable_type v;
int j;

	j=*i;
	v = expr (s,&j, MT_NULL);

	if (v.type==MT_REAL)
		v.numvalue = exp(v.numvalue);
	else if (v.type==MT_INT)
	{
		v.numvalue = exp ((float)v.intvalue);
		v.type=MT_REAL;
	}
	else if (v.type==MT_IMG)
		logexp (&v.imgptr, 2, 0.0);
	else
		macro_error (6,j);

	while (s[j]!=')') j++;
	*i = j+1;
	
	return v;
}


/*
	power  function
	syntax:	var = power (expr1,expr2);
	Raise expr1 to the power of expr2.
	expr1 and expr2 may be either images or scalars.
	If expr1 and expr2 are both scalars, the result is a scalar. In this
	   case, expr1 and expr2 may be either int or real, but the result is
	   always real.
	If expr1 is an image and expr2 is a scalar, each pixel of the image
	   will be raised to the power of expr2. The result is an image.
	If expr1 is a scalar and expr2 an image, each pixel of the result will
	   contain expr1 raised to the power of the corresponding pixel in
	   expr2. The result is an image. This is the function that needs to be
	   used as opposite to log10 (expr1==10).
	   
	example: x = power (10,y);

*/

variable_type mac_power (unsigned char* s, int* i)
{
variable_type x,y,v;
int j;

	j=*i; v=nullvar;
	x = expr (s,&j, MT_NULL);
	if (s[j++]!=',') macro_error (1,j);
	y = expr (s,&j, MT_NULL);

	if (x.type==MT_INT) { x.type=MT_REAL; x.numvalue=(float)x.intvalue; }
	if (y.type==MT_INT) { y.type=MT_REAL; y.numvalue=(float)y.intvalue; }

	if ((x.type==MT_REAL) && (y.type==MT_REAL))
	{
		v.numvalue = pow (x.numvalue, y.numvalue);
		v.type=MT_REAL;
	}
	else if ((x.type==MT_IMG) && (y.type==MT_REAL))
	{
		logexp (&x.imgptr, 4, y.numvalue);
		v.type=MT_IMG; v.imgptr=x.imgptr;
	}
	else if ((x.type==MT_REAL) && (y.type==MT_IMG))
	{
		logexp (&y.imgptr, 5, x.numvalue);
		v.type=MT_IMG; v.imgptr=y.imgptr;
	}
	else
		macro_error (6,j);

	while (s[j]!=')') j++;
	*i = j+1;
	
	return v;
}


/* ...and  a series of string-related functions
   (note that concatenation is handled by the + operator */


/*
	strlen function
	Syntax: intvar = strlen(strexpr);
	Purpose: Returns the length of a string
*/

variable_type mac_strlen (unsigned char* s, int* i)
{
variable_type x;
int j;

	j=*i;
	x = expr (s,&j, MT_STRING);
	while (s[j]!=')') j++;
	*i = j+1;
	
	x.type = MT_INT;
	x.intvalue = strlen (x.strg);
	if (x.strg) free (x.strg);
	x.strg = NULL;
	return x;

}


/*
	strnum function
	Syntax: strvar = strnum (numexpr);
			strvar = strnum (numexpr, numdigs);
			strvar = strnum (numexpr, format);
	Purpose: Converts a numerical value into a string
	Parameters: Numexpr (may be both int or real): Value to be converted
				numdigs (integer): Pad to number of digits (optional)
				format (string): C-style format string (optional)
*/


variable_type mac_strnum (unsigned char* s, int* i)
{
variable_type x,y,v;
char s0[255], f[255];
int j;

	j=*i; v=nullvar;
	x = expr (s,&j, MT_NULL);
	if ((x.type!=MT_INT) && (x.type!=MT_REAL)) macro_error (6,j);

	if (s[j]!=',')		/* No format given, use default */
	{
		if (x.type==MT_INT) sprintf (s0, "%li", x.intvalue);
		if (x.type==MT_REAL) sprintf (s0, "%f", x.numvalue);
	}
	else
	{
		j++; 	/* skip the comma */
		y = expr (s,&j, MT_NULL);
		if (y.type==MT_INT)
		{
			sprintf (f, "%%0%lid", y.intvalue);
			if (x.type==MT_REAL) f[strlen(f)-1] = 'f';
			if (x.type==MT_INT) sprintf (s0, f, x.intvalue);
			if (x.type==MT_REAL) sprintf (s0, f, x.numvalue);
		}
		else if (y.type==MT_STRING)
		{
			if (x.type==MT_INT) sprintf (s0, y.strg, x.intvalue);
			if (x.type==MT_REAL) sprintf (s0, y.strg, x.numvalue);
			free (y.strg);
		}
		else
			macro_error (6,j);
	}
	
	while (s[j]!=')') j++;
	*i = j+1;
	
	v.type = MT_STRING;
	v.strg = calloc (MIN(strlen(s0),1),sizeof (char));
	strcpy (v.strg, s0);
	return v;

}




/*
	substring function
	Syntax: strvar = substring (strexpr,pos1,pos2);
	        strvar = substring (strexpr,"dirname");
	        strvar = substring (strexpr,"filename");
	Purpose: Returns the substring in strexpr from pos1 to pos2 (first form)
	or returns either the directory or base file name from a file path.
*/


variable_type mac_substring (unsigned char* s, int* i)
{
variable_type x,r,v;
int p1, p2,j;
char *fname;


	j=*i; v=nullvar;
	r = expr (s,&j, MT_STRING);
	if (s[j++]!=',') macro_error (1,j);
	v.type = MT_STRING;						/* Define return type */

	/* Now we expect either a numexpr or a string constant */

	if (s[j]=='"')						/* Opening quotes -> "dirname" or "filename" */
	{
		x = expr (s, &j, MT_STRING);	/* Decode the string, although it is a string constant */
		if (!strncasecmp (x.strg, "dir", 3))
		{
			fname = dirname (r.strg);
			v.strg = strdup (fname);
		}
		else if (!strncasecmp (x.strg, "file", 4))
		{
			fname = basename (r.strg);
			v.strg = strdup (fname);
		}
		else
			macro_error (15,j);

	}
	else								/* Must be numerical parameter */
	{
		x = expr (s,&j, MT_INT); p1 = x.intvalue;
		if (s[j++]!=',') macro_error (1,j);
		x = expr (s,&j, MT_INT); p2 = x.intvalue;

		v.strg = calloc (p2-p1+2, sizeof (char));
		if (v.strg==NULL) fatal_memerr();

		if (p1 < strlen (r.strg))
			strncpy (v.strg, &r.strg[p1], p2-p1+1);
	}

	free (r.strg);
	while (s[j]!=')') j++;
	*i = j+1;
	return v;

}


/*
	strpos function
	Syntax: intvar = strpos(needle, haystack);
	Purpose: Returns the position of a string "needle"
	         in a longer string "haystack"
			 or a negative number if not found.
*/


variable_type mac_strpos (unsigned char* s, int* i)
{
variable_type haystack, needle, v;
int j,k;
unsigned char* p;

	j=*i; v = nullvar;
	needle = expr (s,&j, MT_STRING);
	if (s[j++]!=',') macro_error (1,j);
	haystack = expr (s,&j, MT_STRING);
	while (s[j]!=')') j++;
	*i = j+1;
	
	v.type = MT_INT;
	if ((strlen(haystack.strg)==0) || (strlen(needle.strg)==0)
		|| (strlen (needle.strg) > strlen(haystack.strg)))
		v.intvalue = -1;
	else
	{
		for (k=0; k<strlen(haystack.strg); k++) haystack.strg[k] = toupper(haystack.strg[k]);
		for (k=0; k<strlen(needle.strg); k++) needle.strg[k] = toupper(needle.strg[k]);
	
		p = strstr (haystack.strg, needle.strg);
		v.intvalue =  (unsigned long)p - (unsigned long)haystack.strg;
		if ((v.intvalue < 0) ||(v.intvalue) > strlen(haystack.strg)) v.intvalue=-1;
	}

	if (needle.strg) free (needle.strg); 
	if (haystack.strg) free (haystack.strg); 
	return v;

}




/*
	mainimage function
	syntax: imgvar = mainimage;
	Purpose: Gets the image stored in the interactive part
	(the main image) and assigns it to a variable.
*/

variable_type mac_mainimg (unsigned char* s, int* i)
{
variable_type v;

	v = nullvar; v.type = MT_IMG;
	if (!file_ok) macro_error (23,*i);
	clone_image (&mainimg, &v.imgptr);
	return v;
}




/*
	invert function
	syntax:	imgvar = invert (imgexpr);
	Parameter: Image expression to perform the operator on
	Invert subtracts each image value from the maximum.
	Example: i = invert (thisimage);
*/

variable_type mac_invert (unsigned char* s, int* i)
{
variable_type v;
int j;

	j=*i;
	v = expr (s,&j, MT_IMG);
	if (v.type != MT_IMG) macro_error (6,j);

	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;
	
	invert_image (&v.imgptr);
	return v;
}



/*
	sobel function ----- D E P R E C A T E D ------ Use CONVOLVE instead
	syntax:	imgvar = sobel (imgexpr);
	Parameter: Image expression to perform the Sobel operator on
	Sobel computes the gradient image of imgexpr and returns it
	to the expression handler.
	Example: grad = sobel (thisimage);
*/

variable_type mac_sobel (unsigned char* s, int* i)
{
variable_type v;
int j;

	j=*i;
	if (!isalpha1(s[j])) macro_error (1,j);
	v = expr (s,&j, MT_IMG);
	if (v.type != MT_IMG) macro_error (6,j);

	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;
	
	if (v.imgptr.imgtype == B_RGB)
		sobel_rgb (&v.imgptr);
	else
		sobel (&v.imgptr);
	return v;
}

/*
	Convolve function
	syntax:	imgvar = convolve (imgexpr, "filter" [,P]);
	Parameter: Image expression to perform the convolution on,
			"filter" is the filter function, which is one of
			"smooth", "sharpen", "sharpen more", "laplace" for standard 3x3 kernels,
			"sobel", "compass" for the Sobel and compass edge detectors
			"gauss" for generalized Gaussian smoothing (P is neighboorhood size 1...9)
			"LoG" for Laplacian of Gauss second derivative (P is neighboorhood size 1...9)
			"locvar" for local variance (P is neighboorhood size 1...9)
			"locmax" to find local maxima (P is neighboorhood size 1...9)
			"AMMSE" is the adaptive lowpass or minimum mean-squared error filter (P is neighboorhood size 1...9)
			"TSM" is the adaptive three-state median filter (P is a real-valued threshold )
			"ADLF" is the anisotropic diffusion lowpass filter (P1 is a real-valued threshold, P2 is the number of iterations (int) )
			"KUWA" the Kuwahara filter (P is neighboorhood size 1...9)
	Convolve convolves the source image of imgexpr with the filter
	kernel specified and returns the result to the expression handler.
	Example: lapl = convolve (thisimage, "laplace");

	NOTE: "convolve" is a misnomer as this contains most spatial-domain filters
	NOTE that we exclude the local Hoelder operator - it is too experimental 
			and should maybe have its own function.
*/

variable_type mac_convolve (unsigned char* s, int* i)
{
variable_type v,r;
int j,filtfn,p1;
float p2;
char filt[20];

	j=*i;
	/*if (!isalpha1(s[j])) macro_error (1,j);	// TODO: This line needs to go as it limits the possible expressions. Remove everywhere! */
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s, &j, MT_STRING);
	strncpy (filt, r.strg, 19);
	free (r.strg);

	/* extract_string (s, &j, filt, 15); */

	filtfn=0; p1=0; p2=0;										/* to silence a compiler warning */
	if (!strncasecmp (filt,"smooth",6)) filtfn=1;				/* 3x3 simple smoothing filter */
	else if (!strncasecmp (filt,"sharpen more",12)) filtfn=3;	/* NIH "sharpen more" Laplacian-based. Note the order. */
	else if (!strncasecmp (filt,"sharpen",7)) filtfn=2;			/* NIH "sharpen" Laplacian-based */
	else if (!strncasecmp (filt,"laplace",7)) filtfn=4;			/* Pure Laplacian */
	else if (!strncasecmp (filt,"compas",6)) filtfn=5;			/* Compass edge detector */
	else if (!strncasecmp (filt,"sobel",5)) filtfn=6;			/* Sobel edge detector */
	else if (!strncasecmp (filt,"gauss",6)) filtfn=7;			/* Generalized Gaussian, requires P */
	else if (!strncasecmp (filt,"LoG",3)) filtfn=8;				/* LoG, requires P */
	else if (!strncasecmp (filt,"locvar",6)) filtfn=9;			/* Local variance, requires P */
	else if (!strncasecmp (filt,"locmax",6)) filtfn=10;			/* Local maxima, requires P */
	else if (!strncasecmp (filt,"AMMSE",5)) filtfn=11;			/* Adaptive min mean-squared error filter, requires P */
	else if (!strncasecmp (filt,"TSM",3)) filtfn=12;			/* Three-state median filter, float P is threshold */
	else if (!strncasecmp (filt,"ADLF",4)) filtfn=13;			/* Anisotropic Diffusion Lowpass Filter */
	else if (!strncasecmp (filt,"KUWA",4)) filtfn=14;			/* Kuwahara edge-preserving Filter */
	else macro_error (15,j);

	/* Retrieve the additional parameter if required */

	if (((filtfn>=7) && (filtfn<=10)) || (filtfn==14))		/* Gauss to locmax and Kuwahara */
	{
		if (s[j++]!=',') macro_error (1,j);
		r = expr (s,&j,MT_INT); p1 = r.intvalue;
	}
	else if (filtfn==11)					/* AMMSE requires p1 (neighborhood), optional p2 (gamma) */
	{
		p2 = 1.0;
		r = expr (s,&j,MT_INT); p1 = r.intvalue;
		if (s[j++]==',')
		{
			j++;							/* Skip the comma */
			r = expr (s,&j,MT_REAL); p2 = r.numvalue;
		}
		if (p2 < 1.0) p2=1.0; 
		if (p2 > 3.0) p2=3.0;
	}
	else if (filtfn==12)					/* TSM requires real-valued threshold */
	{
		if (s[j++]!=',') macro_error (1,j);
		r = expr (s,&j,MT_REAL); p2 = r.numvalue;
	}
	else if (filtfn==13)					/* ADLF requires real-value threshold and int-valued number of iterations */
	{
		if (s[j++]!=',') macro_error (1,j);
		r = expr (s,&j,MT_REAL); p2 = r.numvalue;
		if (s[j++]!=',') macro_error (1,j);
		r = expr (s,&j,MT_INT); p1 = r.intvalue;
	}

	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;


	if (filtfn <= 4)
		convolve_filter (&v.imgptr, filtfn);
	else if (filtfn == 5)
	{
		if (v.imgptr.imgtype == B_RGB)
			compass_edge_rgb (&v.imgptr);
		else
			compass_edge (&v.imgptr);
	}
	else if (filtfn == 6)
	{
		if (v.imgptr.imgtype == B_RGB)
			sobel_rgb (&v.imgptr);
		else
			sobel (&v.imgptr);
	}
	else if (filtfn == 7)
		gaussian_general (&v.imgptr, p1, 0);
	else if (filtfn == 8)
		gaussian_general (&v.imgptr, p1, 1);
	else if (filtfn == 9)
		local_variance_2D (&v.imgptr, p1);
	else if (filtfn == 10)
		locmax_transform_2d (&v.imgptr, p1);
	else if (filtfn == 11)
		local_adaptive_lowpass (&v.imgptr, p1, p2, 2);		/* Compat mode with Q5, TODO: Allow Donoho's SD */
	else if (filtfn == 12)
		tristate_median_macro (&v.imgptr, p2);
	else if (filtfn == 13)
		anisotropic_diffusion_iter (&v.imgptr, p2, 0.2, p1);
	else if (filtfn == 14)
		kuwahara_2D (&v.imgptr, p1);

	return v;
}


/*
	freichen function
	syntax:	imgvar = freichen (imgexpr, "mode");
	Parameter: Image expression to perform the convolution on,
			"filter" is either "edge" for the Frei & Chen edge
			space, or "line" for the Frei & Chen line space.
	Freichen applies the Frei and Chen edge/line detector algorithm.
	Example: fcedge = freichen (thisimage, "edge");
*/

variable_type mac_freichen (unsigned char* s, int* i)
{
variable_type v,r;
int j,filtfn;
char filt[20];

	j=*i;
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s, &j, MT_STRING);
	strncpy (filt, r.strg, 19);
	free (r.strg);
	/* extract_string (s, &j, filt, 15); */

	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;

	filtfn=0;	/* to silence a compiler warning */
	if (!strncasecmp (filt,"edge",4)) filtfn=0;
	else if (!strncasecmp (filt,"line",4)) filtfn=1;
	else macro_error (15,j);

	frei_chen_filter (&v.imgptr, filtfn);
	return v;
}


/*
	fourierfilt function
	syntax:	imgvar = fourierfilt (imgexpr, "filter", a,b,c);
	        imgvar = fourierfilt (imgexpr, "filter", a,b,c, "opts");
	        imgvar = fourierfilt (imgexpr, "filter", a,b,c, "opts", "opts");
	Parameter: Image expression to perform the convolution on,
			"filter" is either "hhp" or "hlp" or "glp" or "deriv"
			a,b,c,d are the filter parameters (real).
			"opts" is an option string (up to two can be provided)

			The "filter" parameter means:
				"hhp"	Homomorphic HP				requires a (steepness), b(cutoff)
				"bhp"	Butterworth Highpass		requires a (asymmetry), b (cutoff), c (order)
				"blp"	Butterworth lowpass			requires a (asymmetry), b (cutoff), c (order)
				"glp"	Gaussian lowpass			requires a (sigma), b (shoulder cutoff)
				"deriv"	First derivative			requires a (Shepp-Logan-ness)
				"opp"	Oppenheim HP				requires a (DC attenuation), b (steepness)
				"wien"	Wiener filter				requires a (sigma), b (cutoff), c (denominator offset)

			The "opts" parameter can be any of:
				"log"	Apply filter on log data for multiplicative filtering
				"hann"	Apply Hann window
				"hamm"	Apply Hamming window
				"gauss"	Apply Gaussian window
			Clearly, log data and windowing is mutually exclusive (sorry!)

	Fourierfilt performs a fft, applies the specified filter function,
	and transforms it back into the spatial domain.
	Fourierfilt_log does the same operation, but on log data so that
	a background subtraction is rather a background division.
	
	Example: newimg = fourierfilt (thisimage, "hhp", 10.0,0.3,0.2,0.0);
			 newimg = fourierfilt (thisimage, "blp", 10.0,0.3,0.2,0.0, "hamm");

*/

variable_type mac_fourierfilt (unsigned char* s, int* i)
{
variable_type v,r;
int j,filtfn, logflag;
WINDOW_TYPE wndtype;
char filt[20];
char opts1[20], opts2[20];
float a,b,c,d;

	j=*i;
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s, &j, MT_STRING);
	strncpy (filt, r.strg, 19);
	free (r.strg);

	/* Extract required filter coefficients a, b, c (but not all of them are used in all filters) */

	if (s[j++]!=',') macro_error (1,j);
	a = b = c = d = 0.0;
	r = expr (s,&j,MT_REAL); a = r.numvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_REAL); b = r.numvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_REAL); c = r.numvalue;

	logflag=0; wndtype=W_NONE; opts1[0]=0; opts2[0]=0;

	if (s[j]==',') 				/* Optional filter opts */
	{
		j++;
		r = expr (s, &j, MT_STRING);
		strncpy (opts1, r.strg, 19);
		free (r.strg);
	}
	if (s[j]==',') 				/* More? Optional filter opts */
	{
		j++;
		r = expr (s, &j, MT_STRING);
		strncpy (opts2, r.strg, 19);
		free (r.strg);
	}

	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;

	filtfn=0;	/* to silence a compiler warning */
	if (!strncasecmp (filt,"hhp",3)) filtfn=1;
	else if (!strncasecmp (filt,"glp",3)) filtfn=2;
	else if (!strncasecmp (filt,"deriv",5)) filtfn=5;
	else if (!strncasecmp (filt,"blp",3)) filtfn=3;
	else if (!strncasecmp (filt,"bhp",3)) filtfn=4;
	else if (!strncasecmp (filt,"opp",3)) filtfn=7;
	else if (!strncasecmp (filt,"wien",4)) filtfn=6;
	else macro_error (15,j);

	if (strlen (opts1) > 0)
	{
		if (!strncasecmp (opts1, "log", 3)) logflag=1;
		else if (!strncasecmp (opts1,"hann",4)) wndtype=W_HANN;
		else if (!strncasecmp (opts1,"hamm",4)) wndtype=W_HAMMING;
		else if (!strncasecmp (opts1,"gauss",5)) wndtype=W_GAUSS;
		else if (!strncasecmp (opts1,"welch",5)) wndtype=W_WELCH;
		else macro_error (15,j);
	}
	if (strlen (opts2) > 0)
	{
		if (!strncasecmp (opts2, "log", 3)) logflag=1;
		else if (!strncasecmp (opts2,"hann",4)) wndtype=W_HANN;
		else if (!strncasecmp (opts2,"hamm",4)) wndtype=W_HAMMING;
		else if (!strncasecmp (opts2,"gauss",5)) wndtype=W_GAUSS;
		else if (!strncasecmp (opts2,"welch",5)) wndtype=W_WELCH;
		else macro_error (15,j);
	}

	/* In calling these filters, we need to consider the obsolete gain coefficient c.
		It is fixed to 1.0. The fourth coefficient passed is the third macro coefficient. */

	if (!logflag)
		fftfilt_go (&v.imgptr, filtfn,a,b,1.0,c, wndtype);
	else
		fftfilt_log_go (&v.imgptr, filtfn,a,b,1.0,c, wndtype);
	return v;
}




/*
	runlength function
	syntax:	imgvar = runlength (imgexpr, graybins, runbins);
	Parameter: Image expression to perform the convolution on,
			graybins is the number of bins for the gray scale axis
			runbins is the number of bins for the run length axis
	
	runlength computes a run length histogram on a gray scale image.
	A run is the number of consecutive pixels along a straight line
	that fall into the same gray bin. The histogram is two-dimensional
	(y-axis is the gray scale, x axis is the runlength), and the magnitude
	is coded in the image values (surface elevation).
	The returned image is a stack with 4 slices, each corresponding
	to one run angle: 0�, 45�, 90�, 135�.
	In addition, the run length statistics in the order given in 
	Table 8.5 are returned in user[].
	Example: newimg = runlength(inpimg, 16, 16);
*/

variable_type mac_runlength (unsigned char* s, int* i)
{
variable_type v,r;
int j;
int gb, rb;

	j=*i;
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); gb = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); rb = r.intvalue;

	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;

	runlength (&v.imgptr, gb,rb);
	runlength_stats (&v.imgptr, mac_user);
	return v;
}


/*
	skeletonize function
	syntax:	imgvar = skeletonize (imgexpr, thresh);
			imgvar = skeletonize (imgexpr, thresh, slice);
	Parameter: Image expression to perform the skeletonization on
				thresh = Threshold value to binarize the image
	If the third parameter is given, only the specified slice will be 
	skeletonized (3D images only)
	Example: s = skeletonize (thisimage,0);
*/

variable_type mac_skeletonize (unsigned char* s, int* i)
{
variable_type v,r;
int j,z,z1;
float thresh;

	j=*i;
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_REAL); thresh = r.numvalue;
	
	if (s[j]==',') 			/* slice parameter */
	{
		j++;
		r = expr (s,&j,MT_INT);
		z1 = r.intvalue;
		if (z1>=v.imgptr.zmax) macro_error (15,j);
		skeletonize (&v.imgptr, z1, thresh);
	}
	else
	{
		for (z=0; z<v.imgptr.zmax; z++)
			skeletonize (&v.imgptr, z, thresh);
	}

	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;
	
	return v;
}



/*
	morph function
	syntax:	imgvar = morph (imgexpr, "function", neighbors);
	       	imgvar = morph (imgexpr, "function", neighbors, slice);
	Parameter: Image expression to perform the morphological op on
				"function": Morphological operator, one of "erode",
				"dilate", "median", "patch", "bound_erode"
				Neighbors (either 4 or 8) defines the neighborhood
	If the fourth parameter is given, only the specified slice will be 
	filtered (3D images only)
	Example: s = morph (thisimage,"rank",8);
*/

variable_type mac_morph (unsigned char* s, int* i)
{
variable_type v,r;
int j,z,z1,n;
char filt[20];
MORPHOP what;

	j=*i;
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s, &j, MT_STRING);
	strncpy (filt, r.strg, 19);
	free (r.strg);
	/* extract_string (s, &j, filt, 15); */
	
	what = M_BYPASS;
	if (!strncasecmp (filt, "erode",5)) what=M_ERODE;
	else if (!strncasecmp (filt, "dilate",6)) what=M_DILATE;
	else if (!strncasecmp (filt, "rank",4)) what=M_RANK;
	else if (!strncasecmp (filt, "median",6)) what=M_RANK;
	else if (!strncasecmp (filt, "patch",5)) what=M_PATCH;
	else if (!strncasecmp (filt, "bound_erode",11)) what=M_BOUNDERODE;
	else if (!strncasecmp (filt, "cwmf",4)) what=M_CWMF;
	else
		macro_error (15,j);
	

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); n = r.intvalue;
	if ((n!=4) && (n!=8)) macro_error (15,j);	

	if (s[j]==',') 			/* slice parameter */
	{
		j++;
		r = expr (s,&j,MT_INT);
		z1 = r.intvalue;
		if (z1>=v.imgptr.zmax) macro_error (15,j);
		morph_img (&v.imgptr, what, z1, (n==8));
	}
	else
	{
		for (z=0; z<v.imgptr.zmax; z++)
			morph_img (&v.imgptr, what, z, (n==8));
	}

	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;
	
	return v;
}


/*
	threshold function
	syntax:	imgvar = threshold (imgexpr, "method", value);
	       	imgvar = threshold (imgexpr, "method");
	Parameter: Image expression to perform the thresholding on,
			method is one of "binary", "soft", "hard", "otsu", "iterative", "saturate", "satsym", or "clamp".

	Threshold thresholds in one of the following 3 ways:
		1) If method is one of "binary", "soft", "hard", "softsym", "hardsym", "saturate", or "satsym"
			- value needs to be defined. This is the threshold value
			- "binary" creates a pure 1/0 image
			- "soft" subtracts the threshold from the image, but clips negative values to 0
			- "hard" sets everything below the threshold to 0
			- "softsym" sets all values with an absolute value below the threshold to 0
					and reduces the absolute magniture of all values above the threshold
					by the threshold (origin-symmetric soft thresholding)
			- "hardsym" sets all values with an absolute value below the threshold to 0
					this implies that all values more negative than the negative threshold
					remain unchanged (origin-symmetric hard thresholding)
			- "saturate" sets all values above the threshold to the threshold value, 
					and leaves all values below unchanged
			- "satsym" acts like "saturate", but saturates values above the threshold
					and negative values below the negative threshold
		2) If method is one of "otsu", "iterative", "fvalley",
			- value must not be specified
			- For each slice, the threshold is determined by iterative or Otsu's method
			- The image (stack) is binarized as in method "hard" above
		3) If method is "clamp",
			- value needs to be defined. This is the threshold value
			- Clamp acts like "saturate", but uses its own separate function.
			- Should be obsoleted?

	Example: bin = threshold (thisimage, "binary", 115);
			 bin = threshold (myimg, "otsu");
*/

variable_type mac_threshold (unsigned char* s, int* i)
{
variable_type v,r,q;
int j;
char filt[20];
THRESHOLD_TYPE ty;
int tt;

	j=*i;
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	q = expr (s, &j, MT_STRING);
	strncpy (filt, q.strg, 19);
	free (q.strg);
	ty = TH_BINARY;	tt=0; /* to silence a compiler warning */
	if (!strncasecmp (filt,"binary",6)) { ty=TH_BINARY; tt=0; }
	else if (!strncasecmp (filt,"softsym",7)) { ty=TH_SOFTSYM; tt=0; }
	else if (!strncasecmp (filt,"hardsym",7)) { ty=TH_HARDSYM; tt=0; }
	else if (!strncasecmp (filt,"soft",4)) { ty=TH_SOFT; tt=0; }
	else if (!strncasecmp (filt,"hard",4)) { ty=TH_HARD; tt=0; }
	else if (!strncasecmp (filt,"otsu",4)) { ty=TH_HARD; tt=2; }
	else if (!strncasecmp (filt,"iter",4)) { ty=TH_HARD; tt=1; }
	else if (!strncasecmp (filt,"entropy",7)) { ty=TH_HARD; tt=3; }
	else if (!strncasecmp (filt,"fvalley",7)) { ty=TH_HARD; tt=4; }
	else if (!strncasecmp (filt,"clamp",5)) { ty=TH_HARD; tt=5; }
	else if (!strncasecmp (filt,"saturate",8)) { ty=TH_SATURATE; tt=0; }
	else if (!strncasecmp (filt,"satsym",6)) { ty=TH_SATSYM; tt=0; }
	else macro_error (15,j);

	
	if ((tt==0) || (tt==5))					/* Get threshold value for "bin", "soft", & "hard"  and "clamp" */
	{
		if (s[j++]!=',') macro_error (1,j);
		r = expr (s, &j, MT_REAL);			/* the threshold value */
	}
	
	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;


	if (tt==0)
	{
		thresh_return.thresh_type = ty;
		thresh_return.img = &v.imgptr;
		thresh_return.t = r.numvalue;
		thresh_return.r = thresh_return.g = thresh_return.b = thresh_return.t;	/* Make sure we also populate the RGB values */
		do_thresholding();
	}
	else if ((tt==1) || (tt==2) || (tt==3) || (tt==4))
	{
		do_thresholding_universal (&v.imgptr, -1, (tt-1), 0);
	}
	else if (tt==5)
	{
		clamp_image (&v.imgptr, r.numvalue);
	}


	return v;
}


/*
	threshold2 function
	syntax:	imgvar = threshold2 (imgexpr, lthresh, uthresh);
	syntax:	imgvar = threshold2 (imgexpr, lthresh, uthresh, slice);
	Parameter: Image expression to perform the thresholding on,
		lthresh = lower threshold; uthresh = upper threshold
		If slice is selected, the function will only performed on one slice
	Threshold2 is a threshold-"bandpass" which sets to 0 all
	image areas below lthresh or above uthresh and sets
	everything else to 1.
	Example: bin = threshold2 (thisimage, 128, 131);
*/

variable_type mac_threshold2 (unsigned char* s, int* i)
{
variable_type v,rl,ru,r;
int j;
int z,z1;

	j=*i;
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	rl = expr (s, &j, MT_REAL);			/* the threshold value */
	if (s[j++]!=',') macro_error (1,j);
	ru = expr (s, &j, MT_REAL);			/* the threshold value */

	thresh_return.thresh_type = TH_BINARY;
	thresh_return.img = &v.imgptr;
	thresh_return.t = rl.numvalue;

	if (s[j]==',') 			/* slice parameter */
	{
		j++;
		r = expr (s,&j,MT_REAL);
		z1 = (int) r.numvalue;
		if (z1>=v.imgptr.zmax) macro_error (15,j);
		do_thresholding2( rl.numvalue, ru.numvalue, z1);
	}
	else
		do_thresholding2( rl.numvalue, ru.numvalue, -1);


	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;

	return v;
}


/*
	auto_threshold function
	syntax:	numvar = auto_threshold (imgexpr, slice);
	Parameter: Image expression to determine ideal threshold for
		Slice is the image slice to perform the function on.
	Auto_threshold computes the iterative threshold for an image
	under the assumption of two distinct peaks in the histogram
	(e.g. object and background). This function can be used
	a parameter in the threshold function.
	Example: bin = threshold (thisimage, auto_threshold(thisimage,0), "hard");
*/

variable_type mac_auto_threshold (unsigned char* s, int* i)
{
variable_type v,r;
int j;
int z;
float thresh;

	j=*i;
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s, &j, MT_INT);			/* the slice */
	while (s[j]!=')') j++;
	*i = j+1;

	z = r.intvalue;
	thresh = threshold_iterative (&v.imgptr,z);
	freebuf (&v.imgptr);
	r = nullvar;
	r.numvalue = thresh; r.type=MT_REAL;
	return r;
}


/*
	otsu_threshold function
	syntax:	numvar = otsu_threshold (imgexpr, slice);
	Parameter: Image expression to determine ideal threshold for.
		Slice is the image slice to perform the function on.
	Otsu_threshold computes the threshold for an image by Otsu's method
	under the assumption of two distinct peaks in the histogram
	(e.g. object and background). This function can be used as
	a parameter in the threshold function.
	Example: bin = threshold (thisimage, otsu_threshold(thisimage,0), "hard");
*/

variable_type mac_otsu_threshold (unsigned char* s, int* i)
{
variable_type v,r;
int j;
int z;
float thresh;

	j=*i;
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s, &j, MT_INT);			/* the slice */
	while (s[j]!=')') j++;
	*i = j+1;

	z = r.intvalue;
	thresh = otsus_threshold (&v.imgptr,z,0);
	freebuf (&v.imgptr);
	r = nullvar;
	r.numvalue = thresh; r.type=MT_REAL;
	return r;
}



/*
	hysteresis_threshold function
	syntax:	imgvar = hysteresis_threshold (imgexpr, lthresh, uthresh);
	Parameter: Image expression to perform the thresholding on,
		lthresh = lower threshold; uthresh = upper threshold
		If slice is selected, the function will only performed on one slice
	Hysteresis thresholding is a combination of thresholding and region growing.
	Segmentation starts with thresholding above the upper threshold followed by
	region growing of regions above the lower threshold.
	The result is a binary mask.
	Example: bin = hysteresis_threshold (thisimage, 45, 90);
*/

variable_type mac_hysteresis_threshold (unsigned char* s, int* i)
{
variable_type v,rl,ru;
int j;

	j=*i;
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	rl = expr (s, &j, MT_REAL);			/* the threshold value */
	if (s[j++]!=',') macro_error (1,j);
	ru = expr (s, &j, MT_REAL);			/* the threshold value */

	thresh_return.thresh_type = TH_BINARY;
	thresh_return.img = &v.imgptr;
	thresh_return.t = rl.numvalue;

	hysteresis_threshold ( &v.imgptr, rl.numvalue, ru.numvalue);

	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;

	return v;
}




/*
	adaptive_threshold function
	syntax:	imgvar = adaptive_threshold (imgexpr, method, radius, bias);
			imgvar = adaptive_threshold (imgexpr, method, radius, bias, smooth);
	Parameter: Image expression to perform the thresholding on,
		method is any of
			"unsharp": 		Use FFT-based unsharp masking
			"movavg":  		Use moving-average window
			"iterative": 	Use ISODATA in local window
			"otsu": 		Use Otsu's method in local window
		radius is the radius of the local window
		bias is an offset (positive or negative) to globally bias the local threshold
		smooth is an optional parameter, if set to 1, causes Gaussian smoothing of the threshold map
	The result is a hard-thresholded image.
	Example: result_img = adaptive_threshold (thisimage, "iterative", 60, 0);
*/

variable_type mac_adaptive_threshold (unsigned char* s, int* i)
{
variable_type v,q,r;
int j, method,radius, smooth;
float bias;
unsigned char smethod[64];

	j=*i;
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	q = expr (s, &j, MT_STRING);
	strncpy (smethod, q.strg, 12);
	free (q.strg);
	method = 0;		 /* to silence a compiler warning */
	if (!strncasecmp (smethod,"unsharp",7)) { method=0; }
	else if (!strncasecmp (smethod,"movavg",6)) { method=1; }
	else if (!strncasecmp (smethod,"iterative",9)) { method=2; }
	else if (!strncasecmp (smethod,"otsu",4)) { method=3; }

	if (s[j++]!=',') macro_error (1,j);	/* Skip over next comma, mandatory arg */
	r = expr (s, &j, MT_INT);			/* the radius value */
	radius = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s, &j, MT_REAL);			/* the threshold bias value */
	bias = r.numvalue;

	smooth=0;
	if (s[j]==',') 			/* Optional smoothing flag */
	{
		j++;
		r = expr (s,&j,MT_INT);
		smooth = r.intvalue;
	}

	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*i = j+1;

	/* Execute the adaptive threshold function */

	adaptive_threshold ( &v.imgptr, radius, bias, 0, method, smooth);


	return v;
}



/*
	crop
	Syntax: imgvar = crop (imgexpr, x1,y1,x2,y2);
	This procedure crops an image to the absolute coordinates
	(x1,y1) to (x2,y2).
	Example: smallimg = crop (myimg, 10,10,250,250);
*/

variable_type mac_crop (unsigned char* s, int* i)
{
variable_type v,r;
int j,k,x1,y1,x2,y2;

	j = *i; k=0; 
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); x1 = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); y1 = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); x2 = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); y2 = r.intvalue;
	while (s[j]!=')') j++;
	*i = j+1;
	
	do_crop (&v.imgptr, x1,y1,x2,y2);
	
	return v;
}


/*
	rotate function
	Syntax: imgvar = rotate (imgexpr, angle);
	This procedure rotates an image clockwise by angle degrees.
	Example: rotimg = rotate (myimg, 35);
*/

variable_type mac_rotate (unsigned char* s, int* i)
{
variable_type v,r;
int j;

	j = *i; 
	/*if (!isalpha(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_REAL); 
	while (s[j]!=')') j++;
	*i = j+1;
	
	rotate_img (&v.imgptr, r.numvalue);
	
	return v;
}


/*
	resize function
	Syntax: imgvar = resize (imgexpr, newx, newy);
			imgvar = resize (imgexpr, binsize);
	This procedure resizes an image by interpolation or by rebinning.
	If two parameters are provided, the new image size is given 
	by newx x newy pixels. If one parameter is provided, 
	the new image size is reduced by the parameter.
	Example:  new = resize (img, 1024, 1024);
			  binned = resize (img, 2);
*/

variable_type mac_resize (unsigned char* s, int* i)
{
variable_type v,r; 
int xnew,ynew;
int j;

	j = *i; 
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);			/* One parameter is required */
	r = expr (s,&j, MT_INT); xnew = r.intvalue;

	if (s[j]==',') 			/* Second parameter - resize by interpolation */
	{
		j++;
		r = expr (s,&j,MT_INT); ynew = r.intvalue;
		zoom2D (&v.imgptr, xnew,ynew);
	}
	else				/* One parameter only - reduce size by binning */
	{
		if ((xnew<2) || (xnew>4)) 
			macro_error (25, j);
		rebin_img (&v.imgptr, xnew);
	}

	while (s[j]!=')') j++;
	*i = j+1;
	
	return v;
}




/*
	imginfo function
	syntax:	numvar = imginfo (imgvar, "info");
	Parameter: Image expression to get data from.
		"info" determines what image data to return. May be any of
		"xmax",ymax",zmax","minval","maxval","meanval","type".

	imginfo returns some global image information.
	Example: xmax = imginfo (thisimage, "xmax");
	Note that "meanval" computes the mean value in a loop over
		all image pixels (time-consuming)
	Note that "type" returns a numerical value that needs to be
		translated into "char" "short" "float" and "rgb"
*/

variable_type mac_imginfo (unsigned char* s, int* i)
{
variable_type v,r;
int j,k,vidx,x,y,z;
float avg,cnt;
char info[16];


	j = *i; k=0; memset (v.name, 0, sizeof(v.name));
	if (!isalpha1(s[j])) macro_error (1,j);
	
	do
	{
		v.name[k++] = s[j++];
	} 
	while ((k<16) && (isalpha2(s[j])));
	
	vidx = findvar (v.name);
	if (vidx<0)		/* not found error */
		macro_error (16,j);
	if (vars[vidx].type != MT_IMG) macro_error (6,j);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s, &j, MT_STRING);
	strncpy (info, r.strg, 15);
	free (r.strg);

	while (s[j]!=')') j++;
	*i = j+1;

	v = vars[vidx];
	r = nullvar; r.type=MT_INT;	/* Default type */
	if (!strncasecmp (info,"xmax",4)) r.intvalue = v.imgptr.xmax;
	else if (!strncasecmp (info,"ymax",4)) r.intvalue = v.imgptr.ymax;
	else if (!strncasecmp (info,"zmax",4)) r.intvalue = v.imgptr.zmax;
	else if (!strncasecmp (info,"type",4)) r.intvalue = v.imgptr.imgtype;
	else if (!strncasecmp (info,"maxval",6))
	{
		if (v.imgptr.imgtype==B_FLOAT)
		{
			r.numvalue = v.imgptr.maxval.fmax;
			r.type=MT_REAL;
		}
		else
			r.intvalue = v.imgptr.maxval.imax;
	}
	else if (!strncasecmp (info,"minval",6))
	{
		if (v.imgptr.imgtype==B_FLOAT)
		{
			r.numvalue = v.imgptr.minval.fmin;
			r.type=MT_REAL;
		}
		else
			r.intvalue = v.imgptr.minval.imin;
	}
	else if (!strncasecmp (info,"meanval",7))
	{
		avg=0; cnt=0;
		for (z=0; z<v.imgptr.zmax; z++)
			for (y=0; y<v.imgptr.ymax; y++)
				for (x=0; x<v.imgptr.xmax; x++)
				{
					avg+=readbuf_flt (&v.imgptr,x,y,z);
					cnt++;
				}
		r.numvalue = avg/cnt;
		r.type = MT_REAL;
	}
	else
		macro_error (15,j);

	return r;
}


/*
	quantile function
	Syntax: numvar = quantile (imgexpr, slice, quant, numbins, oorflg);
	This function computes the threshold value that
	corresponds to the quant% quantile of the image
	Numbins determines the accuracy through the number of bins.
	For a 8-bit image, numbins should be no more than 256.
	oorflg determines if the zero-value is interpreted as
	out-of-region.
	Example: t = quantile (myimg, 50, 1024, 1);
*/

variable_type mac_quantile (unsigned char* s, int* i)
{
variable_type v,r;
int j,nb,oor,z;
float quant;

	j = *i; 
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); z = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_REAL); quant = r.numvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); nb = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); oor = r.intvalue;
	while (s[j]!=')') j++;
	*i = j+1;
	
	r = nullvar;
	r.numvalue = find_quantile_standalone (&v.imgptr, z,nb,quant,oor);
	r.type = MT_REAL;
	
	freebuf (&v.imgptr);
	
	return r;
}


/*
	substack function
	Syntax: imgvar = substack (imgexpr, zmin, zmax);
	Imgexpr must be a stack. Substack extracts slices from 
	zmin to zmax (inclusive).
	Example: oneslice = substack (myimg, 1,1);
*/

variable_type mac_substack (unsigned char* s, int* i)
{
variable_type v,r,v1;
int j,z1,z2;

	j = *i; 
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); z1 = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); z2 = r.intvalue;
	while (s[j]!=')') j++;
	*i = j+1;
	if (z1>z2) macro_error (15,j);
	
	/* Check if the substack is within the stack bounds of the original image */
	
	if ((z1 >= v.imgptr.zmax) || (z2 >= v.imgptr.zmax)) macro_error (15,j);

	v1=nullvar; v1.type = MT_IMG;
	if (allocate_image (&v1.imgptr, v.imgptr.imgtype, v.imgptr.xmax, v.imgptr.ymax, 1+z2-z1)==-1)
		fatal_memerr();

	copybuf_z (&v.imgptr, &v1.imgptr, z1,z2);	
	freebuf (&v.imgptr);
	make_minmax (&v1.imgptr);
	
	return v1;
}


/*
	rgbtostack function
	Syntax: imgvar = rgbtostack (imgexpr);
	Imgexpr must be a RGB image. Each RGB slice will
	be converted into 3 CHAR slices (changing memory organization,
	not memory usage), and the output image has 3 times as many
	slices as the input image.
	Example: stk = rgbtostack (myimg);
*/

variable_type mac_rgb2stack (unsigned char* s, int* i)
{
variable_type v;
int j;

	j = *i; 
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	while (s[j]!=')') j++;
	*i = j+1;

	if (v.imgptr.imgtype!=B_RGB) macro_error (17,j);
	
	dataconv_expand_rgb (&v.imgptr);
	return v;
}


/*
	imgtorgb function
	Syntax: imgvar = imgtorgb (imgexpr1,imgexpr2,imgexpr3);
	Imgexpr1,2,3 will be converted to graycsale (if needed),
	range-compressed to 0-255, and then joined into one
	RGB image.
	Example: colorful = imgtorgb (redimg, greenimg, blueimg);
*/

variable_type mac_imgtorgb (unsigned char* s, int* i)
{
variable_type ri,gi,bi,v;
int j,x1,y1,z1,x,y,z;
unsigned char *p1;
float buf,min,max;

	j = *i; 
	ri = expr (s,&j, MT_IMG);
	if (s[j++]!=',') macro_error (1,j);
	gi = expr (s,&j, MT_IMG);
	if (s[j++]!=',') macro_error (1,j);
	bi = expr (s,&j, MT_IMG);
	while (s[j]!=')') j++;
	*i = j+1;

	x1 = MAX( MAX (ri.imgptr.xmax, gi.imgptr.xmax), bi.imgptr.xmax);
	y1 = MAX( MAX (ri.imgptr.ymax, gi.imgptr.ymax), bi.imgptr.ymax);
	z1 = MAX( MAX (ri.imgptr.zmax, gi.imgptr.zmax), bi.imgptr.zmax);

	v.type=MT_IMG;
	if (allocate_image (&v.imgptr,B_RGB, x1,y1,z1)==-1)
		fatal_memerr();
	
	p1=v.imgptr.data; 
	imgminmax (&ri.imgptr, &min, &max);
	for (z=0; z<ri.imgptr.zmax; z++)
		for (y=0; y<ri.imgptr.ymax; y++)
			for (x=0; x<ri.imgptr.xmax; x++)
			{
				buf = (readbuf_flt (&ri.imgptr,x,y,z) - min)/(max-min);
				p1[3*(x + mainimg.xmax*(y + mainimg.ymax*z))] = (int)(255.0*buf);
			}
	freebuf (&ri.imgptr);

	imgminmax (&gi.imgptr, &min, &max);
	for (z=0; z<gi.imgptr.zmax; z++)
		for (y=0; y<gi.imgptr.ymax; y++)
			for (x=0; x<gi.imgptr.xmax; x++)
			{
				buf = (readbuf_flt (&gi.imgptr,x,y,z) - min)/(max-min);
				p1[1+3*(x + mainimg.xmax*(y + mainimg.ymax*z))] = (int)(255.0*buf);
			}
	freebuf (&gi.imgptr);

	imgminmax (&bi.imgptr, &min, &max);
	for (z=0; z<bi.imgptr.zmax; z++)
		for (y=0; y<bi.imgptr.ymax; y++)
			for (x=0; x<bi.imgptr.xmax; x++)
			{
				buf = (readbuf_flt (&bi.imgptr,x,y,z) - min)/(max-min);
				p1[2+3*(x + mainimg.xmax*(y + mainimg.ymax*z))] = (int)(255.0*buf);
			}
	freebuf (&bi.imgptr);

	make_minmax (&v.imgptr);
	return v;
}



/*
	combine function
	Syntax: imgvar = combine (imgexpr1, imgexpr2, "method");
	Comines two images to either form a stack or a false-colored image,
	depending on the method. In both cases, imgexpr1 and imgexpr2 must have
	the same x and y size. For a color combination, the Z size must match, too.
		if "method" is "stack", the second image will be stacked onto the first image.
		if "method" is "color", both images will be combined in complementary colors,
			very much like in imgtorgb.
	
	Example: newstk = combine (oldstk, slc, "stack");
*/

variable_type mac_combine (unsigned char* s, int* i)
{
variable_type v,r,v1, v2;
int j, meth;
char buf[32];

	j = *i; 
	v1 = expr (s,&j, MT_IMG);
	if (s[j++]!=',') macro_error (1,j);
	v2 = expr (s,&j, MT_IMG);
	if (s[j++]!=',') macro_error (1,j);

	r = expr (s, &j, MT_STRING);
	strncpy (buf, r.strg, 15);
	free (r.strg);

	if (!strncasecmp (buf,"color",5)) meth=2;
	else if (!strncasecmp (buf,"stack",5)) meth=1;
	else macro_error (15,j);
	while (s[j]!=')') j++;
	*i = j+1;

	if (meth==1)				/* Stack combine */
	{
		if (combine1 (&v1.imgptr, &v2.imgptr) !=0)
			macro_error (12,j);
	}
	else if (meth==2)			/* Color combine */
	{
		if (combine2 (&v1.imgptr, &v2.imgptr) !=0)
			macro_error (12,j);
	}
	
	return v1;
}


/*---------------------------------------------------------------------------*/

/*
	remove_bk function
	Syntax: imgvar = remove_bk (imgexpr, method, r, oorflag);
	        imgvar = remove_bk (imgexpr, method, r, oorflag, slice);
	remove_kb performs background removal on one or all slices on a 2D basis,
	depending on the method selected as follows:
	"movavg" uses a moving-average window to reduce the background.
	The window is circular with raduis r. 
	"unsharp masking" uses FFT Gaussian blurring to create a background mask
	The Gaussian sigma is 400/r for compatibility with the radius, that is, 
	a larger r means weaker filter action.
	"plane" fits a linear plane where the tiles have a size of r.
	"parabolic" fits a 2D parabolic function where the tiles have a size of r.
	Example: oneslice = remove_bk (myimg, "movavg", 40,1);
*/

variable_type mac_remove_bk (unsigned char* s, int* i)
{
variable_type v,r;
int j,radius,z,z1, method;
rgbtriplet bandflg = {1,1,1};
int oorflg = 0;
char meth[20];

	j = *i; 
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	method=0;
	r = expr (s,&j, MT_STRING); strncpy (meth, r.strg, 15); free (r.strg);
	if (!strncasecmp (meth,"movav",5)) method=1;
	else if (!strncasecmp (meth,"unsha",5)) method=2;
	else if (!strncasecmp (meth,"plane",5)) method=3;
	else if (!strncasecmp (meth,"parab",5)) method=4;
	else macro_error (15,j);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); radius = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); oorflg = r.intvalue;

	if (s[j]==',') 			/* slice parameter */
	{
		j++;
		r = expr (s,&j,MT_INT);
		z1 = r.intvalue;
		if (z1>=v.imgptr.zmax) macro_error (15,j);
		mac_handle_background (&v.imgptr, method, radius, z1, bandflg, oorflg);
	}
	else
	{
		for (z=0; z<v.imgptr.zmax; z++)
			mac_handle_background (&v.imgptr, method, radius, z, bandflg, oorflg);
	}
	while (s[j]!=')') j++;
	*i = j+1;

	return v;
}


/*
	fill_holes function
	Syntax: imgvar = fill_holes (imgexpr);
	        imgvar = fill_holes (imgexpr, slice);
	Fill_holes assumes a 8-connected background of value zero
	that starts at (touches) the top left corner. This region 
	will be marked and any remaining holes of zero value filled.
	Example:  filled = fill_holes (fillme, 1);
*/

variable_type mac_fill_holes (unsigned char* s, int* i)
{
variable_type v,r;
int j,z,z1;

	j = *i; 
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j]==',') 			/* slice parameter */
	{
		j++;
		r = expr (s,&j,MT_INT);
		z1 = r.intvalue;
		if (z1>=v.imgptr.zmax) macro_error (15,j);
		fill_holes (&v.imgptr, z1);
	}
	else
	{
		for (z=0; z<v.imgptr.zmax; z++)
			fill_holes (&v.imgptr, z);
	}
	while (s[j]!=')') j++;
	*i = j+1;

	return v;
}



/*
	img_histo function
	Syntax: numvar = imghisto (imgexpr, x1,y1,z1, x2,y2,z2, delta, moment);
			numvar = imghisto (imgexpr, x1,y1,z1, x2,y2,z2, delta);
	
	This is a quickly hacked function which may not be permanent.
	It uses the values of an image, assumed to be the output of
	a run-length analysis, and computes the mean value using
	mean = sum from i=0 to max of (i*p(i))
	
	(x1,y1,z1) and (x2,y2,z2) are the start and end points of a profile
	through the image. Delta is the bin size. Moment is the moment
	number. For moment==0, the area under the curve is computed, while
	for moment > 0 the above equation holds:
	
			[ sum from i=0 to max of p(i)*i^moment ]
	result=	----------------------------------------
			[ sum from i=0 to max of p(i)]
	
*/

variable_type mac_img_histo (unsigned char* s, int* ii)
{
variable_type v,r;
int j,x1,y1,z1,x2,y2,z2,i,numpts,n;
double x,y,z, delta, area, moment, buf;
double dx,dy,dz;


	v = expr (s, ii, MT_IMG);
	j = *ii;

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); x1 = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); y1 = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); z1 = r.intvalue;

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); x2 = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); y2 = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); z2 = r.intvalue;

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_REAL); delta = r.numvalue;

	/* Optional moment parameter */
	
	if (s[j]==',') 
	{
		j++;
		r = expr (s,&j,MT_INT);
		n = r.intvalue;
	}
	else
		n=1;	

	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*ii = j+1;

	dx = x2-x1; dy = y2-y1; dz = z2-z1;
	if ( (fabs(dx)>fabs(dy)) && (fabs(dx)>fabs(dz)) )	/* x-dominant */
		numpts = abs(x2 - x1);
	else if ( (fabs(dy)>fabs(dx)) && (fabs(dy)>fabs(dz)) )	/* y-dominant */
		numpts = abs(y2 - y1);
	else 												/* z-dominant */
		numpts = abs(z2 - z1);

	dy /= numpts;
	dz /= numpts;
	dx /= numpts; 	/* This retains the sign */
	x=x1; y=y1; z=z1;

	area=0; moment=0;
	for (i=0; i<numpts; i++)
	{
		/* TODO: Doesn't work across slices */
	
		buf = ireadbuf (&v.imgptr, x,y,(int)(z+0.5),intp_pref);
		area += buf;
		if (n==0)
			moment += delta*buf;
		else if (n==1)
			moment += (delta*(i+1))*buf;
		else
			moment += pow(delta*(i+1),n)*buf;
		
		x+=dx; y+=dy; z+=dz;
	}
	
	r.numvalue=moment;
	if (n>0)
	{	
		if (area==0.0) r.numvalue=0.0; else r.numvalue = moment/area;
	}

	freebuf (&v.imgptr);
	r.type = MT_REAL;
	return r;

}



/*
	histogram function
	Syntax: numvar = histogram (imgexpr, "value", slice, numbins, oorflg);
	
	This function determines key values in a histogram of
	an image imagexpr.
	"Value" may be any of
	- "mean": The mean value
	- "median": The median value
	- "mode": The mode value
	- ...
	Numbins determines how many bins to use.
	The oorflag specifies if zero values should not be taken into account
*/

variable_type mac_histogram (unsigned char* s, int* ii)
{
variable_type v,r;
char value[20];
int mode,slice,numbins,j,oorflg;

	v = expr (s, ii, MT_IMG);
	j = *ii;

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s, &j, MT_STRING);
	strncpy (value, r.strg, 18);
	free (r.strg);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); slice = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); numbins = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j,MT_INT); oorflg = r.intvalue;

	/* Skip to closing parentheses, update pointer */

	while (s[j]!=')') j++;
	*ii = j+1;

	mode=0;
	if (!strncasecmp (value, "mode",4))
		mode=0;
	else if (!strncasecmp (value, "median",6))
		mode=1;
	else if (!strncasecmp (value, "mean",4))
		mode=2;
	else if (!strncasecmp (value, "variance",8))
		mode=3;
	else if (!strncasecmp (value, "skew",4))
		mode=4;
	else if (!strncasecmp (value, "kurtosis",8))
		mode=5;
	else macro_error (15,j);
	
	/* Now get that value. */
	
	r=nullvar; r.type=MT_REAL;
	r.numvalue = histogram_standalone (&v.imgptr, slice, numbins, mode, oorflg);
	
	freebuf (&v.imgptr);
	return r;
	

}



/*
	dimension function
	Syntax: numvar = dimension (imgexpr, "notion", slice, lrange, urange, lthresh, uthresh, oorflg);
	This function computes a fractal dimension on one specific slice of imgexpr.
	Notion may be one of "boxcount", "minkowsky", "massdimension", "manhattan dimension"
	slice specifies the slice to work on
	lrange and urange determine the scaling. lrange is the minimum box
	   size (box-counting), urange is the maximum box size (box-counting),
	   for example 2 to 128. Both Minkowsky and Mass are different. For Minkowsky,
	   this is the dilation radius to analyze (typical values 1 and 5), and for
	   Mass this is the number of scale reductions. For example 1 to 4 would
	   cover box sizes (squared) 4, 16, 64, 256.
	lthresh and uthresh determine the binarization threshold and are not needed
	   in the massdimension computation
	oorflg determines if zero should be interpreted as out-of-region
	   (massdimension only)
	Example: D = dimension (myimg, "minkowsky", 0, 1, 5, 0, 0, 1);

	NOTE: There is still a problem with the ROI - 
		an undefined ROI may cause problems

*/

variable_type mac_dimension (unsigned char* s, int* i)
{
variable_type v,r;
int lrange, urange, oorflg, z, j;
float lthresh, uthresh;
char notion[25];
fitresult fit;


	j = *i; 
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s, &j, MT_STRING);
	strncpy (notion, r.strg, 24);
	free (r.strg);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); z = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); lrange = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); urange = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_REAL); lthresh = r.numvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_REAL); uthresh = r.numvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); oorflg = r.intvalue;
	while (s[j]!=')') j++;
	*i = j+1;

	if (!strncasecmp (notion, "boxcount",8))
	{
		fit = boxdim2d (&v.imgptr, z, lrange, urange, lthresh, uthresh);
		r.numvalue = fabs(fit.H);
	}
	else if (!strncasecmp (notion, "minkowsky",9))
	{
		fit = minkovskydim (&v.imgptr, z, lrange, urange, lthresh, uthresh);
		r.numvalue = 2.0-fit.H;
	}
	else if (!strncasecmp (notion, "massdim",7))
	{
		fit = mass_dimension (&v.imgptr, z, lrange, urange, oorflg);
		r.numvalue = fabs(fit.H);
	}
	else if (!strncasecmp (notion, "manhattan",9))
	{
		fit = manhattan_boxdim (&v.imgptr, z, lrange, urange, oorflg);
		r.numvalue = 2.0-fabs(fit.H);
	}
	else if (!strncasecmp (notion, "dispersion",10))
	{
		fit = dispersion_boxdim (&v.imgptr, z, lrange, urange, oorflg);
		r.numvalue = 3.0-fabs(fit.H);
	}
	else macro_error (15,j);

	r.type = MT_REAL;
	freebuf (&v.imgptr);
	
	return r;
}




/*
	Lacunarity function
	Syntax: numvar = roseplot (imgexpr, slice, radius);
	This function computes the lacunarity of an image imgexpr
	(2D only, and acts on slice "slice" if it is a 3D image).
	The lacunarity function returns L(r) for one specific radius,
	defined by "radius".
	Both "slice" and "radius" are integer expressions.
	
	Example: Lr = lacunarity (someimg, 0, 10);

*/

variable_type mac_lacunarity (unsigned char* s, int* i)
{
	variable_type v,r;
	int slice, radius, j;

	j = *i; 
	v = expr (s,&j, MT_IMG);		/* Image to analyze */

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); slice = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); radius = r.intvalue;
	while (s[j]!=')') j++;
	*i = j+1;

	r.type = MT_REAL;			/* Return type is a real value */
	r.numvalue = lacunarity (&v.imgptr, slice, radius);
	
	freebuf (&v.imgptr);
	return r;
}





/*
	Rose plot function
	Syntax: numvar = roseplot (imgexpr, startpt, maxlen, numpts);
	This function computes a Hurst exponent in a frequency image.
	The image "imgexpr" must be a FFT, i.e. a two slice real/imaginary
	frequency transform.
	The rose plot fits a straight line into LOG(magnitude) over LOG(frequency)
	where the slope s = 2H+1 = 2D-1.
	Startpt is the lowest frequency to be considered in the fit,
	Maxlen is the number of points from the start point into which to fit a straight line,
	numpts is the number of angles starting at 0 (u axis). So the angular increment
	is 360 degrees/numpts.
	The function returns the average dimension.
	The function also provides ther average dimension in user[0], the
	anisotropy in user[1], and the individual angular points un user[2] and
	subsequent. Numpts is limited to 80.
	
	Example: D = roseplot (fftimg, 10, 50, 24);

*/

variable_type mac_roseplot (unsigned char* s, int* i)
{
	variable_type v,r;
	int startpt, maxlen, numrays, j;

	j = *i; 
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); startpt = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); maxlen = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); numrays = r.intvalue;
	while (s[j]!=')') j++;
	*i = j+1;

	r.type = MT_REAL;
	r.numvalue = rose_plot_for_macro (&v.imgptr, startpt, numrays, maxlen);
	
	freebuf (&v.imgptr);
	if (r.numvalue < 0) macro_error (19, j);
	
	return r;
}



/*
	fft function
	Syntax: imgvar = fft (imgexpr, slice);
	One slice of imgexpr will be Fourier-transformed, and the
	transform returned in imgvar. Imgvar will be a 2-slice stack
	with the real part in the first slice, and the imaginary part
	in the second slice. Imgvar will have the image type B_FLOAT.
	Example: myft = fft (myimg, 0);
*/

variable_type mac_fft (unsigned char* s, int* i)
{
variable_type v,r;
int j,slice;

	j = *i; 
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); slice = r.intvalue;
	while (s[j]!=')') j++;
	*i = j+1;

	fft2d (&v.imgptr, slice);

	return v;
}


/*
	invfft function
	Syntax: imgvar = invfft (imgexpr);
	imgexpr must be a Fourier image, i.e. must have 2 slices
	(first: real component, second: imaginary component) and
	must be of type B_FLOAT. Ifft performs a inverse Fourier-transform
	and returns the result (one slice, real-valued) in imgexpr.
	
	Example: myft = ifft (myimg);
*/

variable_type mac_invfft (unsigned char* s, int* i)
{
variable_type v;
int j;

	j = *i; 
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	while (s[j]!=')') j++;
	*i = j+1;

	if ((v.imgptr.imgtype!=B_FLOAT) || (v.imgptr.zmax!=2))
		macro_error (19,j);

	ifft2d (&v.imgptr);

	return v;
}




/*
	boundary_length function
	Syntax: numvar = boundary_length (imgexpr, zslice);
	This function determines the number of boundary pixels of
	a segmented image with feature(s). Assumptions: a. The image
	is segmented, i.e. background is zero. b. The image has a
	background pixel in the top-left corner. c. The image may have
	more than one feature; in that case, the number of boundary
	pixels for all features will be returned.

	Example: bound = boundary_length (field, 0);
	
*/


variable_type mac_boundary_len (unsigned char* s, int* i)
{
variable_type v,r;
int j, z;

	j = *i; 
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); z = r.intvalue;

	while (s[j]!=')') j++;
	*i = j+1;

	r = nullvar;
	r.intvalue = meas_boundary_pts (&v.imgptr, z);
	r.type = MT_INT;
	freebuf (&v.imgptr);

	return r;
}





/*
	kltransform function
	Syntax: imgvar = kltransform (imgvar1, oorflg);
	Compute the Karhunen-Loeve-transform of a RGB-image imgvar1.
	The Karhunen-Loeve transform delivers the principal component
	of the RGB image (of type B_FLOAT).
	oorflg determines if zero regions will be considered (0)
	or not considered in the transfrmation matrix computation.
	
*/

variable_type mac_kltransform (unsigned char* s, int* i)
{
variable_type v,r;
int j,k,oorflg,vidx;


	/* The image stack */

	j = *i; k=0; v = nullvar;
	if (!isalpha1(s[j])) macro_error (1,j);
	do
	{
		v.name[k++] = s[j++];
	} 
	while ((k<16) && (isalpha2(s[j])));
	
	vidx = findvar (v.name);
	if (vidx<0)		/* not found error */
		macro_error (16,j);
	if (vars[vidx].type != MT_IMG) macro_error (6,j);

	/* The oorflg parameter */	

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); oorflg = r.intvalue;

	while (s[j]!=')') j++;
	*i = j+1;

	/* Go execute. Prepare v to be returned as image */

	v = nullvar;
	v.type=MT_IMG;
	clone_image (&vars[vidx].imgptr, &v.imgptr);
	karhunen_loeve_transform (&v.imgptr, oorflg);

	return v;
}



/*
	skeleton_stats function
	syntax: strvar = skeleton_stats (imgvar);
	syntax: strvar = skeleton_stats (imgvar, prunelen);

	I don't get a break, do I? Yet another ugly hack that may or may not stay.
	This macro computes topology and geometry of a skeleton and returns a
	string variable that contains the report. 
	
	imgvar:	A skeletonized image. The user is responsible for
			passing a skeletonized image, otherwise the results
			are unpredictable. 
			Note: Pixels in this image are lableed according to
			they being a node, branch,link, endpoint.
	prunelen: Integer variable to determine the branch length
			to be pruned. 0 = no pruning. Branches up to prunelen
			are deleted from the skeleton image

	
*/


variable_type mac_skeletonstats (unsigned char* s, int* i)
{
variable_type rtn, r;
int j,k, vidx,prunelen;
char vname[32];
char report[32768];

	j = *i; k=0; 
	if (!isalpha1(s[j])) macro_error (1,j);
	
	do
	{
		vname[k++] = s[j++];
	} 
	while ((k<16) && (isalpha2(s[j])));
	vname[k]=0;
	
	vidx = findvar (vname);

	if (vidx<0)	macro_error (16,j);
	
	/* The prunelen parameter */	

	prunelen=0;
	if (s[j]==',') 			/* Optional prune parameter */
	{
		j++;
		r = expr (s,&j, MT_INT); prunelen = r.intvalue;
	}

	while (s[j]!=')') j++;
	*i = j+1;

	/*
	if (vars[vidx].type != MT_IMG) macro_error (6,j);
	
	/* Highly unlikely that a float or rgb image is a skeleton 
	if ((vars[vidx].imgptr.imgtype == B_RGB)  || (vars[vidx].imgptr.imgtype == B_FLOAT))
		macro_error (17,j);	
	*/

	if (prunelen) skel_prune (&vars[vidx].imgptr, prunelen);

	skel_statistics (&vars[vidx].imgptr, report);
	
	rtn.type = MT_STRING;
	rtn.strg = calloc (3+strlen(report),sizeof(char));
	strcpy (rtn.strg, report);

	return rtn;
}





/*
	cluster2d function
	cluster3d function
	Syntax: 
			imgvar = cluster2d (imgexpr, slice, threshold, minsize, maxclust, neighb, labelopts);
			imgvar = cluster3d (imgexpr, threshold, minsize, maxclust, neighb, labelopts);

	Find clusters in imgexpr (2D, one slice).
	threshold: float - Everything below this value is discarded from the image
	minsize: int - Minimum cluster size. Speckles below this size are deleted
	maxclust: int - Maximum number of clusters to detect
	neighb: int - select 0 (or 4) for four-neighborhood or 1 (or 8)  for 8-nghb
	labelopts: int - Label clusters by 
					0: size rank
					1: pure binary
					2: size			(2D only)
					3: compactness	(2D only)
					4: Aspect		(2D only)
					5: Irregularity	(2D only)

	Note: The cluster's centroid (x,y,z) and one value depending on the
	labelopts is stored in the user array for the first 20 clusters.

	Note: There is no option to use different thresholds in rgb images.
	Either threshold separately or use the same threshold for all panes

	Example: clst = cluster2d (myimg, 0, 1, 3, 32767, 4, 0);
*/

variable_type mac_cluster2d (unsigned char* s, int* i)
{
variable_type v,r;
int j,slice,minsize,maxclust,neighb,labelopts,k,l;
float thresh;
cluster_type* list;


	j = *i; 
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); slice = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_REAL); thresh = r.numvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); minsize = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); maxclust = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); neighb = r.intvalue;
	if ((neighb==0) || (neighb==4))
		neighb=0;
	else if ((neighb==1) || (neighb==8))
		neighb=1;
	else
		macro_error (15,j);
	
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); labelopts = r.intvalue;
	if ((labelopts<0) || (labelopts>5)) macro_error (15,j);

	while (s[j]!=')') j++;
	*i = j+1;

	/* Allocate memory for the list */
	
	list = calloc (maxclust+10, sizeof (cluster_type));
	if (!list) fatal_memerr();

	clust_thresh (&v.imgptr,thresh,thresh,thresh,thresh, 0);
	mac_user[0] = findclust_2d (&v.imgptr, slice, maxclust, neighb, minsize, labelopts, list);

	/* Copy the listed results into the user array - only the first 30 clusters */
	
	l=1;
	for (k=l; k<100; k++) mac_user[k]=0;
	for (k=0; k<MIN(30,maxclust); k++)
	{
		mac_user[l++] = list[k].cog_x;
		mac_user[l++] = list[k].cog_y;
		if (labelopts==0) mac_user[l++] = list[k].size;
		else if (labelopts==1) mac_user[l++] = list[k].compactness;
		else if (labelopts==2) mac_user[l++] = list[k].aspect;
		else if (labelopts==3) mac_user[l++] = list[k].irreg;
	}
	free (list); 

	return v;
}


/* TODO: The 3D function still uses the old algorithm and cannot handle labelopts */

variable_type mac_cluster3d (unsigned char* s, int* i)
{
variable_type v,r;
int j,minsize,maxclust,neighb,labelopts,k,l;
float thresh;
cluster_type* list;

	j = *i; 
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_REAL); thresh = r.numvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); minsize = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); maxclust = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); neighb = r.intvalue;
	if ((neighb==0) || (neighb==4))
		neighb=0;
	else if ((neighb==1) || (neighb==8))
		neighb=1;
	else
		macro_error (15,j);
	
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); labelopts = r.intvalue;
	if ((labelopts<0) || (labelopts>1)) macro_error (15,j);

	while (s[j]!=')') j++;
	*i = j+1;

	/* Allocate memory for the list */
	
	list = calloc (maxclust+10, sizeof (cluster_type));
	if (!list) fatal_memerr();

	clust_thresh (&v.imgptr,thresh,thresh,thresh,thresh, 0);
	mac_user[0] = findclust_3d (&v.imgptr, maxclust, neighb, minsize, labelopts, list);

	/* Copy the listed results into the user array - only the first 20 clusters */
	
	l=1;
	for (k=l; k<100; k++) mac_user[k]=0;
	for (k=0; k<MIN(20,maxclust); k++)
	{
		mac_user[l++] = list[k].cog_x;
		mac_user[l++] = list[k].cog_y;
		mac_user[l++] = list[k].cog_z;
		if (labelopts==0) mac_user[l++] = list[k].size;
		else if (labelopts==1) mac_user[l++] = list[k].compactness;
		else if (labelopts==2) mac_user[l++] = list[k].aspect;
		else if (labelopts==3) mac_user[l++] = list[k].irreg;
	}
	free (list); 


	return v;
}



/*
	clustercount function
	Syntax: 
			numvar = clustercount (imgexpr, threshold, samples, minsize, maxclust, lowdens);

	Determine the maximum number of clusters in dependency of the threshold

	threshold: float - Everything below/above this value is discarded from the image
	samples: int - number of samples to test between the threshold value and the image max value
	minsize: int - Minimum cluster size. Speckles below this size are deleted
	maxclust: int - Maximum number of clusters to detect
	lowdens: int - inverts the threshold selection.
			lowdens=0: A cluster lies between threshold and image max value
			lowdens=1: A cluster lies between image min value and threshold

	The function returns the maximum number of clusters found.
	The corresponding threshold is stored in user[0].

	Note: There is no option to use different thresholds in rgb images.
	Either threshold separately or use the same threshold for all panes
	
	Note: For 3D images, the 3D algorithm is automatically invoked.
	
	Note: This function always assumes 8-connectedness

	Example: maxclust = clustercount (myimg, 120.0, 50, 3, 32767, 0);
*/


variable_type mac_clustercount (unsigned char* s, int* i)
{
variable_type v, r;
int j,numsamp,minsize,maxclust,lowdens;
float thresh;


	j = *i; 
	/*if (!isalpha1(s[j])) macro_error (1,j);	*/
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_REAL); thresh = r.numvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); numsamp = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); minsize = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); maxclust = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); lowdens = (r.intvalue!=0);

	while (s[j]!=')') j++;
	*i = j+1;

	r.intvalue = clust_clustercount (&v.imgptr,thresh,numsamp,minsize,maxclust,lowdens);
	r.type = MT_INT;
	freebuf (&v.imgptr);

	return r;
}




/*
	center function
	Syntax: imgvar = center (imgexpr, mode);
	        imgvar = center (imgexpr, mode, slice);
	
	center determines the centroid of all image features and
	moves them into the center of the image. If mode==0,
	a feature is anything nonzero. If mode==1, the centroid
	is weighted by the image gray values.
	If slice is not specified, the operation will be performed
	on all slices of the image.
	Ideally, there should be only one feature in the image.

	Example:  c = center (centerme, 1);
*/

variable_type mac_center (unsigned char* s, int* i)
{
variable_type v,r;
int j,z,z1,mode;

	j = *i; 
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); mode = r.intvalue;
	
	if (s[j]==',') 			/* slice parameter */
	{
		j++;
		r = expr (s,&j,MT_INT);
		z1 = r.intvalue;
		if (z1>=v.imgptr.zmax) macro_error (15,j);
		center_feat (&v.imgptr, z1, mode);
	}
	else
	{
		for (z=0; z<v.imgptr.zmax; z++)
			center_feat (&v.imgptr, z, mode);
	}
	
	while (s[j]!=')') j++;
	*i = j+1;

	return v;
}



/*
	watershed function
	Syntax: imgvar = watershed (imgexpr, threshold, elemsize, smoothing, slice);
	        imgvar = watershed (imgexpr, threshold, elemsize, smoothing, slice, mode);
	
	The Watershed function performs the watershed transform on one
	slice of the image imgexpr. Note that a 3D stack will be converted
	to one single slice, so a substack operation in a loop is needed
	to do a full 3D Watershed transform (TODO: needs to be changed).
	The additional parameters are:
		threshold: Before the operation, the image is thresholded.
		elemsize: Minimum element size/2 (minimum distance of local maxima)
		smoothing: Apply Gaussian smoothing to EDM to suppress spurious local maxima
		mode: If mode is specified, the watershed transform can be restricted
			to individual steps or functions:
			"uep": Returns the ultimate eroded points
			"edm": the function returns the Euclidean distance map.
			"erode": Perform watershed segmentation, create EDM by iterative erosion
			"euclid": Perform watershed segmentation, create EDM by exhaustive search (circles)
			"skiz": Expand watershed into a SKIZ, return the boundary lines (binary image)
			"rawskiz": One step before SKIZ: Expand the regions by successive dilation, but do not
						stop with pixels that were set in the source image. Instead, continue growing until
						only a 1-pixel boundary to the next region remains.
		The default mode is "Euclidean", and the UEP and EDM options are always implicitly used
		with the Euclidean distance exhaustive search method.

	Example:  	w = watershed (myimg, 1, 8, 0, 0);
				w = watershed (myimg, 1, 8, 6, 0, "uep");

	NOTE THAT THIS FUNCTION MAY BE BUGGY AND NEEDS THOROUGH TESTING 

*/

variable_type mac_watershed (unsigned char* s, int* i)
{
variable_type v,r;
int j,z,mode;
float thresh;
int edm_mode;			/* Which EDM algorithm to use. */
int gauss_size;			/* Gaussian kernel size. Zero for no smoothing */
int envsize;			/* local maximum structuring element size. */


	j = *i; 
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);					/* Required: Digitization threshold */
	r = expr (s,&j, MT_REAL); thresh = r.numvalue;
	if (s[j++]!=',') macro_error (1,j);					/* Required: Element size (radius) */
	r = expr (s,&j, MT_INT); envsize = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);					/* Required: Gaussian kernel size */
	r = expr (s,&j, MT_INT); gauss_size = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);					/* Required: Slice */
	r = expr (s,&j, MT_INT); z = r.intvalue;

	/* Optional mode parameter */

	mode=0;					/* Default: Full watershed segmentation */
	edm_mode=1;				/* Default: EDM via exhaustive search */
	if (s[j]==',') 			/* mode parameter */
	{
		j++;
		r = expr (s,&j,MT_STRING);
		if (!strncasecmp (r.strg, "uep",3)) mode=1;
		else if (!strncasecmp (r.strg, "edm",3)) mode=2;
		else if (!strncasecmp (r.strg, "erode",5)) edm_mode=0;
		else if (!strncasecmp (r.strg, "euclid",6)) edm_mode=1;
		else if (!strncasecmp (r.strg, "rawskiz",7)) mode=3;
		else if (!strncasecmp (r.strg, "skiz",4)) mode=4;
		else macro_error (15,j);
		free (r.strg);
	}

	/* Missing: Option to combine gray value with EDM. Permanently set to zero here. */

	while (s[j]!=')') j++;
	*i = j+1;

	if (mode==0)
		watershed (&v.imgptr, mode, thresh, gauss_size, envsize, 0, z, edm_mode);
	else if (mode==1)
		make_edm (&v.imgptr, mode, thresh, gauss_size, envsize, 0, z, edm_mode);
	else if (mode==2)
		create_UEP (&v.imgptr, mode, thresh, gauss_size, envsize, 0, z, edm_mode);
	else if ((mode==3) || (mode==4))
		watershed (&v.imgptr, mode, thresh, gauss_size, envsize, 0, z, edm_mode);

	/* TODO: Do we have a major memory leak here?? We need to dispose of the edm image, no? */

	return v;
}




/*
	regiongrow function
	Syntax: imgvar = regiongrow (imgexpr, lthreshold, uthreshold, vertexlist);

	Region growing is a segmentation method where a contiguous region
	is grown from one or more seed points. Pixels are accepted as part of
	the feature if they are connected with the feature _and_ fall
	into the threshold range.
	lthreshold and uthreshold define the threshold range
	vertexlist is a sequence of 3 coordinates (x,y,z) and may contain
	multiple seed points.

	Example:  	w = regiongrow (oldimg, 100,150, 128,128,0);

*/

variable_type mac_regiongrow (unsigned char* s, int* i)
{
variable_type v,r;
int j,jsave,numverts;
float lthresh,uthresh;
vertex3d* seeds;

	j = *i; 
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_REAL); lthresh = r.numvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_REAL); uthresh = r.numvalue;
	if (s[j++]!=',') macro_error (1,j);

	/* We need to go over the vertex list twice to determine 
		how many vertices we have */

	numverts=0; jsave=j;
	while (s[j] != ')')
	{
		r = expr (s,&j, MT_INT); 			/* x coordinate */
		if (s[j++]!=',') macro_error (1,j);
		r = expr (s,&j, MT_INT); 			/* y coordinate */
		if (s[j++]!=',') macro_error (1,j);
		r = expr (s,&j, MT_INT); 			/* z coordinate */
		if (s[j]!=',') j++;					/* Skip over comma if it exists */
		numverts++;							/* Count # of vertices */
	}

	seeds = calloc (numverts+1, sizeof (vertex3d));
	if (!seeds) fatal_memerr();

	numverts=0; j=jsave;					/* Now gather the vertices */
	while (s[j] != ')')
	{
		r = expr (s,&j, MT_INT); 			/* x coordinate */
		seeds[numverts].x = r.intvalue;
		if (s[j++]!=',') macro_error (1,j);
		r = expr (s,&j, MT_INT); 			/* y coordinate */
		seeds[numverts].y = r.intvalue;
		if (s[j++]!=',') macro_error (1,j);
		r = expr (s,&j, MT_INT); 			/* z coordinate */
		seeds[numverts].z = r.intvalue;
		if (s[j]!=',') j++;					/* Skip over comma if it exists */
		numverts++;							/* Count # of vertices */
	}
	
	*i = j+1;

	regiongrow_3d (&v.imgptr, seeds, numverts, lthresh, uthresh, 0,0);

	return v;
}




/*
	hillclimb function
	Syntax: imgvar = hillclimb (imgexpr, threshold, locmax_radius, numrays, feature_radius);

	Hill climbing is a variation of the region-growing theme, whereby an image is
	assumed with small features that have a distinct local maximum. 
	From that maximum, the points of steepest slope are found, and region growing
	back to the local maximum fills the feature.
	threshold:			Minimum image value for a pixel to be counted as local maximum
	locmax_radius:		Region in which a local maximum is unique to be counted as local maximum
	numrays:			Number of probing rays to detect steepest slope = number of seed points
	feature radius:		Hard-clipping boundary in case the algorithm "leaks" in the wrong direction

	Example:  	w = hillclimb (oldimg, 10,20, 16, 45);

*/

variable_type mac_hillclimb (unsigned char* s, int* i)
{
variable_type v,r;
int j,jsave,numrays, locmaxrad, featrad;
float thresh,uthresh;
vertex3d* seeds;

	j = *i; 
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_REAL); thresh = r.numvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); locmaxrad = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); numrays = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); featrad = r.intvalue;

	while (s[j]!=')') j++;
	*i = j+1;

	hillclimb_segmentation (&v.imgptr, thresh, locmaxrad, numrays, featrad);

	return v;
}





/* MEASURE function

	Syntax: numvar = measure (imgexpr, oorflg);
	        numvar = measure (imgexpr, oorflag, slice);
	
	The measure function determines average, SD, and area of an image.
	If a ROI is specified, it will be honored.
	If OORFLAG is 1, zero values will be ignored.
	If the slice parameter is given, only the specified slice will
	be processed, otherwise the average of all slices will be determined.
	The function returns the pixel average value, and stores the
	same value in USER[0]. SD can be found in USER[1], and the N
	value (area or counted pixels) in USER[2]. 
	VERY DIFFERENT FOR RGB. In this case, the function returns 0,
	and USER[1] is also 0. USER[2] still contains the pixel count.
	USER[3] through USER[5] contain the RGB averages, and 
	USER[6] through USER[8] contain the RGB standard deviations.
	It is the macro programmer's responsibility to trap the RGB case.

*/


variable_type mac_measure (unsigned char* s, int* i)
{
variable_type v,r;
int j,oormode,slice;
stat_type result;

	j = *i; 
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); oormode = r.intvalue;
	
	slice=-1;
	if (s[j]==',') 			/* slice parameter */
	{
		j++;
		r = expr (s,&j,MT_INT);
		slice = r.intvalue;
	}

	while (s[j]!=')') j++;
	*i = j+1;

	/* Now call the actual service function */

	if (slice<0)			/* all slices */
	{
		result = meas_core (&v.imgptr, 0, v.imgptr.zmax-1, oormode);
	}
	else
	{
		result = meas_core (&v.imgptr, slice, slice, oormode);
	}

	if (result.n > 1)
	{
		mac_user[0] = result.Sx/result.n;
		mac_user[1] = SDEV(result.Sx, result.Sxx, result.n);
		mac_user[2] = result.n;
		mac_user[3] = result.SxR/result.n;
		mac_user[4] = result.SxG/result.n;
		mac_user[5] = result.SxB/result.n;
		mac_user[6] = SDEV(result.SxR, result.SxxR, result.n);
		mac_user[7] = SDEV(result.SxG, result.SxxG, result.n);
		mac_user[8] = SDEV(result.SxB, result.SxxB, result.n);
	}
	else
	{
		for (j=0; j<=8; j++) mac_user[j]=0;
		mac_user[2] = result.n;
	}

	freebuf (&v.imgptr);
	r.type = MT_REAL;
	r.numvalue = mac_user[0];
	return r;

}





/*
	histeq function
	Syntax: imgvar = histeq (imgexpr, numbins, oorflag);

	Histeq performs histogram equalization of an image passed by imgexpr.
	A histogram is computed with [numbins] bins. Its integral is the 
	translation table.
	OORFLAG=1 indicates that zeroes are background and are not processed.
	This is a 2D slice by slice operation.

*/

variable_type mac_histeq (unsigned char* s, int* i)
{
variable_type v,r;
int j,numbins,oorflag;

	j = *i; 
	v = expr (s,&j, MT_IMG);

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); numbins = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); oorflag = r.numvalue;

	while (s[j]!=')') j++;
	*i = j+1;

	histeq (&v.imgptr, numbins, oorflag);


	return v;
}




/*
	DWT2D function
	Syntax: imgvar = DWT2D (imgexpr, "wavelet", idwt, level);
	Compute the discrete wavelet transform of imgexpr.
	"wavelet" is the wavelet type, one of HAAR, DAUB4, DAUB12, DAUB20...
	idwt decides between forward transform (idwt=0) and inverse transform (idwt=1)
	level is the decomposition level.
*/

variable_type mac_dwt (unsigned char* s, int* i)
{
variable_type v,r;
int j;
char wparm[20];
WAVELET_TYPE wvlt;
int idwt, level;


	/* The image stack */

	j = *i; v = nullvar;
	if (!isalpha(s[j])) macro_error (1,j);

	v = expr (s,&j, MT_IMG);		/* Get image to transform */
	
	/* The wavelet parameter */	

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_STRING);	strncpy (wparm, r.strg, 18); free (r.strg);
	wvlt = W_NULL;
	if (!strncasecmp (wparm, "haar",4)) wvlt = W_HAAR;
	else if (!strncasecmp (wparm, "daub4",5)) wvlt = W_DAUB4;
	else if (!strncasecmp (wparm, "daub8",5)) wvlt = W_DAUB8;
	else if (!strncasecmp (wparm, "daub12",6)) wvlt = W_DAUB12;
	else if (!strncasecmp (wparm, "daub20",6)) wvlt = W_DAUB20;
	else if (!strncasecmp (wparm, "coif6",5)) wvlt = W_COIF6;
	else if (!strncasecmp (wparm, "coif12",6)) wvlt = W_COIF12;
	else if (!strncasecmp (wparm, "symlet4",7)) wvlt = W_SYM4;
	else if (!strncasecmp (wparm, "symlet8",5)) wvlt = W_SYM8;
	else if (!strncasecmp (wparm, "symlet12",6)) wvlt = W_SYM12;
	else macro_error (15,j);	

	/* the integer arguments idwt and level */	

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); idwt = r.intvalue;
	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_REAL); level = r.intvalue;

	while (s[j]!=')') j++;
	*i = j+1;

	/* Go execute. Prepare v to be returned as image */

	dwt_macro_driver (&v.imgptr, wvlt, idwt, level);
	v.type=MT_IMG;

	return v;
}




/*
	wvdenoise function
	Syntax: imgvar = wvdenoise (imgexpr, wavelet, shrinkfn, hardsoft);
			imgvar = wvdenoise (imgexpr, wavelet, shrinkfn, hardsoft, decomp, tmult);

	wvdenoise performs wavelet-based denoising by using coefficient shrinking.
	imgexpr is the image to perform the operation on
	wavelet (string) specifies the mother wavelet and must be one of Haar, Daub4, 
			Daub8, Daub12, Daub20, Coif6, Coif12, Symlet4, Symlet8, Symlet12.
	shrinkfn (string) is the shrinking function and must be one of VisuShrink,
			BayesShrink, SUREShrink, or GCV
	hardsoft (integer flag) determines whether to use hard thresholding (1) or
			soft thresholding (0)
	decomp (integer) is the maximum decomposition level, a reasonable value is 3
	tmult (float) is the threshold multiplier. A threshold found by any of the above
			methods is multiplied by this factor before shrinking.

	Example: 	nonoise = wvdenoise (inp, "daub8", "bayes", 1);
				nonoise2 = wvdenoise (inp, "daub8", "bayes", 1, 3, 0.9);

	This is a 2D slice by slice operation.

*/


variable_type mac_wv_denoise (unsigned char* s, int* i)
{
variable_type v,r;
int j,k;
char wparm[20], shrinkfn[20];
WAVELET_TYPE wvlt;
int shrink, level, decomp;
float tmult;


	/* The image stack */

	j = *i; k=0; v = nullvar;
	if (!isalpha1(s[j])) macro_error (1,j);

	v = expr (s,&j, MT_IMG);		/* Get image to transform */
	
	/* The wavelet parameter */	

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_STRING);	strncpy (wparm, r.strg, 18); free (r.strg);
	wvlt = W_NULL;
	if (!strncasecmp (wparm, "haar",4)) wvlt = W_HAAR;
	else if (!strncasecmp (wparm, "daub4",5)) wvlt = W_DAUB4;
	else if (!strncasecmp (wparm, "daub8",5)) wvlt = W_DAUB8;
	else if (!strncasecmp (wparm, "daub12",6)) wvlt = W_DAUB12;
	else if (!strncasecmp (wparm, "daub20",6)) wvlt = W_DAUB20;
	else if (!strncasecmp (wparm, "coif6",5)) wvlt = W_COIF6;
	else if (!strncasecmp (wparm, "coif12",6)) wvlt = W_COIF12;
	else if (!strncasecmp (wparm, "symlet4",7)) wvlt = W_SYM4;
	else if (!strncasecmp (wparm, "symlet8",7)) wvlt = W_SYM8;
	else if (!strncasecmp (wparm, "symlet12",7)) wvlt = W_SYM12;
	else macro_error (15,j);	

	/* The shrink function parameter */

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_STRING);	strncpy (shrinkfn, r.strg, 18); free (r.strg);
	shrink = 0;
	if (!strncasecmp (shrinkfn, "visu",4)) shrink = 0;
	else if (!strncasecmp (shrinkfn, "bayes",5)) shrink = 2;
	else if (!strncasecmp (shrinkfn, "sure",4)) shrink = 4;
	else if (!strncasecmp (shrinkfn, "gcv",3)) shrink = 6;
	else macro_error (15,j);	

	/* the required integer argument hardsoft */	

	if (s[j++]!=',') macro_error (1,j);
	r = expr (s,&j, MT_INT); 
	if (r.intvalue) shrink |=1;

	/* The optional arguments decomp and tmult */

	decomp = 3; tmult = 1.0;
	if (s[j]==',') 			/* decomp parameter */
	{
		j++;
		r = expr (s,&j,MT_INT);
		decomp = r.intvalue;
		if (s[j]==',') 			/* tmult parameter */
		{
			j++;
			r = expr (s,&j,MT_REAL);
			tmult = r.numvalue;
		}
	}

	while (s[j]!=')') j++;
	*i = j+1;

	/* Go execute. Prepare v to be returned as image */

	wvdenoise_macro_driver (&v.imgptr, wvlt, shrink, decomp, tmult);
	v.type=MT_IMG;

	return v;
}



/*
	mip function
	Syntax: imgvar = mip (imgexpr);

	This function returns the maximum intensity projecyion (MIP)
	of a stack parallel to the z axis.

*/

variable_type mac_mk_mip (unsigned char* s, int* i)
{
variable_type v,r;
int j,numbins,oorflag;

	j = *i;
	v = expr (s,&j, MT_IMG);

	while (s[j]!=')') j++;
	*i = j+1;

	makemip (&v.imgptr);


	return v;
}







/******************************************************************************************/





