

/*
 *  randgen.c -- random generators and noise synthesis package
 *  Used to generate all sorts of weird noise, including
 *  fractional Brownian motion, exponential noise, 
 *  and Perlin noise.
 */


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <fftw3.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_rng.h>


#include "defs.h"
#include "application.h"
#include "dlgbox.h"
#include "fft.h"
#include "functions.h"
#include "functions2.h"
#include "randgen.h"
#include "imgbuf.h"





/*********************************************+

	R A N D O M   G E N E R A T O R S
	
***********************************************/



/* Random generator from numerical recipes:

Long period (> 2 � 1018) random number generator of L'Ecuyer with Bays-Durham
shuffle and added safeguards. Returns a uniform random deviate between 0.0 and
1.0 (exclusive of the endpoint values). Call with rndreg a negative integer to
initialize; thereafter, do not alter rndreg between successive deviates in a
sequence. 
RNMX should approximate the largest floating value that is less than 1.

*/


/* For the use of GSL random generators, see

http://www.gnu.org/software/gsl/manual/html_node/Sampling-from-a-random-number-generator.html

and for non-uniform random generators,

http://www.gnu.org/software/gsl/manual/html_node/The-Gaussian-Distribution.html

*/


/* We need a GSL random generator */

const gsl_rng_type *gsl_randgen_type;
gsl_rng *gsl_randgen;




/* Uniform random number between 0 (inclusive) and 1.0 (exclusive) */

float nr_rand2 () { return gsl_rng_uniform (gsl_randgen); }



/* 
	Returns a normally distributed deviate with zero mean and unit variance,
	using uniform deviates as source.
*/


// the gsl version didn't get us anywhere, comment it out

// float gauss_rnd_nogo () { return gsl_ran_ugaussian (gsl_randgen); }

float gauss_rnd ()
{

static int iset=0;
static float gset;
float fac,rsq,v1,v2;

	if (rndreg < 0) iset=0;				/* Reinitialize. */

	if (iset == 0) 						/* We don't have an extra deviate handy, so... */
	{
		do 
		{
			v1=2.0*nr_rand2 ()-1.0;		/* ...pick two uniform numbers in the square */
			v2=2.0*nr_rand2 ()-1.0;		/* extending from -1 to +1 in each direction, */
			rsq=v1*v1+v2*v2;				/* see if they are in the unit circle, */
		} 
		while (rsq >= 1.0 || rsq == 0.0);	/* and if they are not, try again. */

		fac=sqrt(-2.0*log(rsq)/rsq);

		 /* Now make the Box-Muller transformation to get two normal deviates. 
		 	Return one and save the other for next time. */

		gset=v1*fac;
		iset=1;
		return v2*fac;
	} 
	else								/* We have an extra deviate handy, */
	{
		iset=0;
		return gset;
	}

}




/*
	eran1:
	Returns an exponentially distributed, positive, random deviate 
	of unit mean (?), using nr_rand2() as the source of uniform deviates.

	Note about mean: If we use log10, mean is about 0.1, horrible dropoff.
		If we use log_e, mean is about 0.58, better dropoff.
		If needed, we can normalize by dividing by 0.58.

*/


float eran1 ()
{
float buf; 

	do 
		buf = nr_rand2(); 
	while (F_ISZERO(buf)); 
	return -log(buf);
}


/* pnoise, smooth_p_noise, intp_p_noise, perlin_noise:

	Noise generator for supposedly-Perlin noise. 
	Used in the generator module for Perlin landscapes.
	(but: see "real" Perlin noise below)
	This Perlin-style noise generator produces 1/f Perlin noise with zero mean.
	We have four functions. The actual noise generator (pnoise), a deterministic
	pseudo-random generator which returns a pseudo-random number as a function of 
	an x,y vector, smooth_p_noise which runs a [1, 2, 1] kernel over the noise, and 
	intp_p_noise which uses a cosine interpolation to interpolate.
	Finally, perlin_noise synthesizes the Perlin noise value at x and y.
	Note that the spatial arguments of perlin_noise should be normalized to 0...1
	Credit for basic pseudocode: Hugo Elias, http://freespace.virgin.net/hugo.elias/
*/

/* Cosine interpolation (better than linear, not as good as cubic).
	x is the fraction, 0<=x<1; a and b are the numbers to interpolate between */

float cosine_interpolate (float a, float b, float x)
{
float fx;

	fx = (1.0 - cos (x*PI))*0.5;	/* Cosine-swing rising from 0 to 1 as x goes from 0 to 1 */
									/* guarantees that there is no slope discontinuity at vertices */
	return  a*(1.0-fx) + b*fx;
}


float pnoise (int x, int y, int level)		/* Box-distributed not-really-random numbers between -1 and 1 */
{
long n;
int  p0[16] = {	251, 257, 263, 269, 271, 277, 281, 283, 293, 307, 311, 313, 317, 331, 337, 347};	/* prime numbers */
long p2[16] = {	15667, 15671, 15679, 15683, 15727, 15731, 15733, 15737, 15739,
				15749, 15761, 15767, 15773, 15787, 15791, 15797};
long p3[16] = {	789181, 789221, 789227, 789251, 789311, 789323, 789331, 789343, 
				789367, 789377, 789389, 789391, 789407, 789419, 789443, 789473};
long p4[16] = {	1376312017, 1376312071, 1376312087, 1376312099, 1376312107, 1376312173, 1376312183, 1376312257, 
				1376312261, 1376312263, 1376312269,  1376312279, 1376312323, 1376312369, 1376312383, 1376312393};


	level = MIN(level,15);

	n = x + y*p0[level];
	n =  (n<<13) ^ n;
	n = (n * ( SQR(n) * p2[level] + p3[level]) + p4[level]) & 0x7fffffff;
	return 1.0 - (float)n/1073741824.0;

}


float smooth_p_noise (int x, int y, int level)
{
float smoothmask[9] = {0.0625, 0.125, 0.0625, 0.125, 0.25, 0.125, 0.0625, 0.125, 0.0625};		/* 3x3 mask */
int x1,y1,i;
float buf;

	buf = 0.0; i=0;
	for (y1=y-1; y1<=y+1; y1++)
		for (x1=x-1; x1<=x+1; x1++)
			buf += pnoise (x1,y1,level)*smoothmask[i++];

	return buf;
}


float intp_p_noise (float xr, float yr, int level)
{
int x,y;
float fx, fy;
float v1, v2, v3, v4; 		/* Rnd numbers on integer vertices of a square containing xr, yr */
float buf1, buf2;

	x = (int)floor(xr);
	fx= xr-x;
	y = (int)floor(yr);
	fy= yr-y;


	v1 = pnoise (x, y, level);			/* We could use smooth_p_noise here, but it doesn't do very much */
	v2 = pnoise (x+1, y, level);
	v3 = pnoise (x, y+1, level);
	v4 = pnoise (x+1, y+1, level);

	buf1 = cosine_interpolate (v1, v2, fx);				/* Interpolate in x direction */
	buf2 = cosine_interpolate (v3, v4, fx);				/* Interpolate in x direction */
	return cosine_interpolate (buf1, buf2, fy);			/* Interpolate in y direction */

}

/* The actual Perlin noise synthesizer */

float perlin_noise (float x, float y, float H, int octaves)		/* H is the persistence, 0<H<1 */
{
int i;
float freq, amplitude, buf;

	amplitude = 1.0;
	freq = 2.0;			/* This number determines the lowest-level smoothness */
	buf = 0;

	for (i=0; i<octaves; i++)
	{
		buf += intp_p_noise (x*freq, y*freq, i)*amplitude;
		amplitude *= H;			/* Amplitude is power (H,i) */
		freq *= 2;				/* Frequency is power (2,i) */
	}

	return buf;
}


/* There is some discussion about Perlin noise on the web, notably that Hugo Elias' code
	is not really Perlin noise. Perlin's original code follows. It was converted from Java
	found here:
	https://cs.nyu.edu/~perlin/noise/
*/

/* fade is the blending function between two integer points.
	The presented function fade(t) = 6 t^5 - 15 t^4 + 10 t^3
	is valid for 0 <= t <= 1, and f(0)=0 and f(1)=1 (continuity requirement).
	In addition, df/dt vanishes for t=0 and t-1, and d^2 f / dt^2 also vanishes
	for t=0 and for t=1. Therefore, sections that are pieced together with fade(t)
	have a continuous first *and* second derivative.
*/

double fade (double t) 
{ 

	return t * t * t * (t * (t * 6 - 15) + 10); 

}


/* lerp... linear interpolation. Meh. */

double lerp (double t, double a, double b) 
{ 

	return a + t * (b - a); 

}


/* This function is central to Perlin's idea of "gradient noise".
	It calculates the pseudo-random gradient at any arbitrary point
	(x, y, z) inside the cube, i.e, 0 <= x <= 1, 0 <= y <= 1, 0 <= z <= 1
*/

double grad (int hash, double x, double y, double z) 
{
int h;
double u, v, gradval;

	h = hash & 15;		/* Convert lower 4 bits of hash code into 12 gradient directions */
	if (h<8)
		u = x;
	else
		u = y;

	if (h<4)
		v = y;
	else if ((h==12) || (h==14))
		v = x;
	else
		v = z;

	gradval = u;
	if (h & 1)
		gradval = -u;
	if (h & 2)
		gradval -=v;
	else
		gradval += v;

	return gradval;
}


double perlins_noise (double x, double y, double z) 
{
int i, ix, iy, iz;
int a, aa, ab, b, ba, bb;
double u, v, w;
double lt, la, lb, lx, ly, l;
short permutation[] = 
			{ 	151,160,137,91,90,15,
				131,13,201,95,96,53,194,233,7,225,140,36,103,30,69,142,8,99,37,240,21,10,23,
				190, 6,148,247,120,234,75,0,26,197,62,94,252,219,203,117,35,11,32,57,177,33,
				88,237,149,56,87,174,20,125,136,171,168, 68,175,74,165,71,134,139,48,27,166,
				77,146,158,231,83,111,229,122,60,211,133,230,220,105,92,41,55,46,245,40,244,
				102,143,54, 65,25,63,161, 1,216,80,73,209,76,132,187,208, 89,18,169,200,196,
				135,130,116,188,159,86,164,100,109,198,173,186, 3,64,52,217,226,250,124,123,
				5,202,38,147,118,126,255,82,85,212,207,206,59,227,47,16,58,17,182,189,28,42,
				223,183,170,213,119,248,152, 2,44,154,163, 70,221,153,101,155,167, 43,172,9,
				129,22,39,253, 19,98,108,110,79,113,224,232,178,185, 112,104,218,246,97,228,
				251,34,242,193,238,210,144,12,191,179,162,241, 81,51,145,235,249,14,239,107,
				49,192,214, 31,181,199,106,157,184, 84,204,176,115,121,50,45,127, 4,150,254,
				138,236,205,93,222,114,67,29,24,72,243,141,128,195,78,66,215,61,156,180
			};
short p[512];


	/* Initialize the permutation/hashing table. This serves as pseudo-random
		number reservoir. */

	for (i=0; i < 256 ; i++) 
		p[256+i] = p[i] = permutation[i];

	/* Find unit cube that contains point (x,y,z). The part & 0xff
		rolls over large surfaces and thereby limits the size of the
		required permutation table */

	ix = ((int)floor(x)) & 0xff;
	iy = ((int)floor(y)) & 0xff;
	iz = ((int)floor(z)) & 0xff;

	/* Find the point's relative position inside the cube (meaning:
		its fractional part when the cuber vertices are integer coordinates)
		and immediately compute the fade curves for each coordinate */

	x -= floor (x); u = fade (x);
	y -= floor (y); v = fade (y);
	z -= floor (z); w = fade (z);

	/* Hash coordinates of the 8 cube centers */

	a  = p[ix  ] + iy; 
	aa = p[a   ] + iz;
	ab = p[a+1 ] + iz;
	b  = p[ix+1] + iy;
	ba = p[b   ] + iz;
	bb = p[b+1 ] + iz;

	/* Add the blended results from 8 corners of the cube */

	la = lerp (u, grad (p[aa], x, y, z), grad(p[ba], x-1, y, z) );
	lb = lerp (u, grad (p[ab], x, y-1, z), grad(p[bb], x-1, y-1, z) );
	lx = lerp (v, la, lb);
	la = lerp (u, grad (p[aa+1], x, y, z-1), grad(p[ba+1], x-1, y, z-1) );
	lb = lerp (u, grad (p[ab+1], x-1, y-1, z-1), grad(p[bb+1], x-1, y-1, z-1) );
	ly = lerp (v, la, lb);
	l = lerp (w, lx, ly);

	return l;
}



/* Fractal Brownian Motion random generator.
 	A time series representing fractional Brownian motion (fBm)  
 	is generated by the method of successive random additions 
 	It is differenced to generate a series representing fractional 
	Brownian noise (fBn) if requested.
	The algorithm was derived from the FORTRAN-program "sra"
	by G. Raymond and J.B. Bassingthwaighthe, U Washington.
	Input parameters are sd (initial standard deviation) and 
	h (Hurst coefficient).
	This function will always generate time series in powers of 2,
	so we'll generate more than needed and simply cut it off.

	NOTE that this is a 1D function and gives sort of senseless data when
	stored line-by-line in an image. 
*/


void create_fbm (float* timeseries, float sd, float h, int numpoints, char make_fbm)
{
int nlog,n,id,idd,level,i;
float* delta; float* fbm;


	g_assert ( (h>=0) && (h<=1) );
	g_assert ( (sd>0) );

	nlog=1;
	while (pow(2,nlog)<numpoints) nlog++;
	n = pow (2,nlog);							/* Next-largest power-of-two size; nlog is the number of scales */
	if (make_fbm)
		dp (2, "FBM: ");
	else
		dp (2, "FBN: ");
	dp (2, "%d data points requested, next power of 2 is 2^%d, %d points\n",numpoints, nlog,n);
 
	delta = (float*) calloc (nlog+1, sizeof (float));
	fbm   = (float*) calloc (n   +1, sizeof (float));

	/* Delta is the level-dependent variation range */

	for (level=0; level<nlog; level++)
		delta[level] = sd*pow(0.5,level*h)*sqrt(0.5)*sqrt(1 - pow(2,2*h-2));

	/* Define first and last point */

	fbm[0]=0;		
	fbm[n-1] = sd*gauss_rnd();		/* FORTRAN: fbm(n) = sd * randn(3,1.0) */

	/* Go down level-by-level; For each level, generate one new point in the center
		between the existing points (linearly interpolated), then displace that point
		by a level-dependent random distance */

	idd=n; id = n/2;
	for (level=0; level<nlog; level++)
	{
		for (i=id; i<=n-id; i+=idd)
			fbm[i]  =0.5*(fbm[i-id] + fbm[i+id]);		/* New interpolated point in the middle */
			
		for (i=0; i<=n; i+=id)
			fbm[i] = fbm[i] + delta[level] * gauss_rnd();		/* Random displacement */
  
		idd/=2;
		id/=2;
	}

	/* Now generate the dataset. Memory alloc etc is duty of 
	the calling program */
	
	if (make_fbm)
	{
		for (i=0; i<numpoints; i++)
			timeseries[i]= fbm[i];		/* Fractal Brownian Motion */
	}
	else
	{
		timeseries[0] = 2.0*(fbm[0]-fbm[1]);
		timeseries[numpoints-1] = 2.0*(fbm[numpoints-2]-fbm[numpoints-1]);
		for (i=1; i<numpoints-1; i++)
			timeseries[i]= fbm[i+1]-fbm[i-1];	/* Fractional Brownian Noise, 1st derivative of FBM */
	}

	free (delta);
	free (fbm);

}



/* This is the two-dimensional version of the above. */


void fbm_2d_recurse (image_type *img, image_type *mask, float *delta, int x, int y, int sq, int level)
{
int mid;			/* Note that the sq parameter *must* be a power of two */
int n;
float *p1;
unsigned char *p0;


	/* First, set the 5 pixels in a plus-arrangement */

	p0 = mask->data;		
	n = img->xmax;			/* Allows a shorthand in the index expressions */
	p1 = img->data;
	mid = sq/2;

	/* DEBUG: Are the four corner points set? */

// 	if (p0[x+n*y]==0)
// 		dp (2, "FBM recurse, level %d: MISS UL corner (%d,%d)\n",level, x,y);
// 	if (p0[x+sq+n*y]==0)
// 		dp (2, "FBM recurse, level %d: MISS UR corner (%d,%d)\n",level, x+sq,y);
// 	if (p0[x+n*(y+sq)]==0)
// 		dp (2, "FBM recurse, level %d: MISS BL corner (%d,%d)\n",level, x,y+sq);
// 	if (p0[x+sq+n*(y+sq)]==0)
// 		dp (2, "FBM recurse, level %d: MISS BR corner (%d,%d)\n",level, x+sq,y+sq);
// 
// 	if (level<3)
// 		dp (2, "Level %d. Square (%d,%d) to (%d,%d). Corner point values %3.0f %3.0f %3.0f %3.0f\n", level, x,y,x+sq,y+sq, p1[x+n*y], p1[x+sq+n*y], p1[x+n*(y+sq)], p1[x+sq+n*(y+sq)]);
	

	if (p0[x + n*(y+mid)]==0)					/* left margin */
	{
		p0[x + n*(y+mid)]=level+1;
		p1[x + n*(y+mid)]= 0.5*(p1[x+n*y] + p1[x+n*(y+sq)])+delta[level] * gauss_rnd();
	}

	if (p0[x+mid + n*y]==0)					/* top margin */
	{
		p0[x+mid + n*y]=level+1;
		p1[x+mid + n*y]= 0.5*(p1[x+n*y] + p1[x+sq+n*y])+delta[level] * gauss_rnd();
	}

	if (p0[x+sq + n*(y+mid)]==0)				/* right margin */
	{
		p0[x+sq + n*(y+mid)]=level+1;
		p1[x+sq + n*(y+mid)]= 0.5*(p1[x+sq+n*y] + p1[x+sq+n*(y+sq)])+delta[level] * gauss_rnd();
	}

	if (p0[x+mid + n*(y+sq)]==0)				/* bottom margin */
	{
		p0[x+mid + n*(y+sq)]=level+1;
		p1[x+mid + n*(y+sq)]= 0.5*(p1[x+n*(y+sq)] + p1[x+sq+n*(y+sq)])+delta[level] * gauss_rnd();
	}

	if (p0[x+mid + n*(y+mid)]==0)				/* center point */
	{
		p0[x+mid + n*(y+mid)]=level+1;
		p1[x+mid + n*(y+mid)]= 0.25*(p1[x+n*y] + p1[x+n*(y+sq)] + p1[x+sq+n*y] + p1[x+sq+n*y])+SQ2*delta[level] * gauss_rnd();
	}

	/* Next, recurse one level down into the four sub-squares.
		...PROVIDED that we still have a more-than-2-pixel square */

	if (mid >= 2)
	{
		fbm_2d_recurse (img, mask, delta, x, y, mid, level+1);
		fbm_2d_recurse (img, mask, delta, x+mid, y, mid, level+1);
		fbm_2d_recurse (img, mask, delta, x, y+mid, mid, level+1);
		fbm_2d_recurse (img, mask, delta, x+mid, y+mid, mid, level+1);
	}


}




int create_fbm_2d (image_type *img, float sd, float h, int numpoints, char make_fbm)
{
int nlog,n,level,i,x,y;
int xst, yst, midpt, sqsize;
float pxul, pxur, pxbl, pxbr;
float* delta;
float *p1;
image_type mask;
unsigned char *p0;


	g_assert ( (h>=0) && (h<=1) );
	g_assert ( (sd>0) );

	nlog=1; 
	while (pow(2,nlog)<numpoints) nlog++;
	n = pow (2,nlog);							/* Next-largest power-of-two size; nlog is the number of scales */
	if (make_fbm)
		dp (2, "FBM: ");
	else
		dp (2, "FBN: ");
	n++;						/* We need one row & column more than power-of-2 */

	dp (2, "%d data points requested, next power of 2 is 2^%d, --> %d points\n",numpoints, nlog,n);

	if (allocate_image (&mask, B_CHAR, n,n,1)==-1)
		fatal_memerr();
	p0 = mask.data;
 	if (allocate_image (img, B_FLOAT, n,n,1)==-1)
		fatal_memerr();
	p1 = img->data;

	delta = (float*) calloc (nlog+1, sizeof (float));

	/* Delta is the level-dependent variation range */

	for (level=0; level<nlog; level++)
		delta[level] = sd*pow(0.5,level*h+0.5)*sqrt(1 - pow(2,2*h-2));

	/* Define the four corner points of the whole image */

	p1[0] = sd*gauss_rnd();			/* Top left */
	p1[n-1] = sd*gauss_rnd();			/* Top right */
	p1[n*(n-1)] = sd*gauss_rnd();		/* Bottom left */
	p1[n*n-1] = sd*gauss_rnd();		/* Bottom right */

	p0[0] = 1;
	p0[n-1] = 1;
	p0[n*(n-1)] = 1;
	p0[n*n-1] = 1;

	fbm_2d_recurse (img, &mask, delta, 0, 0, n-1, 1);
	make_minmax (img);

	free (delta);
	freebuf (&mask);
	return n;
}





/* Fractional Brownian Motion random generator using spectral synthesis -
	 Function computes a series from specified amplitudes and
	 phases using a back Fourier transform.
	
	 The phases are randomly chosen from between plus/minus pi.
	 The amplitudes are the product of a random number drawn from a
	 Gaussian distribution and a power function (i**(-beta/2)).
	 Treating all the amplitudes as positive numbers, the random
	 part has a mean of approximately 0.80 with standard deviation 0.60.

	The algorithm is derived from the FORTRAN-program "ssm" by
	G. Raymond and J.B. Bassingthwaighthe, U Washington
	
	According to the above reference, the dataset should be
	8 times longer than requested, and only the first n/8 points
	be used (http://nsr.bioeng.washington.edu/Software/index.html)
*/

/* TODO: This is incomplete and must be finished with an IFFT */

#define PHASER	1.813799364	/* PI/sqrt(3) */


void create_fbm_spectral (float *timeseries, float sd, float beta, int numpoints, char make_fbm)
{
int nlog, n,i,j;
long idum;
double amp, phase;
double* fbm;
double* fftdata;

	ShowMessage ("FBM/FBN spectral synthesis is not yet fully implemented");

	nlog=1; idum=-1;
	while (pow(2,nlog)<numpoints) nlog++;
	n = pow (2,nlog);
	if (make_fbm)
		dp (1, "FBM ");
	else
		dp (1, "FBN ");
	dp (1, "(spectral): %d data points requested, next power of 2 is 2^%d, %d points\n",numpoints, nlog,n);
	fbm   = (double*) calloc (n   +1, sizeof (double));
	fftdata= (double*) calloc (4*n+1, sizeof (double));
 
	j=2;	/* We leave the D.C. component zero */
	for (i=0; i<n; i++)
	{
		amp = gauss_rnd () / pow(i+1,beta);	/* Decay amplitude with frequency, use Gaussian rnd */
		phase = 2*PI*(nr_rand2()-0.5);	/* Phase is from equally distributed noise */
		fftdata[j  ] = amp*cos(phase);	/* Real part of freq */
		fftdata[j+1] = amp*sin(phase);	/* Imaginary part of freq */
		fftdata[4*n-j] = amp*cos(phase);	/* Real part of negative freq */
		fftdata[4*n-j+1] = amp*sin(phase);	/* Imaginary part of negative freq */

		fftdata[j]=fftdata[j+1]=amp;
		j+=2;
	}
	
	/*	dft_core (fftdata, n, -1);			/* Inverse FFT */

	/* Post-adjustment - the inverse transformation returns nn*data */
/*	for (i=0; i<n; i++) fftdata[i]/=n;

	/* Now generate the dataset. Memory alloc etc is duty of 
	the calling program.
	The fft array returns data as real - imag - real - ...
	and we simply drop the imaginary parts. */
	
	j=0;
	for (i=0; i<numpoints; i++)
	{
		timeseries[i]= fftdata[j];
		if ((!make_fbm) && (i>0))
			timeseries[i]-=fftdata[j-2];
		j+=2;
	}
	
	free (fftdata);
	free (fbm);
}


/* 	2D Fractional Brownian Motion random generator using spectral synthesis -
	 Function computes a 2D image from specified amplitudes and
	 phases using an inverse Fourier transform in 2D
	 img should be an empty, unallocated placeholder.
*/


void create_2d_fbm_spectral (image_type *img, float beta, int numpoints)
{
int n,x,y;
float amp, phase, re, im, r;
image_type fftimg; float *p1;
long cidx_re, cidx_im;


	n = numpoints/2; cidx_re = n + n*numpoints; cidx_im = cidx_re+SQR(numpoints);
	if (allocate_image (&fftimg, B_FLOAT, numpoints,numpoints,2)==-1)
		fatal_memerr();
	p1 = fftimg.data;
 
	for (y=1; y<n; y++)			/* We start with 1 to leave the D.C. component zero */
	{
		for (x=1; x<n; x++)
		{
		
			/* Choose a random number by magnitude and phase. The random number
				amplitude decays with the frequency. Then convert into re-im. */
			
			r = sqrt(ARITHM(x,y))*PI/n;			/* Frequency = distance from origin */
			amp = gauss_rnd () / pow(r,beta);	/* Decay amplitude with frequency, use Gaussian rnd */
			phase = TWOPI*(nr_rand2()-0.5);		/* Phase is from equally distributed noise */
			re = amp*cos(phase);	/* Real part of freq */
			im = amp*sin(phase);	/* Imaginary part of freq */
			
			/* Stuff the data into the four quadrants of the FFT image.
				Use {re, im} to include the phase or {amp, 0} to have a real-valued F.T.
				Note that we have to obey the Hermitian symmetry F(u,v) = F*(-u,-v) 
				or the result is going to be cr... */
	
			p1[cidx_re + x + y*numpoints] = re;		/* Quadrant I */
			p1[cidx_im + x + y*numpoints] = im;
			p1[cidx_re - x - y*numpoints] = re;		/* Quadrant III */
			p1[cidx_im - x - y*numpoints] = -im;
			
			amp = gauss_rnd () / pow(r,beta);	/* Decay amplitude with frequency, use Gaussian rnd */
			phase = 2*PI*(nr_rand2()-0.5);	/* Phase is from equally distributed noise */
			re = amp*cos(phase);	/* Real part of freq */
			im = amp*sin(phase);	/* Imaginary part of freq */
	
			p1[cidx_re - x + y*numpoints] = re;		/* Quadrant II */
			p1[cidx_im - x + y*numpoints] = im;
			p1[cidx_re + x - y*numpoints] = re;		/* Quadrant IV */
			p1[cidx_im + x - y*numpoints] = -im;

		}
	}
	
	/* Done with the generation, now backtransform using standard IFFT */
	
	ifft2d (&fftimg);
	img->data = NULL;			/* Ensure that the unallocated image does not use free() in handback */
	handback (img, &fftimg);	/* Copy image data into provided spaceholder */

}




/*****************************************************************************************************/




