
/*
 *	Snakes: Contains the algorithms and GUI for active contours
 * (specifically: Greedy snake algorithm)
 */



/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.

		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>


#include "snakes.h"
#include "roi.h"
#include "dlgbox.h"
#include "imgbuf.h"
#include "application.h"
#include "secondwnd.h"
#include "filter.h"
#include "functions.h"
#include "functions2.h"
#include "gtkhelpers.h"


#define MAX_NEIGH	 10
#define MIN_ENERGY	 1000000000000.0
#define MAX_ENERGY	-1000000000000.0


//#define SNXT(I)			(((I)+1) % snake_maxpts)
//#define SPRV(I)			(((I)-1+snake_maxpts) % snake_maxpts)
//#define SNXT2(I)		(((I)+2) % snake_maxpts)
//#define SPRV2(I)		(((I)-2+snake_maxpts) % snake_maxpts)

#define SNXT(I)			((I)+1 >= snake_maxpts? (I)+1-snake_maxpts : (I)+1)
#define SPRV(I)			((I)-1 < 0? (I)-1+snake_maxpts : (I)-1)
#define SNXT2(I)		((I)+2 >= snake_maxpts? (I)+2-snake_maxpts : (I)+2)
#define SPRV2(I)		((I)-2 < 0? (I)-2+snake_maxpts : (I)-2)
#define SWRAP(I)		((I) >= snake_maxpts? (I)-snake_maxpts : ( (I)<0? (I)+snake_maxpts : (I) ))


int neighborhood;	/* 1: 3x3; 2: 5x5, 3: 7x7 etc */
int snake_maxpts;
vertex* snake_points; /* Snake Points, to be copied from ROI */
float *e_alpha,*e_beta, *e_gamma, *e_delta; /* Weights for energies */
float threshold_curvature;
float init_length, stretch_scale, curv_scale, cont_scale;
int threshold_movement;
char snake_statusbar_was_used;
char selected_snake_algorithm;

image_type snake_gradient_image, snake_display_image, snake_gvf_image;
roi_type snake_roi_save;
int restoreroi;
int snake_autoiter;
int snake_use_gvf;
int snake_interpolate_pend;




/*****************************************************************************************************/

/*    ***    ****    *****   *****   ****    *   *          ***    *   *     *     *   *   *****     */
/*   *   *   *   *   *       *       *   *   *   *         *   *   *   *    * *    *  *    *         */
/*   *       *   *   *       *       *   *    * *          *       **  *   *   *   * *     *         */
/*   *  **   ****    ****    ****    *   *     *            ***    * * *   *   *   **      ****      */
/*   *   *   * *     *       *       *   *     *               *   *  **   *****   * *     *         */
/*   *   *   *  *    *       *       *   *     *           *   *   *   *   *   *   *  *    *         */
/*    ***    *   *   *****   *****   ****      *            ***    *   *   *   *   *   *   *****     */

/* This is the simplest form of snake algorithm. In a n by n neighborhood around each vertex,
	the energy is computed that the snake would assume if this vertex was displaced.
	The minimum energy pixel from that neighborhood is chosen as the new vertex point */

/* A note on scaling: Each energy term (stretching, curvature, continuity, image) has its own
	scale factor, adjustable via GUI. To make those sliders adhere to a comparable value range,
	internal scale factors are used. These are empirically determined.
	Stretching: 500/(initial ROI length), can cancel out image size and initial snake size
	Curvature:  5.0, constant, experimentally determined
	Continuity: 1.0, constant, also experimentally determined
	These scale factors wreak havoc with the Williams-Shah formulation and are therefore disabled

*/


/* Returns whether two vertices are one and the same point */

unsigned char v_eq (vertex p1, vertex p2)
{

	return ( (p1.x==p2.x) && (p1.y==p2.y) );
}


/* Distance between snake point i and arbitrary point vertex.
	This is subtly different from calc_stretching() below. */

float calcdist(int i, vertex point2)
{

	/* abs (snake_points[i-1] - P2) */

	point2.x -= snake_points[i].x;
	point2.y -= snake_points[i].y;
	return sqrt( SQR(point2.x) + SQR (point2.y));
}


/* Square deviation from average distance - this is the continuity energy */

float calc_continuity(int i, vertex point, float avg_distance)
{
float r;

	r = avg_distance-calcdist(SPRV(i),point);
	dp (3, "Continuity returns %f\n",SQR(r));
	return SQR(r)*cont_scale;
}


/* The stretching energy is simply the distance between two points.
	Note that the definition implies a variable point around vertex i
	and its distance to the vertex i-1 (thus SPRV)
	Total stretching scales poorly. Not normalizing the sum of distances
	is desirable, because the stretching energy is lowered as the snake
	contracts. Still, some normalization should happen so that
	the initial size or at least image size are canceled out.
*/

float calc_stretching (int i, vertex point)
{
float r;

	r = calcdist(SPRV(i),point) * stretch_scale;
	return r*r;
}



/* Curvature energy: Strong curvature -> high energy
	For some ideas for the value range, consider some angles
	between the segments v(i-1)-v(i) and v(i)-v(i+1):
		  0 degrees (bounce back)	4 curvature
		 45 degrees					3.5
		 90 degrees					2
		135 degrees					0.59
		180 degrees (straight)		0
	We can take this idea further and look at regular polygons.
	The equilateral triangle (3 vertices) has a bending energy of 3
	per vertex, 9 total. A square has 2 per vertex, total of 8.
	A hexagon, 1 per vertex, total of 6. The bending per vertex vanishes
	as the polygon has more and more vertices, and so does the total
	bending energy. Total bending for a circle is zero.
 */

float calc_curvature(int i, vertex p, int algorithm)
{
vertex nextP, prevP;
float v1mag, v2mag;
float curv_x, curv_y, r;


	if (algorithm==1)
	{

		/* Curvature as finite second-order difference normalized by
			the distance of points . Finite difference equation is, for x,

				    x(i)-x(i-1)       x(i+1)-x(i)
			Cx(i)= -------------  -  -------------
				   |v(i)-v(i-1)|     |v(i+1)-v(i)|

			and similar for y. Then C(i) = Cx(i)^2 + Cy(i)^2 is the curvature.
			This formulation places the three successive vertices v0, v1, v2 on a circle
			such that v1 is in the center, and v0 and v2 lie on the circle with radius 1.
			The segments v0-v1 and v1-v2 are at an angle alpha, and the intersections of
			v0-v1 and v1-v2 with the unit circle yield a secant with the distance d from
			the center and thus v1. The curvature, as SQR(curv_x)+SQR(curv_y) is d^2.
			Moreover,
						d = 2r cos (alpha/2)            (where r=1)

			Vertices should always be at least one pixel apart. Sometimes, two vertices
			may collapse onto the same pixel. Those experience the same energies and cannot
			be separated, so this calls for the cleanup squad to remove one of the two.
			Here, we try the next vertex in the same direction and use that for the curvature
			to at least avoid a division by zero error. If the extended distance
			turns out to be zero again, a curvature of zero is returned.
		*/

		prevP = snake_points[SPRV(i)];		/* vertex preceding the current one at i */
		if (v_eq (prevP, p))				/* Distance zero? Not acceptable. */
			prevP = snake_points[SPRV2(i)];	/* Try the preceding-preceding vertex */
		if (v_eq (prevP, p))				/* Still at distace zero? Unlikely, but... */
			return 0;						/* ... in that case, we give up. */
		nextP = snake_points[SNXT(i)];		/* Vertex succeeding the current one at i */
		if (v_eq (nextP, p))				/* And the same set of tests */
			nextP = snake_points[SNXT2(i)];
		if (v_eq (nextP, p))
			return 0;

		v1mag = sqrt ( SQR(prevP.x-p.x) + SQR(prevP.y-p.y));
		v2mag = sqrt ( SQR(nextP.x-p.x) + SQR(nextP.y-p.y));
		curv_x = ((p.x-prevP.x) / v1mag) - ((nextP.x-p.x) / v2mag);
		curv_y = ((p.y-prevP.y) / v1mag) - ((nextP.y-p.y) / v2mag);

    }
	else
	{

		/* Curvature as finite second-order difference, no normalization.
			Therefore not rotation independent. Crude -- not recommended. */

		curv_x = snake_points[SPRV(i)].x - 2*p.x + snake_points[SNXT(i)].x;
		curv_y = snake_points[SPRV(i)].y - 2*p.y + snake_points[SNXT(i)].y;
	}

	r=  SQR(curv_x) + SQR(curv_y);
	dp (3, "Curvature (%d) returns %f\n",algorithm,r);
	return r*curv_scale;
}


/* Compute the average distance between vertices */

float calc_avg_dist ()
{
int i;
float avg_dist;

	avg_dist=0;
	for(i=0; i<snake_maxpts;i++)		/* This loop includes the wraparound link*/
	{
		avg_dist += calcdist(SPRV(i),snake_points[i]);
	}
	avg_dist /= snake_maxpts;
	return avg_dist;
}


/* Energy normalization is a key step in the Williams-Shah algorithm */

void e_normalize (float *energy, int n, float thresh)
{
int i;
float min, max;

	dp (3, "   E: ");
	min = max = energy[0];
	for (i=0; i<n; i++)
	{
		if (min > energy[i]) min=energy[i];
		if (max < energy[i]) max=energy[i];
		dp (3, "%6.2f ",energy[i]);
	}

	if (fabs(max-min) < thresh) max = min+thresh;


	if ( fabs(max-min) < 1e-5)
	{
		dp (3, " -> zeroed! \n");
		for (i=0; i<n; i++)
			energy[i]=0;				/* If there is no gradient, we don't observe this energy */
		return;
	}


	{
		dp (3, " ->     %f to %f ", min, max);
		for (i=0; i<n; i++)
			energy[i]= 100.0*(energy[i]-min)/(max-min);	/* Normalize */
	}
	dp (3, "\n");
}


/* Fill a n by n matrix with gvf energy at pixel x0, y0.
	Caller is responsible for providing the energy array */

void gvf_energy (float *energy, int ngh, int x0, int y0)
{
int x,y,i;
float gvf_x, gvf_y, theta, gvfmag;
float thetapix, rpix, f;

	g_assert (snake_use_gvf);		/* can't do this without gvf */
	gvf_x = readbuf_flt (&snake_gvf_image, x0,y0,0);
	gvf_y = readbuf_flt (&snake_gvf_image, x0,y0,1);

	theta = atan2 (gvf_y, gvf_x);	/* range -pi to +pi */
	gvfmag = sqrt(SQR(gvf_x)+SQR(gvf_y));

	i=0;
	for (y=-ngh; y<=ngh; y++)
		for (x=-ngh; x<=ngh; x++)
		{
			thetapix = atan2 (y,x);
			f = cos(theta-thetapix);
			rpix = sqrt(SQR(x)+SQR(y));
			energy[i++] = -300*rpix*gvfmag*f;
		}

}


/* Sum up the energy over all snake vertices, return total energy */

float snake_compute_energy (image_type* img, int curvalg)
{
float *data;
int i,j,x,y,x1,y1;
float energy, avg_dist, dist;


	g_assert (img->imgtype = B_FLOAT);
	data = img->data;
	g_assert (snake_points!=NULL);

	avg_dist=calc_avg_dist ();		/* Compute average distance between vertices */

	energy=0;
	for(i=0;i<snake_maxpts;i++)
	{
		energy +=		e_alpha[i]*calc_stretching (i, snake_points[i])					/* Stretching component */
					+	e_beta [i]*calc_curvature  (i, snake_points[i], curvalg)		/* Bending energy */
					+	e_delta[i]*calc_continuity (i, snake_points[i], avg_dist);		/* Equally spacing energy */

		/* Now handle the image energy. This is
			- either simply the gradient image value (normalized) with a negative sign
				so that the energy DECREASES towards high gradients
			- or an energy gradient computed from GVF (handled elsewhere) */

		x = snake_points[i].x; y = snake_points[i].y; 			/* Extract vertex to make the rest more readable */

		if (!snake_use_gvf) energy -=	e_gamma[i] * data[x + img->xmax*y];			/* Subtract image energy (negative gradient) */

		/* Finally handle repulsors and attractors */

		if (nattractors>0)
		{
			for (j=0; j<nattractors; j++)
			{
				x1 = snake_attractors[j].x; y1 = snake_attractors[j].y;
				dist = pow(SQR(x-x1)+SQR(y-y1),0.25);					/* Formulate as a sqrt-shaped depression */
				if (dist>50) dist=50;									/* Clamp the distance */
				energy += 30.0*dist;
			}
		}
		if (nrepulsors>0)
		{
			for (j=0; j<nrepulsors; j++)
			{
				x1 = snake_repulsors[j].x; y1 = snake_repulsors[j].y;
				dist = sqrt(SQR(x-x1)+SQR(y-y1));	/* linear dropoff */
				if (dist < 3) dist=3;		/* Cap to prevent energy from getting too high */
				energy += 100.0/dist;
			}
		}

	}

	return energy;
}


/* Purely informative, print the energy components after iterating has finished */

float snake_assess_energy (image_type* img, int iter, int curvalg)
{
int i,x,y;
float avg_dist;
float stretch, bend, contin;		/* Internal energies */
float ienergy;						/* image energy = external energy */
float *data;


	g_assert (snake_points!=NULL);
	g_assert (img->imgtype = B_FLOAT);
	data = img->data;

	avg_dist=calc_avg_dist ();		/* Compute average distance between vertices */

	stretch = bend = contin = ienergy = 0.0;
	for(i=0;i<snake_maxpts;i++)
	{
		stretch +=		e_alpha[i]*calc_stretching (i, snake_points[i]);				/* Stretching component */
		bend	+=		e_beta [i]*calc_curvature  (i, snake_points[i],curvalg);			/* Bending energy */
		contin	+=		e_delta[i]*calc_continuity (i, snake_points[i], avg_dist);		/* Equally spacing energy */
		x = snake_points[i].x; y = snake_points[i].y; 			/* Extract vertex to make the rest more readable */
		ienergy +=		e_gamma[i] * data[x + img->xmax*y];								/*  image energy */
	}
	stretch /= snake_maxpts;
	bend    /= snake_maxpts;
	contin  /= snake_maxpts;
	ienergy  /= snake_maxpts;


	dp (1, "Snake: Iteration %d. Average distance %5.2f, Internal energies   stretching=%6.2f   bending=%6.2f   continuity=%6.2f   Image=%6.2f\n",iter, avg_dist, stretch,bend,contin,ienergy);
	return stretch+bend+contin+ienergy;
}



void dp_matrix (float *mat, int ngh)
{
int i,x,y;

		i=0;
		for (y=-ngh; y<=+ngh; y++)
		{
			dp (3, "      ");
			for (x=-ngh; x<=+ngh; x++)
			{
				dp (3, "%6.2f  ", mat[ i++ ]);
			}
			dp (3, "\n");
		}
}



/************************************************************************************

	TOP LEVEL OF SNAKE ALGORITHM, HAIDEKKER FORMULATION

*************************************************************************************/


int Snake_algorithm_1 (image_type* img, int ngh, float i_alpha, float i_beta, float i_gamma, float i_delta)
{
int movement,i,itercnt,vtx, x,y,x1,y1,dx,dy, n2;
float energy, totalenergy, totalenergy_prev, epsilon;
gboolean abort_snake;
float *data;
float *energy_mat, *gvf_energy_mat;
vertex *disp;
int curvalg;


	g_assert (img->imgtype = B_FLOAT);
	data = img->data;
	g_assert (snake_points!=NULL);

	/* Note. We assign one parameter per vertex to have the ability to adapt them */

	e_alpha = (float*) calloc (snake_maxpts, sizeof (float));
	e_beta  = (float*) calloc (snake_maxpts, sizeof (float));
	e_gamma = (float*) calloc (snake_maxpts, sizeof (float));
	e_delta = (float*) calloc (snake_maxpts, sizeof (float));
	disp = (vertex*) calloc (snake_maxpts, sizeof (vertex));

	n2 = 2*ngh+1;
	energy_mat = (float*) calloc (SQR(n2)+10, sizeof (float));
	gvf_energy_mat = (float*) calloc (SQR(n2)+10, sizeof (float));

	itercnt=0; curvalg=1;
	abort_snake = FALSE;
	totalenergy_prev=-1;		/* Impossible value */

	for(i=0;i<snake_maxpts;i++)
	{
		e_alpha[i] = i_alpha;
		e_beta[i] = i_beta;
		e_gamma[i] = i_gamma;
		e_delta[i] = i_delta;
	}

	/* Greedy snake core. Repeat finding the energy minimum until max iterations exceeded or converged */

	while(!abort_snake)
	{
		//if (calc_avg_dist() < 4.0) curvalg=2;		/* Switch to unnormed if high vertex density */

		/* Pick each vertex and displace over the entire neighborhood. Fill energy matrix */

		movement=0;
		for (vtx=0; vtx < snake_maxpts; vtx++)
		{
			dp (2, "Snake: Iteration %d, Displacing vertex %d\n",itercnt,vtx);
			x1 = snake_points[vtx].x; y1 = snake_points[vtx].y;		/* Center point of matrix */

			i=0;
			for (y=y1-ngh; y<=y1+ngh; y++)
				for (x=x1-ngh; x<=x1+ngh; x++)
				{
					snake_points[vtx].x=x; snake_points[vtx].y=y;		/* new point in neighborhood */
					energy = snake_compute_energy (img,curvalg);		/* Energy for this new point */
					energy_mat[ i++ ] = energy;		/* Fill matrix */
				}

			dp (3, "Snake: This is the energy matrix\n");
			dp_matrix (energy_mat, ngh);

			/* If this is a GVF algorithm, add the gradient energy */

			if (snake_use_gvf)
			{
				gvf_energy (gvf_energy_mat, ngh, x1,y1);
				dp (3, "   GVF-Snake: This is the GVF energy matrix\n");
				dp_matrix (gvf_energy_mat, ngh);
				i=0;
				for (y=-ngh; y<=+ngh; y++)
					for (x=-ngh; x<=+ngh; x++)
					{
						energy_mat[i] += gvf_energy_mat[i]; i++;
					}
			}

			/* Scan the energy matrix for the lowest energy point and displace */

			energy = energy_mat[ngh+n2*ngh]; dx=0; dy=0; i=0;		/* Default to center */
			for (y=-ngh; y<=+ngh; y++)
				for (x=-ngh; x<=+ngh; x++)
				{
					if (energy_mat[i] < energy)		/* Found a minimum? */
					{
						energy = energy_mat[i];
						dx=x; dy=y;
					}
					i++;
				}
			dp (2, "Snake: Minimum energy %6.2f, displaced vertex by %d,%d\n",energy,dx,dy);
			disp[vtx].x=dx; disp[vtx].y=dy;
			snake_points[vtx].x=x1; snake_points[vtx].y=y1;		/* Restore the vertex */

		}

		/* Finally, displace the vertices - deferred until all vertices have been computed */

		for (vtx=0; vtx < snake_maxpts; vtx++)
		{
			dx=disp[vtx].x; dy=disp[vtx].y;

			if ((dx!=0) || (dy!=0))
			{
				snake_points[vtx].x+=dx; snake_points[vtx].y+=dy;
				movement++;
			}
		}
		totalenergy = snake_assess_energy (img, itercnt,curvalg);
		epsilon = fabs(totalenergy-totalenergy_prev); totalenergy_prev=totalenergy;
		dp (2, "Convergence level epsilon: %f\n",epsilon);

		/* One iteration finished. Check if converged or need to abort */
		/* The algorithm tends to oscillate, so there is no point in allowing too many
			iterations. Rather, abort after few iterations and re-iterate manually */

		if ((movement==0) || (itercnt++ > 25)) abort_snake=1;

	}

	free (e_alpha);
	free (e_beta);
	free (e_gamma);
	free (e_delta);
	free (energy_mat); free (gvf_energy_mat);
	free (disp);
	return itercnt;
}



/************************************************************************************

	TOP LEVEL OF SNAKE ALGORITHM, WILLIAMS-SHAH FORMULATION
		See Williams D & Shah M. A Fast Algorithm for Active Contours and
		Curvature Estimation. CVGIP Image Understanding 1992; 55(1): 14-26.

*************************************************************************************/


int Snake_algorithm_2 (image_type* img, int ngh, float i_alpha, float i_beta, float i_gamma, float i_delta)
{
int movement,i,j,itercnt,vtx, x,y,x1,y1,dx,dy, n2;
int curvalg;
float energy;
gboolean abort_snake;
float *data;
float *cenergy, *benergy, *ienergy;		/* Energies in neighborhood: continuity, bending, image */
float avg_dist;
float *curvi;
vertex p;


	g_assert (img->imgtype = B_FLOAT);
	data = img->data;
	g_assert (snake_points!=NULL);

	/* Note. We assign one parameter per vertex to have the ability to adapt them */

	e_alpha = (float*) calloc (snake_maxpts, sizeof (float));
	e_beta  = (float*) calloc (snake_maxpts, sizeof (float));
	e_gamma = (float*) calloc (snake_maxpts, sizeof (float));
	e_delta = (float*) calloc (snake_maxpts, sizeof (float));
	curvi   = (float*) calloc (snake_maxpts, sizeof (float));

	n2 = 2*ngh+1;
	cenergy = (float*) calloc (SQR(n2)+10, sizeof (float));
	benergy = (float*) calloc (SQR(n2)+10, sizeof (float));
	ienergy = (float*) calloc (SQR(n2)+10, sizeof (float));

	itercnt=0; curvalg=1;
	abort_snake = FALSE;


	for(i=0;i<snake_maxpts;i++)
	{
		e_alpha[i] = i_alpha;
		e_beta[i] = i_beta;
		e_gamma[i] = i_gamma;
		e_delta[i] = i_delta;
	}

	/* Greedy snake core. Repeat finding the energy minimum until max iterations exceeded or converged */

	while(!abort_snake)
	{

		/* Pick each vertex and displace over the entire neighborhood. Fill energy matrix */

		movement=0;
		for (j=0; j <= /*sic!*/ snake_maxpts; j++)
		{
			vtx = j; if (vtx==snake_maxpts) vtx=0;		/* Roll over to 0 - we need to treat that vertex twice */
			avg_dist=calc_avg_dist ();		/* Compute average distance between vertices */
			//if (avg_dist<4) curvalg=2;		/* Switch to non-norm'd curvature if snake gets too tight */

			dp (2, "Snake: Iteration %d, Displacing vertex %d\n",itercnt,vtx);
			x1 = snake_points[vtx].x; y1 = snake_points[vtx].y;		/* Center point of matrix */

			i=0;
			for (y=y1-ngh; y<=y1+ngh; y++)
				for (x=x1-ngh; x<=x1+ngh; x++)
				{
					p.x=x; p.y=y;
					cenergy[i] = calc_continuity (vtx, p, avg_dist);
					benergy[i] = calc_curvature (vtx, p, curvalg);
					if (!snake_use_gvf) ienergy[i] = readbuf_flt (img, x,y,0);
					i++;
				}

			if (snake_use_gvf) 	gvf_energy (ienergy, ngh, x1,y1);


			/* Normalize the energy fields to the range [0...1], then
				sum up the energy with the weight factors into ienergy */

			e_normalize (cenergy, SQR(n2), 2.0);
			e_normalize (benergy, SQR(n2), 2.0);
			e_normalize (ienergy, SQR(n2), 5.0);
			dp (3, "Snake: These are the energy matrices in order continuity - bending - internal (normalized)\n");
			i=0;
			for (y=-ngh; y<=+ngh; y++)
			{
				dp (3, "      ");
				for (x=-ngh; x<=+ngh; x++) dp (3, "%5.2f ", cenergy[i++]);
				dp (3, "     "); i-=n2;
				for (x=-ngh; x<=+ngh; x++) dp (3, "%5.2f ", benergy[i++]);
				dp (3, "     "); i-=n2;
				for (x=-ngh; x<=+ngh; x++) dp (3, "%5.2f ", ienergy[i++]);
				dp (3, "\n");
			}
			for (i=0; i<SQR(n2); i++)
				ienergy[i] = e_alpha[vtx]*cenergy[i] + e_beta[vtx]*benergy[i] + e_gamma[vtx]*cenergy[i];

			/* Scan the energy matrix for the lowest energy point and displace IMMEDIATELY */

			energy = ienergy[ngh+n2*ngh]; dx=0; dy=0; i=0;		/* Default to center */
			for (y=-ngh; y<=+ngh; y++)
				for (x=-ngh; x<=+ngh; x++)
				{
					if (ienergy[i] < energy)		/* Found a minimum? */
					{
						energy = ienergy[i];
						dx=x; dy=y;
					}
					i++;
				}
			dp (2, "Snake: Minimum energy %6.2f, displaced vertex by %d,%d\n",energy,dx,dy);
			snake_points[vtx].x+=dx; snake_points[vtx].y+=dy;		/* Displace the vertex */
			if ((dx!=0) || (dy!=0)) movement++;
		}

		/* Almost done now. Williams & Shah propose to set beta to zero if
			(1) the bending energy of a vertex is higher than its neighbors
			(2) the same bending energy is above a threshold
			(3) the edge magnitude is above a threshold */


 		snake_assess_energy (img, itercnt, curvalg);

		/* One iteration finished. Check if converged or need to abort */
		/* The algorithm tends to oscillate, so there is no point in allowing too many
			iterations. Rather, abort after few iterations and re-iterate manually */

		if ((movement==0) || (itercnt++ >= 1)) abort_snake=1;

	}

	free (e_alpha);
	free (e_beta);
	free (e_gamma);
	free (e_delta);
	free (cenergy); free (benergy); free (ienergy);
	free (curvi);
	return itercnt;
}



/************************************************************************************************/
/*                                                                                              */
/*    *****    ***    ****     ***    *****            ***    *   *     *     *   *   *****     */
/*    *       *   *   *   *   *   *   *               *   *   *   *    * *    *  *    *         */
/*    *       *   *   *   *   *       *               *       **  *   *   *   * *     *         */
/*    ****    *   *   ****    *       ****             ***    * * *   *   *   **      ****      */
/*    *       *   *   * *     *       *                   *   *  **   *****   * *     *         */
/*    *       *   *   *  *    *   *   *               *   *   *   *   *   *   *  *    *         */
/*    *        ***    *   *    ***    *****            ***    *   *   *   *   *   *   *****     */

/* This is a simple force-equilibrium formulation. Top-level function has compatible interface,
   therefore ngh and i_delta are dummies.

   Note that a better way to solve this problem was proposed by Kass et al (Kass M, Witkin A,
   Terzopoulos D. Snakes: Active Contour Models. Int J Computer Vision 1988;: 321-331.
   Xu C, Prince JL. Snakes, Shapes, and Gradient Vector Flow, IEEE Trans Image Proc 1998; 7: 359-69. ):

   Use the constant alpha, beta, gamma to create a matrix A as follows:

   | c d e 0 0 0 0 0 ... 0 0 0 a b |
   | b c d e 0 0 0 0 ... 0 0 0 0 a |
   | a b c d e 0 0 0 ... 0 0 0 0 0 |
   | 0 a b c d e 0 0 ... 0 0 0 0 0 |
   | .......
   | .......
   | 0 0 0 0 0 0 0 0 ... a b c d e |
   | e 0 0 0 0 0 0 0 ... 0 a b c d |
   | d e 0 0 0 0 0 0 ... 0 0 a b c |

	This matrix has the dimensions snake_maxpts x snake_maxpts. The coefficients are computed as
		a = beta;
		b = -alpha - 4 beta
		c = 2 alpha + 6 beta + gamma
		d = b
		e = a			under the assumption of nonadaptive alpha, beta

	Next, invert the matrix -> invA.
	Finally, compute the displacements of the vertex vectors X and Y through

	X = invA * (gamma* X + kappa*fx(Xi,Yi));
	Y = invA * (gamma* Y + kappa*fy(Xi,Y1));

	where fx, fy is the external gradient field; either the directional derivatives of the
	image gradient or the forces of the GVF field. Note that X and Y contain non-integer
	coordinates requiring interpolation of the force field.
*/


int Snake_algorithm_3 (image_type* img, int ngh, float i_alpha, float i_beta, float i_gamma, float i_delta)
{
vertexf *fsnake;		/* We need fractional coordinates */
vertexf *disp;			/* and fractional displacements */
float eps, maxeps;		/* To compute convergence */
float dx, dy;			/* The step up-gradient */
float deltat;			/* Time-step delta t, small, constant value */
float x,y,fx,fy;
int i;
int itercnt;

	/* Copy the integer snake vertices into the vertexf field */

	fsnake = (vertexf *) calloc (snake_maxpts, sizeof (vertexf));
	disp = (vertexf *) calloc (snake_maxpts, sizeof (vertexf));
	if (!fsnake || !disp) fatal_memerr();
	for (i=0; i<snake_maxpts; i++)
	{
		fsnake[i].x = snake_points[i].x;
		fsnake[i].y = snake_points[i].y;
	}


	itercnt=0; deltat = 0.1;
	do
	{
		maxeps = 0;
		for (i=0; i<snake_maxpts; i++)
		{

			/* Stretching energy. A second-order term exerts a contractile force */

			x = (fsnake[SPRV(i)].x - 2*fsnake[i].x + fsnake[SNXT(i)].x);
			y = (fsnake[SPRV(i)].y - 2*fsnake[i].y + fsnake[SNXT(i)].y);
			//x = 0.5*(fsnake[SNXT(i)].x - fsnake[SPRV(i)].x);
			//y = 0.5*(fsnake[SNXT(i)].y - fsnake[SPRV(i)].y);		/* A first-order term makes the snake rotate clockwise */
			dx=i_alpha*x; dy=i_alpha*y;
			dp (2, "    Vertex %d: Stretch %5.2f,%5.2f   ",i,x,y);

			/* Bending energy - Neither 3rd nor 4th order gives any reasonable behavior, so for this formulation
			   the beding force can simply be turned off. This is probably related to the primitive numerical
			   implementation, but this is what you find in the literature. */

			//x = (fsnake[SPRV2(i)].x - 4*fsnake[SPRV(i)].x + 4*fsnake[i].x - 4*fsnake[SNXT(i)].x + fsnake[SNXT2(i)].x)/10;
			//y = (fsnake[SPRV2(i)].y - 4*fsnake[SPRV(i)].y + 4*fsnake[i].y - 4*fsnake[SNXT(i)].y + fsnake[SNXT2(i)].y)/10;
			x = (fsnake[SPRV2(i)].x - 2*fsnake[SPRV(i)].x + 2*fsnake[SNXT(i)].x - fsnake[SNXT2(i)].x)/10;
			y = (fsnake[SPRV2(i)].y - 2*fsnake[SPRV(i)].y + 2*fsnake[SNXT(i)].y - fsnake[SNXT2(i)].y)/10;
			dx -= i_beta*x; dy -= i_beta*y;
			dp (2, "   Bend %5.2f,%5.2f   ",x,y);

			/* Image energy, gradient of the gradient image. Using simple Canny`s hoping that we did enough Gaussian smoothing.
				Image has a range from 0 to 100, so divide accordingly */

			fx = fsnake[i].x; fy = fsnake[i].y;
			if (snake_use_gvf)
			{
				x = 0.1*i_gamma * (ireadbuf (&snake_gvf_image,fx,fy,0,intp_pref));
				y = 0.1*i_gamma * (ireadbuf (&snake_gvf_image,fx,fy,1,intp_pref));
			}
			else
			{
				x = 0.1*i_gamma * (ireadbuf (img,fx+1,fy,cur_slice,intp_pref) - ireadbuf (img,fx-1,fy,cur_slice,intp_pref));
				y = 0.1*i_gamma * (ireadbuf (img,fx,fy+1,cur_slice,intp_pref) - ireadbuf (img,fx,fy-1,cur_slice,intp_pref));
			}
			dx += i_gamma*x; dy += i_gamma*y;
			dp (2, "   Image %5.2f,%5.2f   total: %5.2f,%5.2f\n",x,y,dx,dy);

			disp[i].x = dx; disp[i].y = dy; eps = SQR(dx)+SQR(dy);
			if (eps > maxeps) maxeps=eps;
		}

		/* Now apply the displacements */

		for (i=0; i<snake_maxpts; i++)
		{
			fsnake[i].x += disp[i].x*deltat; fsnake[i].y += disp[i].y*deltat;
		}
		dp (2, "Snake: Iteration %d, max Epsilon %5.2f\n",itercnt,maxeps);

		itercnt++;
	}
	while ( (itercnt<25) && (maxeps > 1));		/* Converged if step size less than one pixel */
	dp (1, "Snake converged: Max Epsilon %5.2f after %d iterations\n",maxeps, itercnt);

	/* After convergence, round the new snake data and store back in snake_points */

	for (i=0; i<snake_maxpts; i++)
	{
		snake_points[i].x = (int)(0.5+fsnake[i].x);
		snake_points[i].y = (int)(0.5+fsnake[i].y);
		dp (2, "Snake point %d at %d,%d\n",i,snake_points[i].x, snake_points[i].y);

	}

	//snake_assess_energy (itercnt);		Can't use that here b/c we don't have the adaptive arrays.

	free (fsnake);
	free (disp);
	return itercnt;

}





/******************************************************************

	Interpolate to get a new number of vertices.
	The first function uses simple linear interpolation within
	the segments. The second function makes use of natural
	cubic splines (in functions2.c) for a smoother interpolation.
	TODO: It should be relatively straightforward to introduce
	some form of normalization in which the interpolated seglents
	have approximately the same length...

******************************************************************/


void Snake_vertex_interpol (int newnum)
{
vertex *new_snake_points;
vertex p;
float a;
int i,j;

	new_snake_points = (vertex*) calloc (newnum, sizeof (vertex));
	if (!new_snake_points) fatal_memerr();

	for (i=0; i<newnum; i++)
	{
		a = (float)i*snake_maxpts/(float)newnum;
		j = (int)a;		/* Segment we are in */
		a = a-j;		/* Offset into segment */

		p.x = a*snake_points[SNXT(j)].x + (1.0-a)*snake_points[j].x;
		p.y = a*snake_points[SNXT(j)].y + (1.0-a)*snake_points[j].y;
		new_snake_points[i] = p;
	}

	free (snake_points);
	snake_points = new_snake_points;
	snake_maxpts = newnum;
	if (calcdist(0, snake_points[snake_maxpts-1])< 2) snake_maxpts--;

}



void Snake_vertex_splinterpol (int newnum)
{
spline_block splx, sply;
double *vtx;
vertex *new_snake_points;
vertex p;
double a, dt, avglen, seg_len;
int i,j, v;
unsigned char adaptive = 0;


	/* Build the spline arrays for x and y (these are linear and separable).
		The vtx array phase-shifts the coordinates by two places to add
		a two-point wraparound. This is necessary to ensure continuity
		at the vertex where the contour closes. */

	vtx = calloc (snake_maxpts+10, sizeof(double));
	for (i=-2; i<snake_maxpts+2; i++)					/* Compatibility conversion */
		vtx[i+2] = snake_points[SWRAP(i)].x;			/* Wrap to ensure continuity */
	spline_new (&splx, snake_maxpts+4, vtx);			/* Goal: Get splx prepared */

	for (i=-2; i<snake_maxpts+2; i++)
		vtx[i+2] = snake_points[SWRAP(i)].y;			/* Reuse vtx for y spline */
	spline_new (&sply, snake_maxpts+4, vtx);
	free (vtx);

	/* dt is the interpolated spacing. If we increment an index k by dt,
		its integer part indicates the segment, and its fractional part
		is the t-coordinate for the spline function.
		avglen is the average segment length with the old number of vertices.
		By scaling it with snake_maxpts / newnum (=dt), we obtain the average length
		of interpolated segments. That value can be used to scale the step size
		depending on the uninterpolated segment length and thus more or less
		equalize vertex spacing. */

	dt = (double)snake_maxpts/(double)newnum;
	avglen = calc_avg_dist() * dt;

	/* Create temporary storage for the new vertices of the ROI */

	new_snake_points = (vertex*) calloc (3*newnum, sizeof (vertex));
	if (!new_snake_points) fatal_memerr();

	/* Actual interpolation. Interpolated vertices are indexed by non-integer a,
		and a increments by (old # of vertices)/(new # of vertices).
		Similar to the linear interpolation where interpolation takes place between
		vertices j = int(a) and j+1, we use the same scheme, but the fractional part
		is used on the spline. We merely need to ensure proper rounding into the vertex.
		Also, the 'j+2' in spline_get() reflects the wraparound phase introduced above.
	*/

	dp (2, "Refining snake from %d vertices to %d vertices\n", snake_maxpts, newnum);
	a = 0; i = 0; j=0;

	if (adaptive)		/* This is not yet ready for use, so don't */
	{
		while ((a<snake_maxpts))
		{
			seg_len = calcdist (j, snake_points[SNXT(j)]);
			j = (int)a;		/* Segment we are in */

			v = spline_get (&splx, j+2, a-j);
			p.x = IROUND(v);
			v = spline_get (&sply, j+2, a-j);
			p.y = IROUND(v);

			new_snake_points[i] = p;
			dp (2, "    New vertex %d (%d, %d) in segment %.2f between (%d, %d) and (%d, %d)\n", i, p.x, p.y, a, snake_points[j].x, snake_points[j].y, snake_points[SNXT(j)].x, snake_points[SNXT(j)].y);
			a += dt*avglen/seg_len;		/* adaptive point along the interpolated contour */
			i++;
		}
		newnum=i;
	}
	else
	{
		for (i=0; i<newnum; i++)
		{
			a = i*dt;		/* Nonadaptive point along the interpolated contour */
			j = (int)a;		/* Segment we are in */
			a = a-j;		/* Offset into segment */

			v = spline_get (&splx, j+2, a);
			p.x = IROUND(v);
			v = spline_get (&sply, j+2, a);
			p.y = IROUND(v);

			new_snake_points[i] = p;
			dp (3, "    New vertex %d (%d, %d) in old segment %d between (%d, %d) and (%d, %d)\n", i, p.x, p.y, j, snake_points[j].x, snake_points[j].y, snake_points[SNXT(j)].x, snake_points[SNXT(j)].y);
		}
	}

	/* Now trade places. Activate the new vertices and discard the old ones. */

	free (snake_points);
	spline_free (&splx); spline_free (&sply);
	snake_points = new_snake_points;
	snake_maxpts = newnum;
	if (calcdist(0, snake_points[snake_maxpts-1])< 2) snake_maxpts--;

}


/* Vertex cleanup does two things:
	- Check for any vertices that collapsed to the same point
		and delete one of them
	- Redistribute the vertices slightly to nudge them in the
		direction of equal distances (TODO - not yet implemented)
*/


int snake_vertex_cleanup ()
{
int i, j, cnt;
vertex p;

	/* Check for collapsed vertices. Wraparound ignored */

	i=1; cnt=0;
	while (i < snake_maxpts)
	{
		p = snake_points[SPRV(i)];			/* Vertex preceding i */
		if (v_eq (p, snake_points[i]))		/* Same coordinates, delete the current one */
		{
			dp (1, "Cleanup: Vertices %d and %d have same coordinates (%d, %d)\n", SPRV(i),i, p.x,p.y);
			if (i<snake_maxpts-1)
			{
				for (j=i+1; j<snake_maxpts; j++)
					snake_points[SPRV(j)] = snake_points[j];
			}
			snake_maxpts--;
			cnt++;
		}
		else
			i++;							/* Not collapsed, proceed with next vertex */
	}

	return cnt;
}




/*************************************

	Compute the gradient image
	NB - we used to create a subimage
	determined by ROI size, but for
	the extension to GVF we'd better
	do a full-size gradient

**************************************/



vertex Snake_gradient_image (image_type* img)
{
int bx1, bx2, by1, by2;
float* p1;
float* p2;
float f, fmin, fmax;
int x,y, xf,yf, i,j, x1,y1;
long idx;
vertex rtnv;

	/* Clone the main image; make FLOAT */

	roi_bbox (img, &bx1,&by1,&bx2,&by2);
	bx1 = MAX (0,bx1-30);
	bx2 = MIN (mainimg.xmax, bx2+30);
	by1 = MAX (0,by1-30);
	by2 = MIN (mainimg.ymax, by2+30);
	xf = mainimg.xmax; yf = mainimg.ymax;

	if (allocate_image (img, B_FLOAT, xf,yf, 1)==-1)
		fatal_memerr ();
	p2 = img->data;

	/* Now compute the gradient */

	idx=0;
	for (y=0; y<yf; y++)
		for (x=0; x<xf; x++)
			p2[idx++]=readbuf_flt (&mainimg,x,y,cur_slice);
	make_minmax (img);

	sobel (img);						/* Edge detection */
	gaussian_general (img, 3, 0);		/* 7x7 neighborhood, sigma=0.8 */
	gaussian_general (img, 3, 0);		/* Twice, so we get even stronger broadening of the edges */
	p2 = img->data;						/* Update the pointer for the normalization step */

	/* Finally, normalize the gradient image so that the gradient is squared
		(energy!) and that the maximum is 100 */

	for (y=0; y<yf; y++)
		for (x=0; x<xf; x++)
			p2[x+xf*y] = SQR(p2[x+xf*y]);

	imgminmax (img, &fmin, &fmax);
	for (y=0; y<yf; y++)
		for (x=0; x<xf; x++)
		{
			p2[x+xf*y] = 100.0*(p2[x+xf*y]-fmin)/(fmax-fmin);
		}

	img->minval.fmin=0.0;
	img->maxval.fmax=100.0;
	rtnv.x = 0; rtnv.y = 0;
	return rtnv;
}


/* Display gradient image (or GVF) and vertex position in secondary display */


void snake_markat (image_type *img, int x, int y, rgbtriplet rgb, int shape)
{
int xf,yf;
rgbtriplet *p2;


	xf = img->xmax; yf = img->ymax;
	p2 = img->data;
	p2[x+xf*y] = rgb;			/* Center point is always marked */

	if (shape==1)				/* Make a + sign */
	{
		if (x<xf-1) p2[x+1+xf*y]=rgb;
		if (x>0) 	p2[x-1+xf*y]=rgb;
		if (y<yf-1) p2[x+xf*(y+1)] = rgb;
		if (y>0) 	p2[x+xf*(y-1)] = rgb;
	}
	else if (shape==2)			/* Make an x symbol */
	{
		if (x<xf-1)
		{
			if (y<yf-1) p2[x+1+xf*(y+1)]=rgb;
			if (y>0)	p2[x+1+xf*(y-1)] = rgb;
		}
		if (x>0)
		{
			if (y<yf-1) p2[x-1+xf*(y-1)]=rgb;
			if (y>0)	p2[x-1+xf*(y+1)] = rgb;
		}
	}
}


void snake_display_vertices ()
{
int i,xf, yf, x,y;
int c;
rgbtriplet *p2;
rgbtriplet rgb;
hsvtriplet hsv;
float f1, f2;

	xf = snake_gradient_image.xmax; yf = snake_gradient_image.ymax;

	if (allocate_image (&snake_display_image, B_RGB, xf, yf, 1)==-1)
		fatal_memerr ();
	p2 = snake_display_image.data;
	hsv.sat=255; hsv.val=255;	/* Always full saturation & brightness */

	if (snake_use_gvf)
	{
		for (y=0; y<yf; y++)
			for (x=0; x<xf; x++)
			{
				f1 = readbuf_flt(&snake_gvf_image,x,y,0);
				f2 = readbuf_flt(&snake_gvf_image,x,y,1);
				c = IROUND(sqrt(f1*f1+f2*f2)*1.5);
				if (c<0) c=0; else if (c>255) c=255;
				p2[x+xf*y] = (rgbtriplet) {c,c,c};
			}

	}
	else							/* Display the gradient image */
	{
		for (y=0; y<yf; y++)
			for (x=0; x<xf; x++)
			{
				c = (int)(readbuf_flt(&snake_gradient_image,x,y,0)*255.0/snake_gradient_image.maxval.fmax);
				if (c<0) c=0; else if (c>255) c=255;
				p2[x+xf*y] = (rgbtriplet) {c,c,c};
			}
	}

	rgb = (rgbtriplet){255,255,80};					/* Snake vertices are bright yellow x symbols */
	for (i=0; i < snake_maxpts; i++)
	{
		x=snake_points[i].x; y=snake_points[i].y;
		snake_markat (&snake_display_image, x, y, rgb, 2);
	}

	/* Paint the attractors & repulsors */

	rgb = (rgbtriplet){0,255,128};		/* attractors in green-turqoise hue */
	if (nattractors > 0)
	{
		for (i=0; i<nattractors; i++)
		{
			x = snake_attractors[i].x; y = snake_attractors[i].y;
			snake_markat (&snake_display_image, x, y, rgb, 1);
		}
	}

	rgb = (rgbtriplet){255,128,0};		/* Repulsors in red-orange hue */
	if (nrepulsors > 0)
	{
		for (i=0; i<nrepulsors; i++)
		{
			x = snake_repulsors[i].x; y = snake_repulsors[i].y;
			snake_markat (&snake_display_image, x, y, rgb, 1);
		}
	}

	secondary_display (&snake_display_image, 0, "Snake gradient image");

}


/*************************************************************************

	Code to compute the gradient vector field. The first function is
	generalized enough to serve as a genral filter, the second is
	provided for compatibility reasons.
	For GVF, see Xu C & Prince JL. Gradient Vector Flow: A New External
	Force for Snakes. IEEE Proc Conf Comp Vis Patt Recogn 1997, pp 66-71

	This function works on a single-slice basis and cannot handle RGB.

***************************************************************************/

void gvf_laplacian (float *p, int xf, int yf)
{
float *target;
int x,y,x1,y1,x2,y2;
float buf,sum;
float lapl[3][3] = {{0.5,1,0.5} , {1,-6,1} , {0.5,1,0.5}};

	target = (float *)calloc (xf*yf, sizeof (float));

	for (y=0; y<yf; y++)
		for (x=0; x<xf; x++)
		{
			buf = 0;  sum=0;
			for (y1=-1; y1<=1; y1++)
				for (x1=-1; x1<=1; x1++)
				{
					x2 = x+x1; if (x2<0) x2=0; else if (x2>=xf) x2=xf-1;
					y2 = y+y1; if (y2<0) y2=0; else if (y2>=yf) y2=yf-1;
					buf += p[x2+xf*y2] * lapl[y1+1][x1+1];
				}
			target[x+xf*y] = buf;
		}
	memcpy (p, target, xf*yf*sizeof(float));
	free (target);
}


void compute_gvf (image_type *img, int slice, float mu)
{
image_type edges;			/* The gradient image and directional derivatives */
image_type gvf;				/* The final gradient vector field in x and y direction */
image_type tempimg;			/* Intermediate */

int x,y,xf,yf, x1,y1,x2,y2,itercnt;
float *pe, *pt, *pg;
float smooth_mat[25] = {1,4,6,4,1, 4,16,24,26,4, 6,24,36,24,6, 4,16,24,26,4, 1,4,6,4,1};
float buf1, buf2, fac, deltat, eps1, eps2, epsilon;
float min, max;


	/* Preprocess the input image. Compute Sobel(img) and the directional
		derivatives towards x and y. Note that the edge image must be normalized
		to values <= 1 or the iteration explodes numerically. */

	xf = img->xmax; yf = img->ymax;
	imgminmax (img, &min, &max);
	if (allocate_image (&edges, B_FLOAT, xf, yf, 4)==-1)
		fatal_memerr ();
	pe = edges.data;
	for (y=0; y<yf; y++)
		for (x=0; x<xf; x++)
			pe[x + xf*y] = (readbuf_flt (img,x,y,slice)-min)/(max-min);
	sobel_core ((float *)edges.data, xf,yf);				/* Sobel edge image in Slice 0 */
	convolve ((float *)edges.data, xf,yf, smooth_mat, 5,5);	/* Do some post-filter smoothing */

	for (y=0; y<yf; y++)
		for (x=0; x<xf; x++)
		{
			x1=x-1; x2=x+1; fac=0.5;
			if (x==0)
			{
				x1=x; fac=1;
			}
			else if (x==xf-1)
			{
				x2=x; fac=1;
			}
			buf1 = fac*(pe[x2+xf*y] - pe[x1+xf*y]);		/* Difference in x direction */
			pe[x+xf*y+xf*yf] = buf1;					/* Store df/dx in slice 1 */

			y1=y-1; y2=y+1; fac=0.5;
			if (y==0)
			{
				y1=y; fac=1;
			}
			else if (y==yf-1)
			{
				y2=y; fac=1;
			}
			buf2 = fac*(pe[x+xf*y2] - pe[x+xf*y1]);		/* Difference in y direction */
			pe[x+xf*y+2*xf*yf] = buf2;					/* Store df/dy in slice 2 */

			buf1 = SQR(buf1) + SQR(buf2);
			pe[x+xf*y+3*xf*yf] = buf1;					/* Store (df/dx)^2+(df/dy)^2 in slice 3 */
		}


	/* Allocate gvf (output) memory and temp image memory.
		Initialize gvf to gradient (doesn't help much, though) */

	if (allocate_image (&gvf, B_FLOAT, xf, yf, 2)==-1)
		fatal_memerr ();
	pg = gvf.data;
	for (y=0; y<yf; y++)
		for (x=0; x<xf; x++)
		{
			pg[x+xf*y]      = pe[x+xf*(y+yf)];
			pg[x+xf*(y+yf)] = pe[x+xf*(y+yf+yf)];
		}

	if (allocate_image (&tempimg, B_FLOAT, xf, yf, 1)==-1)
		fatal_memerr ();
	pt = tempimg.data;

	/* Careful when selecting deltat t. Above 0.2513/(mu+0.1774), the iteration develops unstable
		behavior. However, the larger this value, the more gradient you'll get.
		So let's try to stay a safe margin away from that instability border */

	if (mu>1) mu=1;					/* clamp mu */
	deltat = 0.2513/(mu+0.1774)-0.1;
	dp (1, "Computing GVF with mu=%f, delta t=%f\n",mu,deltat);

	itercnt=0;

	/* Ready to start the time-stepping loop */

	do
	{

		 /* One iteration in x direction. Compute mu*Laplace(u) - (u-df/dx)*( (df/dx)^2+(df/dy)^2) */

		eps2 = eps1 = 0.0;
		for (y=0; y<yf; y++)
			for (x=0; x<xf; x++)
				pt[x+xf*y] = pg[x+xf*y]*mu;			/* Step 1: mu*u -> temp */

		gvf_laplacian (pt,xf,yf);					/* Step 2: Laplace(temp) is now mu*Laplace(u) -> temp */

		for (y=0; y<yf; y++)						/* Step 3: Finalize du/dt into temp */
			for (x=0; x<xf; x++)
				pt[x+xf*y] -= (pg[x+xf*y] - pe[x+xf*(y+yf)])*pe[x+xf*(y+3*yf)];

		for (y=0; y<yf; y++)						/* Step 4: Add (du/dt)*dt to u, completes the iteration */
			for (x=0; x<xf; x++)
			{
				buf1 = deltat*pt[x+xf*y];
				pg[x+xf*y] += buf1;
				eps1 = MAX(eps1, fabs(buf1));
			}

		 /* One iteration in y direction. Compute mu*Laplace(v) - (v-df/dy)*( (df/dx)^2+(df/dy)^2) */

		for (y=0; y<yf; y++)
			for (x=0; x<xf; x++)
				pt[x+xf*y] = pg[x+xf*(y+yf)]*mu;			/* Step 1: mu*v -> temp */

		gvf_laplacian (pt,xf,yf);							/* Step 2: Laplace(temp) is now mu*Laplace(v) -> temp */

		for (y=0; y<yf; y++)								/* Step 3: Finalize dv/dt into temp */
			for (x=0; x<xf; x++)
				pt[x+xf*y] -= (pg[x+xf*(y+yf)] - pe[x+xf*(y+2*yf)])*pe[x+xf*(y+3*yf)];

		for (y=0; y<yf; y++)								/* Step 4: Add (dv/dt)*dt to v, completes the iteration */
			for (x=0; x<xf; x++)
			{
				buf2 = deltat*pt[x+xf*y];
				pg[x+xf*(y+yf)] += buf2;
				eps2 = MAX(eps2, fabs(buf2));
			}

		/* Check convergence */

		epsilon = SQR(eps1)+SQR(eps2); itercnt++;
		dp (2, "Snake: Computation of GVF, iteration %d, epsilon %f\n",itercnt, epsilon);

	}
	while ((itercnt < 1000) && (epsilon > 0.0000001));

	freebuf (&tempimg);
	freebuf (&edges);
	make_minmax (&gvf);		imgminmax (&gvf, &min, &max);

	dp (1, "GVF done.\n");

	/* At this point, the gvf has a range from approximately -1 to 1. Expand to -100..100,
		but keep the zero. This means that we need to expand neg. and pos. values separately.*/

	for (y=0; y<yf; y++)
		for (x=0; x<xf; x++)
		{
			buf1 = pg[x+xf*y];
			if (buf1 < 0)
				buf1 = 100.0*buf1/fabs(min);
			else
				buf1 = 100.0*buf1/fabs(max);
			pg[x+xf*y] = buf1;

			buf1 = pg[x+xf*(y+yf)];
			if (buf1 < 0)
				buf1 = 100.0*buf1/fabs(min);
			else
				buf1 = 100.0*buf1/fabs(max);
			pg[x+xf*(y+yf)] = buf1;
		}

	handback (img, &gvf);

}


/* Much like we visualize the gradient image

void visualize_gvf (image_type *gvfimg)
{



}



/*************************************************************************

	Code to compute the Fourier descriptors of the snake.
	The snake can be described by x(i)=a0 + sum (ak*cos(2*pi*k*i)+bk*sin(2*pi*k*i))
	and a similar equation for y(i). Sum runs from k=1 to infinity.

	This function returns the pair (ak,bk) or (ck,dk) for a given k
	for the snake as it is presently set. The general equation is
	ak=(1/pi) sum x(i)* cos (2*pi*k*i/N)
	bk=(1/pi) sum x(i)* sin (2*pi*k*i/N)

	The first descriptor, a0,c0, is the centroid.
	The second descriptor (k=1) is related to size
	The higher descriptors quantify irregularity on different levels.


***************************************************************************/



vertexf snake_fourier_point (int p, vertexf cent, int alg)
{
vertexf rtnval;

	switch (alg)
	{
		case 1:								/* Return point relative to centroid */
			rtnval.x = snake_points[p].x-cent.x;
			rtnval.y = snake_points[p].y-cent.y;
			break;
		case 2:								/* Return vector from previous point to this point */
			rtnval.x = snake_points[p].x - snake_points[SPRV(p)].x;
			rtnval.y = snake_points[p].y - snake_points[SPRV(p)].y;
			break;
		default:							/* ...and case 0; Return the original, unprocessed point */
			rtnval.x = snake_points[p].x;
			rtnval.y = snake_points[p].y;
	}
	return rtnval;
}




vertexf snake_fourier_descriptors (int k, int xy)
{
vertexf rtnval, p;
float centx, centy;
int i, alg;

	alg=0;

	/* Determine centroid */

	centx = centy = 0.0;
	for(i=0; i<snake_maxpts; i++)
	{
		centx += snake_points[i].x;
		centy += snake_points[i].y;
	}
	centx /= snake_maxpts;
	centy /= snake_maxpts;

	rtnval.x=rtnval.y=0.0;

	if (k==0)				/* trap this case b/c we have a different prefactor */
	{
		if (xy)
		{
			for(i=0; i<snake_maxpts; i++)
			{
				p = snake_fourier_point (i, (vertexf){centx,centy}, alg);
				rtnval.x += p.y;											/* c0 */
			}
		}
		else
		{
			for(i=0; i<snake_maxpts; i++)
			{
				p = snake_fourier_point (i, (vertexf){centx,centy}, alg);
				rtnval.x += p.x;											/* a0 */
			}
		}
		rtnval.x /= snake_maxpts;
		return rtnval;
	}

	if (xy)
		for(i=0; i<snake_maxpts; i++)
		{
			p = snake_fourier_point (i, (vertexf){centx,centy}, alg);
			rtnval.x += p.y * cos(2*PI*k*i/snake_maxpts);				/* ck */
			rtnval.y += p.y * sin(2*PI*k*i/snake_maxpts);				/* dk */
		}
	else
		for(i=0; i<snake_maxpts; i++)
		{
			p = snake_fourier_point (i, (vertexf){centx,centy}, alg);
			rtnval.x += p.x * cos(2*PI*k*i/snake_maxpts);				/* ak */
			rtnval.y += p.x * sin(2*PI*k*i/snake_maxpts);				/* bk */
		}


	rtnval.x /= snake_maxpts;
	rtnval.y /= snake_maxpts;
	return rtnval;
}




/******************************************/

/*          ***    *   *    ***           */
/*         *   *   *   *     *            */
/*         *       *   *     *            */
/*         *  **   *   *     *            */
/*         *   *   *   *     *            */
/*         *   *   *   *     *            */
/*          ***     ***     ***           */
/*

	Snake controls GUI

********************************************/

GtkWidget *snake_wnd;
GtkWidget *snake_alphascale;
GtkWidget *snake_betascale;
GtkWidget *snake_gammascale;
GtkWidget *snake_deltascale;
GtkWidget *snake_neighscale;
GtkWidget *snake_numvertscale;
GtkAdjustment* snake_alpha_adj;
GtkAdjustment* snake_beta_adj;
GtkAdjustment* snake_gamma_adj;
GtkAdjustment* snake_delta_adj;
GtkAdjustment* snake_neigh_adj;
GtkAdjustment* snake_numvert_adj;
GtkWidget *snake_statusbar1;

void snake_statusbar_display (const char* text);




/* This gets called whenever a slider has been changed.
	- read control panel values
	- then start the snake algorithm
*/

void snake_finalize_roi ()
{
int i;

	current_roi.bbox_x1 = current_roi.bbox_x2 = snake_points[0].x;
	current_roi.bbox_y1 = current_roi.bbox_y2 = snake_points[0].y;
	for (i=0; i<snake_maxpts; i++)
	{
		current_roi.numvertices = snake_maxpts;
		current_roi.xi[i] = snake_points[i].x;
		current_roi.yi[i] = snake_points[i].y;
		if (current_roi.bbox_x1 > current_roi.xi[i]) current_roi.bbox_x1 = current_roi.xi[i];
		if (current_roi.bbox_x2 < current_roi.xi[i]) current_roi.bbox_x2 = current_roi.xi[i];
		if (current_roi.bbox_y1 > current_roi.yi[i]) current_roi.bbox_y1 = current_roi.yi[i];
		if (current_roi.bbox_y2 < current_roi.yi[i]) current_roi.bbox_y2 = current_roi.yi[i];
	}
	current_roi.roi_mode = ROI_POLY; /* Is no longer a circle but a general polygon */
	current_roi.xi[current_roi.numvertices] = current_roi.xi[0];
	current_roi.yi[current_roi.numvertices] = current_roi.yi[0];
	current_roi.numvertices++;
	current_roi.xi[current_roi.numvertices]=-1;
	current_roi.yi[current_roi.numvertices]=-1;

}


/* Perform one set of iterations of the snake */

void snake_go (GtkWidget* widget, gpointer user_data)
{
float i_alpha, i_beta, i_gamma, i_delta;
int itercnt, cleanupcnt;
char s[255];

	if (snake_interpolate_pend)				/* Need to interpolate a new number of vertices */
	{
		Snake_vertex_splinterpol (snake_interpolate_pend);
		snake_display_vertices ();
		snake_interpolate_pend = 0;
	}

	i_alpha = snake_alpha_adj->value;
	i_beta = snake_beta_adj->value;
	i_gamma = snake_gamma_adj->value;
	i_delta = snake_delta_adj->value;
	/* snake_maxpts = (int) snake_numvert_adj->value; */
	neighborhood = (int) snake_neigh_adj->value;

	threshold_curvature = 0.9;	/* What value? */
	threshold_movement = 1;	 	/* ;Max # of vertices that still move */
	itercnt=0;					/* Silence a compiler warning */

	snake_statusbar_display ("Computing");
	switch (selected_snake_algorithm)
	{
		case 0:
			itercnt = Snake_algorithm_1 (&snake_gradient_image, neighborhood, i_alpha, i_beta, i_gamma, i_delta);
			break;
		case 1:
			itercnt = Snake_algorithm_2 (&snake_gradient_image, neighborhood, i_alpha, i_beta, i_gamma, i_delta);
			break;
		case 2:
			itercnt = Snake_algorithm_3 (&snake_gradient_image, neighborhood, i_alpha, i_beta, i_gamma, i_delta);
			break;
	}

	cleanupcnt = snake_vertex_cleanup ();
	if (cleanupcnt)
		sprintf (s, "Idle after %d iterations, %d duplicates removed", itercnt, cleanupcnt);
	else
		sprintf (s, "Idle after %d iterations", itercnt);
	snake_statusbar_display (s);

	snake_finalize_roi();
	snake_display_vertices ();
}


/* Whenever the Z spinbutton is clicked, this function gets called.
   We need to recompute the gradient image and perform one set of
   snake iterations with snake_go () */

/* 	TODO: Must handle GVF if selected.
	TODO: Huge memory leak. Yuck!
*/

void snake_zcallback ()
{
	dp (2, "Snake Z callback called \n");
	freebuf (&snake_gradient_image);				/* Discard the old gradient image */
	Snake_gradient_image (&snake_gradient_image);	/* Compute gradient from new slice */
	snake_display_vertices ();						/* Let's stay interactive */
	if (snake_autoiter) snake_go (NULL, NULL);


}


/* This is a callback for the GUI and gets called when a relevant
	image event has occurred (such as adding a repulsor) */

void snake_interaction_callback (const char *message)
{
	snake_display_vertices ();						/* Let's stay interactive */
	if (message) snake_statusbar_display (message);
	if (snake_autoiter) snake_go (NULL, NULL);

}

void snake_statusbar_display (const char* text)
{

guint context;

	context = gtk_statusbar_get_context_id((GtkStatusbar*)snake_statusbar1, "snakecontext");

	if (snake_statusbar_was_used) /* pop the last message */
	{
		gtk_statusbar_pop((GtkStatusbar*)snake_statusbar1, context);
		snake_statusbar_was_used=0;
	}

	gtk_statusbar_push((GtkStatusbar*)snake_statusbar1, context, text);
	snake_statusbar_was_used = 1;
}



void snake_slider_changed (GtkWidget* widget, gpointer user_data)
{

	if (snake_autoiter)  snake_go (widget, user_data);

}



void snake_numverts_changed (GtkWidget* widget, gpointer user_data)
{
int newnum;

	newnum = (int) snake_numvert_adj->value;
	snake_interpolate_pend = newnum;			/* Nonzero. Use it as flag and information carrier */


}



void on_snake_cancel_clicked  (GtkButton *button, gpointer user_data)
{
	if (restoreroi) current_roi = snake_roi_save;
	freebuf (&snake_gradient_image);
	snake_active=0;
	zslice_spin_unregister ();
	gtk_widget_destroy (GTK_WIDGET(user_data));
}


/* Apply copies the results of the last snake application into curent ROI */

void on_snake_apply_clicked (GtkButton *button, gpointer user_data)
{

	restoreroi=0;
	freebuf (&snake_gradient_image);
	snake_active=0;
	zslice_spin_unregister ();
	gtk_widget_destroy (GTK_WIDGET(user_data));

}


void on_snake_checkbutton1_toggled (GtkButton *button, gpointer user_data)
{
	snake_autoiter = (*(GtkToggleButton*)user_data).active;

}


void on_snake_checkbutton2_toggled (GtkButton *button, gpointer user_data)
{
GtkWidget *messg;
int togglestate;

	togglestate = gtk_toggle_button_get_active ((GtkToggleButton*)user_data);
	if (togglestate)
	{								/* GVF selected */

		if (snake_use_gvf) return;		/* Don't create GVF twice */
		snake_use_gvf = 1;
		messg = ShowDynMessage ("Creating GVF.\nThis may take some time.","Please wait");
		while (gtk_events_pending())
			gtk_main_iteration();

		clone_image (&mainimg, &snake_gvf_image);
		compute_gvf (&snake_gvf_image, cur_slice, 0.2);
		//vsrtwrite (&snake_gvf_image, "snake_gvf.vflt", 0, snake_gvf_image.zmax-1);
		gtk_widget_destroy (messg);
		snake_display_vertices ();
		gtk_button_set_label (GTK_BUTTON(user_data), "GVF Ready");
	}
	else
	{
		snake_use_gvf = 0;
		gtk_button_set_label (GTK_BUTTON(user_data), "Regular gradient image");
	}
		while (gtk_events_pending())
			gtk_main_iteration();


}


void on_snake_fourier_clicked (GtkButton *button, gpointer user_data)
{
char report[12005];
char buf[256];
int k;
vertexf akbk;

	strcpy (report, "Snake Fourier descriptors\n\nk     ak      bk      ck      dk\n------------------------------------------\n");
	for (k=0; k<10; k++)			/* We arbitrarily end the series at k=9 */
	{
		akbk = snake_fourier_descriptors (k, 0);
		sprintf (buf, "%1d   %7.3f %7.3f",k,akbk.x, akbk.y);
		strncat (report, buf, 12000);
		akbk = snake_fourier_descriptors (k, 1);
		sprintf (buf, " %7.3f %7.3f\n",akbk.x, akbk.y);
		strncat (report, buf, 12000);
	}
	ShowLongMessage (report, "Snake Fourier Descriptors");

}



void create_snake_wnd (void)
{
  GtkWidget *vbox1, *hbox1;
  GtkWidget *snake_table1;
  GtkWidget *snake_label1, *snake_label2, *snake_label3, *snake_label4, *snake_label5, *snake_label6;
  GtkWidget *snake_label7, *snake_label8, *snake_label9;
  GtkWidget *snake_cancel;
  GtkWidget *snake_apply;
  GtkWidget *snake_compute;
  GtkWidget *snake_dspfourier;
  GtkWidget *snake_check1;
  GtkWidget *snake_check2;
  GtkTooltips *tt_snalpha, *tt_snbeta, *tt_sngamma, *tt_sndelta, *tt_snngh, *tt_snvtx, *tt_snautoi, *tt_sngvf;
  GtkTooltips *tt_snroi, *tt_sniter, *tt_snfour, *tt_sncancel;
  int y;

	snake_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (snake_wnd), "snake_wnd", snake_wnd);
	gtk_window_set_title (GTK_WINDOW (snake_wnd), "Snake Controls");
	vbox1 = vbox_in_window (snake_wnd, "vbox1");

	snake_label1 = MakeLabel (snake_wnd,vbox1," ","snake_label1",-1,-1);
	snake_table1 = MakeTable (snake_wnd,vbox1,"snake_table1",2,7);

	y=0;
	if (selected_snake_algorithm==1)
		snake_label2 = MakeLabel (snake_wnd,snake_table1,"Alpha (continuity)","snake_label2",0,y);
	else
		snake_label2 = MakeLabel (snake_wnd,snake_table1,"Alpha (stretching)","snake_label2",0,y);
	snake_alpha_adj = GTK_ADJUSTMENT (gtk_adjustment_new (0.5, 0, 4, 0.1, 1, 0));
	snake_alphascale = MakeHScale (snake_alpha_adj,snake_wnd,snake_table1,"snake_alphascale",1,y);
	tt_snalpha = gtk_tooltips_new ();
	if (selected_snake_algorithm==1)
		gtk_tooltips_set_tip (tt_snalpha, snake_alphascale, "Weight for internal continuity force\n(deviation from average length)\nThis is responsible for shrinking the snake.", NULL);
	else
		gtk_tooltips_set_tip (tt_snalpha, snake_alphascale, "Weight for internal stretching force\n(squared normalized length)\nThis is responsible for shrinking the snake.", NULL);
	y++;

	snake_label3 = MakeLabel (snake_wnd,snake_table1,"Beta (curvature)","snake_label3",0,y);
	snake_beta_adj = GTK_ADJUSTMENT(gtk_adjustment_new (0.5, 0, 4, 0.1, 1, 0));
	snake_betascale = MakeHScale (snake_beta_adj,snake_wnd,snake_table1,"snake_betascale",1,y);
	tt_snbeta = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_snbeta, snake_betascale, "Weight for internal bending force\n(curvature or second derivative)\nThis prevents sharp angles or corners.", NULL);
	y++;

	snake_label4 = MakeLabel (snake_wnd,snake_table1,"Gamma (Image)","snake_label4",0,y);
	snake_gamma_adj = GTK_ADJUSTMENT(gtk_adjustment_new (2.0, 0, 8, 0.1, 1, 0));
	snake_gammascale = MakeHScale (snake_gamma_adj,snake_wnd,snake_table1,"snake_gammascale",1,y);
	tt_sngamma = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_sngamma, snake_gammascale, "Weight for external image force\n(negative image gradient or GVF)", NULL);
	y++;

	snake_label7 = MakeLabel (snake_wnd,snake_table1,"Delta (Spacing)","snake_label7",0,y);
	snake_delta_adj = GTK_ADJUSTMENT(gtk_adjustment_new (0.2, 0, 2, 0.1, 1, 0));
	snake_deltascale = MakeHScale (snake_delta_adj,snake_wnd,snake_table1,"snake_deltascale",1,y);
	tt_sndelta = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_sndelta, snake_deltascale, "Weight for internal continuity force\n(squared deviation from average vertex distance)\nimplies that alpha is normalized length.\nThis prevents vertices from clustering together.", NULL);
	if (selected_snake_algorithm==1)
	{
		gtk_widget_hide (snake_label7);
		gtk_widget_hide (snake_deltascale);
	}
	y++;

	snake_label5 = MakeLabel (snake_wnd,snake_table1,"Neighborhood","snake_label5",0,y);
	snake_neigh_adj = GTK_ADJUSTMENT(gtk_adjustment_new (1, 1, MAX_NEIGH-1, 1, 1, 0));
	snake_neighscale = MakeHScale (snake_neigh_adj,snake_wnd,snake_table1,"snake_neighscale",1,y);
	gtk_scale_set_digits (GTK_SCALE (snake_neighscale), 0);
	tt_snngh = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_snngh, snake_neighscale, "Greedy search neighborhood radius\nLarger radius usually rasults in better convergence\nespecially for non-GVF searches", NULL);
	y++;

	snake_label6 = MakeLabel (snake_wnd,snake_table1,"No. of vertices","snake_label6",0,y);
	snake_numvert_adj = GTK_ADJUSTMENT(gtk_adjustment_new (snake_maxpts, 5, ROI_MAXVERTICES-1, 1, 10, 0));
	snake_numvertscale = MakeHScale (snake_numvert_adj,snake_wnd,snake_table1,"snake_numvertscale",1,y);
	gtk_scale_set_digits (GTK_SCALE (snake_numvertscale), 0);
	tt_snvtx = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_snvtx, snake_numvertscale, "Number of vertices. Changing this number interpolates\nnew vertices into the curve.\nMore vertices usually yield better results.", NULL);
	y++;

	snake_label8 = MakeLabel (snake_wnd,snake_table1,"Auto-iterate","snake_label8",0,y);
	snake_check1 = Checkbutton_in_table (snake_wnd, snake_table1, " ","snake_check1",1,y, NULL);
	tt_snautoi = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_snautoi, snake_check1, "If checked, the snake recomputes automatically\nwhenever a slider is moved", NULL);
	y++;

	snake_label9 = MakeLabel (snake_wnd,snake_table1,"Use GVF","snake_label9",0,y);
	snake_check2 = Togglebutton_in_table (snake_wnd, snake_table1, "Regular gradient image","snake_check2",1,y, NULL);
	tt_sngvf = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_sngvf, snake_check2, "Toggles between gradient and GVF.\nInitial GVF computation may take some time", NULL);
	y++;

	hbox1 = hbox_in_window (snake_wnd, vbox1, "snake_hbox");

	snake_apply  = Button_in_table (snake_wnd,hbox1,"Apply ROI","snake_apply",-1,-1,NULL);
	snake_compute  = Button_in_table (snake_wnd,hbox1,"Iterate","snake_compute",-1,-1,NULL);
	snake_dspfourier  = Button_in_table (snake_wnd,hbox1,"Fourier descriptors","snake_fourier",-1,-1,NULL);
	snake_cancel = Button_in_table (snake_wnd,hbox1,"Cancel","snake_cancel",-1,-1,NULL);
	tt_snroi = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_snroi, snake_apply, "Applies current snake as new ROI and closes this window", NULL);
	tt_sniter = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_sniter, snake_compute, "Computes the new snake under present parameter settings", NULL);
	tt_snfour = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_snfour, snake_dspfourier, "Computes Fourier coefficients of the current snake", NULL);
	tt_sncancel = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_sncancel, snake_cancel, "Exits this window, keeps old ROI", NULL);

	snake_statusbar1 = gtk_statusbar_new ();
	gtk_widget_ref (snake_statusbar1);
	gtk_object_set_data_full (GTK_OBJECT (snake_wnd), "snake_statusbar1", snake_statusbar1,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (snake_statusbar1);
	gtk_box_pack_start (GTK_BOX (vbox1), snake_statusbar1, FALSE, FALSE, 0);

	gtk_signal_connect (GTK_OBJECT (snake_cancel), "clicked",
                    	GTK_SIGNAL_FUNC (on_snake_cancel_clicked), snake_wnd);
	gtk_signal_connect (GTK_OBJECT (snake_compute), "clicked",
                    	GTK_SIGNAL_FUNC (snake_go), snake_wnd);
	gtk_signal_connect (GTK_OBJECT (snake_apply), "clicked",
                    	GTK_SIGNAL_FUNC (on_snake_apply_clicked), snake_wnd);
	gtk_signal_connect (GTK_OBJECT (snake_wnd), "destroy",
                    	GTK_SIGNAL_FUNC (on_snake_cancel_clicked), snake_wnd);
	gtk_signal_connect (GTK_OBJECT (snake_check1), "toggled", GTK_SIGNAL_FUNC (on_snake_checkbutton1_toggled), snake_check1);
	gtk_signal_connect (GTK_OBJECT (snake_check2), "toggled", GTK_SIGNAL_FUNC (on_snake_checkbutton2_toggled), snake_check2);
	gtk_signal_connect (GTK_OBJECT (snake_dspfourier), "clicked",
                    	GTK_SIGNAL_FUNC (on_snake_fourier_clicked), snake_wnd);

	gtk_signal_connect (GTK_OBJECT (snake_alpha_adj), "value_changed",
                    	GTK_SIGNAL_FUNC (snake_slider_changed), NULL);
	gtk_signal_connect (GTK_OBJECT (snake_beta_adj), "value_changed",
                    	GTK_SIGNAL_FUNC (snake_slider_changed), NULL);
	gtk_signal_connect (GTK_OBJECT (snake_gamma_adj), "value_changed",
                    	GTK_SIGNAL_FUNC (snake_slider_changed), NULL);
	gtk_signal_connect (GTK_OBJECT (snake_delta_adj), "value_changed",
                    	GTK_SIGNAL_FUNC (snake_slider_changed), NULL);
	gtk_signal_connect (GTK_OBJECT (snake_neigh_adj), "value_changed",
                    	GTK_SIGNAL_FUNC (snake_slider_changed), NULL);
	gtk_signal_connect (GTK_OBJECT (snake_numvert_adj), "value_changed",
                    	GTK_SIGNAL_FUNC (snake_numverts_changed), NULL);

	gtk_widget_show (snake_wnd);

}




/*************************************************

	Snake: Main

**************************************************/


void snake_main (int snake_type)
{
int i;

	if (snake_active) return;

	if (spider_active)
	{
		ShowMessage ("Snake and Spider segmentation are mutually exclusive");
		return;
	}


	if (((current_roi.roi_mode !=ROI_POLY) && (current_roi.roi_mode!=ROI_CIRCLE))
		|| (current_roi.numvertices<5))
	{
		ShowMessage ("Snake requires polygonal ROI with >= 5 vertices");
		return;
	}

	/* Grab vertices from ROI */

	snake_roi_save = current_roi; restoreroi=1;
	current_roi.roi_mode = ROI_POLY;

	snake_points = (vertex*) calloc (current_roi.numvertices, sizeof (vertex));
	if (!snake_points) fatal_memerr();

	snake_autoiter=0;
	for (i=0; i<current_roi.numvertices; i++)
	{
		snake_points[i].x = current_roi.xi[i];
		snake_points[i].y = current_roi.yi[i];
	}
	snake_maxpts = current_roi.numvertices;
	if (v_eq(snake_points[0],snake_points[snake_maxpts-1])) snake_maxpts--;
	stretch_scale = 1.0; curv_scale = 1.0; cont_scale = 1.0;
	if (snake_type == 1)
	{
		init_length = calc_avg_dist() * snake_maxpts;	/* For stretching energy normalization */
		stretch_scale = 500.0/init_length;
		curv_scale = 5.0;
		cont_scale = 1.0;
		dp (1, "Initialized snake: %d vertices, total length: %.0f pixels, internal scale factors alpha=%f; beta=%f, delta=%f\n", snake_maxpts, init_length, stretch_scale, curv_scale, cont_scale);
	}

	snake_statusbar_was_used=0;
	snake_active=1;
	snake_use_gvf=0;
	snake_interpolate_pend = 0;
	nattractors=0; nrepulsors=0;
	selected_snake_algorithm = snake_type;
	create_snake_wnd ();
	snake_statusbar_display ("Creating gradient image");

	/* Compute, then display, a gradient from the main img */

	Snake_gradient_image (&snake_gradient_image);
	zslice_spin_register ((gpointer)snake_zcallback);

	snake_display_vertices ();

}



