

/* Module: Wavelets filters.
	Please note that this module is highly experimental.


*/

/*		MODULE FOR CRYSTAL IMAGE 0.85 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.

		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */




#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>



#include "application.h"
#include "functions.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "viz.h"
#include "wvlt_core.h"
#include "wvlt_filt.h"






/*		Wavelet-based filtering (denoising).
		As of rev. 1.0.0, this has become a separate function.
		Moreover, the steps of DWT -> shrinking -> IDWT have been separated
		from the image structure so that intermediate results can be maintained
		as double-precision.
*/


/*
		I guess at some point we need a definition of L/H and H/L.
		Since x is the fastest index, it would come first, so that
		a one-level decomposition would be represented by

		+---------+---------+---> x
		|         |         |
		|   L/L   |   H/L   |
		|         |         |
		+---------+---------+
		|         |         |
		|   L/H   |   H/H   |
		|         |         |
		+---------+---------+
		|
		v  y
*/



/* Wavelet highpass filter (background removal).
	The highpass filter operates only on the L/L level. The simplest way to
	deal with this scenario (driven by wavelet_denoise) is to pass the full
	image data buffer in data with dimensions xm, ym, and to determine the
	boundaries of the L/L area with the parameter 'level'.

	The idea behind highpass filtering is that the details (high-frequency content)
	can be found in the lower decomposition levels, H/H, H/L, and L/H. If we leave those alone
	but shrink the L/L coefficients, then we increase the *relative* high-frequency
	energy or content.

	One highpass formulation is nonlinear shrinking in which large coefficients are
	attenuated more than small ones. A sine function (slope=1 for small values of the
	sine argument) advertises itself. The highpass auto-adjusts itself so that
	the median L/L coefficient hits 45 degrees, and its sine is 0.71. Since pi/4 = 0.79,
	the median coefficient is attenuated by ~10%. Highest attenuation would be pi/2 = 1.57
	reduced to sin(pi/2) = 1, i.e., by 36%. This does not appear to be very strong,
	but the effect is rather dramatic.

	(Hm... or would it be more reasonable to reduce the lower-valued L/L coefficients,
	hard-thresholding style?)
*/


void wavelet_highpass (double *data, int xm, int ym, int level, int whp_func, double multiplier)
{
int x,y, x2,y2, i, j, idx;
double *tdata;
double arg, median, max, scalefac;
char sign;


	x2 = xm; y2 = ym;
	for (i=0; i<level; i++)
	{
		x2 = x2 >> 1;					/* Limit to the L/L region of image */
		y2 = y2 >> 1;
	}
	tdata = calloc (x2*y2, sizeof (double));
	dp (2, "Wavelet HP functions: L/L region from (0,0) to (%d,%d).\n", x2-1,y2-1);

	/* Get the coefficients of the L/L region for analysis */

	idx=0; max = 0;
	for (y=0; y<y2; y++)
		for (x=0; x<x2; x++)
		{
			arg = fabs (data[x + xm*y]);
			tdata[idx++] = arg;
			max = MAX (max, arg);
		}

	median = element_finder (tdata, idx/2, idx);
	dp (2, "Wavelet HP functions: Median L/L coefficient is %f, maximum is %f.\n", median, max);

	switch (whp_func)
	{
		case 0:			/* Highpass with large-coefficient attenuation */
						/* Highpass scale factor: Designed so that the median wavelet coefficient scales to pi/4.
							By also using the multiplier, we can shift the scale of coefficients being reduced */

				scalefac = PI / (4.0*median*multiplier);
				for (y=0; y<y2; y++)
					for (x=0; x<x2; x++)
					{
						arg = data[x + xm*y] * scalefac;			/* Sine is an odd function, so the sign is kept correctly */
						if (fabs(arg) > PI/2)
							arg = SIGN(arg)*PI/2;					/* Keep within +/- PI/2 */
						arg = sin(arg);								/* Sine-clipping; sin(arg)=arg for small arg */
						data[x + xm*y] = arg /scalefac;				/* De-normalize and ensure slope=1 for small coeffs */
					}
				break;

		case 1:			/* Highpass with small-coefficient attenuation */
						/* Highpass scale factor: Designed so that the median wavelet coefficient scales to 1.0.
							Multiplier used similar as above. */

				scalefac = multiplier * median;
				for (y=0; y<y2; y++)
					for (x=0; x<x2; x++)
					{
						arg = data[x + xm*y];
						sign = SIGN(arg);					/* The spline is not an odd function, need to preserve sign */
						arg = fabs(arg)/scalefac;			/* Scale so that median coeff becomes 1.0 */
						if ( arg < 1.0 )					/* Large coefficients (above median) are kept */
							arg = (2.0-arg)*arg*arg;		/* ... and smaller coefficients subjected to 2x^2 - x^3 */

						arg = sign*arg * scalefac;			/* Restore sign & undo median normalization */
						data[x + xm*y] = arg;				/* Store back in data array */
					}
				break;
	}

	free (tdata);

}



/* A variant of the highpass. This function can be called from anywhere, and it merely returns
	the isolated L/L area to a callback function, where the caller can perform the
	desired operation. Useful, for example, if the function is related to a different menu item,
	such as background removal. The big deal is that it is integrated with the wavelet infrastructure
	and does not need separate steps for decomposition and reconstruction.
*/


void wavelet_hp_with_callback (double *data, int xm, int ym, int level,
			void (*wvhp_callback)(double *data, int xsize, int ysize) )
{
int x,y, x2,y2, i, j, idx;
double *tdata;


	x2 = xm; y2 = ym;
	for (i=0; i<level; i++)
	{
		x2 = x2 >> 1;					/* Limit to the L/L region of image */
		y2 = y2 >> 1;
	}
	tdata = calloc (x2*y2, sizeof (double));
	dp (2, "Wavelet HP with callback: L/L region from (0,0) to (%d,%d).\n", x2-1,y2-1);

	/* Get the coefficients of the L/L region for analysis */

	idx=0;
	for (y=0; y<y2; y++)
		for (x=0; x<x2; x++)
			tdata[idx++] = data[x + xm*y];

	wvhp_callback (tdata, x2, y2);		/* execute the callback function */

	/* Copy modified tdata back into data */

	for (y=0; y<y2; y++)
		for (x=0; x<x2; x++)
		{
			data[x + xm*y] = tdata[x + x2*y];
		}

	free (tdata);


}






/* Wavelet shrinkage for image denoising.

	2D slice-by-slice wavelet shrinkage with different threshold finding algorithms.
	img must be a wavelet-transformed image with a decomposition given in decomp_level.
	For this reason, we assume the image type to be B_FLOAT (otherwise, we return doing nothing)
	(Note that this function performs the second step out of DWT -> shrinking -> IDWT)

	This function performs wavelet shrinkage according to wshrinkfunc and returns
	the shrunken coefficients for subsequent reconstruction.
	The function is
		0: Universal threshold, soft thresholding
		1: Universal threshold, hard thresholding
		2: Bayes shrink, soft thresholding
		3: Bayes shrink, hard thresholding
		4: SUREShrink, soft thresholding
		5: SUREShrink, hard thresholding
		6: GCV, soft thresholding
		7: GCV, hard thresholding

*/



void wavelet_shrink (double *data, int xm, int ym, int decomp_level, int wshrinkfunc, double tmult)
{
int x,y, x2,y2, i, j;
int level;
double *tdata;
double sigma, T;		/* estimated noise, the threshold */


	sigma = estimate_sigma_2d (data, xm,ym);		/* Robust median estimator in HH (Donoho) */
	dp (2, "    Denoising: Estimate of noise, sigma=%f\n", sigma);

	x2=xm; y2=ym;
	tdata = calloc (xm*ym, sizeof (double));		/* a very generous allocation ... */

	for (level=0; level<decomp_level; level++)
	{
		/* For each level, go over the H/L, H/H, and L/H areas separately.
			This also means having a separate shrinking threshold. */

		i=0;
		for (y=0; y<y2/2; y++)					/* Top right quadrant, H/L */
			for (x=x2/2; x<x2; x++)
				tdata[i++] = data[x + xm*y];

		T = find_wsthreshold (tdata, i, sigma, wshrinkfunc >> 1);
		dp (2, "    Denoising: Shrinking H/L of decomposition level %d with threshold %f\n", level, T);

		for (y=0; y<y2/2; y++)
			for (x=x2/2; x<x2; x++)
				data[x + xm*y] = wv_thresh (data[x + xm*y], T*tmult, wshrinkfunc & 0x01);

		i=0;
		for (y=y2/2; y<y2; y++)					/* Bottom right quadrant, L/L */
			for (x=x2/2; x<x2; x++)
				tdata[i++] = data[x + xm*y];

		T = find_wsthreshold (tdata, i, sigma, wshrinkfunc >> 1);
		dp (2, "    Denoising: Shrinking H/H of decomposition level %d with threshold %f\n", level, T);

		for (y=y2/2; y<y2; y++)
			for (x=x2/2; x<x2; x++)
				data[x + xm*y] = wv_thresh (data[x + xm*y], T*tmult, wshrinkfunc & 0x01);

		i=0;
		for (y=y2/2; y<y2; y++)					/* Bottom left quadrant, L/H */
			for (x=0; x<x2/2; x++)
				tdata[i++] = data[x + xm*y];

		T = find_wsthreshold (tdata, i, sigma, wshrinkfunc >> 1);
		dp (2, "    Denoising: Shrinking L/H of decomposition level %d with threshold %f\n", level, T);

		for (y=y2/2; y<y2; y++)
			for (x=0; x<x2/2; x++)
				data[x + xm*y] = wv_thresh (data[x + xm*y], T*tmult, wshrinkfunc & 0x01);

		x2 /= 2; y2 /= 2;			/* Focus on this level's L/L area for next level */

	}

	free (tdata);

}





/* Now we use the above for an image -- any image. The image data is copied into a
	double-precision buffer and processed there. At the end, the original data type is restored.
	This function loops over all slices of an image and processes it on a per-slice basis.

	NOTE: Since the wavelet filter functions require an image size divisible by 2 for
	each decomposition level, we pad the data buffer here, and we un-pad it when we store the
	filtered result back into the image buffer. The best idea would be to have power-of-two
	image sizes anyway.

	Note also: This function is poorly named. It performs coefficient shrinking, for sure,
	but this includes experimental highpass functions. If a non-NULL callback function is given,
	the function index 'shrinkfunc' is ignored.
*/


void wavelet_denoise (image_type *img, int wfunction, int shrinkfunc, int decomp, double tmult,
					  void (*wvhp_callback)(double *data, int xsize, int ysize) )
{
WAVELET_TYPE wtype;
wavefilt wfilter;
double *data, *p;
double buf;
int xs, ys;							/* Power-of-two padded image */
int x, y, z, i, e, idx, xmym;
rgbtriplet rgb;


	/* Determine the padded side lengths */

	xs = img->xmax;
	ys = img->ymax;
// 	e = 1;
// 	for (i=0; i<decomp+3; i++)
// 	{
// 		if (xs & 0x01) xs++;			/* If not divisible by 2, add 1 to make it an even number */
// 		if (ys & 0x01) ys++;
// 		xs >>= 1;						/* Divide by 2 to examine next level */
// 		ys >>= 1;
// 		e <<= 1;						/* Keep track of the powers of 2 */
// 	}
// 	xs *= e; ys *= e;

	xs = 1;
	while (xs < img->xmax) xs <<= 1;
	ys = 1;
	while (ys < img->ymax) ys <<= 1;

	dp (1, "Wavelet filter functions: Padding image size (%d, %d) to (%d, %d)\n", img->xmax, img->ymax, xs,ys);

	wtype = set_subband (wfunction);
	wfilter = wavelet_prepare (wtype);

	e = 1; if (img->imgtype==B_RGB) e=3;
	xmym = xs * ys;

	data = calloc (3*xmym, sizeof (double));		/* x3 so we can handle RGB in one pass */

	for (z=0; z<img->zmax; z++)
	{
		/* Clear the padded data region from the possible previous filter action */

		for (i=0; i<3*xmym; i++)
			data[i] = 0;

		/* Copy one slice into the data buffer. For RGB types, separate the r,g,b channels */

		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				idx = x + y * xs;
				if (img->imgtype == B_RGB)
				{
					rgb = readbuf_rgb (img, x,y,z);
					data[idx         ] = rgb.red;
					data[idx +   xmym] = rgb.green;
					data[idx + 2*xmym] = rgb.blue;
				}
				else				/* All grayscale data types */
				{
					data[idx] = readbuf_flt (img, x,y,z);
				}
			}

		/* Buffer is ready, execute the DWT as specified. Do it three times if needed for RGB */

		for (i=0; i<e; i++)
		{
			p = &(data[i*xmym]);
			dwt_2d (p, W_FORWARD, xs, ys, decomp, wfilter, *subband);

			if (wvhp_callback != NULL)		/* Callback function given */
				 wavelet_hp_with_callback (p, xs, ys, decomp, wvhp_callback);
			if (shrinkfunc < 8)
				wavelet_shrink (p, xs, ys, decomp, shrinkfunc, tmult);
			else
				wavelet_highpass (p, xs, ys, decomp, shrinkfunc-8, tmult);

			dwt_2d (p, W_INVERSE, xs, ys, decomp, wfilter, *subband);
		}

		/* So far, so good... stuff it back into the original image data structure.
			This may require rounding and clamping. We also drop whatever is in the padded extra space */

		switch (img->imgtype)
		{
			case B_CHAR:
			{
				unsigned char* p1 = img->data;
				idx=z*img->xmax*img->ymax;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						i = IROUND(data[x + xs*y]);
						if (i<0) i=0; else if (i>255) i=255;
						p1[x + img->xmax*y+idx]=i;
					}
				break;
			}
			case B_SHORT:
			{
				short* p1 = img->data;
				idx=z*img->xmax*img->ymax;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						i = IROUND(data[x + xs*y]);
						p1[x + img->xmax*y+idx]=i;
					}
				break;
			}
			case B_FLOAT:
			{
				float* p1 = img->data;
				idx=z*img->xmax*img->ymax;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						p1[x + img->xmax*y + idx] = data[x+xs*y];
					}
				break;
			}
			case B_RGB:
			{
				rgbtriplet* p1 = img->data;
				idx=e*z*img->xmax*img->ymax;
				for (y=0; y<img->ymax; y++)
					for (x=0; x<img->xmax; x++)
					{
						i = IROUND(data[x + xs*y]);
						rgb.red = CLAMP (i, 0, 255);
						i = IROUND(data[x + xs*y + xmym]);
						rgb.green = CLAMP (i, 0, 255);
						i = IROUND(data[x + xs*y + 2*xmym]);
						rgb.blue = CLAMP (i, 0, 255);
						p1[x + img->xmax*y+idx]=rgb;
					}
				break;
			}

		}

		/* Done, proceed to next z-slice */
	}

	/* Finally, recompute min & max and invalidate the interpolation cache */

	make_minmax (img);
	img->uid = next_uid++;

}



/* This short function checks how much padding is needed to get a power-of-two image size.
	If padding results in a major increase in the number of pixels, returns TRUE. */

unsigned char check_dyadic (image_type *img)
{
int xs, ys;
float area_frac;

	xs = 1;
	while (xs < img->xmax) xs <<= 1;
	ys = 1;
	while (ys < img->ymax) ys <<= 1;

	dp (1, "Wavelet filter functions: Padding image size (%d, %d) to (%d, %d)\n", img->xmax, img->ymax, xs,ys);

	/* The area fraction is the fraction of padded pixels in the padded image. Absent padding,
		the area fraction is (obviously) zero. Since a lot of coefficients get sorted, or added up to
		a mean of squares, zero-padded pixels bias the filter metrics.
		I do not know exactly how high a fraction of zero-padded pixels is tolerable, I'll simply
		assume a value that I could imagine to be reasonable. For now, this is 8%.
	*/

	area_frac = ((float)xs * (float)ys - (float)img->xmax * (float)img->ymax) / ((float)xs * (float)ys);
	return (area_frac > 0.08);
}




/*+**********************************+******************+

  ####    #    #   #*#
 #    #   #    #    #
 #        #    #    #
 #   ##   #    #    #
 #    #   #    #    #
 #    #   #    #    #
  ####     ####    #*#

*/




wavefilt wfilt;
int wfunction, wshrinkfunc;
GtkAdjustment *wvlt_decomp_adj, *wvlt_tmult_adj;




void on_wvlt_filt_ok_clicked  (GtkButton *button, gpointer user_data)
{
int decomp_level;
WAVELET_TYPE wtype;
double tmult;
char msg[512];


	undo_nexus (&mainimg);

	decomp_level = wvlt_decomp_adj->value;
	tmult = wvlt_tmult_adj->value;

	wtype = set_subband (wfunction);
	if (wtype!=W_NULL)
	{
		sprintf (msg, "Denoised: %s, ", wvlt_used);
		if ((wshrinkfunc>>1) ==0) strcat (msg, "VisuShrink ");
		else if ((wshrinkfunc>>1) ==1) strcat (msg, "BayesShrink ");
		else if ((wshrinkfunc>>1) ==2) strcat (msg, "SUREShrink ");
		else if ((wshrinkfunc>>1) ==3) strcat (msg, "GCV ");
		else if ((wshrinkfunc>>1) ==4) sprintf (msg, "Highpass: %s, ", wvlt_used);

		if (wshrinkfunc & 0x01) strcat (msg, "(hard)");
		else strcat (msg, "(soft)");

		wavelet_denoise (&mainimg, wfunction, wshrinkfunc,decomp_level, tmult, NULL);
	}
	else
		strcpy (msg, "Error: NULL wavelet filter. How did this happen?");

	make_imlib_buffer_from_main_image (cur_slice);
	redraw(0);
	statusbar_display (msg);

	gtk_widget_destroy (GTK_WIDGET(user_data));
}




void wvlt_filt_shrink_menu (GtkWidget *widget, gpointer user_data)
{

	wshrinkfunc = *(int*) user_data;
	dp (2, "Changed filter function to %d\n", wshrinkfunc);

}


void wvlt_filt_type_menu (GtkWidget *widget, gpointer user_data)
{

	wfunction = *(int*) user_data;
	dp (2, "Changed wavelet to %d\n", wfunction);

}

void on_wvlt_filt_cancel_clicked  (GtkButton *button, gpointer user_data)
{

	gtk_widget_destroy (GTK_WIDGET(user_data));
}







void create_wvlt_filt_window (image_type *img)
{
  GtkWidget *wvlt_window;
  GtkWidget *vbox1, *hbox1, *wvlt_table;
  GtkWidget *wvlt_label2, *wvlt_label3, *wvlt_warnlbl;
  GtkWidget *hsep1;
  GtkWidget *wvlt_funcmenu;
  GtkWidget *wvlt_shrinkmenu;
  GtkWidget *wvlt_typemenu;
  GtkWidget *wvlt_label4, *wvlt_label1, *wvlt_label5;
  GtkWidget *wvlt_tmult;
  GtkWidget *wvlt_okbutton;
  GtkWidget *wvlt_showbutton;
  GtkWidget *wvlt_testbutton;
  GtkWidget *wvlt_cancelbutton;
  GtkWidget *wvlt_typemenu_menu;
  GtkWidget *wvlt_funcmenu_menu;
  GtkWidget *wvlt_shrinkmenu_menu;
  GtkWidget *glade_menuitem;
  GtkWidget *wvlt_decomp;
  GtkTooltips *tt_wvltcancel, *tt_wvltdecomp, *tt_wvltfn, *tt_wvltok, *tt_wvltshow, *tt_wvltshrink, *tt_wvlttest, *tt_wvlttmult, *tt_wvlttype;

  int y;


	wvlt_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (wvlt_window), "wvlt_window", wvlt_window);
	gtk_window_set_title (GTK_WINDOW (wvlt_window), "Wavelet-Based Filters");

	vbox1 = vbox_in_window (wvlt_window,"vbox1");

	if (check_dyadic(img))							/* A lot of padding needed... add a warning. */
	{
		wvlt_warnlbl = MakeLabel (wvlt_window, vbox1, "NOTE: Significant zero-padding\nrequired. Filters may not\nyield reliable results.","wvlt_warnlbl",-1,-1);
	}

	wvlt_table = MakeTable (wvlt_window, vbox1, "table1", 2,6);

	y=0;
	wvlt_label1 = MakeLabel (wvlt_window,wvlt_table,"Wavelet type","wvlt_label1",0,y);
	wvlt_typemenu = Make_Optionmenu (wvlt_window, wvlt_table, "wvlt_typemenu", 1,y);
	wvlt_typemenu_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_type_menu), "Haar", (gpointer) &menuopt0);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_type_menu), "Daubechies-4", (gpointer) &menuopt1);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_type_menu), "Daubechies-8", (gpointer) &menuopt2);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_type_menu), "Daubechies-12", (gpointer) &menuopt3);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_type_menu), "Daubechies-20", (gpointer) &menuopt4);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_type_menu), "Coiflet-6", (gpointer) &menuopt5);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_type_menu), "Coiflet12", (gpointer) &menuopt6);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_type_menu), "Symlet-4", (gpointer) &menuopt7);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_type_menu), "Symlet-8", (gpointer) &menuopt8);
	glade_menuitem = Option_menuitem (wvlt_typemenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_type_menu), "Symlet-12", (gpointer) &menuopt9);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (wvlt_typemenu), wvlt_typemenu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (wvlt_typemenu), 0); wfunction=0;
	gtk_container_set_border_width (GTK_CONTAINER (wvlt_typemenu), 5);
	tt_wvlttype = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvlttype, wvlt_typemenu, "Choose mother wavelet", NULL);

	y++;
	wvlt_label2 = MakeLabel (wvlt_window,wvlt_table,"Decomposition level","wvlt_label2",0,y);
	wvlt_decomp_adj = GTK_ADJUSTMENT(gtk_adjustment_new (3, 1, 10, 1, 1, 0));
	wvlt_decomp = MakeHScale (wvlt_decomp_adj, wvlt_window, wvlt_table,"wvlt_decomp", 1,y);
	gtk_scale_set_digits (GTK_SCALE (wvlt_decomp), 0);
	tt_wvltdecomp = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvltdecomp, wvlt_decomp, "Maximum dyadic decomposition levels", NULL);

	y++;
	wvlt_label4 = MakeLabel (wvlt_window,wvlt_table,"Shrink Function","wvlt_label4",0,y);
	wvlt_shrinkmenu = Make_Optionmenu (wvlt_window, wvlt_table, "wvlt_shrinkmenu", 1,y);
	wvlt_shrinkmenu_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_shrink_menu), "Universal/VisuShrink (soft)", (gpointer) &menuopt0);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_shrink_menu), "Universal/VisuShrink (hard)", (gpointer) &menuopt1);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_shrink_menu), "BayesShrink (soft)", (gpointer) &menuopt2);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_shrink_menu), "BayesShrink (hard)", (gpointer) &menuopt3);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_shrink_menu), "SUREShrink (soft)", (gpointer) &menuopt4);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_shrink_menu), "SUREShrink (hard)", (gpointer) &menuopt5);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_shrink_menu), "GCVT (soft)", (gpointer) &menuopt6);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_shrink_menu), "GCVT (hard)", (gpointer) &menuopt7);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_shrink_menu), "Highpass 1 (experimental)", (gpointer) &menuopt8);
	glade_menuitem = Option_menuitem (wvlt_shrinkmenu_menu, GTK_SIGNAL_FUNC(wvlt_filt_shrink_menu), "Highpass 2 (experimental)", (gpointer) &menuopt9);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (wvlt_shrinkmenu), wvlt_shrinkmenu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (wvlt_shrinkmenu), 0);
	gtk_container_set_border_width (GTK_CONTAINER (wvlt_shrinkmenu), 5);
	gtk_widget_show (wvlt_label4); gtk_widget_show (wvlt_shrinkmenu);
	tt_wvltshrink = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvltshrink, wvlt_shrinkmenu, "Wavelet shrinking operation or threshold operation", NULL);

	y++;
	wvlt_label5 = MakeLabel (wvlt_window,wvlt_table,"Threshold Multiplier","wvlt_label5",0,y);
	wvlt_tmult_adj = GTK_ADJUSTMENT(gtk_adjustment_new (1, 0, 5, 0.1, 1, 0));
	wvlt_tmult = MakeHScale (wvlt_tmult_adj, wvlt_window, wvlt_table,"wvlt_tmult", 1,y);
	gtk_scale_set_digits (GTK_SCALE (wvlt_tmult), 1);
	gtk_widget_show (wvlt_label5); gtk_widget_show (wvlt_tmult);
	tt_wvlttmult = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvlttmult, wvlt_tmult, "Factor to multiply the shrinking theshold with.\nHigher = stronger shrinking.", NULL);


	hsep1 = MakeHsep (wvlt_window, vbox1, "hsep3", -1,-1);
	hbox1 = hbox_in_window (wvlt_window, vbox1, "wvlt_hbox");

	wvlt_okbutton = Button_in_table (wvlt_window, hbox1, "OK", "wvlt_okbutton",-1,-1, NULL);
	wvlt_cancelbutton = Button_in_table (wvlt_window, hbox1, "CANCEL", "wvlt_cancelbutton",-1,-1,NULL);
	tt_wvltok = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvltok, wvlt_okbutton, "Execute the selected operation", NULL);
	tt_wvltcancel = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_wvltcancel, wvlt_cancelbutton, "Abort operation, close window.", NULL);

	gtk_signal_connect (GTK_OBJECT (wvlt_okbutton), "clicked", GTK_SIGNAL_FUNC (on_wvlt_filt_ok_clicked), wvlt_window);
	gtk_signal_connect (GTK_OBJECT (wvlt_cancelbutton), "clicked", GTK_SIGNAL_FUNC (on_wvlt_filt_cancel_clicked), wvlt_window);
	gtk_signal_connect (GTK_OBJECT (wvlt_window), "destroy", GTK_SIGNAL_FUNC (on_wvlt_filt_cancel_clicked), wvlt_window);


	gtk_widget_show (wvlt_window);

}


void wavelet_filter_main ()
{

	create_wvlt_filt_window (&mainimg);

}



/* This is called by the macro interpreter */

void wvdenoise_macro_driver (image_type *img, int wfunction, int shrinkfunc, int decomp, double tmult)
{
WAVELET_TYPE wtype;
wavefilt wfilter;


	wtype = set_subband (wfunction);
	if (wtype!=W_NULL)
	{
		wavelet_denoise (img, wfunction, shrinkfunc, decomp, tmult, NULL);
	}


}



/********************************************************************************/



